/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.basic;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.FrameObserver;
import org.audiochain.model.AbstractChainableAudioDevice;
import org.audiochain.model.AudioDevice;
import org.audiochain.model.AudioDeviceProperty;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.Scale;
import org.audiochain.model.Synchronizable;
import org.audiochain.model.UserInterfaceContext;

/**
 * <p>
 * This is the AudioDevice for the basic settings level, pan, mute and solo.
 * 
 * <p>
 * An AudioDevice has several properties. These properties must be defined as
 * non transient fields. For each field there must be a JavaBean conform setter
 * method. This setter method must do the following things per convention:
 * <li>
 * assign the value to the field (to become stored)
 * <li>assign the value to the reader, if it exists (to update and hear the
 * changes immediately while playback)
 */
public class BasicAudioDevice extends AbstractChainableAudioDevice {
	private static final long serialVersionUID = 1L;
	/**
	 * Persistent fields to let the values survive in a serialization process.
	 * At this place, the initial default values for new AudioDevice instances
	 * can be set.
	 * 
	 * Use the Annotations for label the field and set symbol, unit and value
	 * ranges.
	 * 
	 * Alternatively you do not necessarily have to use the AudioDeviceProperty
	 * and implement the whole stuff by hand. See LatencyAudioDevice for an
	 * example.
	 */
	@AudioDeviceProperty(name = "Mute", symbol = 9676, falseSymbol = 9675)
	private boolean mute = false;
	// manage solo manually
	private boolean solo = false;
	private long soloChangeTime;
	@AudioDeviceProperty(name = "Pan", symbol = 9285, max = 1, min = -1, mid = 0)
	private float pan = 0.0F;
	@AudioDeviceProperty(name = "Level", symbol = 9727, max = 6, min = -40, mid = 0, unit = "dB", scale = Scale.Logarithmic)
	private float level = 0.0F;
	/*
	 * the reference to the recent created reader. You may not worry about
	 * closing, because that is the job of the reader's consumer.
	 */
	private transient BasicDeviceAudioDataReader reader;
	private transient Collection<BasicAudioDeviceListener> listeners;

	@Override
	public AudioDataReaderChainLink createAudioDataReaderChainLink(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		/* initialize the new instance with the restored values */
		reader = new BasicDeviceAudioDataReader(source, level, pan, mute, otherIsSolo());
		return reader;
	}

	/*
	 * To build more usability to the BasicAudioDevice, independently from the
	 * AudioDeviceValue mechanism, the direct call of the setter Methods affect
	 * directly the value of the reader (if it exists).
	 */
	void setLevel(float level) {
		this.level = level;
		if (reader != null) {
			reader.setLevel(level);
		}
	}

	void setPan(float pan) {
		this.pan = pan;
		if (reader != null) {
			reader.setPan(pan);
		}
	}

	void setMute(boolean mute) {
		this.mute = mute;
		if (reader != null) {
			reader.setMute(mute);
		}
	}

	void toggleMute() {
		setMute(!mute);
	}

	boolean isSolo() {
		return solo;
	}

	void setSolo(boolean solo) {
		boolean oldSolo = this.solo;
		if (oldSolo != solo) {
			if (solo) {
				makeSolo();
			} else {
				resetSolo();
			}
			setOtherSoloValues(solo);
		}
	}

	/* set the solo state but does not change the other devices */
	private void makeSolo() {
		if (!solo) {
			solo = true;
			soloChangeTime = System.currentTimeMillis();
			if (this.reader != null) {
				this.reader.setSoloBasedMute(false);
			}
			fireSoloChanged(false, true);
			firePerformanceRelevantAudioDevicePropertyChanged();
		}
	}

	private void resetSolo() {
		boolean oldSolo = solo;
		solo = false;
		soloChangeTime = System.currentTimeMillis();
		if (this.reader != null) {
			this.reader.setSoloBasedMute(false);
		}
		fireSoloChanged(oldSolo, false);
		firePerformanceRelevantAudioDevicePropertyChanged();
	}

	private void muteForOtherSolo() {
		boolean oldSolo = solo;
		solo = false;
		soloChangeTime = System.currentTimeMillis();
		if (this.reader != null) {
			this.reader.setSoloBasedMute(true);
		}
		fireSoloChanged(oldSolo, false);
		firePerformanceRelevantAudioDevicePropertyChanged();
	}

	private void fireSoloChanged(boolean oldSolo, boolean newSolo) {
		if (listeners == null) {
			return;
		}
		for (BasicAudioDeviceListener listener : listeners) {
			listener.soloChanged(oldSolo, newSolo);
		}
	}

	void toggleSolo() {
		setSolo(!solo);
	}

	private void setOtherSoloValues(boolean mySoloState) {
		if (audioProject == null) {
			return;
		}
		Collection<BasicAudioDevice> others = findOtherBasicAudioDevices();
		AudioTrack associatedAudioTrack = getAssociatedAudioTrack();
		for (BasicAudioDevice other : others) {
			// not mute if the other device is on my track
			AudioTrack otherTrack = other.getAssociatedAudioTrack();
			boolean sameTrack = associatedAudioTrack == otherTrack;
			if (mySoloState) {
				if (sameTrack) {
					other.resetSolo();
				} else {
					other.muteForOtherSolo();
				}
			} else {
				other.resetSolo();
			}
		}
	}

	/* Finds the other basic devices of all tracks, but not the master chain. */
	private Collection<BasicAudioDevice> findOtherBasicAudioDevices() {
		Collection<BasicAudioDevice> ret = new HashSet<BasicAudioDevice>();
		for (AudioTrack track : audioProject.getAudioTracks()) {
			Collection<AudioDevice> devices = new LinkedHashSet<AudioDevice>(track.getPlaybackAudioDeviceChain().getAudioDevices());
			devices.addAll(track.getCaptureAudioDeviceChain().getAudioDevices());
			for (AudioDevice audioDevice : devices) {
				// dont mute myself
				if (audioDevice != this && audioDevice instanceof BasicAudioDevice) {
					ret.add((BasicAudioDevice) audioDevice);
				}
			}
		}
		return ret;
	}

	boolean otherIsSolo() {
		if (audioProject != null) {
			// if this audioDevice is not attached to any track, but directly to
			// the project, this audioDevice should not be muted
			AudioTrack associatedAudioTrack = findAssociatedAudioTrack();
			if (associatedAudioTrack == null) {
				return false;
			}
			for (AudioTrack track : audioProject.getAudioTracks()) {
				if (track == associatedAudioTrack) {
					continue;
				}
				Collection<AudioDevice> devices = new LinkedHashSet<AudioDevice>(track.getPlaybackAudioDeviceChain().getAudioDevices());
				devices.addAll(track.getCaptureAudioDeviceChain().getAudioDevices());
				for (AudioDevice audioDevice : devices) {
					if (audioDevice instanceof BasicAudioDevice) {
						BasicAudioDevice other = (BasicAudioDevice) audioDevice;
						if (other.solo) {
							return true;
						}
					}
				}
			}
		}
		return false;
	}

	@Override
	protected UserInterfaceContext<BasicAudioDevice> createUserInterfaceContext() {
		return new BasicAudioDeviceUserInterfaceContext();
	}

	@Override
	public String getName() {
		return "Volume";
	}

	void addBasicAudioDeviceListener(BasicAudioDeviceListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<BasicAudioDeviceListener>();
		}
		listeners.add(listener);
	}

	void removeBasicAudioDeviceListener(BasicAudioDeviceListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	@Override
	public void synchronizeBy(Synchronizable source) {
		super.synchronizeBy(source);
		if (source instanceof BasicAudioDevice) {
			BasicAudioDevice other = (BasicAudioDevice) source;
			if(greaterOrBothZero(other.soloChangeTime, soloChangeTime)) {
				setSolo(other.isSolo());
				soloChangeTime = other.soloChangeTime;
			}
		}
	}
}
