/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.basic;

import java.io.IOException;
import java.util.Arrays;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;


class BasicDeviceAudioDataReader implements AudioDataReaderChainLink {

	private AudioDataReader reader;
	private float gainFactor;
	private float pan;
	private boolean positive;
	private float absolutePanValue;
	private float panMultiplicator;
	private float panGainMultiplicator;
	private boolean mute;
	private boolean soloBasedMute;
	
	BasicDeviceAudioDataReader(AudioDataReader reader, float level, float pan, boolean mute, boolean soloBasedMute) {
		this.reader = reader;
		setLevel(level);
		setPan(pan);
		setMute(mute);
		setSoloBasedMute(soloBasedMute);
	}
	
	@Override
	public int read(int[] data) throws IOException {
		int len = reader.read(data);

		if(mute || soloBasedMute) {
			Arrays.fill(data, 0);
			return len;
		}
		
		if(positive) {
			for (int l = 0, r = 1, n = data.length; l < n; l += 2, r += 2) {
				data[l] = (int)(data[l] * panGainMultiplicator);
				data[r] = (int)(data[r] * gainFactor);
			}
		} else {
			for (int l = 0, r = 1, n = data.length; l < n; l += 2, r += 2) {
				data[l] = (int) (data[l] * gainFactor);
				data[r] = (int) (data[r] * panGainMultiplicator);
			}
		}
		
		return len;
	}
	
	public void setLevel(float level) {
		// level in dB to factor
		float exp = level / 20F;
		this.gainFactor = (float) Math.pow(10D, exp);
		panGainMultiplicator = panMultiplicator * gainFactor;
	}

	public void setPan(float pan) {
		if(pan<-1F) {
			pan = -1F;
		} else if(pan>1F) {
			pan = 1F;
		}
		this.pan = pan;
		positive = pan >= 0F;
		absolutePanValue = Math.abs(pan);
		panMultiplicator = 1 - absolutePanValue;
		panGainMultiplicator = panMultiplicator * gainFactor;
	}

	public float getPan() {
		return pan;
	}

	public void setMute(boolean mute) {
		this.mute = mute;
	}

	public boolean isMute() {
		return mute;
	}

	@Override
	public void close() throws IOException {
		reader.close();
	}

	@Override
	public void stop() {
		if (reader != null) {
			reader.stop();
		}
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;		
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	public boolean isSoloBasedMute() {
		return soloBasedMute;
	}

	public void setSoloBasedMute(boolean soloBasedMute) {
		this.soloBasedMute = soloBasedMute;
	}
}
