/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.drum;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

import org.audiochain.io.AudioDataReader;


public class DrumAudioDataReader implements AudioDataReader {

	private DrumSchema drumSchema;
	private DrumRhythm rhythm;
	private final float frameRate;
	private int framesPerPulse;
	private long currentFramePosition;
	private boolean stopped;
	private long rhythmFrameCount; // the count of frames of the total rhythm
	private Collection<DrumHit> hits;
	private int totalBeatCount;
	// intial value of one prevents zero division
	private int beatsPerPulse = 1;
	private int framesPerBeat;
	private boolean repeat = false;

	public DrumAudioDataReader(DrumSchema drumSchema, DrumRhythm rhythm, float beatsPerMinute, float frameRate) {
		this.drumSchema = drumSchema;
		this.frameRate = frameRate;
		initRhythm(rhythm);
		initPulsesPerMinute(beatsPerMinute);
		computeFramesPerBeat();
	}

	@Override
	public int read(int[] data) {
		// manual end condition
		if (stopped) {
			return -1;
		}

		if (!repeat && this.currentFramePosition >= rhythmFrameCount) {
			return -1;
		}

		Arrays.fill(data, 0);
		int readSamples = readClipBytesForPosition(currentFramePosition % rhythmFrameCount, data);
		this.currentFramePosition += readSamples / 2;
		return readSamples;
	}

	int readClipBytesForPosition(final long startFramePosition, int[] target) {		
		final int targetLength = target.length;
		final long frameLength = targetLength / 2;
		final long endFramePosition = Math.min(startFramePosition + frameLength, rhythmFrameCount);
		for (DrumHit drumHit : hits) {
			long beatStartFrame = drumHit.getStartFrame();
			long hitEndFrame = drumHit.getEndFrame();
			if (beatStartFrame < endFramePosition && hitEndFrame >= startFramePosition) {
				mix(drumHit, startFramePosition, target);
			}
		}

		final int ret = (int) (endFramePosition - startFramePosition) * 2;
		return ret;
	}
	
	void mix(DrumHit hit, long relativeFramePosition, int[] target) {
		long hitStartFrame = hit.getStartFrame();
		int clipStartFramePosition;
		int targetStartFramePosition;
		if(hitStartFrame<relativeFramePosition) {
			clipStartFramePosition = (int) (relativeFramePosition - hitStartFrame);
			targetStartFramePosition = 0;
		} else {
			clipStartFramePosition = 0;
			targetStartFramePosition = (int) (hitStartFrame - relativeFramePosition);
		}
		
		DrumClip clip = drumSchema.getDrumClip(hit.getId());
		int[] clipData = clip.getClipData();
		mix(clipData, clipStartFramePosition * 2, target, targetStartFramePosition * 2);
	}

	void mix(int[] sourceSamples, int sourceStartIndex, int[] targetSamples, int targetStartIndex) {
		int sourceLength = sourceSamples.length - sourceStartIndex;
		int targetLength = targetSamples.length - targetStartIndex;
		int length = Math.min(sourceLength, targetLength);
		for (int i = 0, s = sourceStartIndex, d = targetStartIndex; i < length; i++, s++, d++) {
			int a = sourceSamples[s];
			int b = targetSamples[d];
			targetSamples[d] = a + b;
		}
	}

	@Override
	public void close() {
		stopped = true;
	}

	@Override
	public void stop() {
		stopped = true;
	}

	@Override
	public void seek(long framePosition) {
		currentFramePosition = framePosition;
	}

	public DrumRhythm getRhythm() {
		return rhythm;
	}

	private void initRhythm(DrumRhythm rhythm) {
		this.rhythm = rhythm;
		totalBeatCount = rhythm.getAllBeats().size();
		if (totalBeatCount == 0) {
			throw new IllegalArgumentException("The Rhythm '" + rhythm.getName() + "' has no beats.");
		}
		beatsPerPulse = rhythm.getBeatsPerPulse();
	}
	
	public void setDrumSchemaAndRhythm(DrumSchema drumSchema, DrumRhythm rhythm) {
		this.drumSchema = drumSchema;
		initRhythm(rhythm);
		computeFramesPerBeat();
	}
	
	private void initPulsesPerMinute(float pulsesPerMinute) {
		if (pulsesPerMinute <= 0) {
			pulsesPerMinute = 1;
		}
		float pulsesPerSecond = pulsesPerMinute / 60F; // i.e. 120 ppm / 60s = 2 pps
		framesPerPulse = (int) (frameRate / pulsesPerSecond); // 22050 fpp
	}

	public void setPulsesPerMinute(float pulsesPerMinute) {
		initPulsesPerMinute(pulsesPerMinute);
		computeFramesPerBeat();
	}

	private void computeFramesPerBeat() {
		framesPerBeat = framesPerPulse / beatsPerPulse;
		rhythmFrameCount = totalBeatCount * framesPerBeat;
		computeHitStartAndEndFrames();
	}
	
	private void computeHitStartAndEndFrames() {
		Collection<DrumHit> hits = new ArrayList<DrumHit>();
		long beatStartFrame = 0;
		for (DrumBeat beat : rhythm.getAllBeats()) {
			for (DrumHit drumHit : beat.getHits()) {
				DrumClip clip = drumSchema.getDrumClip(drumHit.getId());
				int frameLength = clip.getLength() / 2;
				long hitEndFrame = beatStartFrame + frameLength;
				
				DrumHit hit = new DrumHit(drumHit.getId());
				hit.setStartFrame(beatStartFrame);
				hit.setEndFrame(hitEndFrame);
				hits.add(hit);
			}
			beatStartFrame += framesPerBeat;
		}
		this.hits = hits;
	}

	public boolean isRepeat() {
		return repeat;
	}

	public void setRepeat(boolean repeat) {
		this.repeat = repeat;
	}

	public DrumSchema getDrumSchema() {
		return drumSchema;
	}

	long getRhythmFrameCount() {
		return rhythmFrameCount;
	}

	int getFramesPerBeat() {
		return framesPerBeat;
	}
}
