/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.drum;

import java.io.File;
import java.io.IOException;

import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.UnsupportedAudioFileException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.InputStreamAudioDataReader;


public class DrumClip {
	private final String name;
	private final File file;
	private int[] clipData;
	private final int length;
	private final float level;
	
	public DrumClip(String name, File file) throws UnsupportedAudioFileException, IOException {
		this(name, file, 1.0F);
	}
	
	public DrumClip(String name, File file, float level) throws UnsupportedAudioFileException, IOException {
		this.name = name;
		this.file = file;
		this.level = level;
		read();
		length = clipData != null ? clipData.length : 0;
	}
	
	DrumClip(String name, int[] clipData) {
		this.name = name;
		this.file = null;
		this.level = 1.0F;
		this.clipData = clipData;
		length = clipData.length;
	}

	public int[] getClipData() {
		return clipData;
	}

	public String getName() {
		return name;
	}

	public File getFile() {
		return file;
	}
	
	private void read() throws UnsupportedAudioFileException, IOException {
		final AudioInputStream ais = AudioSystem.getAudioInputStream(file);
		final AudioDataReader reader = new InputStreamAudioDataReader(ais);
		final int[] data = new int[4096];
		int len;
		while ((len = reader.read(data)) != -1) {
			if (level != 1.0F) {
				multiplyLevelFactor(data, len);
			}
			if (clipData == null) {
				clipData = new int[len];
				System.arraycopy(data, 0, clipData, 0, len);
			} else {
				int clen = clipData.length;
				int[] temp = new int[clen+len];
				System.arraycopy(clipData, 0, temp, 0, clen);
				System.arraycopy(data, 0, temp, clen, len);
				clipData = temp;
			}
		}
		reader.close();
	}
	
	private void multiplyLevelFactor(int[] data, int len) {
		for (int i = 0; i < len; i++) {
			data[i] = (int) (data[i] * level);
		}
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder(name);
		sb.append('=');
		
		if(file!=null) {
			sb.append(file.getAbsolutePath());
		} else if(clipData!=null) {
			sb.append('[');
			boolean first = true;
			for (int sample : clipData) {
				if(!first) {
					sb.append(',');
				}
				sb.append(sample);
				first = false;
			}
			sb.append(']');
		}
		if (level != 1.0F) {
			sb.append(' ').append(level);
		}
		return sb.toString();
	}

	public int getLength() {
		return length;
	}

	public float getLevel() {
		return level;
	}
}
