/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.drum;

import static javax.swing.JFileChooser.*;
import static org.audiochain.ui.gui.CommonComponentFactory.*;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Collection;

import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.xml.parsers.ParserConfigurationException;

import org.audiochain.model.AudioDeviceValue;
import org.audiochain.ui.gui.AudioDeviceValueComponent;
import org.audiochain.ui.gui.Dialog;
import org.xml.sax.SAXException;

public class DrumComponent extends Box implements DrumAudioDeviceListener, ActionListener {
	private static final long serialVersionUID = 1L;
	private final DrumAudioDevice device;
	private JLabel schemaLabel;
	private JComboBox rhythmCombo;
	private boolean fillingRhythmComboBox;
	private boolean assertFilledDrumRhythmCombobox;

	public DrumComponent(DrumAudioDevice device) {
		super(BoxLayout.Y_AXIS);
		this.device = device;
		createComponents();
		initData();
		registerListeners();
	}

	private void createComponents() {
		setOpaque(false);
		schemaLabel = new JLabel();
		schemaLabel.setAlignmentX(CENTER_ALIGNMENT);
		add(schemaLabel);

		rhythmCombo = createComboBox();
		add(rhythmCombo);

		Collection<AudioDeviceValue> audioDeviceValues = device.getAudioDeviceValues();
		if (audioDeviceValues != null) {
			for (AudioDeviceValue audioDeviceValue : audioDeviceValues) {
				String name = audioDeviceValue.getName();
				if (!name.equals("DrumSchema") && !name.equals("Rhythm")) {
					add(new AudioDeviceValueComponent(audioDeviceValue));
				}
			}
		}
	}

	private void initData() {
		setDrumSchemaLabel();
		fillDrumRhythmCombobox();
	}

	private void registerListeners() {
		device.addDrumAudioDeviceListener(this);

		schemaLabel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				String oldFileName = device.getDrumSchemaFileName();
				JFileChooser jfc;
				if (oldFileName != null) {
					jfc = new JFileChooser(device.getAudioProject().getRelativeFile(oldFileName).getParentFile());
				} else {
					jfc = new JFileChooser(device.getAudioProject().getProjectPath());
				}
				while (chooseDrumSchemaFile(jfc)) {
					// show open dialog until the chosen file is valid or the dialog is canceled
				}
			}
		});

		rhythmCombo.addActionListener(this);
	}
	
	private boolean chooseDrumSchemaFile(JFileChooser jfc) {
		int status = jfc.showOpenDialog(this);
		boolean showAgain;
		if (status == APPROVE_OPTION) {
			File file = jfc.getSelectedFile();
			try {
				device.setDrumSchemaFile(file);
				showAgain = false;
			} catch (FileNotFoundException e) {
				Dialog.message(this, "Drum Clip File Not Found", e);
				showAgain = true;
			} catch (IOException e) {
				Dialog.message(this, "Drum File Read Error", e);
				showAgain = true;
			} catch (DrumSchemaFormatException e) {
				Dialog.message(this, "Drum Schema Invalid", e);
				showAgain = true;
			} catch (UnsupportedAudioFileException e) {
				Dialog.message(this, "Drum Clip Format Unsupported", e);
				showAgain = true;
			} catch (SAXException e) {
				Throwable cause = e.getCause();
				if (cause instanceof FileNotFoundException) {
					if (Dialog.okCancel(this, "Select Drum Clip Path", e.getMessage()
							+ " Please use the following file dialog to select the path," +
							" where drum clips can be found on your local hard disk!" +
							" Hint: for example on a Debian GNU/Linux system there might be Hydrogen drumkits at" +
							" '/usr/share/hydrogen/data/drumkits'.")) {
						chooseClipSearchPaths();
					}
				} else {
					Dialog.message(this, "XML Drum Schema Invalid", e);
				}
				showAgain = true;
			} catch (ParserConfigurationException e) {
				Dialog.message(this, "XML Drum Schema Read Error", e);
				showAgain = true;
			}
		} else {
			showAgain = false;
		}
		return showAgain;
	}
	
	private void chooseClipSearchPaths() {
		DrumSettings drumSettings = DrumSettings.getInstance();
		Collection<File> clipSearchPaths = drumSettings.getClipSearchPaths();
		File dir = null;
		if(clipSearchPaths.isEmpty()) {
			dir = new File(device.getAudioProject().getProjectPath());
		} else {
			dir = clipSearchPaths.iterator().next();
		}
		JFileChooser jfc = new JFileChooser(dir);
		jfc.setFileSelectionMode(DIRECTORIES_ONLY);
		int status = jfc.showOpenDialog(this);
		if (status == APPROVE_OPTION) {
			File file = jfc.getSelectedFile();
			drumSettings.addClipSearchPath(file);
		}
	}

	@Override
	public void drumSchemaFileNameChanged(String drumSchemaFileName) {
		fillDrumRhythmCombobox();
		setDrumSchemaLabel();
		revalidate();
	}

	@Override
	public void drumRhythmNameChanged(String drumRhythmName) {
		if (assertFilledDrumRhythmCombobox) {
			fillDrumRhythmCombobox();
		} else {
			setDeviceRhythmToRhythmCombobox();
		}
	}

	/* to listen on rhythmCombo */
	@Override
	public void actionPerformed(ActionEvent evt) {
		if(fillingRhythmComboBox) {
			return;
		}
		String rhythmName = (String) rhythmCombo.getSelectedItem();
		if (rhythmName == null) {
			return;
		}
		device.setDrumRhythm(rhythmName);
	}

	private void fillDrumRhythmCombobox() {
		fillingRhythmComboBox = true;
		rhythmCombo.removeAllItems();
		Collection<String> drumRhythmNames = device.getDrumRhythmNames();
		if (!drumRhythmNames.isEmpty()) {
			for (String rhythmName : drumRhythmNames) {
				rhythmCombo.addItem(rhythmName);
			}
			setDeviceRhythmToRhythmCombobox();
			assertFilledDrumRhythmCombobox = false;
		} else {
			assertFilledDrumRhythmCombobox = true;
		}
		fillingRhythmComboBox = false;
	}

	private void setDeviceRhythmToRhythmCombobox() {
		String drumRhythmName = device.getDrumRhythmName();
		rhythmCombo.setSelectedItem(drumRhythmName);
		rhythmCombo.repaint();
		// set the pattern as tool tip text
		rhythmCombo.setToolTipText(drumRhythmName != null ? drumRhythmName : "");
	}

	private void setDrumSchemaLabel() {
		String drumSchemaFileName = device.getDrumSchemaFileName();
		schemaLabel.setText(drumSchemaFileName != null ? drumSchemaFileName : "<no schema>");
	}
}
