/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.drum;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;

/**
 * 01 02 01 02 01 02 01 02
 * 
 * @author Florian Paul
 * 
 */
public class DrumRhythm {
	private String name;
	private final Collection<DrumSequence> sequenceDefinitions = new LinkedHashSet<DrumSequence>();
	private final Collection<DrumSequence> sequences = new ArrayList<DrumSequence>();
	private Integer beatsPerPulse;

	public DrumRhythm(String name) {
		this.name = name;
	}
	
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public Collection<DrumSequence> getSequences() {
		return sequences;
	}
	
	public Collection<DrumBeat> getAllBeats() {
		Collection<DrumBeat> ret = new ArrayList<DrumBeat>();
		for(DrumSequence sequence : sequences) {
			ret.addAll(sequence.getAllBeats());
		}
		return ret;
	}
	
	public List<DrumPattern> getAllPatterns() {
		List<DrumPattern> ret = new LinkedList<DrumPattern>();
		for(DrumSequence sequence : sequences) {
			ret.addAll(sequence.getPatterns());
		}
		return ret;
	}
	
	public Collection<DrumSequence> getSequenceDefinitions() {
		return sequenceDefinitions;
	}

	public DrumSequence getDrumSequenceByName(String name) {
		for (DrumSequence seq : sequenceDefinitions) {
			if (seq.getName().equals(name)) {
				return seq;
			}
		}
		return null;
	}

	@Override
	public String toString() {
		String linesep = System.getProperty("line.separator");
		StringBuilder sb = new StringBuilder(name);
		sb.append(linesep).append(linesep);
		for(DrumSequence sequenceDefinition : sequenceDefinitions) {
			sb.append(sequenceDefinition.toString()).append(linesep);
		}
		sb.append(linesep);
		for(DrumSequence sequence : sequences) {
			sb.append(sequence.getName());
			sb.append(' ');
		}
		return sb.toString();
	}

	public void setBeatsPerPulse(Integer beatsPerPulse) {
		this.beatsPerPulse = beatsPerPulse;
	}

	public Integer getBeatsPerPulse() {
		return beatsPerPulse;
	}
	
	void validate() throws DrumSchemaFormatException {
		if (name == null) {
			throw new DrumSchemaFormatException("A drum rhythm must have a name.");
		}
		if (sequenceDefinitions.isEmpty()) {
			throw new DrumSchemaFormatException("The drum rhythm '" + name + "' has no sequence definitions.");
		}
		for(DrumSequence seq : sequenceDefinitions) {
			try {
				seq.validate();
			} catch (DrumSchemaFormatException ex) {
				throw new DrumSchemaFormatException(ex.getMessage() + " at rhythm '" + name +"'");
			}
		}
		if (sequences.isEmpty()) {
			throw new DrumSchemaFormatException("The drum rhythm '" + name + "' has no sequences.");
		}
		for(DrumSequence seq : sequences) {
			try {
				seq.validate();
			} catch (DrumSchemaFormatException ex) {
				throw new DrumSchemaFormatException(ex.getMessage() + " at rhythm '" + name +"'");
			}
		}
		if(beatsPerPulse==null) {
			throw new DrumSchemaFormatException("The drum rhythm '" + name + "' has an no beats per pulse.");
		}
	}
}
