/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.drum;

import static java.lang.Math.*;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Stack;

import javax.sound.sampled.UnsupportedAudioFileException;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

public class HydrogenDrumSchemaReader extends DefaultHandler {
	private static final String[] charTags = {
		"bpm", "id", "gain", "drumkit", "filename", "name", "size", "position", "instrument", "note", "patternID"
	};
	private final Collection<File> clipSearchPaths;
	private Stack<String> path;
	private DrumSchema schema;
	private Float bpm;
	private StringBuilder chars;
	private String instrumentId;
	private String drumkit;
	private String filename;
	private Float gain;
	private DrumRhythm rhythm;
	private String patternName;
	private Integer patternSize;
	private Map<Integer, Collection<String>> notePositionToInstrumentIds;
	private Integer notePosition;
	private String noteInstrumentId;
	private Collection<String> sequencePatternIds;
	private int maxBeatCount;
	
	public HydrogenDrumSchemaReader(InputStream xmlInputStream, Collection<File> clipSearchPaths) throws SAXException, IOException, ParserConfigurationException {
		this.clipSearchPaths = clipSearchPaths;
		SAXParserFactory parserFactory = SAXParserFactory.newInstance();
		SAXParser parser = parserFactory.newSAXParser();
		parser.parse(xmlInputStream, this);
	}

	@Override
	public void startDocument() throws SAXException {
		schema = new DrumSchema();
		path = new Stack<String>();
		bpm = null;
	}

	@Override
	public void endDocument() throws SAXException {
		chars = null;
		path = null;
	}

	@Override
	public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
		path.push(qName);
		for (String tag : charTags) {
			if (tag.equalsIgnoreCase(qName)) {
				chars = new StringBuilder();
				break;
			}
		}
		if ("noteList".equalsIgnoreCase(qName)) {
			notePositionToInstrumentIds = new LinkedHashMap<Integer, Collection<String>>();
		}
	}

	@Override
	public void endElement(String uri, String localName, String qName) throws SAXException {
		path.pop();
		if ("bpm".equalsIgnoreCase(qName)) {
			bpm = Float.valueOf(chars.toString());
		} else if (instrumentId != null && drumkit != null && filename != null && "instrument".equalsIgnoreCase(qName) && path.peek().equals("instrumentList")) {
			String filesep = System.getProperty("file.separator");
			File file;
			try {
				file = DrumSchema.searchClipFile(drumkit + filesep + filename, clipSearchPaths.toArray(new File[0]));
			} catch (FileNotFoundException ex) {
				throw new SAXException(ex);
			}
			try {
				DrumClip clip;
				if (gain != null) {
					clip = new DrumClip(instrumentId, file, gain);
				} else {
					clip = new DrumClip(instrumentId, file);
				}
				schema.getClips().add(clip);
			} catch (UnsupportedAudioFileException e) {
				throw new SAXException(e.getMessage() + ": " + file.getName());
			} catch (IOException e) {
				throw new SAXException(e.getMessage() + ": " + file.getName());
			}
			instrumentId = null;
			drumkit = null;
			filename = null;
			gain = null;
		} else if ("id".equalsIgnoreCase(qName)) {
			instrumentId = toId(chars);
		} else if ("drumkit".equalsIgnoreCase(qName)) {
			drumkit = chars.toString();
		} else if ("filename".equalsIgnoreCase(qName)) {
			filename = chars.toString();
		} else if ("gain".equalsIgnoreCase(qName) && path.peek().equals("instrument")) {
			gain = Float.valueOf(chars.toString());
		} else if ("name".equalsIgnoreCase(qName) && path.peek().equals("song")) {
			rhythm = new DrumRhythm(chars.toString());
			schema.getRhythms().add(rhythm);
		} else if ("name".equalsIgnoreCase(qName) && path.peek().equals("pattern")) {
			patternName = chars.toString();
		} else if ("size".equalsIgnoreCase(qName) && path.peek().equals("pattern")) {
			patternSize = Integer.valueOf(chars.toString());
		} else if ("position".equalsIgnoreCase(qName) && path.peek().equals("note")) {
			notePosition = Integer.valueOf(chars.toString());
		} else if ("instrument".equalsIgnoreCase(qName) && path.peek().equals("note")) {
			noteInstrumentId = toId(chars);
		} else if ("note".equalsIgnoreCase(qName)) {
			Collection<String> instrumentIds = notePositionToInstrumentIds.get(notePosition);
			if (instrumentIds == null) {
				instrumentIds = new LinkedHashSet<String>();
				notePositionToInstrumentIds.put(notePosition, instrumentIds);
			}
			instrumentIds.add(noteInstrumentId);
		} else if ("pattern".equalsIgnoreCase(qName)) {
			if (!notePositionToInstrumentIds.isEmpty()) {
				DrumSequence seq = new DrumSequence(toId(patternName));
				DrumPattern pattern = new DrumPattern();
				seq.getPatterns().add(pattern);
				rhythm.getSequenceDefinitions().add(seq);
				Set<Integer> notePositions = notePositionToInstrumentIds.keySet();
				int lowestDivisor = patternSize;
				for (int notePosition : notePositions) {
					if (notePosition > 0) {
						int greatestCommonDivisor = greatestCommonDivisor(notePosition, patternSize);
						lowestDivisor = min(lowestDivisor, greatestCommonDivisor);
					}
				}
				int beatCount = patternSize / lowestDivisor;
				for (int i = 0; i < beatCount; i++) {
					DrumBeat beat = new DrumBeat();
					pattern.getBeats().add(beat);
				}

				List<DrumBeat> beatList = new ArrayList<DrumBeat>(pattern.getBeats());
				for (Entry<Integer, Collection<String>> entry : notePositionToInstrumentIds.entrySet()) {
					int notePosition = entry.getKey();
					int index = notePosition == 0 ? 0 : notePosition / lowestDivisor;
					DrumBeat drumBeat = beatList.get(index);
					Collection<String> instrumentIds = entry.getValue();
					for (String instrumentId : instrumentIds) {
						drumBeat.getHits().add(new DrumHit(instrumentId));
					}
				}
			}

			notePositionToInstrumentIds = null;
			patternName = null;
			patternSize = null;
			notePosition = null;
			noteInstrumentId = null;
		} else if ("patternList".equalsIgnoreCase(qName)) {
			// extend the shorter pattern sizes to the largest pattern size
			for (DrumSequence sequenceDefinition : rhythm.getSequenceDefinitions()) {
				for (DrumPattern drumPattern : sequenceDefinition.getPatterns()) {
					maxBeatCount = max(drumPattern.getBeats().size(), maxBeatCount);
				}
			}

			float calcPulse = maxBeatCount;
			while (calcPulse > 4) {
				calcPulse /= 2;
			}
			if (((int) calcPulse) != calcPulse) {
				throw new SAXException("The number of beats must be a multiple of two and additionally a multiple of either three or four, but " + maxBeatCount + " beats were found.");
			}
			int pulse = (int) calcPulse;
			rhythm.setBeatsPerPulse(maxBeatCount / pulse);

			for (DrumSequence sequenceDefinition : rhythm.getSequenceDefinitions()) {
				for (DrumPattern drumPattern : sequenceDefinition.getPatterns()) {
					List<DrumBeat> beats = drumPattern.getBeats();
					int beatCount = beats.size();
					if (maxBeatCount > beatCount) {
						int missingBeatCount = (maxBeatCount / beatCount) - 1;
						for (int i = 0; i < beatCount; i++) {
							int offset = (1 + missingBeatCount) * i;
							for (int j = 1; j <= missingBeatCount; j++) {
								beats.add(offset + j, new DrumBeat());
							}
						}
					}
				}
			}
		} else if ("patternID".equalsIgnoreCase(qName)) {
			if (sequencePatternIds == null) {
				sequencePatternIds = new LinkedHashSet<String>();
			}
			sequencePatternIds.add(toId(chars));
		} else if ("group".equalsIgnoreCase(qName)) {
			if (sequencePatternIds == null) {
				DrumSequence emptySequence = rhythm.getDrumSequenceByName("empty");
				if (emptySequence == null) {
					emptySequence = new DrumSequence("empty");
					DrumPattern emptyPattern = new DrumPattern();
					List<DrumBeat> beats = emptyPattern.getBeats();
					for (int i = 0; i < maxBeatCount; i++) {
						beats.add(new DrumBeat());
					}
					emptySequence.getPatterns().add(emptyPattern);
					rhythm.getSequenceDefinitions().add(emptySequence);
				}
				rhythm.getSequences().add(emptySequence);
			} else {
				addSequence();
			}
		}

		for (String tag : charTags) {
			if (tag.equalsIgnoreCase(qName)) {
				chars = null;
				break;
			}
		}
	}

	private void addSequence() {
		int len = sequencePatternIds.size();
		if (len > 1) {
			addMergedSequence();
		} else if (len == 1) {
			String sequenceName = sequencePatternIds.iterator().next();
			DrumSequence definedSequence = rhythm.getDrumSequenceByName(sequenceName);
			// only if the defined sequence has patterns (otherwise it will not
			// be registered)
			if (definedSequence != null) {
				rhythm.getSequences().add(definedSequence);
			}
		}
		sequencePatternIds = null;
	}

	private void addMergedSequence() {
		StringBuilder mergedSequenceId = new StringBuilder();
		for (String sequencePatternId : sequencePatternIds) {
			mergedSequenceId.append(sequencePatternId);
		}
		String sequenceName = mergedSequenceId.toString();
		DrumSequence mergedSequence = rhythm.getDrumSequenceByName(sequenceName);
		if (mergedSequence == null) {
			mergedSequence = new DrumSequence(sequenceName);
			rhythm.getSequenceDefinitions().add(mergedSequence);
			DrumPattern mergedPattern = null;
			for (String sequencePatternId : sequencePatternIds) {
				DrumSequence definedSequence = rhythm.getDrumSequenceByName(sequencePatternId);
				// only if the defined sequence has patterns (otherwise it will
				// not be registered)
				if (definedSequence != null) {
					assert definedSequence.getPatterns().size() == 1;
					DrumPattern definedPattern = definedSequence.getPatterns().iterator().next();
					if (mergedPattern == null) {
						mergedPattern = new DrumPattern();
						mergedSequence.getPatterns().add(mergedPattern);
						Collection<DrumBeat> mergedBeats = mergedPattern.getBeats();
						for (int i = 0, n = definedPattern.getBeats().size(); i < n; i++) {
							mergedBeats.add(new DrumBeat());
						}
					}
					mergePatterns(mergedPattern, definedPattern);
				}
			}
		}
		rhythm.getSequences().add(mergedSequence);
	}

	private static void mergePatterns(DrumPattern mergedPattern, DrumPattern definedPattern) {
		Iterator<DrumBeat> mergedBeatIter = mergedPattern.getBeats().iterator();
		Iterator<DrumBeat> definedBeatIter = definedPattern.getBeats().iterator();
		while (mergedBeatIter.hasNext()) {
			assert definedBeatIter.hasNext();
			Collection<DrumHit> definedHits = definedBeatIter.next().getHits();
			Collection<DrumHit> mergedHits = mergedBeatIter.next().getHits();
			for (DrumHit definedHit : definedHits) {
				if (!contains(mergedHits, definedHit)) {
					mergedHits.add(definedHit);
				}
			}

		}
	}

	private static String toId(CharSequence string) {
		return ("#" + string).trim().replace(' ', '_');
	}

	private static boolean contains(Collection<DrumHit> hits, DrumHit hit) {
		for (DrumHit drumHit : hits) {
			if (drumHit.getId().equals(hit.getId())) {
				return true;
			}
		}
		return false;
	}

	@Override
	public void characters(char[] ch, int start, int length) throws SAXException {
		if (chars != null) {
			chars.append(ch, start, length);
		}
	}

	static int greatestCommonDivisor(int a, int b) {
		if (a <= 0 || b <= 0) {
			throw new IllegalArgumentException("The values must be greater zero: " + a + " " + b);
		}
		if (a == b) {
			return a;
		}
		int numerator;
		int denominator;
		if (a > b) {
			numerator = a;
			denominator = b;
		} else {
			numerator = b;
			denominator = a;
		}

		int remainder;
		while ((remainder = numerator % denominator) != 0) {
			numerator = denominator;
			denominator = remainder;
		}

		return denominator;
	}

	public Float getBpm() {
		return bpm;
	}

	public DrumSchema getSchema() {
		return schema;
	}
}
