/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.dynamics;

import java.io.IOException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;


public class CompressorAudioDataReader implements AudioDataReaderChainLink {

	// constant unparameterizable variables
	private final EnvelopeAudioDataReader reader;
	private final float sampleMaximum; // full scale
	private final float sampleRate;

	// variables that are parameterizable by the user
	private float threshold;
	private float ratio = 1.0F;
	private float makeUpGain;
	private float attack;
	private float release;
	private float hold;
	private boolean linked = true;
	private boolean leftDisabled;
	private boolean rightDisabled;

	// variables used for compression operation
	private int absoluteThresholdSample;
	private float makeUpGainFactor = 1F;
	private int holdSamples;
	private int leftPassedHoldSamples;
	private int rightPassedHoldSamples;
	private int linkedPassedHoldSamples;
	private float compressionFactor = 1F;
	private float leftOpCompressionFactor = 1F;
	private float rightOpCompressionFactor = 1F;
	private float linkedOpCompressionFactor = 1F;
	private float opCompressionFactorDecrement;
	private float opCompressionFactorIncrement;

	// just for information
	private int attackSamples;
	private int releaseSamples;
	private float leftGainReduction;
	private float rightGainReduction;

	public CompressorAudioDataReader(EnvelopeAudioDataReader reader, int sampleSize, float sampleRate) {
		this.reader = reader;
		this.sampleMaximum = (float) Math.pow(2D, sampleSize - 1);
		this.sampleRate = sampleRate;
	}

	private void init() {
		float thresholdExponent = threshold / 20F;
		float thresholdFactor = (float) Math.pow(10D, thresholdExponent);
		absoluteThresholdSample = (int) (thresholdFactor * sampleMaximum);

		// convert compression level from dB to a factor
		float compressionLevel = threshold + threshold / -ratio;
		float compressionExponent = compressionLevel / 20F;
		compressionFactor = (float) Math.pow(10D, compressionExponent);

		float attackSamplesAsFloat = attack * sampleRate / 1000F;
		if (attackSamplesAsFloat == 0) {
			opCompressionFactorDecrement = compressionFactor;
		} else {
			opCompressionFactorDecrement = compressionFactor / attackSamplesAsFloat;
		}
		attackSamples = (int) attackSamplesAsFloat;

		float releaseSamplesAsFloat = release * sampleRate / 1000F;
		if (releaseSamplesAsFloat == 0) {
			opCompressionFactorIncrement = compressionFactor;
		} else {
			opCompressionFactorIncrement = compressionFactor / releaseSamplesAsFloat;
		}
		releaseSamples = (int) releaseSamplesAsFloat;
	}

	@Override
	public int read(int[] data) throws IOException {
		int len = reader.read(data);
		int[] envelope = reader.getEnvelope();

		for (int l = 0, r = 1, n = data.length; l < n; l += 2, r += 2) {
			if (linked) {
				final int leftEnvelopeValue = envelope[l];
				final int rightEnvelopeValue = envelope[r];
				final boolean thresholdPassed = leftEnvelopeValue > absoluteThresholdSample || rightEnvelopeValue > absoluteThresholdSample;

				if (thresholdPassed) {
					if (linkedOpCompressionFactor > compressionFactor) {
						linkedOpCompressionFactor -= opCompressionFactorDecrement;
					}
					linkedPassedHoldSamples = 0;
				} else {
					if (linkedPassedHoldSamples < holdSamples) {
						linkedPassedHoldSamples++;
					} else if (linkedOpCompressionFactor < 1) {
						linkedOpCompressionFactor += opCompressionFactorIncrement;
						if (linkedOpCompressionFactor > 1) {
							linkedOpCompressionFactor = 1;
						}
					}
				}
			}

			if (!leftDisabled) {
				// left

				if (linked) {
					leftOpCompressionFactor = linkedOpCompressionFactor;
				} else {
					// decrement or increment calculate compression factor
					final int envelopeValue = envelope[l];
					final boolean thresholdPassed = envelopeValue > absoluteThresholdSample;

					if (thresholdPassed) {
						if (leftOpCompressionFactor > compressionFactor) {
							leftOpCompressionFactor -= opCompressionFactorDecrement;
						}
						leftPassedHoldSamples = 0;
					} else {
						if (leftPassedHoldSamples < holdSamples) {
							leftPassedHoldSamples++;
						} else if (leftOpCompressionFactor < 1) {
							leftOpCompressionFactor += opCompressionFactorIncrement;
							if (leftOpCompressionFactor > 1) {
								leftOpCompressionFactor = 1;
							}
						}
					}
				}

				// compression
				final int sample = data[l];
				final float absoluteSample = Math.abs(sample);
				final float absoluteResult = absoluteSample * leftOpCompressionFactor * makeUpGainFactor;
				final float result = sample < 0 ? -absoluteResult : absoluteResult;
				data[l] = (int) result;
				leftGainReduction = absoluteSample - absoluteResult;
			}

			if (!rightDisabled) {
				// right

				if (linked) {
					rightOpCompressionFactor = linkedOpCompressionFactor;
				} else {
					// decrement or increment calculate compression factor
					final int envelopeValue = envelope[r];
					final boolean thresholdPassed = envelopeValue > absoluteThresholdSample;

					if (thresholdPassed) {
						if (rightOpCompressionFactor > compressionFactor) {
							rightOpCompressionFactor -= opCompressionFactorDecrement;
						}
						rightPassedHoldSamples = 0;
					} else {
						if (rightPassedHoldSamples < holdSamples) {
							rightPassedHoldSamples++;
						} else if (rightOpCompressionFactor < 1) {
							rightOpCompressionFactor += opCompressionFactorIncrement;
							if (rightOpCompressionFactor > 1) {
								rightOpCompressionFactor = 1;
							}
						}
					}
				}

				// compression
				final int sample = data[r];
				final float absoluteSample = Math.abs(sample);
				final float absoluteResult = absoluteSample * rightOpCompressionFactor * makeUpGainFactor;
				final float result = sample < 0 ? -absoluteResult : absoluteResult;
				data[r] = (int) result;
				rightGainReduction = absoluteSample - absoluteResult;
			}
		}

		return len;
	}

	@Override
	public void close() throws IOException {
		reader.close();
	}

	@Override
	public void stop() {
		if (reader != null) {
			reader.stop();
		}
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader.setSource(source);		
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	public float getThreshold() {
		return threshold;
	}

	public void setThreshold(float threshold) {
		if (threshold > 0) {
			threshold = 0;
		}
		this.threshold = threshold;
		init();
	}

	public float getRatio() {
		return ratio;
	}

	public void setRatio(float ratio) {
		if (ratio < 1) {
			ratio = 1;
		}
		this.ratio = ratio;
		init();
	}

	public float getMakeUpGain() {
		return makeUpGain;
	}

	public void setMakeUpGain(float makeUpGain) {
		if (makeUpGain < 0) {
			makeUpGain = 0;
		}
		this.makeUpGain = makeUpGain;
		float exp = makeUpGain / 20F;
		makeUpGainFactor = (float) Math.pow(10D, exp);
	}

	public float getAttack() {
		return attack;
	}

	public void setAttack(float attack) {
		if (attack < 0) {
			attack = 0;
		}
		this.attack = attack;
		init();
	}

	public float getRelease() {
		return release;
	}

	public void setRelease(float release) {
		if (release < 0) {
			release = 0;
		}
		this.release = release;
		init();
	}

	public float getHold() {
		return hold;
	}

	public void setHold(float hold) {
		if (hold < 0) {
			hold = 0;
		}
		this.hold = hold;
		float samples = hold * sampleRate / 1000F;
		holdSamples = (int) samples;
	}

	public float getLeftGainReduction() {
		return leftGainReduction;
	}

	public float getRightGainReduction() {
		return rightGainReduction;
	}

	public float getSampleMaximum() {
		return sampleMaximum;
	}

	public float getSampleRate() {
		return sampleRate;
	}

	public int getAbsThresholdSample() {
		return absoluteThresholdSample;
	}

	public float getMakeUpGainFactor() {
		return makeUpGainFactor;
	}

	public int getAttackSamples() {
		return attackSamples;
	}

	public int getReleaseSamples() {
		return releaseSamples;
	}

	public boolean isLeftDisabled() {
		return leftDisabled;
	}

	public void setLeftDisabled(boolean disableLeft) {
		this.leftDisabled = disableLeft;
	}

	public boolean isRightDisabled() {
		return rightDisabled;
	}

	public void setRightDisabled(boolean disableRight) {
		this.rightDisabled = disableRight;
	}

	public boolean isLinked() {
		return linked;
	}

	public void setLinked(boolean linked) {
		this.linked = linked;
	}

	public float getCompressionFactor() {
		return compressionFactor;
	}

	public EnvelopeAudioDataReader getReader() {
		return reader;
	}
}
