/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.frequency;

import java.io.IOException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;


public class ConvolutionAudioDataReader implements AudioDataReaderChainLink {

	private AudioDataReader reader;
	private double[] impulseResponse;
	private int impulseResponseLength;
	private int[] leftDelayLine;
	private int[] rightDelayLine;
	private int currentDelayLineIndex;

	public ConvolutionAudioDataReader(AudioDataReader reader) {
		this.reader = reader;
	}

	public ConvolutionAudioDataReader(AudioDataReader reader, double[] impulseResponse) {
		this.reader = reader;
		setImpulseResponse(impulseResponse);
	}

	@Override
	public int read(int[] data) throws IOException {
		int len = reader.read(data);
		if (len == -1) {
			return -1;
		}
		for (int l = 0, r = 1; l < len; l += 2, r += 2) {
			if (currentDelayLineIndex >= impulseResponseLength) {
				currentDelayLineIndex = 0;
			}
			leftDelayLine[currentDelayLineIndex] = data[l];
			rightDelayLine[currentDelayLineIndex] = data[r];

			double leftSum = 0D;
			double rightSum = 0D;
			for (int i = currentDelayLineIndex, j = impulseResponseLength - 1; j >= 0; i++, j--) {
				if (i >= impulseResponseLength) {
					i = 0;
				}
				double factor = impulseResponse[j];
				leftSum += factor * leftDelayLine[i];
				rightSum += factor * rightDelayLine[i];
			}
			currentDelayLineIndex++;
			data[l] = (int) leftSum;
			data[r] = (int) rightSum;
		}

		return len;
	}

	@Override
	public void close() throws IOException {
		reader.close();
	}

	@Override
	public void stop() {
		reader.stop();
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	public void setImpulseResponse(double[] impulseResponse) {
		this.impulseResponse = impulseResponse;
		impulseResponseLength = impulseResponse.length;
		leftDelayLine = new int[impulseResponseLength];
		rightDelayLine = new int[impulseResponseLength];
		currentDelayLineIndex = 0;
	}

	public double[] getImpulseResponse() {
		return impulseResponse;
	}
}
