/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.frequency;

import static java.lang.Math.*;
import static java.lang.System.arraycopy;
import static java.util.Arrays.*;

public class DigitalSignalProcessing {

	private static final double log2 = log(2);
	private static final double twoPi = 2 * PI;
	private static final double fourPi = 4 * PI;

	private DigitalSignalProcessing() {

	}

	/**
	 * Multiplies every sample value with a previously calculated factor, so
	 * that the sum of all samples is equal to 1. This is useful for adjusting
	 * impulse responses for use with convolution.
	 * 
	 * @param samples
	 * @return the identical array instance with modified values
	 */
	public static double[] normalizeToSumValueOne(double[] samples) {
		/*
		 * For the filter to have unity gain at DC, the constant K must be
		 * chosen such that the sum of all the samples is equal to one.
		 */
		int n = samples.length;
		double sampleValueSum = 0;
		for (int i = 0; i < n; i++) {
			sampleValueSum += samples[i];
		}
		if (sampleValueSum > 1) {
			double k = 1 / sampleValueSum;
			for (int i = 0; i < n; i++) {
				samples[i] *= k;
			}
		}
		return samples;
	}

	/**
	 * Multiplies every samples value with -1. Additionally adds the value 1 to
	 * the mid value of the array. Therefore the given array must have an odd
	 * length.
	 * 
	 * @param samples
	 *            array with an odd length
	 * @return the identical array instance with modified values
	 */
	public static double[] spectralInversion(double[] samples) {
		int len = samples.length;
		if (len % 2 == 0) {
			throw new IllegalArgumentException("The parameter array must have an odd length.");
		}
		int half = len / 2;
		for (int i = 0; i < len; i++) {
			samples[i] *= -1;
			if (i == half) {
				samples[i] += 1;
			}
		}
		return samples;
	}

	/**
	 * Multiplies all the given samples with the blackman window.
	 * 
	 * @param samples the samples to multiply with the samples of a blackman window.
	 * @return the same array instance as the given parameter
	 */
	public static double[] blackmanWindow(double[] samples) {
		int len = samples.length;
		int m;
		if (len % 2 == 0) {
			m = len;
		} else {
			m = len - 1;
		}

		for (int i = 0; i < len; i++) {
			samples[i] *= 0.42 - 0.5 * cos(twoPi * i / m) + 0.08 * cos(fourPi * i / m);
		}

		return samples;
	}

	/**
	 * Multiplies all the given samples with the hamming window.
	 * 
	 * @param samples the samples to multiply with the samples of a hamming window.
	 * @return the same array instance as the given parameter
	 */
	public static double[] hammingWindow(double[] samples) {
		int len = samples.length;
		int m;
		if (len % 2 == 0) {
			m = len;
		} else {
			m = len - 1;
		}

		for (int i = 0; i < len; i++) {
			samples[i] *= 0.54 - 0.46 * cos(twoPi * i / m);
		}

		return samples;
	}

	/**
	 * Performs a standard convolution by multiplying each sample of the first
	 * array with each sample of the second array.
	 * 
	 * @param x
	 *            any sampled signal
	 * @param h
	 *            the sampled frequency response (or any signal)
	 * @return a new array of the length x.length + h.length - 1.
	 */
	public static double[] convolve(double[] x, double[] h) {
		int xn = x.length;
		int hn = h.length;
		double[] ret = new double[xn + hn - 1];
		for (int i = 0; i < xn; i++) {
			for (int j = 0; j < hn; j++) {
				ret[i + j] += x[i] * h[j];
			}
		}
		return ret;
	}

	public static double[] correlate(double[] x, double[] y) {
		int xn = x.length;
		int yn = y.length;
		int yn1 = yn - 1;
		double[] ret = new double[xn + yn - 1];
		for (int i = 0; i < xn; i++) {
			for (int j = 0; j < yn; j++) {
				ret[i + j] += x[i] * y[yn1 - j];
			}
		}
		return ret;
	}

	public static double[] add(double[]... arrays) {
		int maxLen = 0;
		for (double[] arr : arrays) {
			maxLen = max(maxLen, arr.length);
		}

		double[] ret = new double[maxLen];
		for (double[] arr : arrays) {
			int arrLen = arr.length;
			for (int i = 0; i < maxLen; i++) {
				if (i >= arrLen) {
					break;
				}
				ret[i] += arr[i];
			}
		}
		return ret;
	}
	
	public static double[] multiply(double[]... arrays) {
		int maxLen = 0;
		for (double[] arr : arrays) {
			maxLen = max(maxLen, arr.length);
		}

		double[] ret = new double[maxLen];
		fill(ret, 1);
		for (double[] arr : arrays) {
			int arrLen = arr.length;
			for (int i = 0; i < maxLen; i++) {
				if (i >= arrLen) {
					break;
				}
				ret[i] *= arr[i];
			}
		}
		return ret;
	}
	
	public static double[][] multiply(double[][] x, double[][] h) {
		double[] reX = x[0];
		double[] imX = x[1];
		double[] reH = h[0];
		double[] imH = h[1];
		for (int i = 0, n = reX.length; i < n; i++) {
			double rexi = reX[i];
			double imxi = imX[i];
			double rehi = reH[i];
			double imhi = imH[i];
			reX[i] = rexi * rehi - imxi * imhi;
			imX[i] = imxi * rehi + rexi * imhi;
		}
		return x;
	}
	
	public static double[] integrate(double[] x) {
		int len = x.length;
		double sum = 0;
		for (int i = 0; i < len; i++) {
			sum += x[i];
			x[i] = sum;
		}
		return x;
	}
	
	public static double[][] discreteFourierTransform(double[] x) {
		final int timeLen = x.length;
		final int freqLen = timeLen / 2 + 1;
		double[] re = new double[freqLen];
		double[] im = new double[freqLen];
		final double g = twoPi / timeLen;
		for (int k = 0; k < freqLen; k++) {
			final double h = g * k;
			for (int i = 0; i < timeLen; i++) {
				final double xi = x[i];
				final double w = h * i; // same as twoPi * k * i / timeLen
				re[k] += xi * cos(w);
				im[k] -= xi * sin(w);
			}
		}
		return new double[][]{re, im};
	}
	
	/**
	 * Normalizes the frequency amplitudes and performs the inverse discrete
	 * fourier transform.
	 * 
	 * @param rectangularSpectrum
	 *            the spectrum created by the discrete fourier transform (real
	 *            spectrum in rectangular notation)
	 * @return a new array of length 2 * (spectral array length - 1)
	 */
	public static double[] inverseDiscreteFourierTransform(double[][] rectangularSpectrum) {
		rectangularSpectrum = normalizeSpectralDensity(rectangularSpectrum);
		double[] re = rectangularSpectrum[0];
		double[] im = rectangularSpectrum[1];
		int freqLen = re.length;
		int nhalf = freqLen - 1;
		
		final int n = nhalf * 2;
		final double g = twoPi / n;
		double[] x = new double[n];
		for (int k = 0; k < freqLen; k++) {
			final double h = g * k;
			for (int i = 0; i < n; i++) {
				final double w = h * i; // same as twoPi * k * i / n
				x[i] += re[k] * cos(w);
				x[i] += im[k] * sin(w);
			}
		}
		return x;
	}
	
	public static double[][] normalizeSpectralDensity(double[][] rectangularSpectrum) {
		double[] re = rectangularSpectrum[0];
		double[] im = rectangularSpectrum[1];
		int freqLen = re.length;
		int nhalf = freqLen - 1;
		for (int k = 0; k < freqLen; k++) {
			re[k] /= nhalf;
			im[k] /= -nhalf;
		}
		re[0] /= 2;
		re[nhalf] /= 2;
		return rectangularSpectrum;
	}
	
	public static double[] hypot(double[][] freq) {
		double[] re = freq[0];
		double[] im = freq[1];
		int n = re.length;
		double[] ret = new double[n];
		for (int i = 0; i < n; i++) {
			ret[i] = Math.hypot(re[i], im[i]);
		}
		return ret;
	}
	
	public static double[] arctan(double[][] freq) {
		double[] re = freq[0];
		double[] im = freq[1];
		int n = re.length;
		double[] ret = new double[n];
		for (int i = 0; i < n; i++) {
			ret[i] = atan2(im[i], re[i]);
		}
		return ret;
	}
	
	/**
	 * Converts the spectrum from the rectangular notation (with cosine
	 * amplitudes in the first array and sine amplitudes in the second array) to
	 * polar notation (with magnitude in the first array and phase in the second
	 * array)
	 * 
	 * @param rectangularNotatedSpectrum
	 *            the rectangular notated spectrum
	 * @return the same array instances that are given as parameter, but
	 *         different values. The first array contains the magnitude, the
	 *         second array contains the phases.
	 */
	public static double[][] polarNotation(double[][] rectangularNotatedSpectrum) {
		double[] re = rectangularNotatedSpectrum[0];
		double[] im = rectangularNotatedSpectrum[1];
		for (int i = 0, n = re.length; i < n; i++) {
			double cos = re[i];
			double sin = im[i];
			re[i] = Math.hypot(cos, sin);
			im[i] = atan2(sin, cos);
		}
		return rectangularNotatedSpectrum;
	}
	
	/**
	 * Converts the spectrum from polar notation (with magnitude and phase) to
	 * the rectangular notation (with cosine amplitudes in the first array and
	 * sine amplitudes in the second array)
	 * 
	 * @param polarNotatedSpectrum
	 *            the polar notated spectrum
	 * @return the same array instances that are given as parameter, but
	 *         different values. The first array contains the real part (cosine
	 *         amplitudes), the second array contains the imaginary part (sine
	 *         amplitudes).
	 */
	public static double[][] rectangularNotation(double[][] polarNotatedSpectrum) {
		double[] mag = polarNotatedSpectrum[0];
		double[] phase = polarNotatedSpectrum[1];
		for (int i = 0, n = mag.length; i < n; i++) {
			double m = mag[i];
			double p = phase[i];
			mag[i] = m * cos(p);
			phase[i] = m * sin(p);
		}
		return polarNotatedSpectrum;
	}
	
	/**
	 * This method shifts all the values contained in the given array by the
	 * amount of position indexes revolving at the array bounds.
	 * 
	 * @param x
	 *            the array which values will be moved about the given amount of
	 *            position indexes to the right.
	 * @param amount
	 *            the number of indexes all values of the given array will be
	 *            moved to the right
	 * @return the same array instance with shifted values inside
	 */
	public static double[] shift(double[] x, int amount) {
		int len = x.length;
		amount = abs(amount);
		if (amount > len) {
			throw new IllegalArgumentException("The shift amount (" + amount + ") must not be greater than the array length (" + len + ").");
		}
		double[] tmp = new double[amount];
		arraycopy(x, len - amount, tmp, 0, amount);
		arraycopy(x, 0, x, amount, len - amount);
		arraycopy(tmp, 0, x, 0, amount);
		return x;
	}
	
	/**
	 * Converts the complex spectrum, generated by the fast fourier transform,
	 * into the real spectrum. This operation creates two newly arrays with half
	 * the length of the origin arrays plus one. These are returned within the
	 * same parameter array matrix as return value. The values within these
	 * arrays are the same values at the same positions of the parameter arrays
	 * but truncated in their length by param[0].length / 2 + 1.
	 * 
	 * @param complexSpectrum
	 *            the spectrum that is generated by the fast fourier transform
	 *            in the rectangular notation.
	 * @return the same outer array with the two newly created arrays of shorter
	 *         length and without the redundant spectrum information.
	 */
	public static double[][] realSpectrum(double[][] complexSpectrum) {
		double[] re = complexSpectrum[0];
		double[] im = complexSpectrum[1];
		int n = re.length / 2 + 1;
		double[] re2 = new double[n];
		double[] im2 = new double[n];
		arraycopy(re, 0, re2, 0, n);
		arraycopy(im, 0, im2, 0, n);
		complexSpectrum[0] = re2;
		complexSpectrum[1] = im2;
		return complexSpectrum;
	}
	
	/**
	 * Converts the real spectrum into the complex spectrum used for the inverse
	 * fast fourier transform.
	 * 
	 * This operation creates two newly arrays for the real and imaginary parts of
	 * the given parameter spectrum. These new arrays have twice the length of
	 * the real spec minus one. The real and imaginary part is reverse copied
	 * into the second half of the spec, while the imaginary part becomes
	 * changed in the algebraic sign.
	 * 
	 * @param realSpectrum
	 *            the real spectrum conaining the cosinus and sinus amplitudes
	 *            (rectangular notation).
	 * @return the same outer array instance with two new longer inner arrays of
	 *         length (param[0].length-1) * 2
	 */
	public static double[][] complexSpectrum(double[][] realSpectrum) {
		double[] re = realSpectrum[0];
		double[] im = realSpectrum[1];
		int len = re.length;
		int n = (len - 1) << 1;
		double[] re2 = new double[n];
		double[] im2 = new double[n];
		arraycopy(re, 0, re2, 0, len);
		arraycopy(im, 0, im2, 0, len);
		for(int i=len,j=len-2;i<n;i++,j--) {
			re2[i] = re[j];
			im2[i] = -im[j];
		}
		realSpectrum[0] = re2;
		realSpectrum[1] = im2;
		return realSpectrum;
	}
	
	public static double[][] fastFourierTransform(double[] x) {
		int n = x.length;
		// 'guess' the exponent if the assumed windows size is a power of two
		int exp = (int) round(log(n) / log2);
		// check if window size is a power of two
		if (2 << (exp - 1) != n) {
			throw new IllegalArgumentException("The array length (" + n + ") for the fast fourier transformation must be a power of 2.");
		}
		double[] i = new double[n];
		double[][] freq = new double[][]{x, i};
		fastFourierTransform(freq, true, n, exp);
		return freq;
	}
	
	public static double[] inverseFastFourierTransform(double[][] freq) {
		double[] re = freq[0];
		int n = re.length;
		// 'guess' the exponent if the assumed windows size is a power of two
		int exp = (int) round(log(n) / log2);
		// check if window size is a power of two
		if (2 << (exp - 1) != n) {
			throw new IllegalArgumentException("The array length (" + n + ") for the fast fourier transformation must be a power of 2.");
		}
		// normalize amplitudes
		double[] im = freq[1];
		for (int i = 0; i < n; i++) {
			re[i] /= n;
			im[i] /= n;
		}
		fastFourierTransform(freq, false, n, exp);
		return re;
	}
	
	public static double[][] fastFourierTransform(double[][] freq, boolean forward, int windowSize, int windowExponent) {
		double[] re = freq[0];
		double[] im = freq[1];
		
		int halfWindow = windowSize >> 1;
		
		// bit reversal
		for (int i = 0, j = 0, n = windowSize - 1; i < n; i++) {
			if (i < j) {
				double reTemp = re[i];
				double imTemp = im[i];
				re[i] = re[j];
				im[i] = im[j];
				re[j] = reTemp;
				im[j] = imTemp;
			}
			int k = halfWindow;
			while (k <= j) {
				j -= k;
				k >>= 1;
			}
			j += k;
		}

		// in place fft
		double a = -1;
		double b = 0;
		for (int l = 0, step = 1; l < windowExponent; l++) {
			int level = step;
			step <<= 1;
			double u = 1;
			double v = 0;
			for (int j = 0; j < level; j++) {
				for (int i = j; i < windowSize; i += step) {
					int idx = i + level;
					double t1 = u * re[idx] - v * im[idx];
					double t2 = u * im[idx] + v * re[idx];
					re[idx] = re[i] - t1;
					im[idx] = im[i] - t2;
					re[i] += t1;
					im[i] += t2;
				}
				double complex = u * a - v * b;
				v = u * b + v * a;
				u = complex;
			}
			b = sqrt((1D - a) / 2D);
			if (forward) {
				b = -b;
			}
			a = sqrt((1D + a) / 2D);
		}

		return freq;
	}
	
	public static double[] fastFourierTransformConvolution(double[] x, double[] y) {
		// find the length of power 2 and pad inputs with zeros to avoid circular convolution
		int len = nextPowerOfTwo(x.length + y.length - 1);
		double[] zeroPaddedX = copyOf(x, len);
		double[] zeroPaddedY = copyOf(y, len);
		double[][] freqX = fastFourierTransform(zeroPaddedX);
		double[][] freqY = fastFourierTransform(zeroPaddedY);

		// conversion into real spectrum (and following reconversion to complex spectrum) is optional, it could speed up the multiplication
		freqX = realSpectrum(freqX);
		freqY = realSpectrum(freqY);
		
		freqX = multiply(freqX, freqY);
		
		freqX = complexSpectrum(freqX);
		return inverseFastFourierTransform(freqX);
	}

	/**
	 * Returns the next greater value that is a power of two. For example if the
	 * input is 500 than the output will be 512 and if the input is 513 the
	 * output is 1024.
	 * 
	 * @param value any number for which the next greater value that is a power of two should be found.
	 * @return a number that is a power of two and is greater than the given parameter
	 */
	public static int nextPowerOfTwo(int value) {
		if (value == 0) {
			return 0;
		}
		int exp = (int) (log(value-1) / log2); // additional 'value - 1' because if 'value' is equal to a value that is a power of two we want to get the correct exp too
		value = 2 << exp;
		return value;
	}
	
	/**
	 * Create a gaussian bell signal.
	 * 
	 * @param x the target array for the signal, this array is returned
	 * @param amplitude the height of the bells peak
	 * @param shift the center of the bells peak
	 * @param variance the with of the bell
	 * @return the given parameter array
	 */
	public static double[] createGaussian(double[] x, double amplitude, int shift, double variance) {
		// (a*exp(-((x-b)^2)/(2*c^2)))
		int n = x.length;
		double q = 2 * variance * variance;
		for (int i = 0; i < n; i++) {
			int d = i - shift;
			x[i] = (amplitude * exp(-(d * d) / q));
		}
		return x;
	}
	
	/**
	 * Creates an (ideal) impulse response with a length depending on the given
	 * transision band width which is calculated by the given frame rate.
	 * 
	 * @param frameRate
	 * @param cutOffFrequency
	 * @param transitionBandWidth
	 * @return a non normalized ideal low pass impulse response. After this
	 *         operation, at least the normalization must take place. If there
	 *         should be a windowing process be done, the normalization should
	 *         be done after windowing.
	 */
	public static double[] createLowPassImpulseResponse(float frameRate, float cutOffFrequency, float transitionBandWidth) {
		float nyquist = frameRate / 2;
		/*
		 * The cutoff frequency, fC, is expressed as a fraction of the sampling
		 * rate, a value between 0 and 0.5.
		 */
		float transitionFactor = transitionBandWidth / nyquist;
		/*
		 * The length of the filter kernel is determined by M, which must be an
		 * even integer. The sample number i, is an integer that runs from 0 to
		 * M, resulting in M+1 total points in the filter kernel.
		 * 
		 * The length of the filter kernel, M, determines the transition
		 * bandwidth of the filter, BW.
		 */
		int kernelLength = (int) (4 / transitionFactor); // M
		if (kernelLength % 2 == 0) {
			kernelLength++;
		}
		double[] ret = new double[kernelLength];
		double fc = cutOffFrequency / frameRate;
		double twoPiFreq = twoPi * fc;
		int mHalf = kernelLength / 2;
		for (int i = 0; i < kernelLength; i++) {
			double x = i - mHalf;
			double h;
			if (x != 0) {
				h = sin(twoPiFreq * x) / x;
			} else {
				h = twoPiFreq;
			}
			ret[i] = h;
		}
		return ret;
	}

	public static double[] createLowPass(float frameRate, float cutOffFrequency, float transitionBandWidth) {
		float nyquist = frameRate / 2;
		if (cutOffFrequency <= 0) {
			cutOffFrequency = 1;
		} else if (cutOffFrequency > nyquist) {
			cutOffFrequency = nyquist;
		}
		return normalizeToSumValueOne(blackmanWindow(createLowPassImpulseResponse(frameRate, cutOffFrequency, transitionBandWidth)));
	}

	public static double[] createHiPass(float frameRate, float cutOffFrequency, float transitionBandWidth) {
		float nyquist = frameRate / 2;
		if (cutOffFrequency <= 0) {
			cutOffFrequency = 1;
		} else if (cutOffFrequency > nyquist) {
			cutOffFrequency = nyquist;
		}
		return spectralInversion(normalizeToSumValueOne(blackmanWindow(createLowPassImpulseResponse(frameRate, cutOffFrequency, transitionBandWidth))));
	}

	public static double[] createBandPass(float frameRate, float startCutOffFrequency, float startTransitionBandWidth, float endCutOffFrequency, float endTransitionBandWidth) {
		return convolve(createLowPass(frameRate, endCutOffFrequency, endTransitionBandWidth), createHiPass(frameRate, startCutOffFrequency, startTransitionBandWidth));
	}

	public static double[] createBandPass(float frameRate, float filterFrequency, float filterBandWidth, float transitionBandWidth) {
		float filterBandWidthHalf = filterBandWidth / 2;
		float startCutOffFrequency = filterFrequency - filterBandWidthHalf;
		float endCutOffFrequency = filterFrequency + filterBandWidthHalf;
		return createBandPass(frameRate, startCutOffFrequency, transitionBandWidth, endCutOffFrequency, transitionBandWidth);
	}

	public static double[] createBandStop(float frameRate, float startCutOffFrequency, float startTransitionBandWidth, float endCutOffFrequency, float endTransitionBandWidth) {
		return add(createLowPass(frameRate, startCutOffFrequency, startTransitionBandWidth), createHiPass(frameRate, endCutOffFrequency, endTransitionBandWidth));
	}

	public static double[] createBandStop(float frameRate, float filterFrequency, float filterBandWidth, float transitionBandWidth) {
		float filterBandWidthHalf = filterBandWidth / 2;
		float startCutOffFrequency = filterFrequency - filterBandWidthHalf;
		float endCutOffFrequency = filterFrequency + filterBandWidthHalf;
		return createBandStop(frameRate, startCutOffFrequency, transitionBandWidth, endCutOffFrequency, transitionBandWidth);
	}
}
