/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.frequency;

import static java.util.Arrays.*;
import static org.audiochain.devices.frequency.DigitalSignalProcessing.*;

import java.io.IOException;
import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.DelayedBlockAudioDataReader;
import org.audiochain.io.FrameObserver;


public class FftAudioDataReader extends DelayedBlockAudioDataReader implements AudioDataReaderChainLink {

	private Collection<FftAudioDataReaderListener> listeners;
	private double[] left;
	private double[] right;
	private boolean started;
	private boolean window;
	private double[] windowSamples;

	public FftAudioDataReader(AudioDataReader reader, float frameRate) {
		super(reader, frameRate);
		setFftLength(0);
	}

	public FftAudioDataReader(AudioDataReader reader, float frameRate, FrameObserver frameObserver) {
		super(reader, frameRate, frameObserver);
		setFftLength(0);
	}
	
	@Override
	public int read(int[] data) throws IOException {
		if (!started) {
			fireFftReaderStart();
			started = true;
		}
		return super.read(data);
	}

	@Override
	protected void processDelayedBlock(int[] stereo, int offset, int length) {
		double[] left = this.left;
		double[] right = this.right;
		fill(left, 0);
		fill(right, 0);
		boolean equalChannels = true;
		for (int l = offset, r = l + 1, n = offset+length,j=0; l < n; l += 2, r += 2, j++) {
			int leftSample = stereo[l];
			left[j] = leftSample;
			int rightSample = stereo[r];
			right[j] = rightSample;
			equalChannels = equalChannels && leftSample == rightSample;
		}
		double[][] leftSpectrum = fastFourierTransform(window ? multiply(left, windowSamples) : left);
		fireFftReaderUpdate(leftSpectrum, equalChannels ? leftSpectrum : fastFourierTransform(window ? multiply(right, windowSamples) : right));
	}

	public void addFftReaderListener(FftAudioDataReaderListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<FftAudioDataReaderListener>();
		}
		listeners.add(listener);
	}

	public void addFftReaderListeners(Collection<FftAudioDataReaderListener> listeners) {
		if (listeners == null) {
			return;
		}
		if (this.listeners == null) {
			this.listeners = new LinkedHashSet<FftAudioDataReaderListener>();
		}
		this.listeners.addAll(listeners);
	}

	public void removeFftReaderListener(FftAudioDataReaderListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	public void removeFftReaderListeners(Collection<FftAudioDataReaderListener> listeners) {
		if (this.listeners == null || listeners == null) {
			return;
		}
		this.listeners.removeAll(listeners);
		if (this.listeners.isEmpty()) {
			this.listeners = null;
		}
	}

	private void fireFftReaderUpdate(double[][] leftSpectrum, double[][] rightSpectrum) {
		if (listeners == null) {
			return;
		}
		for (FftAudioDataReaderListener listener : listeners) {
			listener.fftReaderUpdate(leftSpectrum, rightSpectrum);
		}
	}

	private void fireFftLengthChanged(int oldLength, int newLength) {
		if (listeners == null) {
			return;
		}
		for (FftAudioDataReaderListener listener : listeners) {
			listener.fftLengthChanged(oldLength, newLength);
		}
	}

	private void fireFftReaderStart() {
		if (listeners == null) {
			return;
		}
		for (FftAudioDataReaderListener listener : listeners) {
			listener.fftReaderStart();
		}
	}

	private void fireFftReaderStop() {
		if (listeners == null) {
			return;
		}
		for (FftAudioDataReaderListener listener : listeners) {
			listener.fftReaderStop();
		}
	}

	private void fireFftReaderClose() {
		if (listeners == null) {
			return;
		}
		for (FftAudioDataReaderListener listener : new LinkedHashSet<FftAudioDataReaderListener>(listeners)) {
			listener.fftReaderClose();
		}
	}

	@Override
	public void close() throws IOException {
		super.close();
		fireFftReaderClose();
	}

	@Override
	public void stop() {
		super.stop();
		fireFftReaderStop();
	}

	@Override
	public void seek(long framePosition) throws IOException {
		super.seek(framePosition);
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	public int getFftLength() {
		return getBlockSizeInFrames();
	}

	public void setFftLength(int fftLength) {
		int oldLength = getFftLength();
		if (oldLength != fftLength) {
			setBlockSizeInFrames(fftLength);
			left = new double[fftLength];
			right = new double[fftLength];
			windowSamples = new double[fftLength];
			fill(windowSamples, 1);
			hammingWindow(windowSamples);
			fireFftLengthChanged(oldLength, fftLength);
		}
	}

	public boolean isWindow() {
		return window;
	}

	public void setWindow(boolean window) {
		this.window = window;
	}
}
