/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.frequency;

import static java.lang.Math.*;
import static java.util.Arrays.copyOf;
import static org.audiochain.devices.frequency.DigitalSignalProcessing.*;

import java.io.IOException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;


public class FftConvolutionAudioDataReader implements AudioDataReaderChainLink {

	private AudioDataReader reader;
	private double[] impulseResponse;
	private double[][] frequencyImpulseResponse;
	private int impulseResponseLength;
	private int frequencyImpulseResponseLength;
	private double[] leftOverlap;
	private double[] rightOverlap;
	private int overlapStartIndex;
	private int overlapLength;
	private boolean bypass;

	public FftConvolutionAudioDataReader(AudioDataReader reader) {
		this.reader = reader;
	}

	public FftConvolutionAudioDataReader(AudioDataReader reader, double[] impulseResponse) {
		this.reader = reader;
		setImpulseResponse(impulseResponse);
	}

	@Override
	public int read(int[] stereo) throws IOException {
		int readStereoLength = reader.read(stereo);
		if (bypass) {
			return readStereoLength;
		}
		if (readStereoLength == -1) {
			return readFromOverlap(stereo);
		}
		int readMonoLength = readStereoLength >> 1;
		int fftLength = nextPowerOfTwo(readMonoLength + impulseResponseLength - 1);
		double[][] frequencyImpulseResponse = this.frequencyImpulseResponse;
		if (fftLength != frequencyImpulseResponseLength || frequencyImpulseResponse == null) {
			frequencyImpulseResponse = fastFourierTransform(copyOf(impulseResponse, fftLength));
			this.frequencyImpulseResponse = frequencyImpulseResponse;
			frequencyImpulseResponseLength = fftLength;
		}
		
		// zero padded, single channeled and converted to double
		double[] left = new double[fftLength];
		double[] right = new double[fftLength];
		// if channels have equal values, use only left channel to improve the performance
		boolean equalChannels = true;
		for (int l = 0, r = 1, i = 0; l < readStereoLength; l += 2, r += 2, i++) {
			int leftSample = stereo[l];
			left[i] = leftSample;
			int rightSample = stereo[r];
			right[i] = rightSample;
			equalChannels = equalChannels && leftSample == rightSample;
		}
		
		// convolve
		left = inverseFastFourierTransform(multiply(fastFourierTransform(left), frequencyImpulseResponse));
		if (!equalChannels) {
			right = inverseFastFourierTransform(multiply(fastFourierTransform(right), frequencyImpulseResponse));
		}
		
		// add overlap to freshly convolved data, or vice versa if overlap is larger than the current fft operation (common when finishing)
		if (overlapLength - overlapStartIndex > fftLength) {
			for (int i = 0, o = overlapStartIndex; o < overlapLength && i < fftLength; i++, o++) {
				double leftSample = left[i];
				leftOverlap[o] += leftSample;
				rightOverlap[o] += equalChannels ? leftSample : right[i];
			}
		} else {
			for (int i = 0, o = overlapStartIndex; o < overlapLength && i < fftLength; i++, o++) {
				left[i] += leftOverlap[o];
				if(!equalChannels) {
					right[i] += rightOverlap[o];
				}
			}
			leftOverlap = left;
			rightOverlap = equalChannels ? left : right;
			overlapStartIndex = 0;
			overlapLength = fftLength;
		}

		return readFromOverlap(stereo);
	}

	private int readFromOverlap(int[] stereo) {
		int overlapRest = overlapLength - overlapStartIndex;
		if (overlapRest <= 0) {
			return -1;
		}
		int chunk = min(stereo.length / 2, overlapRest);
		for (int i = 0, o = overlapStartIndex;i < chunk; i++, o++) {
			stereo[2 * i] = (int) round(leftOverlap[o]);
			stereo[2 * i + 1] = (int) round(rightOverlap[o]);
		}
		overlapStartIndex += chunk;
		return chunk * 2;
	}
	
	@Override
	public void close() throws IOException {
		reader.close();
	}

	@Override
	public void stop() {
		reader.stop();
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	public void setImpulseResponse(double[] impulseResponse) {
		this.impulseResponse = impulseResponse;
		impulseResponseLength = impulseResponse.length;
		// instruct read method to create new frequency impulse response
		frequencyImpulseResponse = null;
	}

	public double[] getImpulseResponse() {
		return impulseResponse;
	}

	public boolean isBypass() {
		return bypass;
	}

	public void setBypass(boolean bypass) {
		this.bypass = bypass;
	}
}
