/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.frequency;

import static java.util.Arrays.*;
import static org.audiochain.devices.frequency.DigitalSignalProcessing.*;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.FrameObserver;
import org.audiochain.model.AbstractChainableAudioDevice;
import org.audiochain.model.UserInterfaceContext;
import org.audiochain.ui.gui.GenericUserInterfaceContext;

/**
 * 
 * http://www.rane.com/note101.html http://www.astralsound.com/parametric_eq.htm
 * http://www.sengpielaudio.com/calculator-bandwidth.htm
 * 
 * 
 * @author Florian Paul
 * 
 */
public class ParametricEqualizerAudioDevice extends AbstractChainableAudioDevice {
	private static final long serialVersionUID = 1L;
	private float gain;
	private float frequency;
	private float quality;
	private transient FftConvolutionAudioDataReader reader;
	private transient double[] filterKernel;
	private transient int kernelLength = 64;

	@Override
	public final AudioDataReaderChainLink createAudioDataReaderChainLink(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		if (filterKernel == null) {
			createFilterKernel();
		}
		reader = new FftConvolutionAudioDataReader(source, filterKernel);
		enableBypass();
		return reader;
	}

	private void createFilterKernel() {
		double[] freqResponse = new double[kernelLength];
		fill(freqResponse, 1);

		float nyquist = audioProject.getNyquistFrequency();

		double[] gaussian = new double[kernelLength];
		int bw = 100;
		int m = (int) (nyquist / bw + 1);
		int shift = (int) (m * frequency / nyquist);
		createGaussian(gaussian, gain, shift, quality);
		freqResponse = add(freqResponse, gaussian);

		double[][] rectangular = rectangularNotation(new double[][] {
				freqResponse, new double[kernelLength]
		});
		double[] inverseDiscreteFourierTransform = inverseDiscreteFourierTransform(rectangular);
		int m1 = kernelLength + 1;
		double[] shifted = shift(inverseDiscreteFourierTransform, m1 / 2);
		double[] truncated = copyOf(shifted, m1);
		double[] windowed = blackmanWindow(truncated);
		filterKernel = windowed;
	}

	private void enableBypass() {
		if (reader != null) {
			reader.setBypass(gain == 0 && frequency == 0 && quality == 0);
		}
	}

	@Override
	public String getName() {
		return "Parametric EQ";
	}

	@Override
	protected UserInterfaceContext<ParametricEqualizerAudioDevice> createUserInterfaceContext() {
		return new GenericUserInterfaceContext<ParametricEqualizerAudioDevice>();
	}

	public float getGain() {
		return gain;
	}

	public void setGain(float lowValue) {
		this.gain = lowValue;
		createFilterKernel();
		if (reader != null) {
			reader.setImpulseResponse(filterKernel);
			enableBypass();
		}
	}

	public float getFrequency() {
		return frequency;
	}

	public void setFrequency(float midValue) {
		this.frequency = midValue;
		createFilterKernel();
		if (reader != null) {
			reader.setImpulseResponse(filterKernel);
			enableBypass();
		}
	}

	public float getQuality() {
		return quality;
	}

	public void setQuality(float hiValue) {
		this.quality = hiValue;
		createFilterKernel();
		if (reader != null) {
			reader.setImpulseResponse(filterKernel);
			enableBypass();
		}
	}

}
