/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.frequency;

import static java.util.Arrays.*;
import static org.audiochain.devices.frequency.DigitalSignalProcessing.*;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.FrameObserver;
import org.audiochain.model.AbstractChainableAudioDevice;
import org.audiochain.model.AudioDeviceProperty;
import org.audiochain.model.UserInterfaceContext;
import org.audiochain.ui.gui.GenericUserInterfaceContext;

public class ThreeBandEqualizerAudioDevice extends AbstractChainableAudioDevice {
	private static final long serialVersionUID = 1L;
	@AudioDeviceProperty(name="Hi", min=-1, mid=0, max=3, symbol=9716)
	private float hiGain;
	@AudioDeviceProperty(name="Mid", min=-1, mid=0, max=3, symbol=9716)
	private float midGain;
	@AudioDeviceProperty(name="Low", min=-1, mid=0, max=3, symbol=9716)
	private float lowGain = 0.001F;
	private transient FftConvolutionAudioDataReader reader;
	private transient double[] filterKernel;
	private transient int kernelLength = 65;
	private transient float lowCutFrequency = 700;
	private transient float hiCutFrequency = 6800;
	
	@Override
	public final AudioDataReaderChainLink createAudioDataReaderChainLink(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		if (filterKernel == null) {
			createFilterKernel();
		}
		reader = new FftConvolutionAudioDataReader(source, filterKernel);
		enableBypass();
		return reader;
	}

	private void createFilterKernel() {
		double[] freqResponse = new double[kernelLength];
		fill(freqResponse, 1);

		float nyquist = audioProject.getNyquistFrequency();

		int lowCutSample = (int) (lowCutFrequency / nyquist * kernelLength) + 1;
		for (int i = 0; i < lowCutSample; i++) {
			freqResponse[i] = 1 + lowGain;
		}

		int hiCutSample = (int) (hiCutFrequency / nyquist * kernelLength);
		for (int i = lowCutSample; i < hiCutSample; i++) {
			freqResponse[i] = 1 + midGain;
		}

		for (int i = hiCutSample; i < kernelLength; i++) {
			freqResponse[i] = 1 + hiGain;
		}

		double[] kernel = inverseFastFourierTransform(complexSpectrum(rectangularNotation(new double[][] {
				freqResponse, new double[kernelLength]
		})));
		kernel = shift(kernel, (kernelLength + 1) / 2);
		kernel = copyOf(kernel, kernelLength); // truncate
		kernel = blackmanWindow(kernel);
		filterKernel = kernel;
	}

	private void enableBypass() {
		if (reader != null) {
			reader.setBypass(lowGain == 0 && midGain == 0 && hiGain == 0);
		}
	}

	@Override
	public String getName() {
		return "3 Band EQ";
	}

	@Override
	protected UserInterfaceContext<ThreeBandEqualizerAudioDevice> createUserInterfaceContext() {
		return new GenericUserInterfaceContext<ThreeBandEqualizerAudioDevice>();
	}

	public void setLowGain(float lowValue) {
		this.lowGain = lowValue;
		createFilterKernel();
		if (reader != null) {
			reader.setImpulseResponse(filterKernel);
			enableBypass();
		}
	}

	public void setMidGain(float midValue) {
		this.midGain = midValue;
		createFilterKernel();
		if (reader != null) {
			reader.setImpulseResponse(filterKernel);
			enableBypass();
		}
	}

	public void setHiGain(float hiValue) {
		this.hiGain = hiValue;
		createFilterKernel();
		if (reader != null) {
			reader.setImpulseResponse(filterKernel);
			enableBypass();
		}
	}
}
