/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.latency;

import java.io.IOException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.Delay;


public class LatencyAudioDataReader implements AudioDataReaderChainLink {

	private AudioDataReader reader;
	private int latencyInSamples;
	private int latencyInFrames;
	private boolean skipComplete;
	private Delay delay;
	private int leftDelayedSamples;
	private boolean readDelayedSamples;

	public LatencyAudioDataReader(AudioDataReader reader, int latencyInFrames) {
		this.reader = reader;
		setLatencyInFrames(latencyInFrames);
	}

	@Override
	public int read(int[] data) throws IOException {
		if (latencyInSamples == 0 || skipComplete) {
			return reader.read(data);
		} else if (latencyInSamples < 0) {
			// negative delay
			// skip all negative delayed samples
			int leftSamplesToSkip = -latencyInSamples;
			while (true) {
				int read = reader.read(data);
				if (read == -1) {
					skipComplete = true;
					return -1;
				}
				if (leftSamplesToSkip == 0) {
					skipComplete = true;
					return read;
				}
				if (read > leftSamplesToSkip) {
					int len = read - leftSamplesToSkip;
					System.arraycopy(data, leftSamplesToSkip, data, 0, len);
					skipComplete = true;
					return len;
				}
				leftSamplesToSkip -= read;
			}
		} else {
			// positive delay
			Delay delay = this.delay;
			if (delay == null) {
				delay = new Delay(latencyInSamples);
			}
			if (readDelayedSamples) {
				if (leftDelayedSamples == 0) {
					return -1;
				}
				int len = Math.min(data.length, leftDelayedSamples);
				delay.delay(data, 0, len);
				leftDelayedSamples -= len;
				this.delay = delay;
				return len;
			}
			int read = reader.read(data);
			if (read > 0) {
				delay.delay(data, 0, read);
				readDelayedSamples = false;
			} else {
				int len = Math.min(data.length, latencyInSamples);
				delay.delay(data, 0, len);
				readDelayedSamples = true;
				leftDelayedSamples = latencyInSamples - len;
				this.delay = delay;
				return len;
			}
			this.delay = delay;
			return read;
		}
	}

	@Override
	public void close() throws IOException {
		if (reader != null) {
			reader.close();
		}
	}

	public void setLatencyInFrames(int latencyInFrames) {
		this.latencyInFrames = latencyInFrames;
		this.latencyInSamples = latencyInFrames * 2;
		reset();
	}

	public int getLatencyInFrames() {
		return latencyInFrames;
	}

	@Override
	public void stop() {
		if (reader != null) {
			reader.stop();
			readDelayedSamples = true;
			leftDelayedSamples = 0;
		}
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
		reset();
	}

	private void reset() {
		skipComplete = false;
		readDelayedSamples = false;
		delay = null;
	}
}
