/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.latency;

import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.FrameObserver;
import org.audiochain.model.AbstractChainableAudioDevice;
import org.audiochain.model.AudioProject;
import org.audiochain.model.Synchronizable;
import org.audiochain.model.UserInterfaceContext;

/**
 * This is the AudioDevice for the delay. The value of the delay means the
 * number of frames.
 */
public class LatencyAudioDevice extends AbstractChainableAudioDevice {
	private static final long serialVersionUID = 1L;
	static final int MIN_LATENCY = -45000;
	static final int MAX_LATENCY = 45000;
	private int latency;
	private long latencyChangeTime;
	private transient LatencyAudioDataReader reader;
	private transient Collection<LatencyAudioDeviceListener> listeners;
	
	@Override
	public void initialize(AudioProject project) {
		super.initialize(project);
		if (!(userInterfaceContext instanceof LatencyAudioDeviceUserInterfaceContext)) {
			this.userInterfaceContext = null;
		}
	}

	@Override
	public AudioDataReaderChainLink createAudioDataReaderChainLink(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		reader = new LatencyAudioDataReader(source, latency);
		return reader;
	}

	public int getLatency() {
		return latency;
	}

	public void setLatency(int latency) {
		assertValueInRange(latency, MIN_LATENCY, MAX_LATENCY);
		int oldLatency = this.latency;
		if (oldLatency != latency) {
			this.latency = latency;
			if (reader != null) {
				reader.setLatencyInFrames(latency);
			}
			latencyChangeTime = System.currentTimeMillis();
			fireLatencyChanged(oldLatency, latency);
		}
	}

	private void fireLatencyChanged(int oldValue, int newValue) {
		if (listeners != null) {
			for (LatencyAudioDeviceListener listener : listeners) {
				listener.latencyChanged(oldValue, newValue);
			}
		}
		firePerformanceRelevantAudioDevicePropertyChanged();
	}

	@Override
	protected UserInterfaceContext<LatencyAudioDevice> createUserInterfaceContext() {
		return new LatencyAudioDeviceUserInterfaceContext();
	}

	@Override
	public String getName() {
		return "Latency";
	}

	void addLatencyAudioDeviceListener(LatencyAudioDeviceListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<LatencyAudioDeviceListener>();
		}
		listeners.add(listener);
	}

	void removeLatencyAudioDeviceListener(LatencyAudioDeviceListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	@Override
	public void synchronizeBy(Synchronizable source) {
		super.synchronizeBy(source);
		if (source instanceof LatencyAudioDevice) {
			LatencyAudioDevice other = (LatencyAudioDevice) source;
			if (greaterOrBothZero(other.latencyChangeTime, latencyChangeTime)) {
				setLatency(other.latency);
				latencyChangeTime = other.latencyChangeTime;
			}
		}
	}
}
