/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.level;

import static java.lang.Math.*;

import java.io.IOException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.DelayedBlockAudioDataReader;
import org.audiochain.io.FrameObserver;
import org.audiochain.ui.PeakProgrammeMeterEvent;


abstract public class SamplePeakProgrammeMeterAudioDataReader extends DelayedBlockAudioDataReader implements AudioDataReaderChainLink {
	private final PeakProgrammeMeterEvent event;
	private final int sampleMaximum;
	private final double sampleMaximumInvers;
	private int updateIntervalInMilliseconds = 70;
	private int returnTimeInMilliseconds = 1700;
	private float returnAttenuationInDecibel = -20F;
	private int attenuationDecrement;
	private int leftPeakSample;
	private int rightPeakSample;
	private int leftPeakHoldSample;
	private int rightPeakHoldSample;
	private float leftPeakHoldDecibelFactor;
	private int leftPeakHoldTime;
	private float rightPeakHoldDecibelFactor;
	private int rightPeakHoldTime;
	private boolean falling;
	private boolean engaged;

	public SamplePeakProgrammeMeterAudioDataReader(AudioDataReader reader, int sampleSize, float frameRate, FrameObserver frameObserver) {
		super(reader, frameRate, frameObserver);
		event = new PeakProgrammeMeterEvent();
		sampleMaximum = (2 << sampleSize - 2) - 1;
		sampleMaximumInvers = 1D / sampleMaximum;
		calcAttenuationDecrement();
	}

	@Override
	public int read(int[] data) throws IOException {
		int len = super.read(data);

		if (len == -1) {
			if (engaged && !falling) {
				engaged = false;
				fall();
			}
		} else {
			engaged = true;
		}

		return len;
	}

	@Override
	protected void processDelayedBlock(int[] stereo, int offset, int length) {
		int leftBlockPeak = 0;
		int rightBlockPeak = 0;
		for (int l = offset, r = l + 1, n = offset + length; l < n; l += 2, r += 2) {
			leftBlockPeak = max(leftBlockPeak, abs(stereo[l]));
			rightBlockPeak = max(rightBlockPeak, abs(stereo[r]));
		}

		if (leftPeakHoldTime < returnTimeInMilliseconds) {
			leftPeakHoldTime += updateIntervalInMilliseconds;
		} else {
			leftPeakHoldDecibelFactor = 0;
		}

		if (rightPeakHoldTime < returnTimeInMilliseconds) {
			rightPeakHoldTime += updateIntervalInMilliseconds;
		} else {
			rightPeakHoldDecibelFactor = 0;
		}

		leftPeakSample -= attenuationDecrement;
		rightPeakSample -= attenuationDecrement;

		if (leftBlockPeak > leftPeakSample) {
			leftPeakSample = leftBlockPeak;
			leftPeakHoldSample = leftBlockPeak;
			leftPeakHoldDecibelFactor = toDecibelFactor(leftBlockPeak);
			leftPeakHoldTime = 0;
		}
		if (rightBlockPeak > rightPeakSample) {
			rightPeakSample = rightBlockPeak;
			rightPeakHoldSample = rightBlockPeak;
			rightPeakHoldDecibelFactor = toDecibelFactor(rightBlockPeak);
			rightPeakHoldTime = 0;
		}

		event.setLeftClip(leftPeakHoldSample > sampleMaximum);
		event.setRightClip(rightPeakHoldSample > sampleMaximum);
		event.setLeftValue(toDecibelFactor(leftPeakSample));
		event.setRightValue(toDecibelFactor(rightPeakSample));
		event.setLeftPeakValue(leftPeakHoldDecibelFactor);
		event.setRightPeakValue(rightPeakHoldDecibelFactor);
		updateSamplePeakProgrammeMeter(event);
	}

	private float toDecibelFactor(int sample) {
		if (sample > sampleMaximum) {
			return 1F;
		}
		if (sample <= 0) {
			return 0F;
		}
		float sqrtValue = (float) (1D - sqrt(-(log10(sample * sampleMaximumInvers) * 20F)) * 0.1D);
		return sqrtValue;
	}

	abstract protected void updateSamplePeakProgrammeMeter(PeakProgrammeMeterEvent event);

	private void calcAttenuationDecrement() {
		double returnValueAfterReturnTime = pow(10, returnAttenuationInDecibel / 20) * sampleMaximum;
		double returnAttenuation = sampleMaximum - returnValueAfterReturnTime;
		double intervals = (double) returnTimeInMilliseconds / (double) updateIntervalInMilliseconds;
		attenuationDecrement = (int) round(returnAttenuation / intervals);
	}

	@Override
	public void close() throws IOException {
		super.close();
		if (engaged && !falling) {
			fall();
		}
	}

	/* finishes the visualization smoothly with a falling ppm */
	private void fall() {
		if (falling || (leftPeakSample <= 0 && rightPeakSample <= 0)) {
			return;
		}
		falling = true;
		new Thread() {
			@Override
			public void run() {
				int[] empty = new int[0];
				while (falling && (leftPeakSample > 0 || rightPeakSample > 0)) {
					processDelayedBlock(empty, 0, 0);
					try {
						Thread.sleep(updateIntervalInMilliseconds);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				engaged = false;
				falling = false;
			}
		}.start();
	}

	void finishImmediately() {
		falling = false;
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	public int getUpdateIntervalInMilliseconds() {
		return updateIntervalInMilliseconds;
	}

	public void setUpdateIntervalInMilliseconds(int updateIntervalInMilliseconds) {
		this.updateIntervalInMilliseconds = updateIntervalInMilliseconds;
		setBlockSizeInFrames((int) (frameRate * updateIntervalInMilliseconds / 1000));
		calcAttenuationDecrement();
	}

	public int getReturnTimeInMilliseconds() {
		return returnTimeInMilliseconds;
	}

	public void setReturnTimeInMilliseconds(int returnTimeInMilliseconds) {
		this.returnTimeInMilliseconds = returnTimeInMilliseconds;
		calcAttenuationDecrement();
	}

	public float getReturnAttenuationInDecibel() {
		return returnAttenuationInDecibel;
	}

	public void setReturnAttenuationInDecibel(float returnAttenuationInDecibel) {
		this.returnAttenuationInDecibel = returnAttenuationInDecibel;
		calcAttenuationDecrement();
	}
}
