/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.level;

import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.FrameObserver;
import org.audiochain.model.AbstractChainableAudioDevice;
import org.audiochain.model.AudioDeviceProperty;
import org.audiochain.model.AudioDeviceValue;
import org.audiochain.model.FloatAudioDeviceValue;
import org.audiochain.model.UserInterfaceContext;
import org.audiochain.ui.PeakProgrammeMeterEvent;
import org.audiochain.ui.PeakProgrammeMeterVisualizer;

public class SamplePeakProgrammeMeterAudioDevice extends AbstractChainableAudioDevice {
	private static final long serialVersionUID = 1L;
	@AudioDeviceProperty(name = "Update Interval", max = 300, min = 0, symbol = 9655, unit = "ms")
	private int updateIntervalInMilliseconds = 70;
	@AudioDeviceProperty(name = "Return Time", max = 10000, min = 0, symbol = 9655, unit = "ms")
	private int returnTimeInMilliseconds = 1700;
	@AudioDeviceProperty(name = "Return Attenuation", max = 0, symbol = 9655, unit = "dB")
	private float returnAttenuationInDecibel = -20F;
	private transient SamplePeakProgrammeMeterAudioDataReader reader;
	private transient Collection<PeakProgrammeMeterVisualizer> visualizers;

	@Override
	public AudioDataReaderChainLink createAudioDataReaderChainLink(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		if (reader != null) {
			reader.finishImmediately();
			reader = null;
		}
		reader = new SamplePeakProgrammeMeterAudioDataReader(source, sampleSizeInBits, frameRate, frameObserver) {
			@Override
			protected void updateSamplePeakProgrammeMeter(PeakProgrammeMeterEvent event) {
				Collection<PeakProgrammeMeterVisualizer> vis = visualizers;
				if (vis != null) {
					for (PeakProgrammeMeterVisualizer visualizer : vis) {
						visualizer.updatePeakProgrammeMeter(event);
					}
				}
			}
		};
		setUpdateIntervalInMilliseconds(updateIntervalInMilliseconds);
		setReturnTimeInMilliseconds(returnTimeInMilliseconds);
		setReturnAttenuationInDecibel(returnAttenuationInDecibel);
		return reader;
	}

	@Override
	protected void createAudioDeviceValues(Collection<AudioDeviceValue> values) {
		if (audioProject != null) {
			for (AudioDeviceValue value : values) {
				if (value.getName().equals("Return Attenuation")) {
					((FloatAudioDeviceValue) value).setMinValue(-audioProject.getDynamicRange());
				}
			}
		}
	}

	@Override
	public String getName() {
		return "SPPM";
	}

	@Override
	protected UserInterfaceContext<SamplePeakProgrammeMeterAudioDevice> createUserInterfaceContext() {
		return new SamplePeakProgrammeMeterUserInterfaceContext();
	}

	public void setUpdateIntervalInMilliseconds(int updateIntervalInMilliseconds) {
		this.updateIntervalInMilliseconds = updateIntervalInMilliseconds;
		if (reader != null) {
			reader.setUpdateIntervalInMilliseconds(updateIntervalInMilliseconds);
		}
	}

	public void setReturnTimeInMilliseconds(int returnTimeInMilliseconds) {
		this.returnTimeInMilliseconds = returnTimeInMilliseconds;
		if (reader != null) {
			reader.setReturnTimeInMilliseconds(returnTimeInMilliseconds);
		}
	}

	public void setReturnAttenuationInDecibel(float returnAttenuationInDecibel) {
		this.returnAttenuationInDecibel = returnAttenuationInDecibel;
		if (reader != null) {
			reader.setReturnAttenuationInDecibel(returnAttenuationInDecibel);
		}
	}

	public void addPeakProgrammeMeterVisualizer(PeakProgrammeMeterVisualizer visualizer) {
		if (visualizers == null) {
			visualizers = new LinkedHashSet<PeakProgrammeMeterVisualizer>();
		}
		visualizers.add(visualizer);
	}

	public void removePeakProgrammeMeterVisualizer(PeakProgrammeMeterVisualizer visualizer) {
		if (visualizers == null) {
			return;
		}
		visualizers.remove(visualizer);
		if (visualizers.isEmpty()) {
			visualizers = null;
		}
	}
}
