/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.text.DecimalFormat;

import javax.sound.sampled.AudioFormat;

import org.audiochain.io.AudioDataJoint;
import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataWriter;
import org.audiochain.io.RawAudioDataWriter;
import org.audiochain.io.RawFileAudioDataReader;
import org.audiochain.model.AudioChannels;
import org.audiochain.model.AudioMixer;
import org.audiochain.model.AudioProject;
import org.audiochain.model.BinaryFileDataset;
import org.audiochain.model.IdentifierGenerator;

public class AudioFile extends BinaryFileDataset implements Serializable {
	private static final long serialVersionUID = 1L;

	private final String name;
	// frame offset within an AudioFileSequence
	private long frameOffset;
	private long frameOffsetChangeTime;
	private Long latencyFrameOffset;
	private long latencyFrameOffsetChangeTime;
	private long timestamp = System.currentTimeMillis();
	private String identifier = IdentifierGenerator.generateIdentifier();
	private transient AudioProject audioProject;
	private transient AudioFormat audioFormat;
	private transient File file;
	private transient boolean markedAsDeleted;
	private transient Long frameLength;

	/**
	 * For Deserialization only
	 */
	@Deprecated
	public AudioFile() {
		name = null;
	}

	public AudioFile(AudioProject audioProject, String name, long frameOffset) {
		if (name == null) {
			throw new IllegalArgumentException(AudioFile.class.getSimpleName() + " name must not be null.");
		}
		this.audioProject = audioProject;
		this.name = audioProject.getRelativePath(name);
		this.frameOffset = frameOffset;
	}

	public AudioFile(AudioProject audioProject, File file, long frameOffset) {
		this.audioProject = audioProject;
		name = audioProject.getRelativePath(file.getAbsolutePath());
		this.frameOffset = frameOffset;
		if (file.exists() && file.canRead()) {
			long lastModified = file.lastModified();
			if (lastModified > 0) {
				timestamp = lastModified;
			}
		}
	}

	void initialize(AudioProject audioProject) {
		this.audioProject = audioProject;
	}

	@Override
	public File getFile() {
		if (file == null) {
			file = audioProject.getRelativeFile(name);
		}
		return file;
	}

	void markAsDeleted() {
		markedAsDeleted = true;
	}

	boolean isMarkedAsDeleted() {
		return markedAsDeleted;
	}

	boolean delete() {
		if (!isMarkedAsDeleted()) {
			throw new IllegalStateException("File " + getName() + " is not marked as deleted.");
		}
		File file = getFile();
		audioProject = null;
		return file.exists() && file.canWrite() && file.delete();
	}

	public String getName() {
		return name;
	}

	public long getTimestamp() {
		return timestamp;
	}

	public AudioDataReader getAudioDataReader() throws IOException {
		return new RawFileAudioDataReader(getFile(), -getLatencyFrameOffset());
	}

	private AudioFormat getAudioFormat() {
		if (audioFormat == null) {
			File file = getFile();
			if (file != null && file.exists() && file.canRead()) {
				try {
					long byteLength = getByteLength();
					// to prevent an EOF Exception, we try to get the following
					// info, only if the file contains data
					if (byteLength > 0) {
						RawFileAudioDataReader rawFileAudioDataReader = new RawFileAudioDataReader(file);
						audioFormat = rawFileAudioDataReader.getAudioFormat();
						rawFileAudioDataReader.close();
					}
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
			}
		}
		return audioFormat;
	}

	private AudioChannels getAudioChannels() {
		AudioFormat audioFormat = getAudioFormat();
		if (audioFormat != null) {
			return AudioChannels.fromChannelCount(audioFormat.getChannels());
		} else {
			return AudioChannels.Mono;
		}
	}

	private float getFrameRate() {
		AudioFormat audioFormat = getAudioFormat();
		if (audioFormat != null) {
			return audioFormat.getFrameRate();
		} else {
			return -1;
		}
	}

	private int getSampleSizeInBits() {
		AudioFormat audioFormat = getAudioFormat();
		if (audioFormat != null) {
			return audioFormat.getSampleSizeInBits();
		} else {
			return -1;
		}
	}

	public long getFrameLength() {
		if (frameLength != null) {
			return frameLength;
		}
		int sampleSizeInBits = getSampleSizeInBits();
		if (sampleSizeInBits == -1) {
			return -1;
		}
		long byteLength = getByteLength();
		if (byteLength == -1) {
			return -1;
		}
		int channels = getAudioChannels().getChannelCount();
		long lengthWithoutHeader = byteLength - RawFileAudioDataReader.HEADER_SIZE;
		long frameLength = lengthWithoutHeader / (sampleSizeInBits / 8 * channels);
		this.frameLength = frameLength;
		return frameLength;
	}

	public long getByteLength() {
		File file = getFile();
		if (file != null && file.exists() && file.canRead()) {
			return file.length();
		} else {
			return -1;
		}
	}

	static String formatTime(long seconds) {
		long min = seconds / 60L;
		long sec = seconds % 60L;
		return (min < 10 ? "0" : "") + min + ':' + (sec < 10 ? "0" : "") + sec;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(getName());

		sb.append(" ");
		sb.append(getAudioChannels().getSymbol());

		String mb = getMegabyteLabel();
		if (mb != null) {
			sb.append(" ");
			sb.append(mb);
		}

		sb.append(' ');
		long offset = getFrameOffset();
		float frameRate = getFrameRate();
		if (offset != 0) {
			if (frameRate != -1) {
				long sec = offset / (long) frameRate;
				sb.append(formatTime(sec));
				sb.append('/');
			}

			sb.append(offset);
		}

		// force a fresh frame length from file
		this.frameLength = null;
		long framesLength = getFrameLength();
		if (framesLength > 0) {
			if (offset != 0) {
				sb.append("->");
			}

			if (frameRate != -1) {
				long sec = framesLength / (long) frameRate;
				sb.append(formatTime(sec));
				sb.append('/');
			}

			sb.append(framesLength);
		}

		if (latencyFrameOffset != null) {
			sb.append(' ');
			sb.append(latencyFrameOffset);
		}

		return sb.toString().trim();
	}

	String getMegabyteLabel() {
		long bytes = getByteLength();
		if (bytes != -1) {
			return formatMegabytes(bytes);
		}
		return null;
	}

	static String formatMegabytes(Long bytes) {
		DecimalFormat df = new DecimalFormat("#0.0");
		return df.format(bytes / 1000000D) + "MB";
	}

	/**
	 * Save replacing of the origFile with the newFile by creating a bak file
	 * while replace operation.
	 * 
	 * @param origFile
	 *            the file to be replace
	 * @param newFile
	 *            the new file that takes the place of the orig file
	 * @return true if succeed, false if not. In case of no success, the
	 *         original state should be reconstructed savely.
	 */
	public static boolean replace(File origFile, File newFile) {
		boolean ret;
		File bakFile = new File(origFile.getAbsolutePath() + ".bak");
		bakFile.setLastModified(origFile.lastModified());
		newFile.setLastModified(origFile.lastModified());
		if (origFile.renameTo(bakFile)) {
			if (newFile.renameTo(origFile)) {
				if (!bakFile.delete()) {
					System.err.println("Could not delete '" + bakFile + "'.");
				}
				ret = true;
			} else {
				// restore renaming to bak
				if (!bakFile.renameTo(origFile)) {
					System.err.println("Cannot rename '" + bakFile + "' to '" + origFile + "'.");
				}
				ret = false;
			}
		} else {
			System.err.println("Cannot rename '" + origFile + "' to '" + bakFile + "'.");
			ret = false;
		}
		return ret;
	}

	public long getFrameOffset() {
		return frameOffset;
	}

	public void setFrameOffset(long frameOffset) {
		this.frameOffset = frameOffset;
		frameOffsetChangeTime = System.currentTimeMillis();
	}

	public boolean isPositionWithinFrames(long position) {
		long offset = getFrameOffset();
		Long length = getFrameLength();
		Long end = length == null ? null : offset + length;

		if (position < offset || (end != null && position > end)) {
			return false;
		}
		return true;
	}

	public long getLatencyFrameOffset() {
		if (latencyFrameOffset == null) {
			return -AudioMixer.getInstance().getSystemLatencyInFrames();
		}
		return latencyFrameOffset;
	}

	public void setLatencyFrameOffset(long latencyFrameOffset) {
		this.latencyFrameOffset = latencyFrameOffset;
		latencyFrameOffsetChangeTime = System.currentTimeMillis();
	}

	public boolean convert(AudioFormat newAudioFormat) throws IOException {
		File origFile = getFile();
		File convertedFile = new File(origFile.getAbsolutePath() + ".tmp");
		FileOutputStream fos = new FileOutputStream(convertedFile);
		BufferedOutputStream bos = new BufferedOutputStream(fos);
		AudioDataWriter writer = new RawAudioDataWriter(newAudioFormat, bos);
		AudioDataReader reader = getAudioDataReader();
		AudioDataJoint joint = new AudioDataJoint(reader, writer);
		joint.start();
		boolean ret;
		try {
			joint.join();
			bos.close();
			fos.close();

			if (replace(origFile, convertedFile)) {
				ret = true;
			} else {
				ret = false;
			}
		} catch (InterruptedException e) {
			ret = false;
		} finally {
			bos.close();
			fos.close();
		}
		return ret;
	}

	String getIdentifier() {
		return identifier;
	}

	void setIdentifier(String identifier) {
		this.identifier = identifier;
	}

	@Override
	public AudioProject getBinaryDatasetAssociatedAudioProject() {
		return audioProject;
	}

	void setTimestamp(long timestamp) {
		this.timestamp = timestamp;
		getFile().setLastModified(timestamp);
	}

	public long getFrameOffsetChangeTime() {
		return frameOffsetChangeTime;
	}

	public void setFrameOffsetChangeTime(long frameOffsetChangeTime) {
		this.frameOffsetChangeTime = frameOffsetChangeTime;
	}

	public long getLatencyFrameOffsetChangeTime() {
		return latencyFrameOffsetChangeTime;
	}

	public void setLatencyFrameOffsetChangeTime(long latencyFrameOffsetChangeTime) {
		this.latencyFrameOffsetChangeTime = latencyFrameOffsetChangeTime;
	}
}
