/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;

import org.audiochain.io.AudioDataReader;
import org.audiochain.model.IdentifierGenerator;


public class AudioFileSequence implements Serializable {
	private static final long serialVersionUID = 1L;
	private boolean active;
	private String identifier = IdentifierGenerator.generateIdentifier();
	private final Collection<AudioFile> audioFiles = new LinkedHashSet<AudioFile>();
	private Collection<String> removedAudioFileIdentifiers;

	public AudioDataReader getAudioDataReader() {
		return new FileSequenceAudioDataReader(this);
	}

	void markAsDeleted() {
		for (AudioFile file : audioFiles) {
			file.markAsDeleted();
		}
	}

	boolean delete() {
		boolean ret = true;
		for (AudioFile file : audioFiles) {
			ret &= file.delete();
		}
		return ret;
	}
	
	boolean isMarkedAsDeleted() {
		boolean ret = true;
		for (AudioFile file : audioFiles) {
			ret &= file.isMarkedAsDeleted();
		}
		return ret;
	}

	public void addAudioFile(AudioFile audioFile) {
		this.audioFiles.add(audioFile);
	}

	public void removeAudioFile(AudioFile audioFile) {
		this.audioFiles.remove(audioFile);
	}

	public Collection<AudioFile> getAudioFiles() {
		return new LinkedHashSet<AudioFile>(audioFiles);
	}

	public long getMaxFrame() {
		long ret = 0;
		for (AudioFile file : audioFiles) {
			long fileOffset = file.getFrameOffset();
			long fileLength = file.getFrameLength();
			long fileEnd = fileOffset + fileLength;
			ret = Math.max(ret, fileEnd);
		}
		return ret;
	}
	
	/**
	 * Masked frames are marked with '#'.
	 * The Brackets are the files in their order within audioFiles.
	 * Earlier recorded files are hidden by the later recorded files.
	 * <pre>
	 *   3   7   11   16  20   25
	 * ..|...|...|....|...|....|..
	 *   [ 5 ]   [ 4  ]   [ 6  ]
	 * [ ##### 3 #]    [1 #][###2]
	 * |          |    |   |     |
	 * 0          12   17  21    27
	 * </pre>
	 * @return
	 */
	public long getFrameLength() {
		long ret = 0;
		Collection<AudioFile> maskingFiles = new LinkedHashSet<AudioFile>();
		List<AudioFile> reversed = new ArrayList<AudioFile>(audioFiles);
		Collections.reverse(reversed);
		for (AudioFile file : reversed) {
			final long fileOffset = file.getFrameOffset();
			final long fileLength = file.getFrameLength();
			final long fileEnd = fileOffset + fileLength;
			long visibleLength = fileLength;
			
			for (AudioFile maskingFile : maskingFiles) {
				final long maskingOffset = maskingFile.getFrameOffset();
				final long maskingLength = maskingFile.getFrameLength();
				final long maskingEnd = maskingOffset + maskingLength;
				
				final long maskedOffset = Math.max(fileOffset, maskingOffset);
				final long maskedEnd = Math.min(fileEnd, maskingEnd);
				if(maskedEnd>maskedOffset) {
					final long maskedLength = maskedEnd - maskedOffset;
					visibleLength -= maskedLength;
				}
			}
			ret += visibleLength;
			
			maskingFiles.add(file);
		}
		return ret;
	}

	Collection<AudioFile> getFilesByRange(long begin, long end) {
		Collection<AudioFile> ret = new LinkedHashSet<AudioFile>();
		for (AudioFile file : audioFiles) {
			long fileOffset = file.getFrameOffset();
			long fileLength = file.getFrameLength();
			long fileEnd = fileOffset + fileLength;
			if (begin >= fileOffset && begin < fileEnd) {
				ret.add(file);
			} else if (begin < fileOffset && end > fileOffset) {
				ret.add(file);
			}
		}
		return ret;
	}
	
	AudioFile getAudibleFileAtFrame(long framePosition) {
		Collection<AudioFile> filesAtFrame = getFilesByRange(framePosition, framePosition);
		if (filesAtFrame.isEmpty()) {
			return null;
		}
		/* The last recorded file is the audible/visible file */
		AudioFile[] files = filesAtFrame.toArray(new AudioFile[0]);
		return files[files.length - 1];
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();

		sb.append('{');
		boolean first = true;
		for (AudioFile file : audioFiles) {
			if (!first) {
				sb.append(", ");
			}
			sb.append(file.toString());
			first = false;
		}
		sb.append('}');

		return sb.toString();
	}
	
	public boolean isActive() {
		return active;
	}

	public void setActive(boolean active) {
		this.active = active;
	}

	public void setLatencyFrameOffset(long latencyFrameOffset) {
		for (AudioFile file : audioFiles) {
			file.setLatencyFrameOffset(latencyFrameOffset);
		}
	}

	public long getByteLength() {
		long ret = 0;
		for (AudioFile file : audioFiles) {
			long byteLength = file.getByteLength();
			if(byteLength>-1) {
				ret += byteLength;
			}
		}
		return ret;
	}
	
	public long getLatestTimestamp() {
		long timestamp = 0;
		for(AudioFile audioFile : getAudioFiles()) {
			Long ts = audioFile.getTimestamp();
			if (ts != null && ts > timestamp) {
				timestamp = ts;
			}
		}
		return timestamp;
	}
	
	void addRemovedAudioFileIdentifier(String removedAudioFileIdentifier) {
		if (removedAudioFileIdentifiers == null) {
			removedAudioFileIdentifiers = new LinkedHashSet<String>();
		}
		removedAudioFileIdentifiers.add(removedAudioFileIdentifier);
	}
	
	boolean isRemoved(AudioFile audioFile) {
		return removedAudioFileIdentifiers != null && removedAudioFileIdentifiers.contains(audioFile.getIdentifier());
	}
	
	AudioFile findAudioFileByIdentifier(String identifier) {
		for (AudioFile audioFile : audioFiles) {
			if(audioFile.getIdentifier().equals(identifier)) {
				return audioFile;
			}
		}
		return null;
	}

	boolean allAudioFilesPresent() {
		if (audioFiles.isEmpty()) {
			return false;
		}
		for (AudioFile audioFile : audioFiles) {
			if(!audioFile.getFile().exists()) {
				return false;
			}
		}
		return true;
	}

	String getIdentifier() {
		return identifier;
	}

	void setIdentifier(String identifier) {
		this.identifier = identifier;
	}
}
