/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import static org.audiochain.io.AudioDataUtil.*;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.audiochain.io.AudioDataReader;


public class FileSequenceAudioDataReader implements AudioDataReader {

	private final AudioFileSequence audioFileSequence;
	private long currentFramePosition;
	private Map<AudioFile, AudioDataReader> openAudioDataReaders = new HashMap<AudioFile, AudioDataReader>();
	private boolean stopped;
	private final long maxFrame;

	public FileSequenceAudioDataReader(AudioFileSequence audioFileSequence) {
		this.audioFileSequence = audioFileSequence;
		this.maxFrame = audioFileSequence.getMaxFrame();
	}

	@Override
	public int read(int[] sampleData) throws IOException {
		if (stopped) {
			return -1;
		}
		final long framePosition = this.currentFramePosition;
		if (framePosition >= maxFrame) {
			return -1;
		}

		final int dataLengthInSamples = sampleData.length;
		Arrays.fill(sampleData, 0);
		final long endFrame = framePosition + dataLengthInSamples / 2;
		Collection<AudioFile> filesInRange = audioFileSequence.getFilesByRange(framePosition, endFrame);
		// later recorded files are masking earlier files
		for (AudioFile file : filesInRange) {
			AudioDataReader reader = getAudioDataReader(file);
			final long fileOffset = file.getFrameOffset();
			if (framePosition >= fileOffset) {
				final long readerPosition = framePosition - fileOffset;
				reader.seek(readerPosition);
				readFully(reader, sampleData);
			} else {
				final int readerBeginPosition = (int) (fileOffset - framePosition);
				final int sampleBufferSize = dataLengthInSamples - readerBeginPosition * 2;
				final int[] sampleBuffer = new int[sampleBufferSize];
				reader.seek(0);
				final int readSampleLength = readFully(reader, sampleBuffer);
				if (readSampleLength > 0) {
					System.arraycopy(sampleBuffer, 0, sampleData, readerBeginPosition, readSampleLength);
				}
			}
		}

		this.currentFramePosition = endFrame;

		return dataLengthInSamples;
	}

	private AudioDataReader getAudioDataReader(AudioFile file) throws IOException {
		AudioDataReader audioDataReader = openAudioDataReaders.get(file);
		if (audioDataReader == null) {
			audioDataReader = file.getAudioDataReader();
			openAudioDataReaders.put(file, audioDataReader);
		}
		return audioDataReader;
	}

	@Override
	public void close() throws IOException {
		for (AudioDataReader reader : openAudioDataReaders.values()) {
			if (reader != null) {
				reader.close();
			}
		}
		openAudioDataReaders.clear();
	}

	@Override
	public void stop() {
		stopped = true;
	}

	@Override
	public void seek(long framePosition) {
		currentFramePosition = framePosition;
	}

}
