/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.TargetDataLine;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataUtil;
import org.audiochain.io.AudioDataWriter;
import org.audiochain.io.AudioIOException;
import org.audiochain.io.FrameObserver;
import org.audiochain.io.LineAudioDataReader;
import org.audiochain.io.LineFrameObserver;
import org.audiochain.io.RawAudioDataWriter;
import org.audiochain.io.RawFileAudioDataReader;
import org.audiochain.model.AudioDeviceChain;
import org.audiochain.model.AudioMixer;
import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.Playback;
import org.audiochain.model.SourceAudioDevice;


class Recording extends Playback {
	static enum RecordingStatus {
		New, Initialized, Started, Capturing, CapturingAndWaitingForPlayer, Recording, Stopped, Finished
	}
	private int[] recordingBuffer;
	private AudioTrack recordingTrack;
	private LineAudioDataReader captureLineReader;
	private AudioDataReader recordingReader;
	private FrameObserver captureLineFrameObserver;
	private TargetDataLine captureLine;
	private AudioDataWriter recordingWriter;
	private AudioFile recordingAudioFile;
	private AudioFileSequence recordingAudioFileSequence;
	private RecordingStatus recordingStatus = RecordingStatus.New;
	private RecordingAudioDevice recordingAudioDevice;

	Recording(AudioProject project) throws AudioIOException {
		super(project);
		recordingBuffer = new int[AudioMixer.getInstance().getCaptureBufferSizeInSamples()];
		obtainRecordingTrack();
		createCaptureLines();
		
		createCaptureAudioDataReader();
		
		setNextRecordingStatus(RecordingStatus.Initialized);

		CaptureThread captureThread = new CaptureThread();
		captureThread.start();

		// wait until capturing has started
		while (captureLine.getLongFramePosition() == 0 || recordingStatus!=RecordingStatus.Capturing || !captureLine.isRunning()) {
			Thread.yield();
		}
	}

	private void createCaptureLines() throws AudioIOException {
		// capture format is the projects format (the tracks format is used to write the file)
		AudioFormat captureFormat = project.getAudioFormat();
		captureLineReader = new LineAudioDataReader(captureFormat, AudioMixer.getInstance().getCaptureMixer().getMixer());
		captureLineReader.setClosable(false);
		captureLineReader.setStopable(false);
		captureLine = captureLineReader.getLine();
		captureLineFrameObserver = new LineFrameObserver(captureLine);
	}

	private void reinit() throws IOException {
		initPlayback();
		recordingBuffer = new int[AudioMixer.getInstance().getCaptureBufferSizeInSamples()];
		obtainRecordingTrack();
		if(recordingReader!=null) {
			recordingReader.stop();
			recordingReader.close();
		}
		createCaptureAudioDataReader();
	}
	
	private void createCaptureAudioDataReader() {
		AudioDeviceChain captureAudioDeviceChain = recordingTrack.getCaptureAudioDeviceChain();
		recordingReader = captureAudioDeviceChain.createReversedAudioDataReaderChain(captureLineReader, project.getFrameRate(), project.getSampleSizeInBits(), captureLineFrameObserver);
	}

	private void obtainRecordingTrack() {
		Collection<AudioTrack> recordTracks = project.getRecordTracks();
		if(recordTracks.isEmpty()) {
			throw new IllegalArgumentException("One track must be selected for recording.");
		}
		recordingTrack = recordTracks.iterator().next();
		SourceAudioDevice sourceAudioDevice = recordingTrack.getSourceAudioDevice();
		if(sourceAudioDevice instanceof RecordingAudioDevice) {
			recordingAudioDevice = (RecordingAudioDevice) sourceAudioDevice;
		} else {
			throw new IllegalArgumentException("The selected track must have a " + RecordingAudioDevice.DEVICE_NAME + " as source audio device to be recordable.");
		}
	}

	@Override
	public void start(long framePosition) throws IOException, AudioIOException {
		// in the case that the project has changed in any way
		reinit();
		this.startFramePosition = framePosition;
		captureLineFrameObserver.setFramePosition(framePosition);
		
		DateFormat dateFormat = new SimpleDateFormat("yyMMdd_HHmmss_S");
		String time = dateFormat.format(new Date());
		String namePrefix = recordingTrack.getName();
		String audioFileName = (namePrefix != null ? namePrefix.replaceAll("\\W", "") + "_" : "") + time + RawFileAudioDataReader.FILE_NAME_ENDING;
		recordingAudioFile = new AudioFile(project, audioFileName, startFramePosition);
		recordingAudioFileSequence = recordingAudioDevice.addNewAudioFile(recordingAudioFile);
		recordingAudioDevice.setActiveAudioFileSequence(recordingAudioFileSequence);
		File recFile = recordingAudioFile.getFile();
		BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(recFile));
		AudioFormat fileFormat = recordingAudioDevice.getAudioFormat();
		recordingWriter = new RawAudioDataWriter(fileFormat, bos);

		setNextRecordingStatus(RecordingStatus.CapturingAndWaitingForPlayer);
		super.start(framePosition);
		// wait until the recording has started, in the case of present play
		// track, this can only be after playback has started
		while (getPlaybackStatus().ordinal() < PlaybackStatus.Playing.ordinal()) {
			Thread.yield();
		}
		fireRecordingStarted();
	}

	@Override
	public void stop() {
		super.stop();
		// stop means stop recording and stay in capture status for a new reinitialization and a further recording
		if (recordingStatus == RecordingStatus.Capturing) {
			return;
		}
		setNextRecordingStatus(RecordingStatus.Capturing);
		if(recordingWriter!=null) {
			try {
				recordingWriter.close();
			} catch (IOException e) {
				addError(e);
				throw new IllegalStateException(e);
			} finally {
				recordingWriter = null;
			}
		}
		fireRecordingStopped();
	}

	@Override
	public void close() throws IOException {
		setStatusStopped();
		if(captureLineReader!=null) {
			captureLineReader.setClosable(true);
			captureLineReader.setStopable(true);
			captureLineReader.stop();
			captureLineReader.close();
		}
	}

	@Override
	public void join() {
		super.join();
		while (recordingStatus != RecordingStatus.Finished) {
			Thread.yield();
		}
	}
	
	@Override
	public void joinStopped() {
		super.joinStopped();
		while (recordingStatus != RecordingStatus.Capturing) {
			Thread.yield();
		}
	}

	@Override
	public void seek(long framePosition) {
		// nop
	}

	private void capture() throws IOException {
		setNextRecordingStatus(RecordingStatus.Capturing);
		try {
			int[] captureBuffer = new int[64];
			while (recordingStatus == RecordingStatus.Capturing || recordingStatus == RecordingStatus.CapturingAndWaitingForPlayer || recordingStatus == RecordingStatus.Recording) {
				switch(recordingStatus) {
					case CapturingAndWaitingForPlayer:
						// wait until playback is playing or has finished
						if (getPlaybackStatus().ordinal() >= PlaybackStatus.Playing.ordinal()) {
							setNextRecordingStatus(RecordingStatus.Recording);
						}
						//$FALL-THROUGH$
					case Capturing:
						if (recordingReader.read(captureBuffer) == -1) {
							setStatusStopped();
						}
						break;
					case Recording:
						int len = recordingReader.read(recordingBuffer);
						if (len == -1) {
							setStatusStopped();
						} else {
							AudioDataWriter writer = recordingWriter;
							// writer may have been already closed when recording has been stopped
							if (writer != null) {
								writer.write(recordingBuffer, 0, len);
							}
						}
						fireRecordingUpdated();
						break;
				}
			}
		} finally {
			setStatusStopped();
		}
	}
	
	private void setStatusStopped() {
		if(recordingStatus == RecordingStatus.Capturing || recordingStatus == RecordingStatus.CapturingAndWaitingForPlayer || recordingStatus == RecordingStatus.Recording) {
			setNextRecordingStatus(RecordingStatus.Stopped);
		}
	}

	private void closeCaptureResources() {
		try {
			recordingReader.close();
		} catch (IOException e) {
			addError(e);
			throw new IllegalStateException(e);
		}
	}

	private class CaptureThread extends Thread {
		@Override
		public void run() {
			setNextRecordingStatus(RecordingStatus.Started);
			setPriority(MAX_PRIORITY);
			setName(getClass().getSimpleName());
			try {
				capture();
			} catch (Throwable e) {
				addError(e);
				throw new IllegalStateException(e);
			} finally {
				closeCaptureResources();
				setNextRecordingStatus(RecordingStatus.Finished);
				fireRecordingFinished();
			}
		}
	}

	private void setNextRecordingStatus(RecordingStatus newStatus) {
		if (newStatus.ordinal() > recordingStatus.ordinal()) {
			recordingStatus = newStatus;
		} else if (recordingStatus == RecordingStatus.Capturing && newStatus == RecordingStatus.Initialized) {
			recordingStatus = newStatus;
		} else if (recordingStatus == RecordingStatus.Recording && newStatus == RecordingStatus.Capturing) {
			recordingStatus = newStatus;
		} else if (recordingStatus == RecordingStatus.CapturingAndWaitingForPlayer && newStatus == RecordingStatus.Capturing) {
			recordingStatus = newStatus;
		} else {
			throw new IllegalStateException(RecordingStatus.class.getSimpleName() + " '" + recordingStatus + "' cannot be followed by '" + newStatus + "'.");
		}
	}

	private void fireRecordingStarted() {
		firePerformanceStarted();
	}

	private void fireRecordingUpdated() {
		firePerformanceUpdated();
	}

	private void fireRecordingStopped() {
		firePerformanceStopped();
	}

	private void fireRecordingFinished() {
		firePerformanceFinished();
	}

	@Override
	protected void firePlaybackStarted() {
		// prevent playback from fire events
	}
	
	@Override
	protected void firePlaybackUpdated() {
		// prevent playback from fire events
	}
	
	@Override
	protected void firePlaybackStopped() {
		// prevent playback from fire events
	}
	
	@Override
	protected void firePlaybackFinished() {
		// prevent playback from fire events
	}

	@Override
	public FrameObserver getFrameObserver() {
		return captureLineFrameObserver;
	}

	RecordingStatus getRecordingStatus() {
		return recordingStatus;
	}

	@Override
	public boolean isRecording() {
		return recordingStatus == RecordingStatus.Recording || recordingStatus == RecordingStatus.CapturingAndWaitingForPlayer || recordingStatus == RecordingStatus.Started;
	}
	
	@Override
	public boolean isRunning() {
		return isRecording();
	}

	AudioFile getRecordingAudioFile() {
		return recordingAudioFile;
	}

	AudioFileSequence getRecordingAudioFileSequence() {
		return recordingAudioFileSequence;
	}
	
	@Override
	protected void dumpReaderChain() throws IOException {
		super.dumpReaderChain();
		System.out.println();
		System.out.println("Recording reader chain:");
		AudioDataUtil.dumpAudioDataReaderChain(recordingReader, System.out);
	}
}
