/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;

import javax.sound.sampled.AudioFormat;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioIOException;
import org.audiochain.io.FrameObserver;
import org.audiochain.model.AbstractAudioDevice;
import org.audiochain.model.AudioChannels;
import org.audiochain.model.AudioProject;
import org.audiochain.model.BinaryDataset;
import org.audiochain.model.BinarySynchronizable;
import org.audiochain.model.Performable;
import org.audiochain.model.PerformanceListener;
import org.audiochain.model.SourceAudioDevice;
import org.audiochain.model.Synchronizable;
import org.audiochain.model.UserInterfaceContext;

public class RecordingAudioDevice extends AbstractAudioDevice implements SourceAudioDevice, BinarySynchronizable {
	private static final long serialVersionUID = 1L;
	static final int FRAME_OFFSET_INCREMENT_IN_MILLIS = 10;
	static final String DEVICE_NAME = "Recorder";
	private Collection<AudioFileSequence> audioFileSequences = new LinkedHashSet<AudioFileSequence>();
	private AudioChannels audioChannels = AudioChannels.Mono;
	private AudioFileSequenceCreationPolicy sequenceCreationPolicy = AudioFileSequenceCreationPolicy.NewAudioFileSequence;
	private Collection<String> detachedAudioFileSequenceIdentifiers;
	private Collection<String> deletedAudioFileSequenceIdentifiers;
	private long lastAudioFileSequenceActivationChangeTime;
	private transient Collection<RecordingAudioDeviceListener> listeners;
	private transient int audioFileFrameOffsetIncrement;

	@Override
	public void initialize(AudioProject audioProject) {
		super.initialize(audioProject);
		for(AudioFileSequence seq : getAudioFileSequences()) {
			for(AudioFile file : seq.getAudioFiles()) {
				file.initialize(audioProject);
			}
		}
		audioFileFrameOffsetIncrement = audioProject.getFrameRate() * FRAME_OFFSET_INCREMENT_IN_MILLIS / 1000;
		audioProject.addAudioProjectLifecycleListener(new UnsavedAudioFileAttendant(this));
	}

	@Override
	public void destroy() {
		deleteAllAudioFileSequences();
		super.destroy();
	}

	@Override
	public AudioDataReader createAudioDataReader(float frameRate, int sampleSizeInBits, FrameObserver frameObserver) throws IOException {
		AudioFileSequence activeAudioFileSequence = getActiveAudioFileSequence();
		if (activeAudioFileSequence == null) {
			return null;
		}
		return activeAudioFileSequence.getAudioDataReader();
	}

	@Override
	protected UserInterfaceContext<RecordingAudioDevice> createUserInterfaceContext() {
		return new RecordingAudioDeviceUserInterfaceContext();
	}

	AudioFileSequence addNewAudioFile(AudioFile newAudioFile) {
		AudioFileSequence ret = addAudioFile(newAudioFile);
		fireNewAudioFileAdded(newAudioFile);
		return ret;
	}
	
	AudioFileSequence addAudioFile(String audioFileName) {
		AudioFile audioFile = new AudioFile(audioProject, audioFileName, audioProject.getFramePosition());
		AudioFileSequence ret = addAudioFile(audioFile);
		fireBinaryChanged(audioFile);
		return ret;
	}

	AudioFileSequence addAudioFile(File file) {
		AudioFile audioFile = new AudioFile(audioProject, file, audioProject.getFramePosition());
		AudioFileSequence ret = addAudioFile(audioFile);
		fireBinaryChanged(audioFile);
		return ret;
	}

	private AudioFileSequence addAudioFile(AudioFile audioFile) {
		switch (sequenceCreationPolicy) {
		case NewAudioFileSequence:
		default:
			AudioFileSequence newSequence = new AudioFileSequence();
			newSequence.addAudioFile(audioFile);
			addAudioFileSequence(newSequence);
			return newSequence;
		case SameAudioFileSequence:
			AudioFileSequence activeSequence = this.getActiveAudioFileSequence();
			if (activeSequence == null) {
				activeSequence = new AudioFileSequence();
				activeSequence.addAudioFile(audioFile);
				addAudioFileSequence(activeSequence);
				return activeSequence;
			} else {
				activeSequence.addAudioFile(audioFile);
				fireAudioFileSequenceUpdated(activeSequence);
				return activeSequence;
			}
		}
	}

	AudioFileSequence getActiveAudioFileSequence() {
		for (AudioFileSequence audioFileSequence : getAudioFileSequences()) {
			if (audioFileSequence.isActive()) {
				return audioFileSequence;
			}
		}
		return null;
	}

	void setActiveAudioFileSequence(AudioFileSequence audioFileSequence) {
		lastAudioFileSequenceActivationChangeTime = System.currentTimeMillis();
		for (AudioFileSequence otherAudioFileSequence : getAudioFileSequences()) {
			otherAudioFileSequence.setActive(false);
		}
		audioFileSequence.setActive(true);
		fireAudioFileSequenceActivated(audioFileSequence);
		firePerformanceRelevantAudioDevicePropertyChanged();
		notifyProjectForChanges();
	}

	void setLatestAudioFileSequenceActive() {
		Collection<AudioFileSequence> audioFileSequences = getAudioFileSequences();
		if (audioFileSequences.isEmpty()) {
			return;
		}
		long timestamp = 0;
		AudioFileSequence latestSequence = null;
		for (AudioFileSequence audioFileSequence : audioFileSequences) {
			Long ts = audioFileSequence.getLatestTimestamp();
			if (ts != null && ts > timestamp) {
				latestSequence = audioFileSequence;
				timestamp = ts;
			}
		}
		if (latestSequence == null) {
			AudioFileSequence[] arr = audioFileSequences.toArray(new AudioFileSequence[0]);
			latestSequence = arr[arr.length - 1];
		}
		setActiveAudioFileSequence(latestSequence);
	}

	void addAudioFileSequence(AudioFileSequence audioFileSequence) {
		if (!getAudioFileSequences().contains(audioFileSequence)) {
			audioFileSequences.add(audioFileSequence);
			fireAudioFileSequenceAdded(audioFileSequence);
			notifyProjectForChanges();
		}
	}

	void deleteAllAudioFileSequences() {
		for (AudioFileSequence audioFileSequence : new HashSet<AudioFileSequence>(getAudioFileSequences())) {
			deleteAudioFileSequence(audioFileSequence);
		}
	}

	void deleteActiveAudioFileSequence() {
		AudioFileSequence audioFileSequence = getActiveAudioFileSequence();
		if (audioFileSequence != null) {
			deleteAudioFileSequence(audioFileSequence);
		}
	}

	void deleteAudioFile(String name) {
		for (AudioFileSequence seq : getAudioFileSequences()) {
			for (AudioFile file : seq.getAudioFiles()) {
				String fileName = file.getName();
				if (fileName.equals(name)) {
					seq.removeAudioFile(file);
					file.markAsDeleted();
					fireAudioFileDeleted(file);
					if (seq.getAudioFiles().isEmpty()) {
						deleteAudioFileSequence(seq);
						return;
					}
				}
			}
		}
	}

	void deleteAudioFileSequences(Collection<AudioFileSequence> sequences) {
		for (AudioFileSequence audioFileSequence : sequences) {
			deleteAudioFileSequence(audioFileSequence);
		}
	}

	void deleteAudioFileSequence(AudioFileSequence audioFileSequence) {
		audioFileSequence.markAsDeleted();
		fireAudioFileSequenceDeleted(audioFileSequence);
		addDeletedAudioFileSequenceIdentifier(audioFileSequence.getIdentifier());
		removeAudioFileSequence(audioFileSequence);
	}

	void detachAudioFileSequence(AudioFileSequence audioFileSequence) {
		addDetachedAudioFileSequenceIdentifier(audioFileSequence.getIdentifier());
		removeAudioFileSequence(audioFileSequence);
	}

	private void addDetachedAudioFileSequenceIdentifier(String removedAudioFileSequenceIdentifier) {
		if (detachedAudioFileSequenceIdentifiers == null) {
			detachedAudioFileSequenceIdentifiers = new LinkedHashSet<String>();
		}
		detachedAudioFileSequenceIdentifiers.add(removedAudioFileSequenceIdentifier);
	}

	private boolean isDetached(AudioFileSequence audioFileSequence) {
		if (detachedAudioFileSequenceIdentifiers != null) {
			return detachedAudioFileSequenceIdentifiers.contains(audioFileSequence.getIdentifier());
		}
		return false;
	}

	private void addDeletedAudioFileSequenceIdentifier(String removedAudioFileSequenceIdentifier) {
		if (deletedAudioFileSequenceIdentifiers == null) {
			deletedAudioFileSequenceIdentifiers = new LinkedHashSet<String>();
		}
		deletedAudioFileSequenceIdentifiers.add(removedAudioFileSequenceIdentifier);
	}

	private boolean isDeleted(AudioFileSequence audioFileSequence) {
		if (deletedAudioFileSequenceIdentifiers != null) {
			return deletedAudioFileSequenceIdentifiers.contains(audioFileSequence.getIdentifier());
		}
		return false;
	}

	private void removeAudioFileSequence(AudioFileSequence audioFileSequence) {
		audioFileSequences.remove(audioFileSequence);
		fireAudioFileSequenceRemoved(audioFileSequence);
		firePerformanceRelevantAudioDevicePropertyChanged();
		notifyProjectForChanges();
	}

	void incrementAudibleFileFrameOffset() {
		moveAudibleFileFrameOffset(audioFileFrameOffsetIncrement);
	}

	void decrementAudibleFileFrameOffset() {
		moveAudibleFileFrameOffset(-audioFileFrameOffsetIncrement);
	}

	private void moveAudibleFileFrameOffset(long amount) {
		AudioFile file = getAudibleFile();
		if (file != null) {
			long oldFrameOffset = file.getFrameOffset();
			long newFrameOffset = oldFrameOffset + amount;
			setFrameOffset(file, newFrameOffset);
		}
	}

	void setAudibleFileFrameOffset(long frameOffset) {
		AudioFile file = getAudibleFile();
		if (file != null) {
			setFrameOffset(file, frameOffset);
		}
	}

	private void setFrameOffset(AudioFile file, long frameOffset) {
		long oldFrameOffset = file.getFrameOffset();
		if (oldFrameOffset != frameOffset) {
			file.setFrameOffset(frameOffset);
			fireAudioFileFrameOffsetChanged(file, frameOffset, oldFrameOffset);
			firePerformanceRelevantAudioDevicePropertyChanged();
			notifyProjectForChanges();
		}
	}

	private AudioFile getAudibleFile() {
		long framePosition = audioProject.getFramePosition();
		AudioFileSequence activeAudioFileSequence = getActiveAudioFileSequence();
		if (activeAudioFileSequence != null) {
			return activeAudioFileSequence.getAudibleFileAtFrame(framePosition);
		}
		return null;
	}

	Collection<AudioFileSequence> getUnusedAudioFileSequences() {
		Collection<AudioFileSequence> copyOfSequenceList = new HashSet<AudioFileSequence>(getAudioFileSequences());
		if (copyOfSequenceList.isEmpty()) {
			return Collections.unmodifiableCollection(copyOfSequenceList);
		}
		AudioFileSequence usedSequence = getActiveAudioFileSequence();
		copyOfSequenceList.remove(usedSequence);
		return Collections.unmodifiableCollection(copyOfSequenceList);
	}

	void setPreviousAudioFileSequenceActive() {
		List<AudioFileSequence> reversedAudioFileSequences = new ArrayList<AudioFileSequence>(getAudioFileSequences());
		Collections.reverse(reversedAudioFileSequences);
		boolean lastWasActive = false;
		for (AudioFileSequence audioFileSequence : reversedAudioFileSequences) {
			if (lastWasActive) {
				setActiveAudioFileSequence(audioFileSequence);
				return;
			}
			lastWasActive = audioFileSequence.isActive();
		}
	}

	void setNextAudioFileSequenceActive() {
		boolean lastWasActive = false;
		Collection<AudioFileSequence> audioFileSequences = getAudioFileSequences();
		for (AudioFileSequence audioFileSequence : audioFileSequences) {
			if (lastWasActive) {
				setActiveAudioFileSequence(audioFileSequence);
				return;
			}
			lastWasActive = audioFileSequence.isActive();
		}
		if (!lastWasActive && !audioFileSequences.isEmpty()) {
			setActiveAudioFileSequence(audioFileSequences.iterator().next());
		}
	}

	@Override
	public long getMaxFrame() {
		AudioFileSequence activeAudioFileSequence = getActiveAudioFileSequence();
		if (activeAudioFileSequence != null) {
			return activeAudioFileSequence.getMaxFrame();
		}
		return -1;
	}

	@Override
	public boolean isAudioDataAvailable() {
		Collection<AudioFileSequence> sequences = getAudioFileSequences();
		if (sequences.isEmpty()) {
			return false;
		}
		for (AudioFileSequence audioFileSequence : sequences) {
			if (!audioFileSequence.allAudioFilesPresent()) {
				return false;
			}
		}
		return true;
	}

	Collection<AudioFileSequence> getAudioFileSequences() {
		Collection<AudioFileSequence> ret = new LinkedHashSet<AudioFileSequence>();
		for (AudioFileSequence seq : audioFileSequences) {
			if (!seq.isMarkedAsDeleted()) {
				ret.add(seq);
			}
		}
		return Collections.unmodifiableCollection(ret);
	}

	private void fireAudioFileSequenceAdded(AudioFileSequence sequence) {
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.audioFileSequenceAdded(sequence);
			}
		}
	}

	private void fireAudioFileSequenceRemoved(AudioFileSequence sequence) {
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.audioFileSequenceRemoved(sequence);
			}
		}
	}

	private void fireAudioFileSequenceActivated(AudioFileSequence sequence) {
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.audioFileSequenceActivated(sequence);
			}
		}
	}

	private void fireAudioFileSequenceUpdated(AudioFileSequence sequence) {
		notifyProjectForChanges();
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.audioFileSequenceUpdated(sequence);
			}
		}
	}

	private void fireAudioFileFrameOffsetChanged(AudioFile audioFile, long newFrameOffset, long oldFrameOffset) {
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.audioFileFrameOffsetChanged(audioFile, newFrameOffset, oldFrameOffset);
			}
		}
	}

	private void notifyProjectForChanges() {
		audioProject.changeFrameLength();
	}

	private void fireNewAudioFileAdded(AudioFile audioFile) {
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.newAudioFileAdded(audioFile);
			}
		}
	}

	private void fireAudioFileDeleted(AudioFile audioFile) {
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.audioFileDeleted(audioFile);
			}
		}
	}

	private void fireAudioFileSequenceDeleted(AudioFileSequence sequence) {
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.audioFileSequenceDeleted(sequence);
			}
		}
	}

	void addRecordingAudioDeviceListener(RecordingAudioDeviceListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<RecordingAudioDeviceListener>();
		}
		listeners.add(listener);
	}

	void removeRecordingAudioDeviceListener(RecordingAudioDeviceListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	AudioFormat getAudioFormat() {
		int sampleRate = audioProject.getSampleRate();
		int sampleSizeInBits = audioProject.getSampleSizeInBits();
		int channelCount = audioChannels.getChannelCount();
		int frameSizeInBytes = audioProject.getFrameSizeInBytes();
		int frameRate = audioProject.getFrameRate();
		boolean bigEndian = audioProject.isBigEndian();
		AudioFormat ret = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, sampleRate, sampleSizeInBits, channelCount, frameSizeInBytes, frameRate, bigEndian);
		return ret;
	}

	@Override
	public boolean isRecordable() {
		return true;
	}

	AudioChannels getAudioChannels() {
		return audioChannels;
	}

	void setAudioChannels(AudioChannels audioChannels) {
		AudioChannels oldAudioChannels = this.audioChannels;
		if (oldAudioChannels != audioChannels) {
			this.audioChannels = audioChannels;
			fireAudioChannelsChanged(oldAudioChannels, audioChannels);
			firePersistableAudioDevicePropertyChanged();
		}
	}

	private void fireAudioChannelsChanged(AudioChannels oldAudioChannels, AudioChannels newAudioChannels) {
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.audioChannelsChanged(oldAudioChannels, newAudioChannels);
			}
		}
	}

	AudioFileSequenceCreationPolicy getSequenceCreationPolicy() {
		return sequenceCreationPolicy;
	}

	void setSequenceCreationPolicy(AudioFileSequenceCreationPolicy sequenceCreationPolicy) {
		AudioFileSequenceCreationPolicy oldAudioFileSequenceCreationPolicy = this.sequenceCreationPolicy;
		if(oldAudioFileSequenceCreationPolicy!=sequenceCreationPolicy) {
			this.sequenceCreationPolicy = sequenceCreationPolicy;
			fireSequenceCreationPolicyChanged(oldAudioFileSequenceCreationPolicy, sequenceCreationPolicy);
			firePersistableAudioDevicePropertyChanged();
		}
	}
	
	private void fireSequenceCreationPolicyChanged(AudioFileSequenceCreationPolicy oldPolicy, AudioFileSequenceCreationPolicy newPolicy) {
		if (listeners != null) {
			for (RecordingAudioDeviceListener listener : listeners) {
				listener.audioFileSequenceCreationPolicyChanged(oldPolicy, newPolicy);
			}
		}
	}

	@Override
	public Performable getPerformable() throws AudioIOException {
		final Recording ret = new Recording(audioProject);
		ret.addPerformanceListener(new PerformanceListener() {
			long lastUpdate;

			@Override
			public void performanceUpdated() {
				long now = System.currentTimeMillis();
				if (now - 1000 > lastUpdate) {
					fireAudioFileSequenceUpdated(findAudioFileSequence(ret.getRecordingAudioFile()));
					lastUpdate = now;
				}
			}

			@Override
			public void performanceStopped() {
				fireAudioFileSequenceUpdated(findAudioFileSequence(ret.getRecordingAudioFile()));
			}

			@Override
			public void performanceStarted() {
				AudioFileSequence recordingAudioFileSequence = ret.getRecordingAudioFileSequence();
				setActiveAudioFileSequence(recordingAudioFileSequence);
				fireAudioFileSequenceUpdated(recordingAudioFileSequence);
			}

			@Override
			public void performanceFinished() {
				ret.removePerformanceListener(this);
				AudioFile recordingAudioFile = ret.getRecordingAudioFile();
				if (recordingAudioFile != null) {
					// only if recording has happened, it was otherwise
					// capturing
					fireAudioFileSequenceUpdated(findAudioFileSequence(recordingAudioFile));
					fireBinaryChanged(recordingAudioFile);
				}
			}
		});
		return ret;
	}

	private AudioFileSequence findAudioFileSequence(AudioFile audioFile) {
		for (AudioFileSequence seq : getAudioFileSequences()) {
			for (AudioFile file : seq.getAudioFiles()) {
				if (file.equals(audioFile)) {
					return seq;
				}
			}
		}
		return null;
	}

	@Override
	public String getName() {
		return DEVICE_NAME;
	}

	protected RecordingAudioDeviceProjectProperties getProjectProperties() {
		return audioProject.getAudioDeviceProperties(getClass(), RecordingAudioDeviceProjectProperties.class);
	}

	private AudioFileSequence findAudioFileSequenceByIdentifier(String identifier) {
		for (AudioFileSequence seq : audioFileSequences) {
			if (seq.getIdentifier().equals(identifier)) {
				return seq;
			}
		}
		return null;
	}

	@Override
	public void synchronizeBy(Synchronizable source) {
		// sync audio file sequences
		if (source instanceof RecordingAudioDevice) {
			RecordingAudioDevice sourceDevice = (RecordingAudioDevice) source;
			setSequenceCreationPolicy(sourceDevice.getSequenceCreationPolicy());
			setAudioChannels(sourceDevice.getAudioChannels());
			Collection<AudioFileSequence> sourceAudioFileSequences = sourceDevice.getAudioFileSequences();
			Collection<AudioFileSequence> audioFileSequences = getAudioFileSequences();

			for (AudioFileSequence sourceAudioFileSequence : sourceAudioFileSequences) {
				String sourceAudioFileSequenceIdentifier = sourceAudioFileSequence.getIdentifier();
				if (isDetached(sourceAudioFileSequence) || isDeleted(sourceAudioFileSequence)) {
					// ignore this sequence on this side
					continue;
				}
				AudioFileSequence audioFileSequence = findAudioFileSequenceByIdentifier(sourceAudioFileSequenceIdentifier);
				if (audioFileSequence == null) {
					audioFileSequence = new AudioFileSequence();
					audioFileSequence.setIdentifier(sourceAudioFileSequenceIdentifier);
					synchronizeAudioFileSequence(sourceAudioFileSequence, audioFileSequence);
					if (audioFileSequence.allAudioFilesPresent()) {
						addAudioFileSequence(audioFileSequence);
						syncSequenceActivation(sourceDevice, sourceAudioFileSequence, audioFileSequence);
					}
				} else {
					synchronizeAudioFileSequence(sourceAudioFileSequence, audioFileSequence);
					syncSequenceActivation(sourceDevice, sourceAudioFileSequence, audioFileSequence);
				}
			}

			for (AudioFileSequence audioFileSequence : new LinkedHashSet<AudioFileSequence>(audioFileSequences)) {
				if (sourceDevice.isDetached(audioFileSequence)) {
					detachAudioFileSequence(audioFileSequence);
				} else if (sourceDevice.isDeleted(audioFileSequence)) {
					deleteAudioFileSequence(audioFileSequence);
				}
			}
		}
	}

	private void synchronizeAudioFileSequence(AudioFileSequence sourceAudioFileSequence, AudioFileSequence targetAudioFileSequence) {
		if (!targetAudioFileSequence.getIdentifier().equals(sourceAudioFileSequence.getIdentifier())) {
			return;
		}
		Collection<AudioFile> sourceAudioFiles = sourceAudioFileSequence.getAudioFiles();
		Collection<AudioFile> audioFiles = targetAudioFileSequence.getAudioFiles();
		for (AudioFile sourceAudioFile : sourceAudioFiles) {
			if (targetAudioFileSequence.isRemoved(sourceAudioFile)) {
				// ignore this sequence on this side
				continue;
			}
			String sourceIdentifier = sourceAudioFile.getIdentifier();
			AudioFile audioFile = targetAudioFileSequence.findAudioFileByIdentifier(sourceIdentifier);
			if (audioFile == null) {
				audioFile = new AudioFile(audioProject, sourceAudioFile.getName(), sourceAudioFile.getFrameOffset());
				audioFile.setIdentifier(sourceIdentifier);
				synchronizeAudioFile(sourceAudioFile, audioFile);
				// only add new files, if the binary sync has supplied the
				// corresponding file
				if (audioFile.getFile().exists()) {
					targetAudioFileSequence.addAudioFile(audioFile);
					fireNewAudioFileAdded(audioFile);
				}
			} else {
				synchronizeAudioFile(sourceAudioFile, audioFile);
			}
		}
		for (AudioFile audioFile : new LinkedHashSet<AudioFile>(audioFiles)) {
			if (sourceAudioFileSequence.isRemoved(audioFile)) {
				targetAudioFileSequence.removeAudioFile(audioFile);
			}
		}
		fireAudioFileSequenceUpdated(targetAudioFileSequence);
	}

	private void synchronizeAudioFile(AudioFile source, AudioFile target) {
		if (target.getIdentifier().equals(source.getIdentifier())) {
			target.setTimestamp(source.getTimestamp());
			long frameOffsetChangeTime = source.getFrameOffsetChangeTime();
			if (greaterOrBothZero(frameOffsetChangeTime, target.getFrameOffsetChangeTime())) {
				setFrameOffset(target, source.getFrameOffset());
				target.setFrameOffsetChangeTime(frameOffsetChangeTime);
			}
			long latencyFrameOffsetChangeTime = source.getLatencyFrameOffsetChangeTime();
			if (greaterOrBothZero(latencyFrameOffsetChangeTime, target.getLatencyFrameOffsetChangeTime())) {
				target.setLatencyFrameOffset(source.getLatencyFrameOffset());
				target.setLatencyFrameOffsetChangeTime(latencyFrameOffsetChangeTime);
			}
		}
	}

	private void syncSequenceActivation(RecordingAudioDevice sourceDevice, AudioFileSequence sourceAudioFileSequence, AudioFileSequence audioFileSequence) {
		if (greaterOrBothZero(sourceDevice.lastAudioFileSequenceActivationChangeTime, lastAudioFileSequenceActivationChangeTime) && sourceAudioFileSequence.isActive()) {
			setActiveAudioFileSequence(audioFileSequence);
			lastAudioFileSequenceActivationChangeTime = sourceDevice.lastAudioFileSequenceActivationChangeTime;
		}
	}

	@Override
	public Collection<BinaryDataset> getAllBinaryDatasets() {
		Collection<BinaryDataset> ret = new LinkedHashSet<BinaryDataset>();
		Collection<AudioFileSequence> audioFileSequences = getAudioFileSequences();
		for (AudioFileSequence audioFileSequence : audioFileSequences) {
			Collection<AudioFile> audioFiles = audioFileSequence.getAudioFiles();
			for (AudioFile audioFile : audioFiles) {
				if (audioFile.getFile().exists()) {
					ret.add(audioFile);
				}
			}
		}
		return ret;
	}
}
