/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import static javax.swing.SwingConstants.*;
import static org.audiochain.devices.recording.RecordingAudioDevice.*;
import static org.audiochain.io.RawFileAudioDataReader.*;
import static org.audiochain.ui.gui.CommonComponentFactory.*;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import java.io.IOException;
import java.text.DateFormat;
import java.util.Collection;
import java.util.Date;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.filechooser.FileFilter;

import org.audiochain.model.AudioChannels;
import org.audiochain.model.AudioProjectAdapter;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.gui.ImageButtonComponent;

public class RecordingAudioDeviceComponent extends Box {
	private static final long serialVersionUID = 1L;
	private final RecordingAudioDevice device;
	private final DateFormat df = DateFormat.getDateInstance();
	private final DateFormat tf = DateFormat.getTimeInstance();

	public RecordingAudioDeviceComponent(RecordingAudioDevice device) {
		super(BoxLayout.Y_AXIS);
		this.device = device;
		init();
	}

	private void init() {
		add(createToolbar());
		add(createSequenceInfoBox());
		add(createSequenceCombobox());
	}

	private Box createToolbar() {
		Box toolbarBox = Box.createHorizontalBox();
		toolbarBox.add(Box.createHorizontalGlue());
		createSequenceCreationPolicyButtons(toolbarBox);
		createDeleteUnusedSequencesButton(toolbarBox);
		createDeleteSequenceButton(toolbarBox);
		createAddSequenceButton(toolbarBox);
		createRemoveSequenceButton(toolbarBox);
		createChannelButtons(toolbarBox);
		return toolbarBox;
	}

	private void createChannelButtons(Box toolbarBox) {
		final ImageButtonComponent monoButton = new ImageButtonComponent("org/audiochain/ui/gui/mono-item.png");
		monoButton.setToolTipText("New files will be recorded in Mono");
		monoButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		final ImageButtonComponent stereoButton = new ImageButtonComponent("org/audiochain/ui/gui/stereo-item.png");
		stereoButton.setToolTipText("New files will be recorded in Stereo");
		stereoButton.setReleasedBorder(BorderFactory.createEmptyBorder());

		device.addRecordingAudioDeviceListener(new RecordingAudioDeviceAdapter() {
			@Override
			public void audioChannelsChanged(AudioChannels oldAudioChannels, AudioChannels newAudioChannels) {
				monoButton.setVisible(newAudioChannels == AudioChannels.Mono);
				stereoButton.setVisible(newAudioChannels == AudioChannels.Stereo);
			}
		});

		MouseListener channelsMouseListener = new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				device.setAudioChannels(device.getAudioChannels().getToggle());
			}
		};
		monoButton.addMouseListener(channelsMouseListener);
		stereoButton.addMouseListener(channelsMouseListener);
		toolbarBox.add(monoButton);
		toolbarBox.add(stereoButton);
		AudioChannels audioChannels = device.getAudioChannels();
		monoButton.setVisible(audioChannels == AudioChannels.Mono);
		stereoButton.setVisible(audioChannels == AudioChannels.Stereo);
	}

	private void createAddSequenceButton(Box toolbarBox) {
		final ImageButtonComponent addButton = new ImageButtonComponent("org/audiochain/devices/recording/add-file-item.png");
		addButton.setToolTipText("Add Audio File");
		addButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		addButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				String path = device.getProjectProperties().getLastAddedAudioFilePath();
				if (path == null) {
					path = device.getAudioProject().getProjectPath();
				}
				JFileChooser jfc = new JFileChooser(new File(path));
				jfc.setFileFilter(new FileFilter() {
					@Override
					public boolean accept(File f) {
						if (f.isDirectory()) {
							return true;
						}
						/* only add the files, that are not associated with this device yet */
						Collection<AudioFileSequence> audioFileSequences = device.getAudioFileSequences();
						for (AudioFileSequence audioFileSequence : audioFileSequences) {
							Collection<AudioFile> audioFiles = audioFileSequence.getAudioFiles();
							for (AudioFile audioFile : audioFiles) {
								if (audioFile.getFile().getAbsolutePath().equals(f.getAbsolutePath())) {
									return false;
								}
							}
						}
						return f.isFile() && f.getName().endsWith(FILE_NAME_ENDING);
					}

					@Override
					public String getDescription() {
						return "*" + FILE_NAME_ENDING;
					}
				});
				int status = jfc.showOpenDialog(null);
				if (status == JFileChooser.APPROVE_OPTION) {
					File file = jfc.getSelectedFile();
					AudioFileSequence sequence = device.addAudioFile(file);
					device.setActiveAudioFileSequence(sequence);
					device.getProjectProperties().setLastAddedAudioFilePath(file.getParentFile().getAbsolutePath());
				}
			}
		});
		toolbarBox.add(addButton);
	}

	private void createRemoveSequenceButton(Box toolbarBox) {
		final ImageButtonComponent removeButton = new ImageButtonComponent("org/audiochain/devices/recording/remove-file-item.png");
		removeButton.setToolTipText("Remove Audio File from Recording Device without deleting the file on disk");
		removeButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		removeButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				AudioFileSequence activeAudioFileSequence = device.getActiveAudioFileSequence();
				if (activeAudioFileSequence != null) {
					device.detachAudioFileSequence(activeAudioFileSequence);
					device.setNextAudioFileSequenceActive();
				}
			}
		});
		toolbarBox.add(removeButton);
	}

	private void createSequenceCreationPolicyButtons(Box toolbarBox) {
		final ImageButtonComponent singleFileButton = new ImageButtonComponent("org/audiochain/devices/recording/file-item.png");
		singleFileButton.setToolTipText("Create files in new sequences");
		singleFileButton.setReleasedBorder(BorderFactory.createEmptyBorder());

		final ImageButtonComponent sameSequenceButton = new ImageButtonComponent("org/audiochain/devices/recording/sequence-item.png");
		sameSequenceButton.setToolTipText("Create files within the same sequence");
		sameSequenceButton.setReleasedBorder(BorderFactory.createEmptyBorder());

		device.addRecordingAudioDeviceListener(new RecordingAudioDeviceAdapter() {
			@Override
			public void audioFileSequenceCreationPolicyChanged(AudioFileSequenceCreationPolicy oldPolicy, AudioFileSequenceCreationPolicy newPolicy) {
				singleFileButton.setVisible(newPolicy == AudioFileSequenceCreationPolicy.NewAudioFileSequence);
				sameSequenceButton.setVisible(newPolicy == AudioFileSequenceCreationPolicy.SameAudioFileSequence);
			}
		});

		MouseListener mouseListener = new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				device.setSequenceCreationPolicy(device.getSequenceCreationPolicy().getToggle());
			}
		};
		singleFileButton.addMouseListener(mouseListener);
		sameSequenceButton.addMouseListener(mouseListener);
		toolbarBox.add(singleFileButton);
		toolbarBox.add(sameSequenceButton);
		singleFileButton.setVisible(device.getSequenceCreationPolicy() == AudioFileSequenceCreationPolicy.NewAudioFileSequence);
		sameSequenceButton.setVisible(device.getSequenceCreationPolicy() == AudioFileSequenceCreationPolicy.SameAudioFileSequence);
	}

	private void createDeleteSequenceButton(Box toolbarBox) {
		final ImageButtonComponent deleteButton = new ImageButtonComponent("org/audiochain/devices/recording/trash-item.png");
		deleteButton.setToolTipText("Delete sequence");
		deleteButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		deleteButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				AudioFileSequence audioFileSequence = device.getActiveAudioFileSequence();
				if (audioFileSequence != null) {
					if (Dialog.okCancel(RecordingAudioDeviceComponent.this, "Delete Audio Data Sequence", "The Audio Data Sequence '" + audioFileSequence
							+ "' will be removed and all audio files that are associated with this sequence will be deleted. Please ensure, that these "
							+ "audio files are not associated with any other tracks after deletion!")) {
						device.deleteAudioFileSequence(audioFileSequence);
						device.setNextAudioFileSequenceActive();
					}
				}
			}
		});
		toolbarBox.add(deleteButton);
	}

	private void createDeleteUnusedSequencesButton(Box toolbarBox) {
		final ImageButtonComponent removeUnusedButton = new ImageButtonComponent("org/audiochain/devices/recording/rubber-item.png");
		removeUnusedButton.setToolTipText("Delete unused sequences");
		removeUnusedButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		removeUnusedButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				Collection<AudioFileSequence> unusedAudioFileSequences = device.getUnusedAudioFileSequences();
				if (!unusedAudioFileSequences.isEmpty()) {
					if (Dialog.okCancel(RecordingAudioDeviceComponent.this, "Remove Unused Audio Data Sequences", "The Audio Data Sequences '" + unusedAudioFileSequences
							+ "' will be removed. All audio files that are associated with this sequence will be deleted.")) {
						device.deleteAudioFileSequences(unusedAudioFileSequences);
					}
				}
			}
		});
		toolbarBox.add(removeUnusedButton);
	}

	private JComboBox createSequenceCombobox() {
		RecordingAudioDeviceSequenceComboboxModel sequenceComboboxModel = new RecordingAudioDeviceSequenceComboboxModel(device);
		device.addRecordingAudioDeviceListener(sequenceComboboxModel);
		final JComboBox sequenceCombobox = createComboBox(sequenceComboboxModel);
		RecordingAudioDeviceListener listener = new RecordingAudioDeviceAdapter() {
			@Override
			public void audioFileSequenceActivated(AudioFileSequence sequence) {
				updateTooltip();
			}

			@Override
			public void audioFileSequenceUpdated(AudioFileSequence sequence) {
				updateTooltip();
			}

			@Override
			public void audioFileSequenceAdded(AudioFileSequence sequence) {
				updateTooltip();
			}

			@Override
			public void audioFileSequenceRemoved(AudioFileSequence sequence) {
				updateTooltip();
			}

			@Override
			public void audioFileFrameOffsetChanged(AudioFile audioFile, long newFrameOffset, long oldFrameOffset) {
				updateTooltip();
			}

			private void updateTooltip() {
				AudioFileSequence activeAudioFileSequence = device.getActiveAudioFileSequence();
				if (activeAudioFileSequence != null) {
					sequenceCombobox.setToolTipText(activeAudioFileSequence.toString().replace(AudioChannels.Mono.getSymbol(), 'M').replace(AudioChannels.Stereo.getSymbol(), 'S'));
				} else {
					sequenceCombobox.setToolTipText(null);
				}
			}
		};
		device.addRecordingAudioDeviceListener(listener);
		listener.audioFileSequenceActivated(null);

		return sequenceCombobox;
	}

	private Component createSequenceInfoBox() {
		Box infoBox = Box.createVerticalBox();
		final JLabel dateLabel = new JLabel();
		dateLabel.setAlignmentX(CENTER_ALIGNMENT);
		infoBox.add(dateLabel);
		final JLabel timeLabel = new JLabel();
		timeLabel.setAlignmentX(CENTER_ALIGNMENT);
		infoBox.add(timeLabel);
		final JLabel sizeLabel = new JLabel();
		sizeLabel.setAlignmentX(CENTER_ALIGNMENT);
		infoBox.add(sizeLabel);
		final JTextField frameOffsetField = createFrameOffsetField();
		RecordingAudioDeviceListener updateLabelListener = new RecordingAudioDeviceAdapter() {
			@Override
			public void audioFileSequenceActivated(AudioFileSequence sequence) {
				updateLabels();
			}

			@Override
			public void audioFileSequenceAdded(AudioFileSequence sequence) {
				updateLabels();
			}

			@Override
			public void audioFileSequenceUpdated(AudioFileSequence sequence) {
				updateLabels();
			}

			@Override
			public void audioFileSequenceRemoved(AudioFileSequence sequence) {
				updateLabels();
			}
			
			@Override
			public void audioFileFrameOffsetChanged(AudioFile audioFile, long newFrameOffset, long oldFrameOffset) {
				updateFrameOffsetField(frameOffsetField);
			}

			private void updateLabels() {
				AudioFileSequence activeAudioFileSequence = device.getActiveAudioFileSequence();
				if (activeAudioFileSequence != null) {
					Date date = new Date(activeAudioFileSequence.getLatestTimestamp());
					dateLabel.setText(df.format(date));
					timeLabel.setText(tf.format(date));
					sizeLabel.setText(AudioFile.formatMegabytes(activeAudioFileSequence.getByteLength()));
					updateFrameOffsetField(frameOffsetField);
				} else {
					// set a space to force visibility of the empty label on gui
					dateLabel.setText(" ");
					timeLabel.setText(" ");
					sizeLabel.setText(" ");
					frameOffsetField.setText(" ");
				}
			}
		};
		device.addRecordingAudioDeviceListener(updateLabelListener);
		// to initialize the labels at startup
		updateLabelListener.audioFileSequenceActivated(null);
		device.getAudioProject().addAudioProjectChangeListener(new AudioProjectAdapter() {
			@Override
			public void audioProjectFramePositionChanged(long oldValue, long newValue) throws IOException {
				updateFrameOffsetField(frameOffsetField);
			}
		});
		final ImageButtonComponent incrementFrameOffsetButton = new ImageButtonComponent("org/audiochain/devices/recording/arrowright-item.png");
		incrementFrameOffsetButton.setToolTipText("Increment The Audible File's Frame Offset By " + FRAME_OFFSET_INCREMENT_IN_MILLIS + "ms");
		incrementFrameOffsetButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		incrementFrameOffsetButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				device.incrementAudibleFileFrameOffset();
			}
		});

		final ImageButtonComponent decrementFrameOffsetButton = new ImageButtonComponent("org/audiochain/devices/recording/arrowleft-item.png");
		decrementFrameOffsetButton.setToolTipText("Decrement The Audible File's Frame Offset By " + FRAME_OFFSET_INCREMENT_IN_MILLIS + "ms");
		decrementFrameOffsetButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		decrementFrameOffsetButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				device.decrementAudibleFileFrameOffset();
			}
		});
		Box frameOffsetBox = Box.createHorizontalBox();
		infoBox.add(frameOffsetBox);
		frameOffsetBox.add(decrementFrameOffsetButton);
		frameOffsetBox.add(frameOffsetField);
		frameOffsetBox.add(incrementFrameOffsetButton);
		return infoBox;
	}

	private JTextField createFrameOffsetField() {
		final JTextField field = new JTextField();
		field.setToolTipText("Frame Offset Of The Currently Audible File");
		field.setAlignmentX(CENTER_ALIGNMENT);
		field.setHorizontalAlignment(CENTER);
		Dimension dim = new Dimension(field.getPreferredSize());
		dim.width = Integer.MAX_VALUE;
		field.setMaximumSize(dim);
		field.setBorder(null);
		field.setOpaque(false);
		field.setFocusable(false);
		field.addKeyListener(new KeyAdapter() {
			@Override
			public void keyReleased(KeyEvent event) {
				if (event.getKeyCode() == 10) {
					String text = field.getText();
					if (text != null && !(text = text.trim()).isEmpty()) {
						try {
							long frameOffset = Long.parseLong(text);
							device.setAudibleFileFrameOffset(frameOffset);
						} catch (NumberFormatException ex) {
							Dialog.message(RecordingAudioDeviceComponent.this, "Wrong Frame Offset Value", "The Value '" + text + "' is not a valid frame offset. Please use numbers only!");
						}
					}
					field.setFocusable(false);
				}
			}
		});
		field.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				field.setFocusable(true);
				field.grabFocus();
			}
		});
		return field;
	}

	private void updateFrameOffsetField(final JTextField frameOffsetLabel) {
		AudioFileSequence activeAudioFileSequence = device.getActiveAudioFileSequence();
		if (activeAudioFileSequence != null) {
			AudioFile audibleFileAtFrame = activeAudioFileSequence.getAudibleFileAtFrame(device.getAudioProject().getFramePosition());
			frameOffsetLabel.setText(audibleFileAtFrame == null ? " " : String.valueOf(audibleFileAtFrame.getFrameOffset()));
		}
	}
}
