/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import static org.audiochain.devices.recording.AudioFile.*;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.swing.ComboBoxModel;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;


public class RecordingAudioDeviceSequenceComboboxModel extends RecordingAudioDeviceAdapter implements ComboBoxModel {

	private final RecordingAudioDevice device;
	private Collection<ListDataListener> listDataListeners;
	private Map<AudioFileSequence, FileComboboxItem> sequenceItemMap;
	private final float frameRate;

	public RecordingAudioDeviceSequenceComboboxModel(RecordingAudioDevice device) {
		this.device = device;
		frameRate = device.getAudioFormat().getFrameRate();
	}

	@Override
	public int getSize() {
		return device.getAudioFileSequences().size();
	}

	@Override
	public Object getElementAt(int index) {
		AudioFileSequence[] array = device.getAudioFileSequences().toArray(new AudioFileSequence[0]);
		int len = array.length;
		if (index >= 0 && index < len) {
			AudioFileSequence audioFileSequence = array[index];
			return getFileComboboxItem(audioFileSequence);
		}
		return null;
	}

	@Override
	public void addListDataListener(ListDataListener listener) {
		if (listDataListeners == null) {
			listDataListeners = new LinkedHashSet<ListDataListener>();
		}
		listDataListeners.add(listener);
	}

	@Override
	public void removeListDataListener(ListDataListener listener) {
		if (listDataListeners == null) {
			return;
		}
		listDataListeners.remove(listener);
	}

	@Override
	public void setSelectedItem(Object item) {
		if (sequenceItemMap == null) {
			return;
		}
		Set<Entry<AudioFileSequence, FileComboboxItem>> entrySet = sequenceItemMap.entrySet();
		for (Entry<AudioFileSequence, FileComboboxItem> entry : entrySet) {
			if (item.equals(entry.getValue())) {
				device.setActiveAudioFileSequence(entry.getKey());
			}
		}
	}

	@Override
	public Object getSelectedItem() {
		return getFileComboboxItem(device.getActiveAudioFileSequence());
	}

	@Override
	public void audioFileSequenceAdded(AudioFileSequence sequence) {
		fireChange();
	}

	@Override
	public void audioFileSequenceRemoved(AudioFileSequence sequence) {
		fireChange();
	}

	@Override
	public void audioFileSequenceActivated(AudioFileSequence sequence) {
		fireChange();
	}

	@Override
	public void audioFileSequenceUpdated(AudioFileSequence sequence) {
		fireChange();
	}

	@Override
	public void audioFileFrameOffsetChanged(AudioFile audioFile, long newFrameOffset, long oldFrameOffset) {
		fireChange();
	}

	private void fireChange() {
		if (listDataListeners != null) {
			ListDataEvent event = new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, 0, getSize());
			for (ListDataListener listener : listDataListeners) {
				listener.contentsChanged(event);
			}
		}
	}

	private FileComboboxItem getFileComboboxItem(AudioFileSequence audioFileSequence) {
		if (audioFileSequence == null) {
			return null;
		}
		if (sequenceItemMap == null) {
			sequenceItemMap = new LinkedHashMap<AudioFileSequence, RecordingAudioDeviceSequenceComboboxModel.FileComboboxItem>();
		}
		FileComboboxItem ret = sequenceItemMap.get(audioFileSequence);
		if (ret == null) {
			ret = new FileComboboxItem(audioFileSequence);
			sequenceItemMap.put(audioFileSequence, ret);
		}
		return ret;
	}

	private class FileComboboxItem {
		final AudioFileSequence sequence;

		FileComboboxItem(AudioFileSequence sequence) {
			this.sequence = sequence;
		}

		@Override
		public String toString() {
			long frames = sequence.getFrameLength();
			if (frames == -1) {
				frames = 0;
			}
			float sec = frames / frameRate;
			return formatTime((long) sec);
		}
	}
}
