/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;

import org.audiochain.model.AbstractCommandLineUserInterface;
import org.audiochain.model.AudioChannels;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.SourceAudioDevice;

class RecordingCommandLineUserInterface extends AbstractCommandLineUserInterface {
	private static enum Command {
		AudioChannel(AudioChannels.values()),
		SequenceCreationPolicy(AudioFileSequenceCreationPolicy.values()),
		PrevSeq("Activate the previous audio file of the selected track."), 
		NextSeq("Activate the next audio file of the selected track."),
		AddFile("<filename>", "Add the <filename> to the selected track."),
		RemoveFile("<filename>", "Remove the <filename> from the selected track and the file system."),
		Latency("<frames>", "Set the latency for the selected audio file of the selected track to <frames>."),
		RemoveActiveFile("Remove the active audio file from the selected track and the file system."),
		MoveFile("<number>", "Move the active audio file from the selected track to track <number>."),
		ConvertFile("Converts the audio format of the active audio data sequence of the selected track to the track's current audio format and the raw file format."),
		IncrementFrameOffset("Increments the frame offset of the audible file at the current frame position."),
		DecrementFrameOffset("Decrements the frame offset of the audible file at the current frame position."),
		RemoveUnused("Removes all audio data sequences and files, that are not marked as the active sequence of a track.");
		
		private final String help;
		private String param;
		
		private Command(String help) {
			this.help = help;
		}
		
		private Command(Object[] values) {
			this.help = String.valueOf(Arrays.asList(values));
		}
		
		private Command(String param, String help) {
			this(help);
			this.param = param;
		}
		
		static Command parse(String str) {
			for (Command c : values()) {
				if (c.name().equalsIgnoreCase(str)) {
					return c;
				}
			}
			return null;
		}
	}
	
	private final RecordingAudioDevice recordingAudioDevice;
	private String audioFileSequenceStringRepresentation;

	RecordingCommandLineUserInterface(RecordingAudioDevice recordingAudioDevice) {
		this.recordingAudioDevice = recordingAudioDevice;
		recordingAudioDevice.addRecordingAudioDeviceListener(new RecordingAudioDeviceAdapter(){
			@Override
			public void audioFileSequenceAdded(AudioFileSequence sequence) {
				createAudioFileSequenceStringRepresentation();
			}

			@Override
			public void audioFileSequenceRemoved(AudioFileSequence sequence) {
				createAudioFileSequenceStringRepresentation();
			}

			@Override
			public void audioFileSequenceActivated(AudioFileSequence sequence) {
				createAudioFileSequenceStringRepresentation();
			}

			@Override
			public void audioFileSequenceUpdated(AudioFileSequence sequence) {
				createAudioFileSequenceStringRepresentation();
			}

			@Override
			public void audioFileFrameOffsetChanged(AudioFile audioFile, long oldFrameOffset, long newFrameOffset) {
				createAudioFileSequenceStringRepresentation();
			}
		});
		createAudioFileSequenceStringRepresentation();
	}

	@Override
	public String getDisplayString() {
		StringBuilder sb = new StringBuilder();
		sb.append(recordingAudioDevice.getAudioChannels().getSymbol());
		sb.append(' ');
		sb.append(recordingAudioDevice.getSequenceCreationPolicy().getSymbol());
		sb.append(' ');
		sb.append(audioFileSequenceStringRepresentation);
		return sb.toString();
	}

	private void createAudioFileSequenceStringRepresentation() {
		StringBuilder sb = new StringBuilder();
		Collection<AudioFileSequence> audioDataSequences = recordingAudioDevice.getAudioFileSequences();
		String linesep = System.getProperty("line.separator");
		if (!audioDataSequences.isEmpty()) {
			sb.append(" " + (char) 9991 + " ");
			for (AudioFileSequence seq : audioDataSequences) {
				sb.append(linesep).append("        ");
				if (seq.isActive()) {
					sb.append("[" + seq + "] ");
				} else {
					sb.append(seq + " ");
				}
			}
		}
		audioFileSequenceStringRepresentation = sb.toString();
	}

	@Override
	public boolean modifyAudioDevice(String propertyName, String newValue) throws Exception {
		Command cmd = Command.parse(propertyName);
		if (cmd == null) {
			return false;
		}
		switch (cmd) {
		case AudioChannel:
			recordingAudioDevice.setAudioChannels(AudioChannels.valueOf(newValue));
			break;
		case SequenceCreationPolicy:
			recordingAudioDevice.setSequenceCreationPolicy(AudioFileSequenceCreationPolicy.valueOf(newValue));
			break;
		case PrevSeq:
			recordingAudioDevice.setPreviousAudioFileSequenceActive();
			break;
		case NextSeq:
			recordingAudioDevice.setNextAudioFileSequenceActive();
			break;
		case AddFile:
			AudioFileSequence sequence = recordingAudioDevice.addAudioFile(newValue);
			recordingAudioDevice.setActiveAudioFileSequence(sequence);
			break;
		case RemoveFile:
			recordingAudioDevice.deleteAudioFile(newValue);
			break;
		case Latency:
			long latency = Long.parseLong(newValue.trim());
			AudioFileSequence activeAudioFileSequence = recordingAudioDevice.getActiveAudioFileSequence();
			activeAudioFileSequence.setLatencyFrameOffset(latency);
			break;
		case RemoveActiveFile:
			recordingAudioDevice.deleteActiveAudioFileSequence();
			recordingAudioDevice.setLatestAudioFileSequenceActive();
			break;
		case MoveFile:
			int targetNumber = Integer.parseInt(newValue);
			moveActiveAudioFileToTrackWithNumber(targetNumber);
			break;
		case ConvertFile:
			convertActiveFile();
			break;
		case IncrementFrameOffset:
			recordingAudioDevice.incrementAudibleFileFrameOffset();
			break;
		case DecrementFrameOffset:
			recordingAudioDevice.decrementAudibleFileFrameOffset();
			break;
		case RemoveUnused:
			Collection<AudioFileSequence> unusedAudioFileSequences = recordingAudioDevice.getUnusedAudioFileSequences();
			recordingAudioDevice.deleteAudioFileSequences(unusedAudioFileSequences);
			break;
		default:
			return false;
		}

		return true;
	}

	@Override
	public String getHelpString() {
		StringBuilder sb = new StringBuilder();
		for (Command c : Command.values()) {
			if (c.param == null) {
				switch(c) {
				case AudioChannel:
					sb.append(format(c.name(), recordingAudioDevice.getAudioChannels().getSymbol() + " (" + recordingAudioDevice.getAudioChannels().name() + ')', c.help));
					break;
				case SequenceCreationPolicy:
					sb.append(format(c.name(), recordingAudioDevice.getSequenceCreationPolicy().getSymbol() + " (" + recordingAudioDevice.getSequenceCreationPolicy().name() + ')', c.help));
					break;
				default:
					sb.append(format(c.name(), c.help));
					break;
				}
			} else {
				sb.append(format(c.name() + " " + c.param, c.help));
			}
		}
		return sb.toString();
	}

	private void convertActiveFile() throws IOException {
		AudioFileSequence activeAudioFileSequence = recordingAudioDevice.getActiveAudioFileSequence();
		for (AudioFile audioFile : activeAudioFileSequence.getAudioFiles()) {
			audioFile.convert(recordingAudioDevice.getAudioFormat());
		}
	}

	private void moveActiveAudioFileToTrackWithNumber(int targetNumber) {
		AudioTrack[] allTracks = recordingAudioDevice.getAudioProject().getAudioTracks().toArray(new AudioTrack[0]);
		if (allTracks.length < targetNumber && targetNumber > 0) {
			SourceAudioDevice targetTrackSequenceDevice = allTracks[targetNumber - 1].getSourceAudioDevice();
			if (targetTrackSequenceDevice instanceof RecordingAudioDevice) {
				AudioFileSequence activeAudioFileSequence = recordingAudioDevice.getActiveAudioFileSequence();
				((RecordingAudioDevice) targetTrackSequenceDevice).addAudioFileSequence(activeAudioFileSequence);
				((RecordingAudioDevice) targetTrackSequenceDevice).setActiveAudioFileSequence(activeAudioFileSequence);
				recordingAudioDevice.detachAudioFileSequence(activeAudioFileSequence);
				recordingAudioDevice.setLatestAudioFileSequenceActive();
			}
		}
	}
}
