/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectLifecycleAdapter;

/**
 * This class watches for unsaved audio files that will be deleted from hard disk when the project becomes saved.
 * An audio file becomes marked for deletion when the user wishes to delete it, but the file will remain on hard drive until
 * the user stores the project.
 * The other case handled by this class is, when some audio files will be recorded, but the audio project does not become saved.
 * When the project will be closed without saving, all recorded files will be deleted by this class.
 */
public class UnsavedAudioFileAttendant extends AudioProjectLifecycleAdapter {
	private final RecordingAudioDevice device;
	private Collection<AudioFileSequence> deletedAudioFileSequences = new LinkedHashSet<AudioFileSequence>();
	private Collection<AudioFile> deletedAudioFiles = new LinkedHashSet<AudioFile>();
	// for recorded or synchronized audio files that are referenced in an
	// unsaved project
	private Collection<AudioFile> recordedButUnsavedAudioFiles = new LinkedHashSet<AudioFile>();
	
	UnsavedAudioFileAttendant(RecordingAudioDevice device) {
		this.device = device;
		device.addRecordingAudioDeviceListener(new RecordingAudioDeviceAdapter(){
			@Override
			public void newAudioFileAdded(AudioFile audioFile) {
				recordedButUnsavedAudioFiles.add(audioFile);
			}
			
			@Override
			public void audioFileDeleted(AudioFile audioFile) {
				deletedAudioFiles.add(audioFile);
			}
			
			@Override
			public void audioFileSequenceDeleted(AudioFileSequence sequence) {
				deletedAudioFileSequences.add(sequence);
			}
		});
	}

	@Override
	public void audioProjectStored(AudioProject audioProject) {
		for (AudioFileSequence deletableSequence : deletedAudioFileSequences) {
			deletableSequence.delete();
		}
		deletedAudioFileSequences.clear();
		for (AudioFile deletableFile : deletedAudioFiles) {
			deletableFile.delete();
		}
		deletedAudioFiles.clear();
		// all newly recorded files are now saved with the project file
		recordedButUnsavedAudioFiles.clear();
	}

	@Override
	public void audioProjectClosed(AudioProject audioProject) {
		// remove all newly recorded files, that are not saved with the project
		// file
		for (AudioFile recordedButUnsavedAudioFile : recordedButUnsavedAudioFiles) {
			removeAudioFileFromAudioFileSequence(recordedButUnsavedAudioFile);
			recordedButUnsavedAudioFile.markAsDeleted();
			recordedButUnsavedAudioFile.delete();
		}
	}

	private void removeAudioFileFromAudioFileSequence(AudioFile recordedButUnsavedAudioFile) {
		for (AudioFileSequence potentialParent : device.getAudioFileSequences()) {
			if (potentialParent.getAudioFiles().contains(recordedButUnsavedAudioFile)) {
				potentialParent.removeAudioFile(recordedButUnsavedAudioFile);
			}
		}
	}
}
