/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.reverb;

import java.util.Arrays;

/**
 * 
 * @author Florian Paul
 * @see https://ccrma.stanford.edu/~jos/pasp/Schroeder_Allpass_Sections.html
 * @see http://valhalladsp.wordpress.com/2009/05/30/schroeder-reverbs-the-forgotten-algorithm/
 */
public class AllPass {
	private float leftDecay;
	private float rightDecay;
	private StereoDelay delay;
	private int[] g1, g2;
	private Integer bufferLength;
	
	public AllPass(StereoDelay delay, float leftDecay, float rightDecay) {
		this.delay = delay;
		this.leftDecay = leftDecay;
		this.rightDecay = rightDecay;
	}

	public void allpass(int[] data) {
		int len = data.length;
		
		if (bufferLength==null) {
			bufferLength = len;
			// buffer initialization
			g1 = new int[bufferLength];
			g2 = new int[bufferLength];
		} else if (bufferLength < len) {
			// increase buffers lengths without data loss 
			int[] newBuffer = new int[len];
			System.arraycopy(g1, 0, newBuffer, 0, bufferLength);
			g1 = newBuffer;
			
			newBuffer = new int[len];
			System.arraycopy(g2, 0, newBuffer, 0, bufferLength);
			g2 = newBuffer;
			bufferLength = len;
		}
		
		for (int l = 0, r = 1; l < len; l += 2, r += 2) {
			g1[l] = (int) (data[l] + g2[l] * leftDecay);
			g1[r] = (int) (data[r] + g2[r] * rightDecay);
		}

		for (int l = 0, r = 1; l < len; l += 2, r += 2) {
			data[l] = (int) (g2[l] + g1[l] * -leftDecay);
			data[r] = (int) (g2[r] + g1[r] * -rightDecay);
		}

		// mix new g1 to g2 for the next iteration
		g2 = Arrays.copyOf(g1, bufferLength);
		delay.delay(g2);
	}

	public float getLeftDecay() {
		return leftDecay;
	}

	public void setLeftDecay(float leftDecay) {
		this.leftDecay = leftDecay;
	}

	public float getRightDecay() {
		return rightDecay;
	}

	public void setRightDecay(float rightDecay) {
		this.rightDecay = rightDecay;
	}

	public StereoDelay getDelay() {
		return delay;
	}

	public void setDelay(StereoDelay delay) {
		this.delay = delay;
	}
}
