/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.reverb;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;


public class ReverbAudioDataReader implements AudioDataReaderChainLink {
	static final float[] DELAY_TIME_MULTIPLIER = {
			0.997F, 0.757F, 0.557F, 0.433F, 0.353F, 0.197F, 0.61F
	/*
	 * 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67,
	 * 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149,
	 * 151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 199, 211, 223, 227,
	 * 229, 233, 239, 241, 251, 257, 263, 269, 271, 277, 281, 283, 293, 307,
	 * 311, 313, 317, 331, 337, 347, 349, 353, 359, 367, 373, 379, 383, 389,
	 * 397, 401, 409, 419, 421, 431, 433, 439, 443, 449, 457, 461, 463, 467,
	 * 479, 487, 491, 499, 503, 509, 521, 523, 541, 547, 557, 563, 569, 571,
	 * 577, 587, 593, 599, 601, 607, 613, 617, 619, 631, 641, 643, 647, 653,
	 * 659, 661, 673, 677, 683, 691, 701, 709, 719, 727, 733, 739, 743, 751,
	 * 757, 761, 769, 773, 787, 797, 809, 811, 821, 823, 827, 829, 839, 853,
	 * 857, 859, 863, 877, 881, 883, 887, 907, 911, 919, 929, 937, 941, 947,
	 * 953, 967, 971, 977, 983, 991, 997
	 */
	};
	private AudioDataReader reader;
	private final float frameRate;
	private int leftDelayTimeInMilliseconds;
	private int rightDelayTimeInMilliseconds;
	private float leftDecay;
	private float rightDecay;
	private boolean swapLeftDecayCoefficientSign = true;
	private boolean swapRightDecayCoefficientSign = true;
	private float dryRatio;
	private Collection<AllPass> allPasses;
	private int allPassCount;
	private boolean finishing;
	private boolean finished;

	public ReverbAudioDataReader(AudioDataReader reader, float frameRate) {
		this.reader = reader;
		this.frameRate = frameRate;
		initAllPasses();
	}

	@Override
	public int read(int[] data) throws IOException {
		if (finished) {
			return -1;
		}
		int len;
		if (finishing) {
			len = data.length;
			Arrays.fill(data, 0);
		} else {
			len = reader.read(data);
			if (len == -1) {
				finishing = true;
				len = data.length;
			}
		}

		int[] dryData = Arrays.copyOf(data, len);

		for (AllPass allPass : allPasses) {
			allPass.allpass(data);
		}

		for (int i = 0; i < len; i++) {
			data[i] = (int) ((1F - dryRatio) * data[i] + dryRatio * dryData[i]);
		}

		if (finishing) {
			// continue as long as the reverberation is still audible
			int maxSample = 0;
			for (int sample : data) {
				maxSample = Math.max(maxSample, Math.abs(sample));
			}
			if (maxSample < 10) {
				finished = true;
				return -1;
			}
		}

		return len;
	}

	private synchronized void initAllPasses() {
		Collection<AllPass> allPasses = new ArrayList<AllPass>();
		for (int i = 0; i < allPassCount; i++) {
			float multiplier = DELAY_TIME_MULTIPLIER[i];
			int leftDelayTime = (int) (leftDelayTimeInMilliseconds * multiplier);
			int rightDelayTime = (int) (rightDelayTimeInMilliseconds * multiplier);
			StereoDelay delay = new StereoDelay(frameRate, leftDelayTime, rightDelayTime);
			AllPass allPass = new AllPass(delay, leftDecay, rightDecay);
			allPasses.add(allPass);
		}
		this.allPasses = allPasses;
		setLeftDecay();
		setRightDecay();
	}

	private void setLeftDecay() {
		boolean toggle = false;
		for (AllPass allPass : allPasses) {
			toggle = !toggle;
			allPass.setLeftDecay(swapLeftDecayCoefficientSign ? (toggle ? -leftDecay : leftDecay) : leftDecay);
		}
	}

	private void setRightDecay() {
		boolean toggle = false;
		for (AllPass allPass : allPasses) {
			toggle = !toggle;
			allPass.setRightDecay(swapRightDecayCoefficientSign ? (toggle ? -rightDecay : rightDecay) : rightDecay);
		}
	}

	@Override
	public void close() throws IOException {
		reader.close();
		if (allPasses != null) {
			allPasses.clear();
		}
	}

	@Override
	public void stop() {
		finished = true;
		reader.stop();
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
		finishing = false;
		finished = false;
	}

	@Override
	public void setSource(AudioDataReader source) {
		reader = source;
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	public int getLeftDelayTimeInMilliseconds() {
		return leftDelayTimeInMilliseconds;
	}

	public void setLeftDelayTimeInMilliseconds(int leftDelayTimeInMilliseconds) {
		int newDelayTime = leftDelayTimeInMilliseconds;
		if (newDelayTime != this.leftDelayTimeInMilliseconds) {
			this.leftDelayTimeInMilliseconds = newDelayTime;
			initAllPasses();
		}
	}

	public void setRightDelayTimeInMilliseconds(int rightDelayTimeInMilliseconds) {
		int newDelayTime = rightDelayTimeInMilliseconds;
		if (newDelayTime != this.rightDelayTimeInMilliseconds) {
			this.rightDelayTimeInMilliseconds = newDelayTime;
			initAllPasses();
		}
	}

	public float getLeftDecay() {
		return leftDecay;
	}

	public void setLeftDecay(float leftDecay) {
		float newDecay = leftDecay;
		if (newDecay != this.leftDecay) {
			this.leftDecay = newDecay;
			setLeftDecay();
		}
	}

	public float getRightDecay() {
		return rightDecay;
	}

	public void setRightDecay(float rightDecay) {
		float newDecay = rightDecay;
		if (newDecay != this.rightDecay) {
			this.rightDecay = newDecay;
			setRightDecay();
		}
	}

	public int getRightDelayTimeInMilliseconds() {
		return rightDelayTimeInMilliseconds;
	}

	public int getAllPassCount() {
		return allPassCount;
	}

	public void setAllPassCount(int allPassCount) {
		int newCount = allPassCount;
		if (newCount != this.allPassCount) {
			if (newCount > DELAY_TIME_MULTIPLIER.length) {
				newCount = DELAY_TIME_MULTIPLIER.length;
			}
			this.allPassCount = newCount;
			initAllPasses();
		}
	}

	public boolean isSwapLeftDecayCoefficientSign() {
		return swapLeftDecayCoefficientSign;
	}

	public void setSwapLeftDecayCoefficientSign(boolean swapLeftDecayCoefficientSign) {
		this.swapLeftDecayCoefficientSign = swapLeftDecayCoefficientSign;
	}

	public boolean isSwapRightDecayCoefficientSign() {
		return swapRightDecayCoefficientSign;
	}

	public void setSwapRightDecayCoefficientSign(boolean swapRightDecayCoefficientSign) {
		this.swapRightDecayCoefficientSign = swapRightDecayCoefficientSign;
	}

	public float getDryRatio() {
		return dryRatio;
	}

	public void setDryRatio(float dryRatio) {
		this.dryRatio = dryRatio;
	}
}
