/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;

public class AudioDataJoint extends Thread {
	private static final int defaultBufferSize = 4096;
	private Collection<AudioDataJointListener> listeners;
	private Map<AudioDataReader, AudioDataWriter> members;
	private AudioDataReader reader;
	private AudioDataWriter writer;
	private int[] buffer;

	public AudioDataJoint(AudioDataReader reader, AudioDataWriter writer) {
		this(reader, writer, defaultBufferSize);
	}

	public AudioDataJoint(AudioDataReader reader, AudioDataWriter writer, int bufferSize) {
		this.reader = reader;
		this.writer = writer;
		if (bufferSize <= 0 || bufferSize >= Integer.MAX_VALUE) {
			bufferSize = defaultBufferSize;
		}
		this.buffer = new int[bufferSize];
	}

	public AudioDataJoint(Map<AudioDataReader, AudioDataWriter> members) {
		this(members, defaultBufferSize);
	}

	public AudioDataJoint(Map<AudioDataReader, AudioDataWriter> members, int bufferSize) {
		if (members.isEmpty()) {
			throw new IllegalArgumentException("Members must no be empty.");
		}
		if (members.size() == 1) {
			Entry<AudioDataReader, AudioDataWriter> entry = members.entrySet().iterator().next();
			reader = entry.getKey();
			writer = entry.getValue();
		} else {
			this.members = new LinkedHashMap<AudioDataReader, AudioDataWriter>(members);
		}
		if (bufferSize <= 0 || bufferSize >= Integer.MAX_VALUE) {
			bufferSize = defaultBufferSize;
		}
		this.buffer = new int[bufferSize];
	}

	@Override
	public void run() {
		setName(getClass().getSimpleName());
		fireStarted();
		if (members == null) {
			try {
				int len;
				while ((len = reader.read(buffer)) != -1) {
					writer.write(buffer, 0, len);
					fireUpdate(reader, writer, buffer, len);
				}
			} catch (IOException e) {
				throw new RuntimeException(e);
			} finally {
				try {
					writer.close();
				} catch (IOException e) {
					e.printStackTrace();
				} finally {
					try {
						reader.close();
					} catch (IOException e) {
						e.printStackTrace();
					} finally {
						fireClosed(reader);
						fireClosed(writer);
						fireFinished();
						reader = null;
						writer = null;
						buffer = null;
					}
				}
			}
		} else {
			try {
				while (!members.isEmpty()) {
					Map<AudioDataReader, AudioDataWriter> finishedMembers = new LinkedHashMap<AudioDataReader, AudioDataWriter>();
					for (Entry<AudioDataReader, AudioDataWriter> entry : members.entrySet()) {
						AudioDataReader reader = entry.getKey();
						AudioDataWriter writer = entry.getValue();
						int len = reader.read(buffer);
						if (len == -1) {
							finishedMembers.put(reader, writer);
						} else {
							writer.write(buffer, 0, len);
						}
						fireUpdate(reader, writer, buffer, len);
					}

					for (Entry<AudioDataReader, AudioDataWriter> entry : finishedMembers.entrySet()) {
						AudioDataReader reader = entry.getKey();
						AudioDataWriter writer = entry.getValue();
						members.remove(reader);
						try {
							reader.close();
						} catch (IOException e) {
							e.printStackTrace();
						} finally {
							fireClosed(reader);
						}

						try {
							writer.close();
						} catch (IOException e) {
							e.printStackTrace();
						} finally {
							fireClosed(writer);
						}
					}
				}
			} catch (IOException e) {
				throw new RuntimeException(e);
			} finally {
				try {
					closeAll();
				} finally {
					fireFinished();
					members = null;
					buffer = null;
				}
			}
		}
	}

	private void closeAll() {
		Map<Object, IOException> broken = new LinkedHashMap<Object, IOException>();
		for (Entry<AudioDataReader, AudioDataWriter> entry : members.entrySet()) {
			AudioDataReader reader = entry.getKey();
			try {
				reader.close();
			} catch (IOException e) {
				broken.put(reader, e);
			} finally {
				fireClosed(reader);
			}
			AudioDataWriter writer = entry.getValue();
			try {
				writer.close();
			} catch (IOException e) {
				broken.put(writer, e);
			} finally {
				fireClosed(writer);
			}
		}
		if (!broken.isEmpty()) {
			throw new RuntimeException("One or more Errors occured while closing the AudioDataReaders and AudioDataWriters: " + broken);
		}
	}

	public void addAudioDataJointListener(AudioDataJointListener listener) {
		if (listeners == null) {
			listeners = new ArrayList<AudioDataJointListener>();
		}
		listeners.add(listener);
	}

	public void removeAudioDataJointListener(AudioDataJointListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
		if (listeners.isEmpty()) {
			listeners = null;
		}
	}

	protected void fireStarted() {
		if (listeners == null) {
			return;
		}
		for (AudioDataJointListener listener : new ArrayList<AudioDataJointListener>(listeners)) {
			listener.audioDataJointStarted();
		}
	}

	protected void fireUpdate(AudioDataReader reader, AudioDataWriter writer, int[] buffer, int len) {
		if (listeners == null) {
			return;
		}
		for (AudioDataJointListener listener : new ArrayList<AudioDataJointListener>(listeners)) {
			listener.audioDataJointUpdate(reader, writer, buffer, len);
		}
	}

	protected void fireClosed(AudioDataReader reader) {
		if (listeners == null) {
			return;
		}
		for (AudioDataJointListener listener : new ArrayList<AudioDataJointListener>(listeners)) {
			listener.audioDataReaderClosed(reader);
		}
	}

	protected void fireClosed(AudioDataWriter writer) {
		if (listeners == null) {
			return;
		}
		for (AudioDataJointListener listener : new ArrayList<AudioDataJointListener>(listeners)) {
			listener.audioDataWriterClosed(writer);
		}
	}

	protected void fireFinished() {
		if (listeners == null) {
			return;
		}
		for (AudioDataJointListener listener : new ArrayList<AudioDataJointListener>(listeners)) {
			listener.audioDataJointFinished(this);
		}
	}
}
