/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import java.io.Closeable;
import java.io.IOException;

/**
 * The AudioDataReader is the essential base for all audio processing classes. It represents the
 * source for any audio data. The read method has to fill that unified audio data in form of an
 * int array. This array represents a stereo pcm sample array.
 * 
 * The stop method has to finish the reading process. The read method should return -1 to signal
 * to its caller, that the read process provices no more audio data.
 * 
 * The close method should release any resources and make the reader available for garbage collection.
 * 
 * @author Florian Paul
 *
 */
public interface AudioDataReader extends Closeable {
	/**
	 * Reads a packet of unified audio data. The unified audio data is a stereo pcm sample array.
	 * The samples of the left and right audio channel are alternating with the array positions.
	 * The int value of the array at index zero represents the first sample of the left channel, the int value
	 * of the array at index one is the first sample of the right channel, the int value of the array at index two
	 * is the second sample of the left channel, and so on.
	 * 
	 * @param data
	 * @return the read length of the data. If this value is -1, this reader has reached the end and there are no
	 * more audio samples to read. In this case, the reader should be closed by the caller.
	 * @throws IOException if any Exceptions occur, the implementation should have the possibility to throw an exception.
	 */
	public int read(int[] data) throws IOException;
	public void stop();
	
	/**
	 * Locates the absolute position by the given parameter. The reader must adjust itself to deliver the next read samples
	 * from the new  frame position.
	 * @param framePosition
	 * @throws IOException
	 */
	public void seek(long framePosition) throws IOException;
}
