/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;

public class FileAudioDataWriter implements AudioDataWriter {
	
	private int[] buffer;
	private int bufferSize;
	private AudioFormat format;
	private final boolean monoConversion;
	private AudioFileFormat.Type fileType;
	private OutputStream outputStream;
	private final boolean bigEndian;
	private final Thread writeThread;
	private final Thread readThread;
	private boolean closed;
	private boolean readThreadFinished;
	private volatile int bufferedLength;
	private volatile int consumed;
	private Long predictedLengthInFrames;
	private Throwable readThreadAbort;

	public FileAudioDataWriter(AudioFormat format, AudioFileFormat.Type fileType, OutputStream outputStream) {
		this.format = format;
		monoConversion = format.getChannels() == 1;
		this.fileType = fileType;
		this.outputStream = outputStream;
		this.bigEndian = format.isBigEndian();
		writeThread = Thread.currentThread();
		readThread = new ReadThread();
		readThread.start();
	}

	@Override
	public int write(int[] stereoData, int offset, int length) throws IOException {
		if(closed) {
			throw new IllegalStateException(getClass().getSimpleName()+" is already closed.");
		}
		if(readThreadFinished) {
			if (readThreadAbort != null) {
				throw new RuntimeException(readThreadAbort);
			}
			throw new IllegalStateException(readThread.getClass().getSimpleName()+" has already finished.");
		}
		if (buffer == null) {
			bufferSize = stereoData.length;
			if(monoConversion) {
				bufferSize = bufferSize / 2;
			}
			buffer = new int[bufferSize];
		}
		
		waitForReader();
		
		/* stereo to mono conversion */
		if(monoConversion) {
			/* buffer is in this case mono (half size of stereoData) */
			bufferedLength = Math.min(bufferSize, length / 2);
			for (int i = 0; i < bufferSize; i++) {
				int a = stereoData[2*i];
				int b = stereoData[2*i+1];
				buffer[i] = (a + b) / 2;
			}
		} else {
			/* stereo signal */
			bufferedLength = Math.min(bufferSize, length);
			
			System.arraycopy(stereoData, offset, buffer, 0, bufferedLength);
		}
		
		// new data was written to buffer
		consumed = 0;
		notifyRead();
		
		return bufferedLength;
	}
	
	private class BlockingInputStream extends InputStream {
		
		@Override
		public int read(byte[] bytes, int offsetInBytes, int lenghInBytes) throws IOException {
			while(consumed>=bufferedLength) {
				if(closed) {
					// have nothing to read and writer is closed
					notifyWrite();
					return -1;
				}
				// have nothing to read, waiting for writer
				try {
					synchronized (readThread) {
						notifyWrite();
						readThread.wait();
					}
				} catch (InterruptedException e) {
					throw new IOException(e);
				}
			}
			// have anything to read
			int offset = offsetInBytes / 2;
			int nonConsumedInBuffer = bufferedLength-consumed;
			int requestedLength = lenghInBytes / 2;
			int n = Math.min(requestedLength, nonConsumedInBuffer);
			if(bigEndian) {
				for (int i = offset; i < n; i++) {
					int d = buffer[i+consumed];
					byte msb = (byte) (255 & (d >> 8));
					byte lsb = (byte) (255 & d);
					bytes[2 * i] = msb;
					bytes[2 * i + 1] = lsb;
				}
			} else {
				for (int i = offset; i < n; i++) {
					int d = buffer[i+consumed];
					byte msb = (byte) (255 & (d >> 8));
					byte lsb = (byte) (255 & d);
					bytes[2 * i] = lsb;
					bytes[2 * i + 1] = msb;
				}
			}
			
			consumed += n;
			int ret = n * 2 - offset * 2;
			return ret;
		}

		@Override
		public int read() {
			throw new IllegalStateException("read not supported");
		}
	}
	
	private void waitForReader() throws IOException {
		while(consumed<bufferedLength) {
			try {
				synchronized (writeThread) {
					notifyRead();
					writeThread.wait(100);
				}
			} catch (InterruptedException e) {
				throw new IOException(e);
			}
		}
	}
	
	private void notifyRead() {
		synchronized (readThread) {
			readThread.notify();
		}
	}
	
	private void notifyWrite() {
		synchronized (writeThread) {
			writeThread.notify();
		}
	}

	@Override
	public void close() throws IOException {
		Thread.yield();
		waitForReader();
		closed = true;
		notifyRead();
		while(!readThreadFinished) {
			waitForReader();
		}
	}
	
	private class ReadThread extends Thread {
		@Override
		public void run() {
			setName(getClass().getSimpleName());
			InputStream is = new BlockingInputStream();
			long length = predictedLengthInFrames!=null?(monoConversion?predictedLengthInFrames/2:predictedLengthInFrames):AudioSystem.NOT_SPECIFIED;
			AudioInputStream ais = new AudioInputStream(is, format, length);
			try {
				AudioSystem.write(ais, fileType, outputStream);
			} catch (Throwable e) {
				readThreadAbort = e;
				throw new RuntimeException(e);
			} finally {
				try {
					ais.close();
				} catch (IOException e) {
					e.printStackTrace();
				} finally {
					try {
						outputStream.close();
					} catch (IOException e) {
						e.printStackTrace();
					} finally {
						readThreadFinished = true;
					}
				}
			}
		}
	}
	
	public boolean hasReadThreadFinished() {
		return readThreadFinished;
	}

	public Long getPredictedLengthInFrames() {
		return predictedLengthInFrames;
	}

	public void setPredictedLengthInFrames(Long predictableLength) {
		this.predictedLengthInFrames = predictableLength;
	}
}
