/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import java.io.IOException;

/**
 * This reader adjusts the samples by multiplying each by a factor, that is
 * varying continuously, dependent on the maximum sample of a read chunk. The
 * factor is incrementally adjusted within an amount of samples. After a mix of
 * two signals, there may be great samples, above the clipping boundary. This
 * reader can be used to push the samples beyond the clipping boundary. The
 * factor for this pushing multiplication is calculated automatically. So this
 * reader should be positioned after any other mixing reader, or those which
 * create great sample values.
 * 
 * @author Florian Paul
 * 
 */
public class LimitingAudioDataReader implements AudioDataReaderChainLink {

	private AudioDataReader reader;
	private final float sampleMaximum; // full scale
	// this factor should always be between 0.5 and 1.0
	// it is calculated automatically by the greatest sum of samples
	private Float factor;

	public LimitingAudioDataReader(AudioDataReader reader, int sampleSize) {
		this.reader = reader;
		this.sampleMaximum = (float) Math.pow(2D, sampleSize - 1);
	}

	@Override
	public int read(int[] data) throws IOException {
		final int len = reader.read(data);
		int maxSample = 0;

		// within this count of samples the factor adjustment takes
		// place
		Integer factorAdjustmentSamples = null;
		for (int i = 0; i < len; i++) {
			int abs = Math.abs(data[i]);
			if (factorAdjustmentSamples == null && abs >= sampleMaximum) {
				factorAdjustmentSamples = i;
			}
			maxSample = Math.max(maxSample, abs);
		}
		if (factorAdjustmentSamples == null) {
			factorAdjustmentSamples = len;
		} else if (factorAdjustmentSamples <= 1) {
			factorAdjustmentSamples = 2;
		}
		float preferredFactor;
		if (maxSample >= sampleMaximum) {
			preferredFactor = sampleMaximum / maxSample;
		} else {
			preferredFactor = 1F;
		}

		if (factor == null) {
			factor = preferredFactor;
		}

		float factorAdjustment = (preferredFactor - factor) / (factorAdjustmentSamples - 1);

		for (int i = 0; i < len; i++) {
			if (factorAdjustmentSamples > i) {
				if (factorAdjustment > 0 && factor <= preferredFactor) {
					factor += factorAdjustment;
				} else if (factorAdjustment < 0 && factor > preferredFactor) {
					factor += factorAdjustment;
				}
			} else if (factor == 1) {
				break;
			}
			data[i] *= factor;
		}
		return len;
	}

	@Override
	public void close() throws IOException {
		reader.close();
	}

	@Override
	public void stop() {
		reader.stop();
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;		
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}
}
