/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import java.io.IOException;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.TargetDataLine;


public class LineAudioDataReader extends UnifiedAudioDataReader {

	private final TargetDataLine line;
	private AudioFormat format;
	private boolean stopable = true;
	private boolean closable = true;

	public LineAudioDataReader(AudioFormat format) throws AudioIOException {
		this.format = format;
		try {
			this.line = AudioSystem.getTargetDataLine(format);
			init();
		} catch (IllegalArgumentException e) {
			throw new AudioIOException(e);
		} catch (LineUnavailableException e) {
			throw new AudioIOException(e);
		}
	}

	public LineAudioDataReader(AudioFormat format, Mixer mixer) throws AudioIOException {
		this.format = format;
		try {
			this.line = AudioSystem.getTargetDataLine(format, mixer.getMixerInfo());
			init();
		} catch (IllegalArgumentException e) {
			throw new AudioIOException(e);
		} catch (LineUnavailableException e) {
			throw new AudioIOException(e);
		}
	}

	private void init() throws LineUnavailableException {
		if (!line.isOpen()) {
			line.open(format);
		}
		if (!line.isRunning()) {
			line.start();
		}
	}

	@Override
	protected int read(byte[] data, int offset, int length) {
		int ret = line.read(data, offset, length);
		if (ret == 0 && !line.isRunning()) {
			return -1;
		}
		return ret;
	}

	@Override
	public void close() throws IOException {
		stop();
		if (!closable) {
			return;
		}
		if (line.isOpen()) {
			line.close();
		}
		super.close();
	}

	@Override
	public void stop() {
		if (!stopable) {
			return;
		}
		if (line.isRunning()) {
			line.stop();
			line.flush();
		}
	}

	@Override
	public void seek(long samplePosition) {
		throw new UnsupportedOperationException("Seek is not supported by " + getClass().getName());
	}

	public void addLineListener(LineListener listener) {
		line.addLineListener(listener);
	}

	public void removeLineListener(LineListener listener) {
		line.removeLineListener(listener);
	}

	public TargetDataLine getLine() {
		return line;
	}

	@Override
	protected AudioFormat getAudioFormat() {
		return format;
	}

	public boolean isStopable() {
		return stopable;
	}

	public void setStopable(boolean stopable) {
		this.stopable = stopable;
	}

	public boolean isClosable() {
		return closable;
	}

	public void setClosable(boolean closable) {
		this.closable = closable;
	}
}
