/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioFormat.Encoding;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.SourceDataLine;


public class LineAudioDataWriter implements AudioDataWriter {
	
	private final SourceDataLine line;
	private boolean closable = true;
	private boolean initialized;
	private boolean bigEndian;
	private final AudioFormat format;
	private byte[] audioBytes;
	private int len;

	public LineAudioDataWriter(SourceDataLine line) throws AudioIOException {
		this.line = line;
		this.format = line.getFormat();
		try {
			init();
		} catch (LineUnavailableException e) {
			throw new AudioIOException(e);
		}
	}

	public LineAudioDataWriter(AudioFormat format) throws AudioIOException {
		this(format, null);
	}

	public LineAudioDataWriter(AudioFormat format, Mixer mixer) throws AudioIOException {
		try {
			if (mixer != null) {
				this.line = AudioSystem.getSourceDataLine(format, mixer.getMixerInfo());
			} else {
				this.line = AudioSystem.getSourceDataLine(format);
			}
			this.format = format;
			init();
		} catch (LineUnavailableException e) {
			throw new AudioIOException(e);
		}
	}
	
	private void init() throws LineUnavailableException {
		if (!line.isOpen()) {
			line.open();
		}
		if(!line.isRunning()) {
			line.start();
		}
		
		if(!format.getEncoding().equals(Encoding.PCM_SIGNED) && !format.getEncoding().equals(Encoding.PCM_UNSIGNED)) {
			throw new IllegalArgumentException("The audio encoding must be PCM.");
		}
	}

	@Override
	public int write(int[] data, int offset, int length) {
		if (!initialized) {
			len = data.length << 1;
			audioBytes = new byte[len];
			bigEndian = format.isBigEndian();
			initialized = true;
		}

		if(bigEndian) {
			for (int i = 0, l = 0, r = 1; i < length; i++, l += 2, r += 2) {
				int d = data[i];
				byte msb = (byte) (255 & (d >> 8));
				byte lsb = (byte) (255 & d);
				audioBytes[l] = msb;
				audioBytes[r] = lsb;
			}
		} else {
			for (int i = 0, l = 0, r = 1; i < length; i++, l += 2, r += 2) {
				int d = data[i];
				byte msb = (byte) (255 & (d >> 8));
				byte lsb = (byte) (255 & d);
				audioBytes[l] = lsb;
				audioBytes[r] = msb;
			}
		}
		
		int ret = line.write(audioBytes, offset << 1, length << 1);
		return ret >> 1;
	}
	
	@Override
	public void close() {
		if(!closable) {
			return;
		}
		while (line.isRunning()) {
			line.drain();
			line.stop();
		}
		while (line.isOpen()) {
			line.flush();
			line.close();
		}
	}
	
	public void addLineListener(LineListener listener) {
		line.addLineListener(listener);
	}
	
	public void removeLineListener(LineListener listener) {
		line.removeLineListener(listener);
	}

	public SourceDataLine getLine() {
		return line;
	}

	public boolean isClosable() {
		return closable;
	}

	public void setClosable(boolean closable) {
		this.closable = closable;
	}
}
