/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import java.io.IOException;

import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.TargetDataLine;

/**
 * This is an observer for Source and Target Data Lines. It is synchronized by
 * the real underlying audio line for capture or playback.
 * 
 * The main feature of this class is to obtain the frame position at any time.
 * 
 * If the underlying lines become stopped, this observer holds the last maximum
 * position.
 * 
 * This observer has an additional offset, that is the relative difference
 * between the user requested frame position and the physical line.
 * 
 * @author Florian Paul
 */
public class LineFrameObserver implements FrameObserver, AudioDataReaderChainLink {

	private final SourceDataLine sourceDataLine;
	private final TargetDataLine targetDataLine;
	private long additionalOffset;
	private long sourcePosition;
	private long targetPosition;
	private long readerFramePosition;
	private long latencyInFrames;
	private int fix;
	private int measurementsUntilFix = 50;
	private int latencySmoothingDifferenceInFrames = 200;
	private boolean latencyFixed;
	private AudioDataReader source;
	private boolean started;

	public LineFrameObserver(SourceDataLine sourceDataLine) {
		if (sourceDataLine == null) {
			throw new IllegalArgumentException("The SourceDataLine must not be null.");
		}
		this.sourceDataLine = sourceDataLine;
		this.targetDataLine = null;
	}

	public LineFrameObserver(TargetDataLine targetDataLine) {
		if (targetDataLine == null) {
			throw new IllegalArgumentException("The TargetDataLine must not be null.");
		}
		this.sourceDataLine = null;
		this.targetDataLine = targetDataLine;
	}

	@Override
	public long getCurrentFramePosition() {
		return getLineFramePosition() + additionalOffset;
	}

	public long getLineFramePosition() {
		if (sourceDataLine != null) {
			long pos = sourceDataLine.getLongFramePosition();
			sourcePosition = Math.max(pos, sourcePosition);
			return sourcePosition;
		}

		if (targetDataLine != null) {
			long pos = targetDataLine.getLongFramePosition();
			targetPosition = Math.max(pos, targetPosition);
			return targetPosition;
		}
		throw new IllegalStateException("Either SourceDataLine or TargetDataLine must not be null.");
	}

	@Override
	public void setFramePosition(long framePosition) {
		long linePosition = getLineFramePosition();
		additionalOffset = framePosition - linePosition;
		readerFramePosition = framePosition;
	}

	@Override
	public long getReaderFramePosition() {
		return readerFramePosition;
	}

	@Override
	public long getLatencyInFrames() {
		if (latencyFixed) {
			return latencyInFrames;
		}
		long lat = readerFramePosition - getCurrentFramePosition();
		if (lat < 0) {
			// usual case when recording line in use
			fix++;
			latencyFixed = fix >= measurementsUntilFix;
			return 0;
		}
		// precision of a specific amount of frames to smooth the change
		if (lat > latencyInFrames + latencySmoothingDifferenceInFrames) {
			// usual case for playback line
			latencyInFrames = lat;
			fix = 0;
		} else {
			fix++;
		}
		// due to performance reasons the fix is finished after a specific amount of measurements
		latencyFixed = fix >= measurementsUntilFix;
		return latencyInFrames;
	}

	public void setLatencyInFrames(long latency) {
		// self repairing 10% off
		latency *= 0.9F;
		if (latency < 0) {
			latency = 0;
		}
		this.latencyInFrames = latency;
		fix = 0;
		latencyFixed = false;
	}

	@Override
	public int read(int[] data) throws IOException {
		if(!started) {
			started = true;
		}
		int readSamples = source.read(data);
		readerFramePosition += readSamples >> 1;
		return readSamples;
	}
	
	@Override
	public void seek(long framePosition) throws IOException {
		if(started) {
			latencyFixed = true;
		}
		setFramePosition(framePosition);
		source.seek(framePosition);
	}

	@Override
	public void stop() {
		source.stop();
	}
	
	@Override
	public void close() throws IOException {
		source.close();
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.source = source;
	}

	@Override
	public AudioDataReader getSource() {
		return source;
	}
}
