/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import java.io.IOException;
import java.io.OutputStream;

import javax.sound.sampled.AudioFormat;

public class RawAudioDataWriter implements AudioDataWriter {

	private final boolean monoConversion;
	private OutputStream outputStream;
	private final boolean bigEndian;

	public RawAudioDataWriter(AudioFormat format, OutputStream outputStream) throws IOException {
		monoConversion = format.getChannels() == 1;
		this.outputStream = outputStream;
		this.bigEndian = format.isBigEndian();
		byte[] bytes = RawFileAudioDataReader.toSpec(format).getBytes("UTF-8");
		assert bytes.length==RawFileAudioDataReader.HEADER_SIZE;
		outputStream.write(bytes);
	}

	@Override
	public int write(int[] stereoData, int stereoOffset, int stereoLength) throws IOException {
		if (outputStream == null) {
			throw new IllegalStateException(getClass().getSimpleName() + " is already closed.");
		}

		int byteLength = monoConversion ? stereoLength : stereoLength * 2;
		byte[] bytes = new byte[byteLength];

		if (monoConversion) {
			/* buffer is in this case mono (half size of stereoData) */
			if (bigEndian) {
				for (int i = stereoOffset, n = stereoLength / 2; i < n; i++) {
					int a = stereoData[2 * i];
					int b = stereoData[2 * i + 1];
					int d = (a + b) / 2;
					byte msb = (byte) (255 & (d >> 8));
					byte lsb = (byte) (255 & d);
					bytes[2 * i] = msb;
					bytes[2 * i + 1] = lsb;
				}
			} else {
				for (int i = stereoOffset, n = stereoLength / 2; i < n; i++) {
					int a = stereoData[2 * i];
					int b = stereoData[2 * i + 1];
					int d = (a + b) / 2;
					byte msb = (byte) (255 & (d >> 8));
					byte lsb = (byte) (255 & d);
					bytes[2 * i] = lsb;
					bytes[2 * i + 1] = msb;
				}
			}
		} else {
			if(bigEndian) {
				for (int i = stereoOffset; i < stereoLength; i++) {
					int d = stereoData[i];
					byte msb = (byte) (255 & (d >> 8));
					byte lsb = (byte) (255 & d);
					bytes[2 * i] = msb;
					bytes[2 * i + 1] = lsb;
				}
			} else {
				for (int i = stereoOffset; i < stereoLength; i++) {
					int d = stereoData[i];
					byte msb = (byte) (255 & (d >> 8));
					byte lsb = (byte) (255 & d);
					bytes[2 * i] = lsb;
					bytes[2 * i + 1] = msb;
				}
			}
		}
		
		outputStream.write(bytes);

		return stereoLength;
	}

	@Override
	public void close() throws IOException {
		outputStream.flush();
		outputStream.close();
		outputStream = null;
	}
}
