/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Arrays;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.sound.sampled.AudioFormat;

public class RawFileAudioDataReader extends UnifiedAudioDataReader {
	
	public static final String FILE_NAME_ENDING = ".raw";
	public static final int HEADER_SIZE = 16;
	private final RandomAccessFile raf;
	private final long fileLength;
	private final AudioFormat format;
	private final int framesToSampleBytesFactor;
	private final long additionalFrameOffset;
	private boolean stopped;
	private long negativeAdditionalFrameOffset;
	
	public RawFileAudioDataReader(File file) throws IOException {
		this(file, 0);
	}
	
	public RawFileAudioDataReader(File file, long additionalOffset) throws IOException {
		raf = new RandomAccessFile(file, "r");
		this.fileLength = raf.length();
		this.additionalFrameOffset = additionalOffset;
		format = readAudioFormat();
		this.framesToSampleBytesFactor = format.getSampleSizeInBits() / 8 * format.getChannels();
		seek(0);
	}

	@Override
	protected int read(byte[] data, int offset, int length) throws IOException {
		if(stopped) {
			return -1;
		}
		
		long negativeAdditionalFrameOffset = this.negativeAdditionalFrameOffset;
		
		int emptySampleBytes = 0;
		if (negativeAdditionalFrameOffset < 0) {
			final long filePointer = raf.getFilePointer();
			final long currentFrameInFile = (filePointer - HEADER_SIZE) / framesToSampleBytesFactor;
			if (negativeAdditionalFrameOffset < currentFrameInFile) {
				final int emptyFrames = (int) Math.abs(currentFrameInFile + negativeAdditionalFrameOffset);
				emptySampleBytes = emptyFrames * framesToSampleBytesFactor;
				if(emptySampleBytes>=length) {
					Arrays.fill(data, (byte)0);
					this.negativeAdditionalFrameOffset += length / framesToSampleBytesFactor;
					return length;
				}
			}
		}
		
		int readOffset = offset + emptySampleBytes;
		int readLength = length - emptySampleBytes;
		
		Arrays.fill(data, 0, emptySampleBytes, (byte)0);
		
		int read = raf.read(data, readOffset, readLength);
		return read + emptySampleBytes;
	}

	@Override
	public void close() throws IOException {
		raf.close();
	}

	@Override
	public void stop() {
		stopped = true;
	}

	@Override
	public void seek(long framePosition) throws IOException {
		if (framePosition < 0) {
			framePosition = 0;
		}
		negativeAdditionalFrameOffset = framePosition + additionalFrameOffset;
		
		long pos = framePosition + additionalFrameOffset;
		if (pos < 0) {
			pos = 0;
		}
		long seek = pos * framesToSampleBytesFactor + HEADER_SIZE;
		if (seek > fileLength) {
			seek = fileLength;
		}
		raf.seek(seek);
	}
	
	@Override
	public AudioFormat getAudioFormat() {
		return format;
	}
	
	private AudioFormat readAudioFormat() throws IOException {
		byte[] header = new byte[HEADER_SIZE];
		long currentPos = raf.getFilePointer();
		raf.seek(0);
		raf.readFully(header);
		raf.seek(currentPos);
		AudioFormat format = fromSpec(new String(header, "UTF-8"));
		return format;
	}

	/*
	 * Creates a String format like this: PCMS0116BE044100
	 */
	static String toSpec(AudioFormat format) {
		StringBuilder sb = new StringBuilder();
		if(format.getEncoding()==AudioFormat.Encoding.PCM_SIGNED) {
			sb.append("PCMS");
		} else if(format.getEncoding()==AudioFormat.Encoding.PCM_UNSIGNED) {
			sb.append("PCMU");
		} else if(format.getEncoding()==AudioFormat.Encoding.ALAW) {
			sb.append("ALAW");
		} else if(format.getEncoding()==AudioFormat.Encoding.ULAW) {
			sb.append("ULAW");
		}
		sb.append(fillUp(format.getChannels(), 2));
		sb.append(fillUp(format.getSampleSizeInBits(), 2));
		sb.append(format.isBigEndian()?"BE":"LE");
		sb.append(fillUp((int)format.getSampleRate(), 6));
		return sb.toString();
	}
	
	private static String fillUp(int num, int count) {
		StringBuilder sb = new StringBuilder();
		sb.append(num);
		while(sb.length()<count) {
			sb.insert(0, "0");
		}
		return sb.toString();
	}
	
	/*
	 * Parses something like this: PCMS0116BE044100
	 */
	private static AudioFormat fromSpec(String spec) {
		Pattern pattern = Pattern.compile("^(\\D\\D\\D\\D)(\\d\\d)(\\d\\d)(\\D\\D)(\\d\\d\\d\\d\\d\\d)$");
		Matcher m = pattern.matcher(spec);
		if(m.matches()) {
			AudioFormat.Encoding encoding;
			float sampleRate;
			int sampleSizeInBits;
			int channels;
			int frameSize;
			float frameRate;
			boolean bigEndian;
			String enc = m.group(1);
			if(enc.equals("PCMS")) {
				encoding = AudioFormat.Encoding.PCM_SIGNED;
			} else if(enc.equals("PCMU")) {
				encoding = AudioFormat.Encoding.PCM_UNSIGNED;
			} else if(enc.equals("ALAW")) {
				encoding = AudioFormat.Encoding.ALAW;
			} else if(enc.equals("ULAW")) {
				encoding = AudioFormat.Encoding.ULAW;
			} else {
				throw new IllegalArgumentException("Encoding '"+enc+"' not supported.");
			}
			channels = Integer.parseInt(m.group(2));
			sampleSizeInBits = Integer.parseInt(m.group(3));
			String endianess = m.group(4);
			bigEndian = endianess.equals("BE");
			sampleRate = Integer.parseInt(m.group(5));
			frameSize = sampleSizeInBits * channels / 8;
			frameRate = sampleRate;
			AudioFormat format = new AudioFormat(encoding, sampleRate, sampleSizeInBits, channels, frameSize, frameRate, bigEndian);
			return format;
		}
		throw new IllegalArgumentException("Spec '"+spec+"' not supported.");
	}

	public int getFramesToSamplesFactor() {
		AudioFormat audioFormat = getAudioFormat();
		int channels = audioFormat.getChannels();
		int factor = audioFormat.getSampleSizeInBits() / 8 * channels;
		return factor;
	}
}
