/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.io;

import java.io.IOException;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioFormat.Encoding;

abstract public class UnifiedAudioDataReader implements AudioDataReader {
	private enum AudioType {
		PCM_16_2_BE(16, 2, true, null), 
		PCM_16_2_LE(16, 2, false, null), 
		PCM_16_1_BE(16, 1, true, null), 
		PCM_16_1_LE(16, 1, false, null), 
		PCM_SIGNED_8_2(8, 2, null, true), 
		PCM_UNSIGNED_8_2(8, 2, null, false), 
		PCM_SIGNED_8_1(8, 1, null, true), 
		PCM_UNSIGNED_8_1(8, 1, null, false);
		
		final int sampleSizeInBits;
		final int channels;
		final Boolean bigEndian;
		final Boolean signed;
		
		private AudioType(int sampleSizeInBits, int channels, Boolean bigEndian, Boolean signed) {
			this.sampleSizeInBits = sampleSizeInBits;
			this.channels = channels;
			this.bigEndian = bigEndian;
			this.signed = signed;
		}
		
		public static AudioType fromAudioFormat(AudioFormat format) {
			for(AudioType type : values()) {
				if(type.matches(format)) {
					return type;
				}
			}
			throw new IllegalArgumentException("Unsupported AudioFormat " +  format);
		}
		
		private boolean matches(AudioFormat format) {
			if(format.getChannels()!=channels) {
				return false;
			}
			if(format.getSampleSizeInBits()!=sampleSizeInBits) {
				return false;
			}
			if(bigEndian!=null) {
				if(bigEndian && !format.isBigEndian()) {
					return false;
				}
				if(!bigEndian && format.isBigEndian()) {
					return false;
				}
			}
			if(signed!=null) {
				if(signed && !format.getEncoding().equals(Encoding.PCM_SIGNED)) {
					return false;
				}
				if(!signed && !format.getEncoding().equals(Encoding.PCM_UNSIGNED)) {
					return false;
				}
			}
			return true;
		}
	}
	
	private AudioType audioType;
	private byte[] audioBytes;
	private int audioBytesLength;

	@Override
	public final int read(int[] data) throws IOException {
		if (audioType == null) {
			audioType = AudioType.fromAudioFormat(getAudioFormat());
		}
		switch(audioType) {
			case PCM_16_2_BE:
				return readPcm_16_2_be(data);
			case PCM_16_2_LE:
				return readPcm_16_2_le(data);
			case PCM_16_1_BE:
				return readPcm_16_1_be(data);
			case PCM_16_1_LE:
				return readPcm_16_1_le(data);
			default: 
				throw new IllegalStateException("AutioType " + audioType + " is not supported.");
		}
	}
	
	private int readPcm_16_2_be(int[] data) throws IOException {
		int len = data.length;
		if (audioBytes == null || audioBytesLength != len * 2) {
			if(len%2==1) {
				throw new IllegalArgumentException("The data array must have an even dimension. " + len + " is odd.");
			}
			audioBytesLength = len * 2;
			audioBytes = new byte[audioBytesLength];
		}
		int readByteCount = read(audioBytes, 0, audioBytesLength);
		if(readByteCount==-1) {
			return -1;
		}
		int ret = readByteCount / 2;
        
		for (int i = 0; i < ret; i++) {
			int msb = audioBytes[2 * i];
			int lsb = audioBytes[2 * i + 1];
			data[i] = msb << 8 | (255 & lsb);
		}
        
		return ret;
	}
	
	private int readPcm_16_2_le(int[] data) throws IOException {
		int len = data.length;
		if (audioBytes == null || audioBytesLength != len * 2) {
			if(len%2==1) {
				throw new IllegalArgumentException("The data array must have an even dimension. " + len + " is odd.");
			}
			audioBytesLength = len * 2;
			audioBytes = new byte[audioBytesLength];
		}
		int readByteCount = read(audioBytes, 0, audioBytesLength);
		if(readByteCount==-1) {
			return -1;
		}
		int ret = readByteCount / 2;
        
		for (int i = 0; i < ret; i++) {
			int lsb = audioBytes[2 * i];
			int msb = audioBytes[2 * i + 1];
			data[i] = msb << 8 | (255 & lsb);
		}
        
		return ret;
	}
	
	private int readPcm_16_1_be(int[] data) throws IOException {
		int len = data.length;
		if (audioBytes == null || audioBytesLength != len) {
			if(len%2==1) {
				throw new IllegalArgumentException("The data array must have an even dimension. " + len + " is odd.");
			}
			audioBytesLength = len;
			audioBytes = new byte[audioBytesLength];
		}
		int readByteCount = read(audioBytes, 0, audioBytesLength);
		if(readByteCount==-1) {
			return -1;
		}
		
		for (int i = 0, n = readByteCount / 2; i < n; i++) {
			int msb = audioBytes[2 * i];
			int lsb = audioBytes[2 * i + 1];
			int d = msb << 8 | (255 & lsb);
			data[2*i] = d;
			data[2*i+1] = d;
		}
        
		return readByteCount;
	}
	
	private int readPcm_16_1_le(int[] data) throws IOException {
		int len = data.length;
		if (audioBytes == null || audioBytesLength != len) {
			if(len%2==1) {
				throw new IllegalArgumentException("The data array must have an even dimension. " + len + " is odd.");
			}
			audioBytesLength = len;
			audioBytes = new byte[audioBytesLength];
		}
		int readByteCount = read(audioBytes, 0, audioBytesLength);
		if(readByteCount==-1) {
			return -1;
		}
		
		for (int i = 0, n = readByteCount / 2; i < n; i++) {
			int lsb = audioBytes[2 * i];
			int msb = audioBytes[2 * i + 1];
			int d = msb << 8 | (255 & lsb);
			data[2*i] = d;
			data[2*i+1] = d;
		}
        
		return readByteCount;
	}
	
	abstract protected int read(byte[] data, int offset, int length) throws IOException;
	
	abstract protected AudioFormat getAudioFormat() throws IOException;
	
	@Override
	public void close() throws IOException {
		audioBytes = null;
	}

	public AudioType getAudioType() {
		return audioType;
	}
}
