/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import static org.audiochain.model.ObjectStore.*;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

public abstract class AbstractAudioDevice implements AudioDevice {
	private static final long serialVersionUID = 1L;
	private String identifier = IdentifierGenerator.generateIdentifier();
	protected UserInterfaceContext<? extends AudioDevice> userInterfaceContext;
	private Map<String, Long> audioDeviceValueLastChangeTimes = new LinkedHashMap<String, Long>();
	// audioProject must be transient, because while sync the audio device
	// should be sent to remote only
	protected transient AudioProject audioProject;
	private transient Collection<BinarySynchronizationListener> binarySynchronizationListeners;
	private transient Collection<AudioDeviceListener> audioDeviceListeners;
	private transient Collection<AudioDeviceValue> audioDeviceValues;

	public AudioProject getAudioProject() {
		return audioProject;
	}

	@Override
	public void initialize(AudioProject project) {
		this.audioProject = project;
	}

	@Override
	public void destroy() {
		Collection<AudioDeviceValue> audioDeviceValues = getAudioDeviceValues();
		if (audioDeviceValues != null) {
			for (AudioDeviceValue audioDeviceValue : audioDeviceValues) {
				audioDeviceValue.clearListeners();
			}
		}
		audioProject = null;
		userInterfaceContext = null;
	}

	@Override
	public final Collection<AudioDeviceValue> getAudioDeviceValues() {
		if (audioDeviceValues == null) {
			audioDeviceValues = new LinkedHashSet<AudioDeviceValue>();
			try {
				createAnnotatedAudioDeviceValues();
			} catch (IllegalArgumentException e) {
				throw new IllegalStateException(e);
			} catch (IllegalAccessException e) {
				throw new IllegalStateException(e);
			}
			createAudioDeviceValues(audioDeviceValues);
		}
		return audioDeviceValues;
	}

	protected void createAudioDeviceValues(@SuppressWarnings("unused") Collection<AudioDeviceValue> audioDeviceValues) {
	}

	@SuppressWarnings("rawtypes")
	private void createAnnotatedAudioDeviceValues() throws IllegalArgumentException, IllegalAccessException {
		Collection<Field> fields = collectFields(this);
		for (Field field : fields) {
			field.setAccessible(true);
			if (field.isAnnotationPresent(AudioDeviceProperty.class)) {
				AudioDeviceProperty adp = field.getAnnotation(AudioDeviceProperty.class);
				String fieldName = field.getName();
				if (Modifier.isStatic(field.getModifiers())) {
					throw new IllegalStateException("The field " + fieldName + " is annotated with " + AudioDeviceProperty.class.getSimpleName() + " and therefore must not be static.");
				}
				if (Modifier.isTransient(field.getModifiers())) {
					throw new IllegalStateException("The field " + fieldName + " is annotated with " + AudioDeviceProperty.class.getSimpleName() + " and therefore must not be transient.");
				}
				final Method setter = findSetter(fieldName);
				if (setter == null) {
					throw new IllegalStateException("The field " + fieldName + " is annotated with " + AudioDeviceProperty.class.getSimpleName()
							+ " and therefore must have a JavaBean conform setter.");
				}
				String name = adp.name();
				Object value = field.get(this);
				Class<?> fieldType = field.getType();
				if (fieldType.isAssignableFrom(String.class)) {
					StringAudioDeviceValue adv = new StringAudioDeviceValue(this, name, (String) value);
					adv.addAudioDeviceValueListener(new AudioDeviceValueListener() {
						@Override
						public void audioDeviceValueChanged(AudioDevice audioDevice, AudioDeviceValue audioDeviceValue, String oldValue, String newValue) {
							try {
								setter.invoke(AbstractAudioDevice.this, newValue);
								firePerformanceRelevantAudioDevicePropertyChanged();
							} catch (IllegalArgumentException e) {
								throw new IllegalStateException(e);
							} catch (IllegalAccessException e) {
								throw new IllegalStateException(e);
							} catch (InvocationTargetException e) {
								throw new IllegalStateException(e);
							}
						}
					});
					adv.setUnit(adp.unit());
					adv.setSymbol(adp.symbol());
					audioDeviceValues.add(adv);
				} else if (fieldType.isAssignableFrom(Integer.class) || fieldType.isAssignableFrom(Integer.TYPE)) {
					IntAudioDeviceValue adv = new IntAudioDeviceValue(this, name, (Integer) value);
					adv.addAudioDeviceValueListener(new AudioDeviceValueListener() {
						@Override
						public void audioDeviceValueChanged(AudioDevice audioDevice, AudioDeviceValue audioDeviceValue, String oldValue, String newValue) {
							try {
								setter.invoke(AbstractAudioDevice.this, newValue != null ? new Integer(newValue) : null);
								firePerformanceRelevantAudioDevicePropertyChanged();
							} catch (IllegalArgumentException e) {
								throw new IllegalStateException(e);
							} catch (IllegalAccessException e) {
								throw new IllegalStateException(e);
							} catch (InvocationTargetException e) {
								throw new IllegalStateException(e);
							}
						}
					});
					adv.setUnit(adp.unit());
					adv.setSymbol(adp.symbol());
					if (!Float.isNaN(adp.min())) {
						adv.setMinValue((int) adp.min());
					}
					if (!Float.isNaN(adp.max())) {
						adv.setMaxValue((int) adp.max());
					}
					if (!Float.isNaN(adp.mid())) {
						adv.setMidValue((int) adp.mid());
					}
					
					audioDeviceValues.add(adv);
				} else if (fieldType.isAssignableFrom(Float.class) || fieldType.isAssignableFrom(Float.TYPE)) {
					FloatAudioDeviceValue adv = new FloatAudioDeviceValue(this, name, (Float) value);
					adv.addAudioDeviceValueListener(new AudioDeviceValueListener() {
						@Override
						public void audioDeviceValueChanged(AudioDevice audioDevice, AudioDeviceValue audioDeviceValue, String oldValue, String newValue) {
							try {
								setter.invoke(AbstractAudioDevice.this, newValue != null ? new Float(newValue) : null);
								firePerformanceRelevantAudioDevicePropertyChanged();
							} catch (IllegalArgumentException e) {
								throw new IllegalStateException(e);
							} catch (IllegalAccessException e) {
								throw new IllegalStateException(e);
							} catch (InvocationTargetException e) {
								throw new IllegalStateException(e);
							}
						}
					});
					adv.setUnit(adp.unit());
					adv.setSymbol(adp.symbol());
					if (!Float.isNaN(adp.min())) {
						adv.setMinValue(adp.min());
					}
					if (!Float.isNaN(adp.max())) {
						adv.setMaxValue(adp.max());
					}
					if (!Float.isNaN(adp.mid())) {
						adv.setMidValue(adp.mid());
					}
					adv.setScale(adp.scale());
					audioDeviceValues.add(adv);
				} else if (fieldType.isAssignableFrom(Boolean.class) || fieldType.isAssignableFrom(Boolean.TYPE)) {
					BooleanAudioDeviceValue adv = new BooleanAudioDeviceValue(this, name, (Boolean) value);
					adv.addAudioDeviceValueListener(new AudioDeviceValueListener() {
						@Override
						public void audioDeviceValueChanged(AudioDevice audioDevice, AudioDeviceValue audioDeviceValue, String oldValue, String newValue) {
							try {
								setter.invoke(AbstractAudioDevice.this, newValue != null ? new Boolean(newValue) : null);
								firePerformanceRelevantAudioDevicePropertyChanged();
							} catch (IllegalArgumentException e) {
								throw new IllegalStateException(e);
							} catch (IllegalAccessException e) {
								throw new IllegalStateException(e);
							} catch (InvocationTargetException e) {
								throw new IllegalStateException(e);
							}
						}
					});
					adv.setUnit(adp.unit());
					adv.setSymbol(adp.symbol());
					adv.setFalseSymbol(adp.falseSymbol());
					audioDeviceValues.add(adv);
				} else if (fieldType.isAssignableFrom(Enum.class) || fieldType.isEnum()) {
					@SuppressWarnings({
						"unchecked"
					})
					final Class<? extends Enum> enumClass = (Class<? extends Enum>) fieldType;
					Enum<?> enumValue = (Enum<?>) value;
					Enum<?>[] options = (Enum[]) fieldType.getEnumConstants();
					@SuppressWarnings("unchecked")
					EnumAudioDeviceValue<?> adv = new EnumAudioDeviceValue(this, name, enumValue, fieldType, options);
					adv.addAudioDeviceValueListener(new AudioDeviceValueListener() {
						@SuppressWarnings("unchecked")
						@Override
						public void audioDeviceValueChanged(AudioDevice audioDevice, AudioDeviceValue audioDeviceValue, String oldValue, String newValue) {
							try {
								setter.invoke(AbstractAudioDevice.this, newValue != null ? Enum.valueOf(enumClass, newValue) : null);
								firePerformanceRelevantAudioDevicePropertyChanged();
							} catch (IllegalArgumentException e) {
								throw new IllegalStateException(e);
							} catch (IllegalAccessException e) {
								throw new IllegalStateException(e);
							} catch (InvocationTargetException e) {
								throw new IllegalStateException(e);
							}
						}
					});
					adv.setUnit(adp.unit());
					adv.setSymbol(adp.symbol());
					audioDeviceValues.add(adv);
				} else {
					throw new IllegalStateException("Field '" + fieldName + "' in class '" + this.getClass().getName() + "' is annotated with " + AudioDeviceProperty.class.getSimpleName()
							+ " but has an unsupported value type '" + fieldType + "'.");
				}
			}
		}
	}

	private Method findSetter(String fieldName) {
		Collection<Method> methods = ObjectStore.collectMethods(this);
		fieldName = fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1);
		for (Method method : methods) {
			String methodName = method.getName();

			if (methodName.equals("set" + fieldName) && method.getParameterTypes().length == 1) {
				method.setAccessible(true);
				return method;
			}
		}
		return null;
	}

	public AudioTrack findAssociatedAudioTrack() {
		for (AudioTrack track : audioProject.getAudioTracks()) {
			Collection<AudioDevice> devices = new LinkedHashSet<AudioDevice>(track.getPlaybackAudioDeviceChain().getAudioDevices());
			devices.addAll(track.getCaptureAudioDeviceChain().getAudioDevices());
			devices.add(track.getSourceAudioDevice());
			for (AudioDevice audioDevice : devices) {
				if (audioDevice == this) {
					return track;
				}
			}
		}
		return null;
	}

	public AudioTrack getAssociatedAudioTrack() {
		AudioTrack ret = findAssociatedAudioTrack();
		if (ret == null) {
			throw new IllegalStateException("No " + AudioTrack.class.getSimpleName() + " associated to " + AudioDevice.class.getSimpleName() + " " + getName());
		}
		return ret;
	}

	@Override
	public final UserInterfaceContext<? extends AudioDevice> getUserInterfaceContext() {
		if (userInterfaceContext == null) {
			userInterfaceContext = createUserInterfaceContext();
		}
		return userInterfaceContext;
	}

	abstract protected UserInterfaceContext<? extends AudioDevice> createUserInterfaceContext();

	@Override
	public String getIdentifier() {
		return identifier;
	}

	@Override
	public void setIdentifier(String identifier) {
		this.identifier = identifier;
	}

	@Override
	public void synchronizeBy(Synchronizable source) {
		if (source instanceof AudioDevice) {
			AudioDevice sourceAudioDevice = (AudioDevice) source;
			Collection<AudioDeviceValue> sourceAudioDeviceValues = sourceAudioDevice.getAudioDeviceValues();
			// only if this type of audio device has audio device values
			if (sourceAudioDeviceValues != null) {
				Collection<AudioDeviceValue> audioDeviceValues = getAudioDeviceValues();
				for (AudioDeviceValue sourceAudioDeviceValue : sourceAudioDeviceValues) {
					for (AudioDeviceValue audioDeviceValue : audioDeviceValues) {
						audioDeviceValue.synchronizeBy(sourceAudioDeviceValue);
					}
				}
			}
		}
	}

	@Override
	public long getAudioDeviceValueLastChangeTime(String audioDeviceValueName) {
		Long ret = audioDeviceValueLastChangeTimes.get(audioDeviceValueName);
		if (ret == null) {
			return 0;
		}
		return ret;
	}

	@Override
	public void setAudioDeviceValueLastChangeTime(String audioDeviceValueName, long time) {
		// zero is default and no store worthy information
		if (time != 0) {
			audioDeviceValueLastChangeTimes.put(audioDeviceValueName, time);
		}

		// clean up, remove all zero entries
		Collection<String> toRemove = new HashSet<String>();
		Set<Entry<String, Long>> entrySet = audioDeviceValueLastChangeTimes.entrySet();
		for (Entry<String, Long> entry : entrySet) {
			if (entry.getValue() == 0) {
				toRemove.add(entry.getKey());
			}
		}
		for (String name : toRemove) {
			audioDeviceValueLastChangeTimes.remove(name);
		}
	}

	protected void firePerformanceRelevantAudioDevicePropertyChanged() {
		if (audioDeviceListeners == null) {
			return;
		}
		for (AudioDeviceListener listener : new LinkedHashSet<AudioDeviceListener>(audioDeviceListeners)) {
			listener.performanceRelevantAudioDevicePropertyChanged(this);
		}
		firePersistableAudioDevicePropertyChanged();
	}

	protected void firePersistableAudioDevicePropertyChanged() {
		if (audioDeviceListeners == null) {
			return;
		}
		for (AudioDeviceListener listener : new LinkedHashSet<AudioDeviceListener>(audioDeviceListeners)) {
			listener.persistableAudioDevicePropertyChanged(this);
		}
	}

	@Override
	public void addAudioDeviceListener(AudioDeviceListener listener) {
		if (audioDeviceListeners == null) {
			audioDeviceListeners = new LinkedHashSet<AudioDeviceListener>();
		}
		audioDeviceListeners.add(listener);
	}

	@Override
	public void removeAudioDeviceListener(AudioDeviceListener listener) {
		if (audioDeviceListeners == null) {
			return;
		}
		audioDeviceListeners.remove(listener);
		if (audioDeviceListeners.isEmpty()) {
			audioDeviceListeners = null;
		}
	}

	public void addBinarySynchronizationListener(BinarySynchronizationListener listener) {
		if (binarySynchronizationListeners == null) {
			binarySynchronizationListeners = new LinkedHashSet<BinarySynchronizationListener>();
		}
		binarySynchronizationListeners.add(listener);
	}

	public void removeBinarySynchronizationListener(BinarySynchronizationListener listener) {
		if (binarySynchronizationListeners == null) {
			return;
		}
		binarySynchronizationListeners.remove(listener);
		if (binarySynchronizationListeners.isEmpty()) {
			binarySynchronizationListeners = null;
		}
	}

	protected void fireBinaryChanged(BinaryDataset binaryDataset) {
		if (binarySynchronizationListeners == null) {
			return;
		}
		for (BinarySynchronizationListener listener : binarySynchronizationListeners) {
			listener.binaryChanged(binaryDataset);
		}
	}

	protected boolean isEqual(Object o1, Object o2) {
		if (o1 == null && o2 == null) {
			return true;
		}
		return o1 != null && o1.equals(o2);
	}

	protected boolean greaterOrBothZero(long source, long target) {
		return source > target || (source == 0 && target == 0);
	}

	protected void assertValueInRange(int value, int min, int max) {
		assertValueInRange((float) value, (float) min, (float) max);
	}

	protected void assertValueInRange(float value, float min, float max) {
		if (value < min) {
			throw new IllegalArgumentException("Value must not be lower than " + min + ". Current value is " + value);
		}
		if (value < min || value > max) {
			throw new IllegalArgumentException("Value must not be greater than " + max + ". Current value is " + value);
		}
	}
}
