/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import java.util.Collection;
import java.util.LinkedHashSet;

abstract public class AbstractChainableAudioDevice extends AbstractAudioDevice implements ChainableAudioDevice {
	private static final long serialVersionUID = 1L;
	private boolean enabled = true;
	private long enabledChangeTime;
	private transient Collection<ChainableAudioDeviceListener> listeners;

	@Override
	public void setEnabled(boolean enabled) {
		boolean oldValue = this.enabled;
		this.enabled = enabled;
		if (oldValue != enabled) {
			enabledChangeTime = System.currentTimeMillis();
			fireEnabledChanged(oldValue, enabled);
		}
	}

	@Override
	public boolean isEnabled() {
		return enabled;
	}

	private void fireEnabledChanged(boolean oldValue, boolean newValue) {
		if (listeners == null) {
			return;
		}
		for (ChainableAudioDeviceListener listener : listeners) {
			listener.audioDeviceEnabledChanged(this, oldValue, newValue);
		}
	}

	@Override
	public void addChainableAudioDeviceListener(ChainableAudioDeviceListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<ChainableAudioDeviceListener>();
		}
		listeners.add(listener);
	}

	@Override
	public void removeChainableAudioDeviceListener(ChainableAudioDeviceListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	@Override
	public void synchronizeBy(Synchronizable source) {
		super.synchronizeBy(source);

		if (source instanceof AbstractChainableAudioDevice) {
			AbstractChainableAudioDevice sourceAbstractChainableAudioDevice = (AbstractChainableAudioDevice) source;
			if (sourceAbstractChainableAudioDevice.enabledChangeTime > enabledChangeTime || (sourceAbstractChainableAudioDevice.enabledChangeTime == 0 && enabledChangeTime == 0)) {
				setEnabled(sourceAbstractChainableAudioDevice.isEnabled());
				enabledChangeTime = sourceAbstractChainableAudioDevice.enabledChangeTime;
			}
		}
	}
}
