/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import java.awt.Component;
import java.util.Collection;
import java.util.LinkedHashSet;

abstract public class AbstractUserInterfaceContext<D extends AudioDevice> implements UserInterfaceContext<D> {
	private static final long serialVersionUID = 1L;
	private boolean hidden;
	private transient Component component;
	private transient CommandLineUserInterface commandLineUserInterface;
	private transient Collection<UserInterfaceContextListener> listeners;

	@Override
	public final void setHidden(boolean hidden) {
		boolean oldValue = this.hidden;
		this.hidden = hidden;
		if (oldValue != hidden) {
			fireHiddenChanged(oldValue, hidden);
		}
	}

	@Override
	public final boolean isHidden() {
		return hidden;
	}

	@Override
	public final Component getComponent(D audioDevice) {
		if (component == null) {
			component = createComponent(audioDevice);
		}
		return component;
	}

	abstract protected Component createComponent(D audioDevice);

	@Override
	public final CommandLineUserInterface getCommandLineUserInterface(D audioDevice) {
		if (commandLineUserInterface == null) {
			commandLineUserInterface = createCommandLineUserInterface(audioDevice);
		}
		return commandLineUserInterface;
	}

	abstract protected CommandLineUserInterface createCommandLineUserInterface(D audioDevice);

	protected void fireHiddenChanged(boolean oldValue, boolean newValue) {
		if (listeners != null) {
			for (UserInterfaceContextListener listener : listeners) {
				listener.hiddenChanged(oldValue, newValue);
			}
		}
	}

	@Override
	public void addUserInterfaceContextChangeListener(UserInterfaceContextListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<UserInterfaceContextListener>();
		}
		listeners.add(listener);
	}

	@Override
	public void removeUserInterfaceContextChangeListener(UserInterfaceContextListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}
}
