/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import java.io.Serializable;
import java.util.Collection;

import org.audiochain.io.AudioDataReader;
import org.audiochain.ui.gui.GenericUserInterfaceContext;


/**
 * <p>This Interface is the extension point for every audio processing device. It's main
 * role is to create an instance of a specific {@link AudioDataReader} instance and to keep all
 * values (changed by the user or by the device itself) in sync with itself (the {@link AudioDevice} implementation instance)
 * and with the most recent created {@link AudioDataReader} instance.
 *
 * <p>The {@link AudioDataReader} instance must be transient. All relevant values which have to be
 * made persistent, can be hold in non transient serializable instance variables.
 * 
 * <p>The device vendor must implement one of the {@link AudioDevice} extensions: {@link SourceAudioDevice} or {@link ChainableAudioDevice}.
 * The main difference is, that {@link ChainableAudioDevice}s can be disabled and moved in their order within the {@link AudioDeviceChain}.
 * The {@link SourceAudioDevice} is a special {@link AudioDevice} for playback. It is the first {@link AudioDevice} in the {@link AudioDeviceChain}
 * and can just exist once per {@link AudioTrack}.
 * 
 */
public interface AudioDevice extends Serializable, Synchronizable {
	
	/**
	 * Initializes this audioDevice with the audio project.
	 * This is a replacement for the constructor. The audio device must not expect to be added to any audio track at this moment.
	 * 
	 * @param project for general access and possibility for the device to register any listeners on the project.
	 */
	public void initialize(AudioProject audioProject);
	
	/**
	 * The device should release any resources and prepare to be removed forever.
	 */
	public void destroy();
	
	/**
	 * Returns a collection of AudioControlValues for manipulation of the corresponding AudioDataReader instance.
	 * If the setValue() method is invoked on the AudioControlValues, the AudioDataReader must react on the changes.
	 * 
	 * @return a collection of AudioDeviceValue that are logically linked to the AudioDataReader instance.
	 */
	public Collection<AudioDeviceValue> getAudioDeviceValues();
	
	/**
	 * An audio device should provide a user interface context implementation. The generic implementation {@link GenericUserInterfaceContext} can be used. The audio device
	 * should hold a persistable instance of this object.
	 * 
	 * @return a context object for ui operations
	 */
	public UserInterfaceContext<? extends AudioDevice> getUserInterfaceContext();
	
	/**
	 * Returns a logical name for this audio device.
	 * @return
	 */
	public String getName();
	
	/**
	 * A unique ID for each AudioDevice instance for detecting remote objects.
	 * @param identifier
	 */
	public void setIdentifier(String identifier);
	
	/**
	 * Returns the timestamp (millis after 1970) of the last change that was made to the AudioDeviceValue with the given name.
	 * @param audioDeviceValueName
	 * @return
	 */
	public long getAudioDeviceValueLastChangeTime(String audioDeviceValueName);
	
	/**
	 * Sets the timestamp (millis after 1970) of the last change that was made to the AudioDeviceValue with the given name for storage.
	 * @param audioDeviceValueName
	 * @param time
	 */
	public void setAudioDeviceValueLastChangeTime(String audioDeviceValueName, long time);
	
	/**
	 * Adds the {@link AudioDeviceListener} to this {@link AudioDevice}.
	 * @param listener
	 */
	public void addAudioDeviceListener(AudioDeviceListener listener);
	
	/**
	 * Removes the {@link AudioDeviceListener} from this {@link AudioDevice}.
	 * @param listener
	 */
	public void removeAudioDeviceListener(AudioDeviceListener listener);
}
