/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import static java.util.Collections.*;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;

/**
 * This registry reads all {@link AudioDevice} implementation classes from the
 * file 'audioDeviceClasses'. This should be present on the classpath. Per
 * default, the file contains the following entries:
 * 
 * <code><pre>
 * # Chainable Audio Devices
 * org.audiochain.devices.compound.BasicSppmAudioDevice
 * org.audiochain.devices.level.SamplePeakProgrammeMeterAudioDevice
 * org.audiochain.devices.dynamics.CompressorAudioDevice
 * org.audiochain.devices.basic.BasicAudioDevice
 * org.audiochain.devices.reverb.ReverbAudioDevice
 * org.audiochain.devices.frequency.ThreeBandEqualizerAudioDevice
 * org.audiochain.devices.frequency.LowPassFilterAudioDevice
 * org.audiochain.devices.frequency.HiPassFilterAudioDevice
 * org.audiochain.devices.frequency.BandPassFilterAudioDevice
 * org.audiochain.devices.frequency.BandStopFilterAudioDevice
 * org.audiochain.devices.latency.LatencyAudioDevice
 * org.audiochain.devices.frequency.FftAnalyzerAudioDevice
 * 
 * # Source Audio Devices
 * org.audiochain.devices.recording.RecordingAudioDevice
 * org.audiochain.devices.drum.DrumAudioDevice
 * org.audiochain.devices.frequency.SignalGeneratorAudioDevice
 * </pre></code>
 * 
 * Your own {@link AudioDevice} implementation classes can be added to this file
 * and will be automatically available on the project or tracks. For more
 * information on writing {@link AudioDevice}s see {@link AudioDevice},
 * {@link ChainableAudioDevice} and {@link SourceAudioDevice}.
 */
public class AudioDeviceRegistry {
	private static final String DEVICE_CLASS_DEFINITION_FILE_NAME = "audioDeviceClasses";
	private static AudioDeviceRegistry instance;
	private final Collection<Class<? extends AudioDevice>> audioDeviceClasses;
	private Map<String, Class<? extends AudioDevice>> deviceNameClassMap;
	private Collection<AudioDeviceFactory> chainableAudioDeviceFactories;

	private AudioDeviceRegistry() throws AudioDeviceRegistryException, IOException, ClassNotFoundException, InstantiationException, IllegalAccessException {
		audioDeviceClasses = new LinkedHashSet<Class<? extends AudioDevice>>();
		init();
	}

	public static synchronized AudioDeviceRegistry getInstance() throws AudioDeviceRegistryException, IOException, ClassNotFoundException, InstantiationException, IllegalAccessException {
		if (instance == null) {
			instance = new AudioDeviceRegistry();
		}
		return instance;
	}

	private void init() throws AudioDeviceRegistryException, IOException, ClassNotFoundException, InstantiationException, IllegalAccessException {
		InputStream is = getClass().getResourceAsStream("/" + DEVICE_CLASS_DEFINITION_FILE_NAME);
		if (is == null) {
			throw new IllegalStateException("File '" + DEVICE_CLASS_DEFINITION_FILE_NAME + "' not found in classpath.");
		}

		BufferedReader br = new BufferedReader(new InputStreamReader(is));
		// only for looking up device names while creating the class list
		deviceNameClassMap = new HashMap<String, Class<? extends AudioDevice>>();
		try {
			String line;
			while ((line = br.readLine()) != null) {
				line = line.trim();
				if (!line.isEmpty() && !line.startsWith("#")) {
					loadDeviceClass(line);
				}
			}

			String additionalAudioDeviceClassesString = System.getProperty("additionalAudioDeviceClasses");
			if (additionalAudioDeviceClassesString != null && !additionalAudioDeviceClassesString.trim().isEmpty()) {
				additionalAudioDeviceClassesString = additionalAudioDeviceClassesString.trim();
				String[] additionalAudioDeviceClasses = additionalAudioDeviceClassesString.split(",");
				for (String additionalAudioDeviceClass : additionalAudioDeviceClasses) {
					loadDeviceClass(additionalAudioDeviceClass);
				}
			}

			loadOptionalAudioDeviceFactories();
		} finally {
			deviceNameClassMap = null;
			br.close();
			is.close();
		}
	}

	private void loadDeviceClass(String className) throws ClassNotFoundException, AudioDeviceRegistryException, InstantiationException, IllegalAccessException {
		@SuppressWarnings("unchecked")
		Class<? extends AudioDevice> anyClass = (Class<? extends AudioDevice>) Class.forName(className);
		if (!AudioDevice.class.isAssignableFrom(anyClass)) {
			throw new AudioDeviceRegistryException("Class '" + anyClass.getName() + "' is not a type of " + AudioDevice.class.getName() + ". Please remove it from the "
					+ DEVICE_CLASS_DEFINITION_FILE_NAME + " file!");
		}
		String deviceName = anyClass.newInstance().getName();
		if (deviceNameClassMap.containsKey(deviceName)) {
			throw new AudioDeviceRegistryException("The " + AudioDevice.class.getSimpleName() + " name '" + deviceName + "' is already defined for '" + deviceNameClassMap.get(deviceName) + "'.");
		}
		audioDeviceClasses.add(anyClass);
		deviceNameClassMap.put(deviceName, anyClass);
	}

	private void loadOptionalAudioDeviceFactories() throws ClassNotFoundException, InstantiationException, IllegalAccessException {
		String audioDeviceFactoryClassesString = System.getProperty("audioDeviceFactoryClasses");
		if (audioDeviceFactoryClassesString != null && !audioDeviceFactoryClassesString.trim().isEmpty()) {
			audioDeviceFactoryClassesString = audioDeviceFactoryClassesString.trim();
			String[] audioDeviceFactoryClasses = audioDeviceFactoryClassesString.split(",");
			for (String audioDeviceFactoryClass : audioDeviceFactoryClasses) {
				Class<?> factoryClass = Class.forName(audioDeviceFactoryClass);
				if (AudioDeviceFactory.class.isAssignableFrom(factoryClass)) {
					if (chainableAudioDeviceFactories == null) {
						chainableAudioDeviceFactories = new LinkedHashSet<AudioDeviceFactory>();
					}
					AudioDeviceFactory factory = (AudioDeviceFactory) factoryClass.newInstance();
					chainableAudioDeviceFactories.add(factory);
				}
			}
		}
	}

	public Collection<Class<? extends AudioDevice>> getAudioDeviceClasses() {
		return unmodifiableCollection(audioDeviceClasses);
	}

	@SuppressWarnings("unchecked")
	public Collection<Class<ChainableAudioDevice>> getChainableAudioDeviceClasses() {
		Collection<Class<ChainableAudioDevice>> ret = new LinkedHashSet<Class<ChainableAudioDevice>>();
		for (Class<? extends AudioDevice> audioDeviceClass : audioDeviceClasses) {
			if (ChainableAudioDevice.class.isAssignableFrom(audioDeviceClass)) {
				ret.add((Class<ChainableAudioDevice>) audioDeviceClass);
			}
		}
		return ret;
	}

	@SuppressWarnings("unchecked")
	public Collection<Class<SourceAudioDevice>> getSourceAudioDeviceClasses() {
		Collection<Class<SourceAudioDevice>> ret = new LinkedHashSet<Class<SourceAudioDevice>>();
		for (Class<? extends AudioDevice> audioDeviceClass : audioDeviceClasses) {
			if (SourceAudioDevice.class.isAssignableFrom(audioDeviceClass)) {
				ret.add((Class<SourceAudioDevice>) audioDeviceClass);
			}
		}
		return ret;
	}

	public Class<ChainableAudioDevice> getChainableAudioDeviceClassByName(String deviceName) throws InstantiationException, IllegalAccessException {
		return findAudioDeviceClassByDeviceName(deviceName, getChainableAudioDeviceClasses());
	}

	public Class<SourceAudioDevice> getSourceAudioDeviceClassByName(String deviceName) throws InstantiationException, IllegalAccessException {
		return findAudioDeviceClassByDeviceName(deviceName, getSourceAudioDeviceClasses());
	}

	private static <E extends AudioDevice> Class<E> findAudioDeviceClassByDeviceName(String deviceName, Collection<Class<E>> classes) throws InstantiationException, IllegalAccessException {
		for (Class<E> klass : classes) {
			AudioDevice device = klass.newInstance();
			if (device.getName().equals(deviceName)) {
				return klass;
			}
		}
		return null;
	}

	public Collection<AudioDeviceFactory> getChainableAudioDeviceFactories() {
		if (chainableAudioDeviceFactories == null) {
			return Collections.unmodifiableCollection(new HashSet<AudioDeviceFactory>());
		}
		return Collections.unmodifiableCollection(chainableAudioDeviceFactories);
	}
}
