/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import static javax.sound.sampled.AudioSystem.*;
import static org.audiochain.io.AudioDataUtil.*;

import java.io.IOException;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.Mixer.Info;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.TargetDataLine;

import org.audiochain.io.AudioIOException;
import org.audiochain.io.LineAudioDataReader;
import org.audiochain.io.LineAudioDataWriter;

/**
 * Saves the properties of the systems audio mixer. This class represents the
 * audio mixer and holds the user chosen mixer selection for playback and
 * capture mixer. It is a singleton, that stores itself in a file in the users
 * home directory.
 */
public class AudioMixer implements Serializable {
	private static final long serialVersionUID = 1L;
	private static final String PROPS_FILE = "mixer.properties";
	private static AudioMixer instance;
	private String selectedPlaybackMixerName;
	private String selectedCaptureMixerName;
	private int systemLatencyInFrames = 900;
	private int playbackBufferSizeInSamples = 2048;
	private int captureBufferSizeInSamples = 2048;
	private boolean linePreinitialization = true;
	private boolean automaticMixerDetection = true;
	private transient HardwareMixer detectedPlaybackMixer;
	private transient HardwareMixer detectedCaptureMixer;
	private long playbackLineLatencyInFrames;

	AudioMixer() {

	}

	public static synchronized AudioMixer getInstance() {
		if (instance == null) {
			try {
				instance = (AudioMixer) ObjectStore.read(PROPS_FILE);
			} catch (Exception e) {
				// can be ignored, new instance with default values will be used
				e.printStackTrace();
			}
			if (instance == null) {
				instance = new AudioMixer();
			}
			instance.init();
		}
		return instance;
	}

	private void init() {
		if (automaticMixerDetection) {
			try {
				detectMixers();
			} catch (AudioIOException e) {
				// not important for now. the user has to choose another line.
			}
		}
		if (linePreinitialization) {
			try {
				initLine();
			} catch (Throwable e) {
				// not important for now. the user has to choose another line.
			}
		}
	}

	/*
	 * initialization of the output line causes latency. Therefore the line must
	 * be initialized once after jvm startup.
	 */
	private void initLine() throws LineUnavailableException, AudioIOException {
		int len = 16384;
		AudioFormat format = createCDAudioFormat();
		SourceDataLine sourceDataLine = AudioSystem.getSourceDataLine(format, getPlaybackMixer().getMixerInfo());
		TargetDataLine targetDataLine = AudioSystem.getTargetDataLine(format, getCaptureMixer().getMixerInfo());
		sourceDataLine.open();
		targetDataLine.open();
		sourceDataLine.start();
		targetDataLine.start();
		byte[] outbuf = new byte[len];
		byte[] inbuf = new byte[len];
		for (int i = 0; i < 10; i++) {
			sourceDataLine.write(outbuf, 0, len);
			targetDataLine.read(inbuf, 0, len);
		}
		sourceDataLine.drain();
		sourceDataLine.stop();
		targetDataLine.stop();
		sourceDataLine.close();
		targetDataLine.close();
	}

	public void store() {
		try {
			ObjectStore.storeInSettingsDirectory(this, PROPS_FILE);
		} catch (IOException e) {
			throw new IllegalStateException(e);
		}
	}

	public HardwareMixer getPlaybackMixer() throws AudioIOException {
		if (automaticMixerDetection) {
			if (detectedPlaybackMixer == null) {
				detectMixers();
			}
			return detectedPlaybackMixer;
		} else {
			return getMixerByName(selectedPlaybackMixerName);
		}
	}

	public HardwareMixer getCaptureMixer() throws AudioIOException {
		if (automaticMixerDetection) {
			if (detectedCaptureMixer == null) {
				detectMixers();
			}
			return detectedCaptureMixer;
		} else {
			return getMixerByName(selectedCaptureMixerName);
		}
	}

	public static HardwareMixer getMixerByName(String name) throws AudioIOException {
		// returns the default mixer if none is configured
		if (name == null) {
			return getDefaultHardwareMixer();
		}
		Collection<HardwareMixer> mixerInfo = getAvailableMixers();
		for (HardwareMixer mixer : mixerInfo) {
			if (name.equals(mixer.getName())) {
				return mixer;
			}
		}

		throw new AudioIOException("Audio Mixer '" + name + "' is not available.");
	}

	private static HardwareMixer getDefaultHardwareMixer() {
		Mixer defaultMixer = AudioSystem.getMixer(null);
		return new HardwareMixer(defaultMixer.getMixerInfo());
	}

	private static Collection<HardwareMixer> getAvailableMixers() {
		Collection<HardwareMixer> ret = new HashSet<HardwareMixer>();
		ret.addAll(getAvailablePlaybackMixers());
		ret.addAll(getAvailableCaptureMixers());
		return ret;
	}

	public static Collection<HardwareMixer> getAvailablePlaybackMixers() {
		Collection<HardwareMixer> ret = new LinkedHashSet<HardwareMixer>();
		Info[] mixerInfos = AudioSystem.getMixerInfo();
		for (Info info : mixerInfos) {
			Mixer mixer = AudioSystem.getMixer(info);
			if (mixer.getSourceLineInfo().length > 0) {
				ret.add(new HardwareMixer(info));
			}
		}
		return ret;
	}

	public static Collection<HardwareMixer> getAvailableCaptureMixers() {
		Collection<HardwareMixer> ret = new LinkedHashSet<HardwareMixer>();
		Info[] mixerInfos = AudioSystem.getMixerInfo();
		for (Info info : mixerInfos) {
			Mixer mixer = AudioSystem.getMixer(info);
			if (mixer.getTargetLineInfo().length > 0) {
				ret.add(new HardwareMixer(info));
			}
		}
		return ret;
	}

	public void detectMixers() throws AudioIOException {
		Collection<Mixer.Info> orderedMixerInfos = new LinkedHashSet<Mixer.Info>();
		Info[] allMixerInfos = AudioSystem.getMixerInfo();
		// try those mixer named "default" first
		for (Mixer.Info info : allMixerInfos) {
			if (info.getName().startsWith("default")) {
				orderedMixerInfos.add(info);
				break;
			}
		}
		// secondly try the null (Java default) mixer
		orderedMixerInfos.add(null);
		// finally try all other provided mixers
		orderedMixerInfos.addAll(Arrays.asList(allMixerInfos));

		for (Mixer.Info playbackMixerInfo : orderedMixerInfos) {
			for (Mixer.Info captureMixerInfo : orderedMixerInfos) {
				if (testMixerCombination(playbackMixerInfo, captureMixerInfo)) {
					if (playbackMixerInfo == null) {
						detectedPlaybackMixer = getDefaultHardwareMixer();
					} else {
						detectedPlaybackMixer = new HardwareMixer(playbackMixerInfo);
					}
					if (captureMixerInfo == null) {
						detectedCaptureMixer = getDefaultHardwareMixer();
					} else {
						detectedCaptureMixer = new HardwareMixer(captureMixerInfo);
					}
					return;
				}
			}
		}

		if (detectedPlaybackMixer == null || detectedCaptureMixer == null) {
			throw new AudioIOException("The hardware mixer for playback and capture can not be detected automatically.");
		}
	}

	public static boolean testMixerCombination(HardwareMixer playbackMixer, HardwareMixer captureMixer) {
		Mixer.Info playbackMixerInfo = playbackMixer == null ? null : playbackMixer.getMixerInfo();
		Mixer.Info captureMixerInfo = captureMixer == null ? null : captureMixer.getMixerInfo();
		return testMixerCombination(playbackMixerInfo, captureMixerInfo);
	}

	private static boolean testMixerCombination(Mixer.Info playbackMixerInfo, Mixer.Info captureMixerInfo) {
		AudioFormat format = createCDAudioFormat();
		boolean ret = true;
		try {
			Mixer mixer = AudioSystem.getMixer(playbackMixerInfo);
			Info mixerInfo = mixer.getMixerInfo();
			SourceDataLine sourceLine = AudioSystem.getSourceDataLine(format, mixerInfo);
			try {
				sourceLine.open(format);
				sourceLine.start();
				byte[] testBytes = new byte[4096];
				sourceLine.write(testBytes, 0, testBytes.length);

				/* test if capturing is possible while the other line is playing back */
				try {
					TargetDataLine targetLine = AudioSystem.getTargetDataLine(format, AudioSystem.getMixer(captureMixerInfo).getMixerInfo());
					try {
						targetLine.open(format);
						targetLine.start();

						long startTime = System.currentTimeMillis();
						while (targetLine.available() <= 0) {
							if (System.currentTimeMillis() - startTime > 3000) {
								throw new LineUnavailableException("Timeout while trying to read from line " + targetLine.getLineInfo());
							}
							Thread.sleep(100);
						}
						targetLine.read(testBytes, 0, testBytes.length);
						sourceLine.write(testBytes, 0, testBytes.length);
						targetLine.read(testBytes, 0, testBytes.length);
						sourceLine.write(testBytes, 0, testBytes.length);
						targetLine.stop();
					} finally {
						if (targetLine.isRunning()) {
							targetLine.stop();
						}
						if (targetLine.isOpen()) {
							targetLine.close();
						}
					}
				} catch (Throwable e) {
					ret = false;
				}

				sourceLine.drain();
				sourceLine.stop();
			} finally {
				if (sourceLine.isRunning()) {
					sourceLine.stop();
				}
				if (sourceLine.isOpen()) {
					sourceLine.close();
				}
			}
		} catch (Throwable e) {
			ret = false;
		}
		return ret;
	}

	public void setSelectedPlaybackMixer(HardwareMixer selectedPlaybackMixer) {
		selectedPlaybackMixerName = selectedPlaybackMixer == null ? null : selectedPlaybackMixer.getName();
	}

	public void setSelectedCaptureMixer(HardwareMixer selectedCaptureMixer) {
		this.selectedCaptureMixerName = selectedCaptureMixer == null ? null : selectedCaptureMixer.getName();
	}

	public int getSystemLatencyInFrames() {
		return systemLatencyInFrames;
	}

	public void setSystemLatencyInFrames(int systemLatencyInFrames) {
		this.systemLatencyInFrames = systemLatencyInFrames;
	}

	public static AudioFormat createAudioFormat(int sampleRate, int sampleSizeInBits, AudioChannels channels, boolean bigEndian) {
		int channelCount = channels.getChannelCount();
		int frameSizeInBytes = sampleSizeInBits * channelCount / 8;
		int frameRate = sampleRate;
		AudioFormat format = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, sampleRate, sampleSizeInBits, channelCount, frameSizeInBytes, frameRate, bigEndian);
		return format;
	}

	public void verifyAudioFormat(int sampleRate, int sampleSizeInBits, AudioChannels channels, boolean bigEndian) throws LineUnavailableException, IOException, AudioIOException {
		verifyAudioFormat(createAudioFormat(sampleRate, sampleSizeInBits, channels, bigEndian));
	}

	public synchronized void verifyAudioFormat(AudioFormat format) throws LineUnavailableException, IOException, AudioIOException {
		SourceDataLine sourceDataLine = null;
		TargetDataLine targetDataLine = null;
		try {
			sourceDataLine = getSourceDataLine(format, getPlaybackMixer().getMixerInfo());
			sourceDataLine.open();
			targetDataLine = getTargetDataLine(format, getCaptureMixer().getMixerInfo());
			targetDataLine.open();
		} finally {
			if (sourceDataLine != null && sourceDataLine.isOpen()) {
				sourceDataLine.close();
			}
			if (targetDataLine != null && targetDataLine.isOpen()) {
				targetDataLine.close();
			}
		}
		LineAudioDataReader lineAudioDataReader = null;
		try {
			lineAudioDataReader = new LineAudioDataReader(format);
			lineAudioDataReader.read(new int[1024]);
		} finally {
			if (lineAudioDataReader != null) {
				lineAudioDataReader.close();
			}
		}
		LineAudioDataWriter lineAudioDataWriter = null;
		try {
			lineAudioDataWriter = new LineAudioDataWriter(format);
			lineAudioDataWriter.write(new int[1024], 0, 1024);
		} finally {
			if (lineAudioDataWriter != null) {
				lineAudioDataWriter.close();
			}
		}
	}

	public boolean isLinePreinitialization() {
		return linePreinitialization;
	}

	public void setLinePreinitialization(boolean preinitializeLines) {
		this.linePreinitialization = preinitializeLines;
	}

	public boolean isAutomaticMixerDetection() {
		return automaticMixerDetection;
	}

	public void setAutomaticMixerDetection(boolean automaticMixerDetection) {
		this.automaticMixerDetection = automaticMixerDetection;
	}

	public HardwareMixer getDetectedPlaybackMixer() {
		return detectedPlaybackMixer;
	}

	public HardwareMixer getDetectedCaptureMixer() {
		return detectedCaptureMixer;
	}

	public int getPlaybackBufferSizeInSamples() {
		return playbackBufferSizeInSamples;
	}

	public void setPlaybackBufferSizeInSamples(int bufferSizeInSamples) {
		this.playbackBufferSizeInSamples = bufferSizeInSamples;
	}

	public int getCaptureBufferSizeInSamples() {
		return captureBufferSizeInSamples;
	}

	public void setCaptureBufferSizeInSamples(int captureBufferSizeInSamples) {
		this.captureBufferSizeInSamples = captureBufferSizeInSamples;
	}

	public long getPlaybackLineLatencyInFrames() {
		return playbackLineLatencyInFrames;
	}

	public void setPlaybackLineLatencyInFrames(long playbackLineLatencyInFrames) {
		this.playbackLineLatencyInFrames = playbackLineLatencyInFrames;
	}

	public HardwareMixer getSelectedPlaybackMixer() throws AudioIOException {
		if (selectedPlaybackMixerName == null) {
			return null;
		}
		return getMixerByName(selectedPlaybackMixerName);
	}

	public HardwareMixer getSelectedCaptureMixer() throws AudioIOException {
		if (selectedCaptureMixerName == null) {
			return null;
		}
		return getMixerByName(selectedCaptureMixerName);
	}
}
