/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import static java.lang.Math.*;
import static org.audiochain.model.AudioDeviceChain.*;
import static org.audiochain.model.AudioMixer.*;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;

import javax.sound.sampled.AudioFormat;

public class AudioProject implements Serializable, BinaryDatasetProvider {
	private static final long serialVersionUID = 1L;
	public static final String FILEENDING = ".audio";
	private String name;
	private long nameChangeTime;
	private int sampleRate = 44100;
	private long sampleRateChangeTime;
	private int sampleSizeInBits = 16;
	private long sampleSizeInBitsChangeTime;
	private boolean bigEndian = true;
	private long bigEndianChangeTime;
	private AudioChannels channels = AudioChannels.Stereo;
	private long channelsChangeTime;
	private Collection<AudioTrack> tracks;
	private AudioDeviceChain masterAudioDeviceChain;
	private String projectPath;
	private String templateFileName;
	private long framePosition;
	private Map<Class<? extends AudioDevice>, Serializable> audioDeviceProperties;
	private String identifier = IdentifierGenerator.generateIdentifier();
	private Collection<String> removedAudioTrackIdentifiers;
	private long lastSyncTime; // this gets updated every time this project
								// becomes synced by another project
	private transient String fileSeparator;
	private transient File projectFile;
	private transient Collection<AudioProjectListener> audioProjectChangeListeners;
	private transient Collection<AudioProjectLifecycleListener> audioProjectLifecycleListeners;
	private transient AudioProject templateProject;
	private transient PerformanceManager performanceManager;

	/**
	 * For Deserialization only
	 */
	@Deprecated
	AudioProject() {
	}

	public AudioProject(String projectPath) {
		masterAudioDeviceChain = new AudioDeviceChain(this);
		this.projectPath = projectPath;
	}

	public Collection<AudioTrack> getPlaybackTracks() {
		Collection<AudioTrack> ret = new LinkedHashSet<AudioTrack>();
		for (AudioTrack track : getAudioTracks()) {
			SourceAudioDevice sourceAudioDevice = track.getSourceAudioDevice();
			if (!track.isSelectedForRecording() && sourceAudioDevice != null && sourceAudioDevice.isAudioDataAvailable()) {
				ret.add(track);
			}
		}
		return ret;
	}

	public Collection<AudioTrack> getRecordTracks() {
		Collection<AudioTrack> ret = new LinkedHashSet<AudioTrack>();
		for (AudioTrack track : getAudioTracks()) {
			if (track.isSelectedForRecording()) {
				ret.add(track);
			}
		}
		return ret;
	}

	public static AudioProject load(File file) throws IOException, IllegalAccessException, ClassNotFoundException, InstantiationException, NoSuchMethodException, InvocationTargetException {
		FileReader fileReader = new FileReader(file);
		BufferedReader bufferedReader = new BufferedReader(fileReader);
		AudioProject ret = (AudioProject) ObjectStore.readObject(bufferedReader);
		bufferedReader.close();
		fileReader.close();
		ret.setProjectFile(file);
		if (ret.getProjectPath() == null) {
			if (file.getParent() != null) {
				ret.setProjectPath(file.getParent());
			} else {
				// set to current directory
				ret.setProjectPath(new File(new File("").getAbsolutePath()).getAbsolutePath());
			}
		}
		ret.initializeAudioDevices();
		return ret;
	}

	/*
	 * The AudioDevices may register listeners to observ the project. The
	 * initialization of the devices is an entry point for the devices to
	 * register any listeners or do other things to initialize their transient
	 * fields.
	 */
	void initializeAudioDevices() {
		for (AudioTrack track : getAudioTracks()) {
			SourceAudioDevice sourceAudioDevice = track.getSourceAudioDevice();
			if (sourceAudioDevice != null) {
				sourceAudioDevice.initialize(this);
			}
			initializeAudioDevices(track.getPlaybackAudioDeviceChain());
			initializeAudioDevices(track.getCaptureAudioDeviceChain());
		}
		initializeAudioDevices(getMasterAudioDeviceChain());
	}

	private void initializeAudioDevices(AudioDeviceChain chain) {
		if (chain == null) {
			return;
		}
		for (ChainableAudioDevice device : chain.getAudioDevices()) {
			device.initialize(this);
		}
	}

	public void store() throws IOException {
		store(true);
	}

	public synchronized void store(boolean createBackupFile) throws IOException {
		if (projectFile == null) {
			throw new IllegalStateException("There is no project file associated with this project (" + getUniqueName() + ").");
		}
		store(projectFile, createBackupFile);
	}

	private synchronized void store(File file, boolean createBackupFile) throws IOException {
		if (fileSeparator == null) {
			fileSeparator = System.getProperty("file.separator");
		}
		if (projectPath != null && projectPath.equals(file.getParent() + fileSeparator)) {
			projectPath = null;
		}

		String filePathWithoutFileEnding = file.getAbsolutePath();
		if (filePathWithoutFileEnding.endsWith(FILEENDING)) {
			filePathWithoutFileEnding = filePathWithoutFileEnding.substring(0, filePathWithoutFileEnding.length() - FILEENDING.length());
		}
		File tempFile = new File(filePathWithoutFileEnding + ".tmp");
		FileWriter fileWriter = new FileWriter(tempFile);
		PrintWriter pw = new PrintWriter(fileWriter);
		ObjectStore.writeObject(this, pw);
		pw.close();
		fileWriter.close();

		File projectFile = new File(filePathWithoutFileEnding + FILEENDING);
		if (createBackupFile) {
			File bakFile = new File(filePathWithoutFileEnding + ".bak");
			if (bakFile.exists() && !bakFile.delete()) {
				System.err.println("backup file can not be deleted");
			}
			if (projectFile.exists() && !projectFile.renameTo(bakFile)) {
				System.err.println("project file can not be renamed to backup file");
			}
		}

		if (projectFile.exists() && !projectFile.delete()) {
			System.err.println("project file can not be deleted");
		}

		if (!tempFile.renameTo(projectFile)) {
			throw new RuntimeException("Can not rename the file '" + tempFile.getAbsolutePath() + "' to '" + projectFile.getAbsolutePath() + "'");
		}
		setProjectFile(projectFile);
	}

	public AudioProject getTemplateProject() throws IOException, IllegalAccessException, ClassNotFoundException, InstantiationException, NoSuchMethodException, InvocationTargetException {
		if (templateProject == null) {
			if (templateFileName != null) {
				templateProject = load(new File(templateFileName));
			}
		}
		return templateProject;
	}

	/**
	 * Creates and adds a new track to this project. When a template is
	 * provided, this method tries to make a copy of the template's track at the
	 * same number of the list of given templates tracks. If the template has
	 * fewer tracks than this project, the last of the template tracks is taken
	 * for creating the copy.
	 * 
	 * @param projectTemplate
	 *            the template project from where the template tracks are taken
	 *            from.
	 * @return the new track. This track is already added to the project by this
	 *         method.
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 */
	public AudioTrack addNewAudioTrack(AudioProject projectTemplate) throws InstantiationException, IllegalAccessException {
		AudioTrack track = null;
		if (projectTemplate != null) {
			AudioTrack[] trackTemplates = projectTemplate.getAudioTracks().toArray(new AudioTrack[0]);
			int projectTrackCount = getAudioTracks().size();
			// try to pick the same track number
			if (trackTemplates.length > projectTrackCount) {
				AudioTrack template = trackTemplates[projectTrackCount];
				track = addNewAudioTrack(template);
			} else if (trackTemplates.length > 0) {
				// take the last template track as template
				AudioTrack template = trackTemplates[trackTemplates.length - 1];
				track = addNewAudioTrack(template);
			}
		}
		if (track == null) {
			track = new AudioTrack(this);
		}

		return track;
	}

	public AudioTrack addNewAudioTrack() throws InstantiationException, IllegalAccessException {
		AudioProject templateAudioProject;
		try {
			templateAudioProject = getTemplateProject();
		} catch (Exception ex) {
			templateAudioProject = null;
		}
		return addNewAudioTrack(templateAudioProject);
	}

	public Collection<AudioTrack> getAudioTracks() {
		if (tracks == null) {
			return Collections.unmodifiableCollection(new LinkedHashSet<AudioTrack>());
		}
		return Collections.unmodifiableCollection(tracks);
	}

	public void addAudioTrack(AudioTrack track) {
		if (tracks == null) {
			tracks = new LinkedHashSet<AudioTrack>();
		}
		if (tracks.contains(track)) {
			throw new IllegalArgumentException("This project does already contain the track.");
		}
		String trackIdentifier = track.getIdentifier();
		for (AudioTrack audioTrack : tracks) {
			if (audioTrack.getIdentifier().equals(trackIdentifier)) {
				throw new IllegalArgumentException("This project does already contain a track with the identifier '" + trackIdentifier + "'.");
			}
		}
		tracks.add(track);
		track.setAudioProject(this);
		fireAudioTrackAdded(track);
	}

	private void fireAudioTrackAdded(AudioTrack track) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : new LinkedHashSet<AudioProjectListener>(audioProjectChangeListeners)) {
				listener.audioTrackAdded(track);
			}
		}
	}

	public void removeAudioTrack(AudioTrack track) {
		if (tracks != null) {
			track.delete();
			tracks.remove(track);
			if (removedAudioTrackIdentifiers == null) {
				removedAudioTrackIdentifiers = new LinkedHashSet<String>();
			}
			removedAudioTrackIdentifiers.add(track.getIdentifier());
			if (tracks.isEmpty()) {
				tracks = null;
			}
			fireAudioTrackRemoved(track);
		}
	}

	public boolean isRemoved(AudioTrack track) {
		return removedAudioTrackIdentifiers != null && removedAudioTrackIdentifiers.contains(track.getIdentifier());
	}

	private void fireAudioTrackRemoved(AudioTrack track) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : new LinkedHashSet<AudioProjectListener>(audioProjectChangeListeners)) {
				listener.audioTrackRemoved(track);
			}
		}
	}

	public void decrementAudioTrackPositionIndex(AudioTrack audioTrack) {
		moveAudioTrackPosition(audioTrack, false);
	}

	public void incrementAudioTrackPositionIndex(AudioTrack audioTrack) {
		moveAudioTrackPosition(audioTrack, true);
	}

	private void moveAudioTrackPosition(AudioTrack audioTrack, boolean increment) {
		if (tracks != null && tracks.contains(audioTrack)) {
			if (movePositionIndex(audioTrack, tracks, increment)) {
				fireAudioTrackMoved(audioTrack);
			}
		}
	}

	private void fireAudioTrackMoved(AudioTrack track) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : new LinkedHashSet<AudioProjectListener>(audioProjectChangeListeners)) {
				listener.audioTrackMoved(track);
			}
		}
	}

	public static AudioProject createAudioProject(String projectPath, AudioProject template) throws InstantiationException, IllegalAccessException {
		AudioProject ret = new AudioProject(projectPath);
		if (template != null) {
			ret.bigEndian = template.bigEndian;
			ret.channels = template.channels;
			ret.name = template.name;
			ret.sampleRate = template.sampleRate;
			ret.sampleSizeInBits = template.sampleSizeInBits;

			AudioDeviceChain masterChain = ret.getMasterAudioDeviceChain();
			for (ChainableAudioDevice templateAudioDevice : template.getMasterAudioDeviceChain().getAudioDevices()) {
				masterChain.addAudioDevice(ret.createAudioDevice(templateAudioDevice));
			}

			for (AudioTrack templateTrack : template.getAudioTracks()) {
				ret.addNewAudioTrack(templateTrack);
			}
		}
		return ret;
	}

	public String getProjectPath() {
		if (fileSeparator == null) {
			fileSeparator = System.getProperty("file.separator");
		}
		if (projectPath == null && projectFile != null) {
			projectPath = projectFile.getParent();
		}
		if (projectPath != null && !projectPath.endsWith(fileSeparator)) {
			projectPath += fileSeparator;
		}
		return projectPath;
	}

	public void setProjectPath(String path) {
		String oldPath = this.projectPath;
		this.projectPath = path;
		if (notEqual(oldPath, path)) {
			fireProjectPathChanged(oldPath, path);
		}
	}

	private void fireProjectPathChanged(String oldValue, String newValue) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : audioProjectChangeListeners) {
				listener.audioProjectPathChanged(oldValue, newValue);
			}
		}
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		String oldName = this.name;
		this.name = name;
		if (notEqual(oldName, name)) {
			nameChangeTime = System.currentTimeMillis();
			fireNameChanged(oldName, name);
		}
	}

	private void fireNameChanged(String oldValue, String newValue) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : audioProjectChangeListeners) {
				listener.audioProjectNameChanged(oldValue, newValue);
			}
		}
	}

	public int getSampleRate() {
		return sampleRate;
	}

	public void setSampleRate(int sampleRate) {
		int oldSampleRate = this.sampleRate;
		this.sampleRate = sampleRate;
		if (oldSampleRate != sampleRate) {
			sampleRateChangeTime = System.currentTimeMillis();
			fireSampleRateChanged(oldSampleRate, sampleRate);
		}
	}

	private void fireSampleRateChanged(int oldValue, int newValue) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : audioProjectChangeListeners) {
				listener.audioProjectSampleRateChanged(oldValue, newValue);
			}
		}
	}

	public int getSampleSizeInBits() {
		return sampleSizeInBits;
	}

	public void setSampleSizeInBits(int sampleSizeInBits) {
		int oldSampleSize = this.sampleSizeInBits;
		this.sampleSizeInBits = sampleSizeInBits;

		if (oldSampleSize != sampleSizeInBits) {
			sampleSizeInBitsChangeTime = System.currentTimeMillis();
			fireSampleSizeInBitsChanged(oldSampleSize, sampleSizeInBits);
		}
	}

	private void fireSampleSizeInBitsChanged(int oldValue, int newValue) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : audioProjectChangeListeners) {
				listener.audioProjectSampleSizeInBitsChanged(oldValue, newValue);
			}
		}
	}

	public AudioChannels getChannels() {
		return channels;
	}

	public void setChannels(AudioChannels channels) {
		AudioChannels oldChannels = this.channels;
		this.channels = channels;
		if (oldChannels != channels) {
			channelsChangeTime = System.currentTimeMillis();
			fireChannelsChanged(oldChannels, channels);
		}
	}

	private void fireChannelsChanged(AudioChannels oldValue, AudioChannels newValue) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : audioProjectChangeListeners) {
				listener.audioProjectChannelsChanged(oldValue, newValue);
			}
		}
	}

	public boolean isBigEndian() {
		return bigEndian;
	}

	public void setBigEndian(boolean defaultBigEndian) {
		boolean oldBigEndian = this.bigEndian;
		this.bigEndian = defaultBigEndian;
		if (oldBigEndian != bigEndian) {
			bigEndianChangeTime = System.currentTimeMillis();
			fireBigEndianChanged(oldBigEndian, bigEndian);
		}
	}

	private void fireBigEndianChanged(boolean oldValue, boolean newValue) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : audioProjectChangeListeners) {
				listener.audioProjectBigEndianChanged(oldValue, newValue);
			}
		}
	}

	public File getProjectFile() {
		return projectFile;
	}

	public void setProjectFile(File projectFile) {
		this.projectFile = projectFile;
	}

	public long getFramePosition() {
		return framePosition;
	}

	public void setFramePosition(long framePosition) throws IOException {
		long oldFramePosition = this.framePosition;
		if (oldFramePosition != framePosition) {
			this.framePosition = framePosition;
			fireFramePositionChanged(oldFramePosition, framePosition);
		}
	}

	private void fireFramePositionChanged(long oldFramePosition, long newFramePosition) throws IOException {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : audioProjectChangeListeners) {
				listener.audioProjectFramePositionChanged(oldFramePosition, newFramePosition);
			}
		}
	}

	public long getMaxFrame() {
		long ret = 0;
		for (AudioTrack track : getAudioTracks()) {
			if (!track.isEnabled()) {
				continue;
			}
			SourceAudioDevice sourceAudioDevice = track.getSourceAudioDevice();
			if (sourceAudioDevice != null) {
				long trackLen = sourceAudioDevice.getMaxFrame();
				ret = Math.max(ret, trackLen);
			}
		}
		return ret;
	}

	public void changeFrameLength() {
		fireFrameLengthChanged();
	}

	private void fireFrameLengthChanged() {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : new LinkedHashSet<AudioProjectListener>(audioProjectChangeListeners)) {
				listener.audioProjectFrameLengthChanged();
			}
		}
	}

	public long getSeconds() {
		long frameLength = getMaxFrame();
		if (frameLength < 0) {
			return 0;
		}
		long rate = getFrameRate();
		return frameLength / rate;
	}

	public long toFrames(long seconds) {
		return getFrameRate() * seconds;
	}

	public long toSeconds(long frames) {
		return frames / getFrameRate();
	}

	public void addAudioProjectChangeListener(AudioProjectListener listener) {
		if (this.audioProjectChangeListeners == null) {
			audioProjectChangeListeners = new LinkedHashSet<AudioProjectListener>();
		}
		audioProjectChangeListeners.add(listener);
	}

	public void removeAudioProjectChangeListener(AudioProjectListener listener) {
		if (this.audioProjectChangeListeners == null) {
			return;
		}
		audioProjectChangeListeners.remove(listener);
	}

	public AudioDeviceChain getMasterAudioDeviceChain() {
		return masterAudioDeviceChain;
	}

	public int getFrameSizeInBytes() {
		return sampleSizeInBits * channels.getChannelCount() / 8;
	}

	public int getFrameRate() {
		return sampleRate;
	}

	public float getNyquistFrequency() {
		return sampleRate / 2F;
	}

	public float getDynamicRange() {
		return (float) (20D * log10((2 << sampleSizeInBits - 2) - 1));
	}

	public AudioFormat getAudioFormat() {
		return createAudioFormat(sampleRate, sampleSizeInBits, channels, bigEndian);
	}

	static boolean notEqual(Object o1, Object o2) {
		return (o1 == null && o2 != null) || (o1 != null && !o1.equals(o2));
	}

	/**
	 * Identifies this AudioProject instance at runtime.
	 */
	public String getUniqueName() {
		StringBuilder sb = new StringBuilder();
		if (name != null) {
			sb.append(name);
			if (projectFile != null) {
				sb.append(" (" + projectFile.getAbsolutePath() + ")");
			} else {
				sb.append(" (" + projectPath + ")");
			}
		} else if (projectFile != null) {
			sb.append(projectFile.getAbsolutePath());
		} else {
			sb.append(projectPath);
		}

		return sb.toString();
	}

	public String getTemplateFileName() {
		return templateFileName;
	}

	public void setTemplateFileName(String templateFile) {
		String oldTemplateFileName = this.templateFileName;
		this.templateFileName = templateFile;
		if (notEqual(oldTemplateFileName, templateFile)) {
			fireTemplateFileNameChanged(oldTemplateFileName, templateFile);
		}
	}

	private void fireTemplateFileNameChanged(String oldValue, String newValue) {
		if (audioProjectChangeListeners != null) {
			for (AudioProjectListener listener : audioProjectChangeListeners) {
				listener.audioProjectTemplateFileNameChanged(oldValue, newValue);
			}
		}
	}

	public void fireAudioProjectOpened() {
		if (audioProjectLifecycleListeners == null) {
			return;
		}
		for (AudioProjectLifecycleListener listener : audioProjectLifecycleListeners) {
			listener.audioProjectOpened(this);
		}
	}

	public void fireAudioProjectClosed() {
		if (audioProjectLifecycleListeners == null) {
			return;
		}
		for (AudioProjectLifecycleListener listener : audioProjectLifecycleListeners) {
			listener.audioProjectClosed(this);
		}
	}

	public void fireAudioProjectStored() {
		if (audioProjectLifecycleListeners == null) {
			return;
		}
		for (AudioProjectLifecycleListener listener : audioProjectLifecycleListeners) {
			listener.audioProjectStored(this);
		}
	}

	public void fireAudioProjectActivated() {
		if (audioProjectLifecycleListeners == null) {
			return;
		}
		for (AudioProjectLifecycleListener listener : audioProjectLifecycleListeners) {
			listener.audioProjectActivated(this);
		}
	}

	public void addAudioProjectLifecycleListener(AudioProjectLifecycleListener listener) {
		if (audioProjectLifecycleListeners == null) {
			audioProjectLifecycleListeners = new LinkedHashSet<AudioProjectLifecycleListener>();
		}
		audioProjectLifecycleListeners.add(listener);
	}

	public void removeAudioProjectLifecycleListener(AudioProjectLifecycleListener listener) {
		if (audioProjectLifecycleListeners == null) {
			return;
		}
		audioProjectLifecycleListeners.remove(listener);
	}

	public PerformanceManager getPerformanceManager() {
		if (performanceManager == null) {
			performanceManager = new PerformanceManager(this);
		}
		return performanceManager;
	}

	public <D extends AudioDevice, S extends Serializable> S getAudioDeviceProperties(Class<D> deviceClass, Class<S> valueClass) {
		if (audioDeviceProperties == null) {
			audioDeviceProperties = new HashMap<Class<? extends AudioDevice>, Serializable>();
		}

		@SuppressWarnings("unchecked")
		S value = (S) audioDeviceProperties.get(deviceClass);
		if (value == null) {
			try {
				value = valueClass.newInstance();
			} catch (InstantiationException e) {
				throw new IllegalArgumentException(e);
			} catch (IllegalAccessException e) {
				throw new IllegalArgumentException(e);
			}
			audioDeviceProperties.put(deviceClass, value);
		}
		return value;
	}

	public AudioTrack addNewAudioTrack(AudioTrack templateAudioTrack) throws InstantiationException, IllegalAccessException {
		AudioTrack ret = new AudioTrack(this);
		ret.setDescription(templateAudioTrack.getDescription());
		ret.setName(templateAudioTrack.getName());
		SourceAudioDevice templateSourceAudioDevice = templateAudioTrack.getSourceAudioDevice();
		if (templateSourceAudioDevice != null) {
			ret.setSourceAudioDevice(createAudioDevice(templateSourceAudioDevice));
		}
		for (ChainableAudioDevice templatePlaybackAudioDevice : templateAudioTrack.getPlaybackAudioDeviceChain().getAudioDevices()) {
			ret.getPlaybackAudioDeviceChain().addAudioDevice(createAudioDevice(templatePlaybackAudioDevice));
		}
		for (ChainableAudioDevice templateCaptureAudioDevice : templateAudioTrack.getCaptureAudioDeviceChain().getAudioDevices()) {
			ret.getCaptureAudioDeviceChain().addAudioDevice(createAudioDevice(templateCaptureAudioDevice));
		}
		return ret;
	}

	private ChainableAudioDevice createAudioDevice(ChainableAudioDevice template) throws InstantiationException, IllegalAccessException {
		ChainableAudioDevice ret = (ChainableAudioDevice) createAudioDeviceAndValues(template);
		ret.setEnabled(template.isEnabled());
		return ret;
	}

	private SourceAudioDevice createAudioDevice(SourceAudioDevice template) throws InstantiationException, IllegalAccessException {
		return (SourceAudioDevice) createAudioDeviceAndValues(template);
	}

	private AudioDevice createAudioDeviceAndValues(AudioDevice template) throws InstantiationException, IllegalAccessException {
		AudioDevice ret;
		if (template instanceof Copyable) {
			ret = (AudioDevice) ((Copyable<?>) template).copy();
		} else {
			ret = template.getClass().newInstance();
		}
		ret.initialize(this);
		ret.synchronizeBy(template);

		UserInterfaceContext<? extends AudioDevice> retUserInterfaceContext = ret.getUserInterfaceContext();
		UserInterfaceContext<? extends AudioDevice> templateUserInterfaceContext = template.getUserInterfaceContext();
		if (retUserInterfaceContext != null && templateUserInterfaceContext != null) {
			retUserInterfaceContext.setHidden(templateUserInterfaceContext.isHidden());
		}
		return ret;
	}

	@Override
	public String getIdentifier() {
		return identifier;
	}

	public void setIdentifier(String identifier) {
		this.identifier = identifier;
	}

	public AudioDevice findAudioDeviceByIdentifier(String identifier) {
		Collection<AudioTrack> audioTracks = getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			AudioDevice audioDevice = audioTrack.findAudioDeviceByIdentifier(identifier);
			if (audioDevice != null) {
				return audioDevice;
			}
		}
		return getMasterAudioDeviceChain().findChainableAudioDeviceByIdentifier(identifier);
	}

	public ChainableAudioDevice findChainableAudioDeviceByIdentifier(String identifier) {
		Collection<AudioTrack> audioTracks = getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			ChainableAudioDevice audioDevice = audioTrack.findChainableAudioDeviceByIdentifier(identifier);
			if (audioDevice != null) {
				return audioDevice;
			}
		}
		return getMasterAudioDeviceChain().findChainableAudioDeviceByIdentifier(identifier);
	}

	public AudioTrack findAudioTrackByIdentifier(String identifier) {
		Collection<AudioTrack> audioTracks = getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			if (audioTrack.getIdentifier().equals(identifier)) {
				return audioTrack;
			}
		}
		return null;
	}

	public Collection<AudioDevice> getAllAudioDevices() {
		Collection<AudioDevice> ret = new LinkedHashSet<AudioDevice>();
		Collection<AudioTrack> audioTracks = getAudioTracks();
		ret.addAll(getMasterAudioDeviceChain().getAudioDevices());
		for (AudioTrack audioTrack : audioTracks) {
			ret.add(audioTrack.getSourceAudioDevice());
			ret.addAll(audioTrack.getPlaybackAudioDeviceChain().getAudioDevices());
			ret.addAll(audioTrack.getCaptureAudioDeviceChain().getAudioDevices());
		}
		return ret;
	}

	public AudioDeviceChain findAudioDeviceChainByIdentifier(String identifier) {
		Collection<AudioTrack> audioTracks = getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			AudioDeviceChain audioDeviceChain = audioTrack.findAudioDeviceChainByIdentifier(identifier);
			if (audioDeviceChain != null) {
				return audioDeviceChain;
			}
		}
		if (masterAudioDeviceChain.getIdentifier().equals(identifier)) {
			return masterAudioDeviceChain;
		}
		return null;
	}

	@Override
	public void synchronizeBy(Synchronizable source) throws InstantiationException, IllegalAccessException {
		if (source instanceof AudioProject && source.getIdentifier().equals(getIdentifier())) {
			AudioProject sourceProject = (AudioProject) source;
			if (greaterOrBothZero(sourceProject.nameChangeTime, nameChangeTime)) {
				setName(sourceProject.getName());
				nameChangeTime = sourceProject.nameChangeTime;
			}
			if (greaterOrBothZero(sourceProject.bigEndianChangeTime, bigEndianChangeTime)) {
				setBigEndian(sourceProject.isBigEndian());
				bigEndianChangeTime = sourceProject.bigEndianChangeTime;
			}
			if (greaterOrBothZero(sourceProject.channelsChangeTime, channelsChangeTime)) {
				setChannels(sourceProject.getChannels());
				channelsChangeTime = sourceProject.channelsChangeTime;
			}
			if (greaterOrBothZero(sourceProject.sampleRateChangeTime, sampleRateChangeTime)) {
				setSampleRate(sourceProject.getSampleRate());
				sampleRateChangeTime = sourceProject.sampleRateChangeTime;
			}
			if (greaterOrBothZero(sourceProject.sampleSizeInBitsChangeTime, sampleSizeInBitsChangeTime)) {
				setSampleSizeInBits(sourceProject.getSampleSizeInBits());
				sampleSizeInBitsChangeTime = sourceProject.sampleSizeInBitsChangeTime;
			}

			Collection<AudioTrack> sourceTracks = sourceProject.getAudioTracks();
			for (AudioTrack sourceTrack : sourceTracks) {
				if (isRemoved(sourceTrack)) {
					// ignore this removed track, the source does not yet know
					// about
					// it's deletion
					continue;
				}
				String sourceTrackIdentifier = sourceTrack.getIdentifier();
				AudioTrack targetTrack = findAudioTrackByIdentifier(sourceTrackIdentifier);
				if (targetTrack == null) {
					targetTrack = new AudioTrack(this);
					targetTrack.setIdentifier(sourceTrackIdentifier);
					// this status is only synced on new instances because it is
					// user specific and has no further relevance
					targetTrack.setHidden(sourceTrack.isHidden());
					targetTrack.getPlaybackAudioDeviceChain().setIdentifier(sourceTrack.getPlaybackAudioDeviceChain().getIdentifier());
					targetTrack.getCaptureAudioDeviceChain().setIdentifier(sourceTrack.getCaptureAudioDeviceChain().getIdentifier());
				}
				targetTrack.synchronizeBy(sourceTrack);
			}
			Collection<AudioTrack> targetTracks = getAudioTracks();
			for (AudioTrack targetTrack : new LinkedHashSet<AudioTrack>(targetTracks)) {
				if (sourceProject.isRemoved(targetTrack)) {
					removeAudioTrack(targetTrack);
				}
			}
			// asserts that the master device chains have the same identifier
			masterAudioDeviceChain.setIdentifier(sourceProject.masterAudioDeviceChain.getIdentifier());
			masterAudioDeviceChain.synchronizeBy(sourceProject.masterAudioDeviceChain);
		}
		lastSyncTime = System.currentTimeMillis();
	}

	static boolean greaterOrBothZero(long source, long target) {
		return source > target || (source == 0 && target == 0);
	}

	@Override
	public Collection<BinaryDataset> getAllBinaryDatasets() {
		Collection<BinaryDataset> ret = new HashSet<BinaryDataset>();
		Collection<AudioTrack> audioTracks = getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			ret.addAll(audioTrack.getAllBinaryDatasets());
		}
		ret.addAll(masterAudioDeviceChain.getAllBinaryDatasets());
		return ret;
	}

	public long getLastSyncTime() {
		return lastSyncTime;
	}

	public void incrementFramePosition() throws IOException {
		long position = getFramePosition();
		position += getFrameRate();
		long maxFrame = getMaxFrame();
		if (position >= maxFrame) {
			position = maxFrame;
		}
		setFramePosition(position);
	}

	public void decrementFramePosition() throws IOException {
		long position = getFramePosition();
		position -= getFrameRate();
		if (position < 0) {
			position = 0;
		}
		setFramePosition(position);
	}

	public String getRelativePath(String fileName) {
		File file = new File(fileName);
		if (file.isAbsolute()) {
			String parent = file.getParent();
			String projectPath = getProjectPath();
			if (fileSeparator == null) {
				fileSeparator = System.getProperty("file.separator");
			}
			if (projectPath.equals(parent + fileSeparator)) {
				return file.getName();
			} else if (parent.startsWith(projectPath)) {
				return fileName.substring(projectPath.length());
			}
			return fileName;
		} else {
			return fileName;
		}
	}

	public File getRelativeFile(String fileName) {
		File file = new File(fileName);
		if (file.isAbsolute()) {
			return file;
		}
		return new File(getProjectPath() + fileName);
	}
	
	Monitorable findMonitor(AudioDevice device) {
		Collection<AudioTrack> audioTracks = getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			Monitorable monitor = audioTrack.findMonitor(device);
			if (monitor != null) {
				return monitor;
			}
		}
		Monitorable monitor = masterAudioDeviceChain.findMonitor(device);
		return monitor;
	}
}
