/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import static org.audiochain.model.AudioProject.*;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;

import org.audiochain.devices.recording.RecordingAudioDevice;
import org.audiochain.io.AudioIOException;

public class AudioTrack implements Serializable, BinaryDatasetProvider {
	private static final long serialVersionUID = 1L;
	private String name;
	private long nameChangeTime;
	private String description;
	private long descriptionChangeTime;
	private SourceAudioDevice sourceAudioDevice;
	private long sourceAudioDeviceReplaceTime;
	private AudioDeviceChain playbackAudioDeviceChain;
	private AudioDeviceChain captureAudioDeviceChain;
	private AudioProject audioProject;
	private boolean hidden;
	private boolean enabled = true;
	private long enabledChangeTime;
	private String identifier = IdentifierGenerator.generateIdentifier();
	private transient boolean selectedForRecording;
	private transient Collection<AudioTrackListener> audioTrackChangeListeners;

	/**
	 * For Deserialization only
	 */
	@Deprecated
	AudioTrack() {
	}

	public AudioTrack(AudioProject project) {
		playbackAudioDeviceChain = new AudioDeviceChain(project);
		captureAudioDeviceChain = new AudioDeviceChain(project);
		project.addAudioTrack(this);
	}

	public void delete() {
		if (sourceAudioDevice != null) {
			sourceAudioDevice.destroy();
		}
		playbackAudioDeviceChain.removeAllAudioDevices();
		captureAudioDeviceChain.removeAllAudioDevices();
		if (audioTrackChangeListeners != null) {
			audioTrackChangeListeners.clear();
		}
	}

	public void selectForRecording() throws IOException, AudioIOException {
		for (AudioTrack otherTrack : audioProject.getAudioTracks()) {
			otherTrack.setSelectedForRecording(false);
		}
		setSelectedForRecording(true);
	}

	public void unselectForRecording() throws IOException, AudioIOException {
		setSelectedForRecording(false);
	}

	public boolean isSelectedForRecording() {
		return selectedForRecording;
	}

	private void setSelectedForRecording(boolean selectedForRecording) throws IOException, AudioIOException {
		boolean oldValue = this.selectedForRecording;
		this.selectedForRecording = selectedForRecording;
		if (oldValue != selectedForRecording) {
			fireSelectedForRecordingChanged(oldValue, selectedForRecording);
		}
	}

	private void fireSelectedForRecordingChanged(boolean oldValue, boolean newValue) throws IOException, AudioIOException {
		if (audioTrackChangeListeners != null) {
			for (AudioTrackListener listener : new LinkedHashSet<AudioTrackListener>(audioTrackChangeListeners)) {
				listener.audioTrackSelectedForRecordingChanged(this, oldValue, newValue);
			}
		}
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		String oldName = this.name;
		this.name = name;
		if (notEqual(oldName, name)) {
			nameChangeTime = System.currentTimeMillis();
			fireNameChanged(oldName, name);
		}
	}

	private void fireNameChanged(String oldValue, String newValue) {
		if (audioTrackChangeListeners != null) {
			for (AudioTrackListener listener : new LinkedHashSet<AudioTrackListener>(audioTrackChangeListeners)) {
				listener.audioTrackNameChanged(this, oldValue, newValue);
			}
		}
	}

	public String getDescription() {
		return description;
	}

	public void setDescription(String description) {
		String oldDescription = this.description;
		this.description = description;
		if (notEqual(oldDescription, description)) {
			descriptionChangeTime = System.currentTimeMillis();
			fireDescriptionChanged(oldDescription, description);
		}
	}

	public boolean isHidden() {
		return hidden;
	}

	public void setHidden(boolean hidden) {
		boolean oldValue = this.hidden;
		if (oldValue != hidden) {
			this.hidden = hidden;
			fireHiddenChanged(oldValue, hidden);
		}
	}

	private void fireHiddenChanged(boolean oldValue, boolean newValue) {
		if (audioTrackChangeListeners != null) {
			for (AudioTrackListener listener : new LinkedHashSet<AudioTrackListener>(audioTrackChangeListeners)) {
				listener.audioTrackHiddenChanged(this, oldValue, newValue);
			}
		}
	}

	public boolean isEnabled() {
		return enabled;
	}

	public void setEnabled(boolean enabled) {
		boolean oldValue = this.enabled;
		playbackAudioDeviceChain.setEnabled(enabled);
		captureAudioDeviceChain.setEnabled(enabled);
		if (oldValue != enabled) {
			this.enabled = enabled;
			enabledChangeTime = System.currentTimeMillis();
			audioProject.changeFrameLength();
			fireEnabledChanged(oldValue, enabled);
		}
	}

	private void fireEnabledChanged(boolean oldValue, boolean newValue) {
		if (audioTrackChangeListeners != null) {
			for (AudioTrackListener listener : new LinkedHashSet<AudioTrackListener>(audioTrackChangeListeners)) {
				listener.audioTrackEnabledChanged(this, oldValue, newValue);
			}
		}
	}

	private void fireDescriptionChanged(String oldValue, String newValue) {
		if (audioTrackChangeListeners != null) {
			for (AudioTrackListener listener : new LinkedHashSet<AudioTrackListener>(audioTrackChangeListeners)) {
				listener.audioTrackDescriptionChanged(this, oldValue, newValue);
			}
		}
	}

	/**
	 * Get the currently set {@link SourceAudioDevice}. Per default this value
	 * equals to the {@link RecordingAudioDevice}.
	 * 
	 * @return the currently set {@link SourceAudioDevice}.
	 */
	public SourceAudioDevice getSourceAudioDevice() {
		return sourceAudioDevice;
	}

	/**
	 * Defines the {@link SourceAudioDevice} for this track.
	 * 
	 * @param sourceAudioDevice
	 *            this value is mandatory and an
	 *            {@link IllegalArgumentException} is thrown if this parameter
	 *            is null.
	 */
	public void setSourceAudioDevice(SourceAudioDevice sourceAudioDevice) {
		SourceAudioDevice oldDevice = this.sourceAudioDevice;
		boolean differentInstance = notEqual(oldDevice, sourceAudioDevice);
		if (differentInstance) {
			if (oldDevice != null) {
				oldDevice.destroy();
			}
			this.sourceAudioDevice = sourceAudioDevice;
			this.sourceAudioDeviceReplaceTime = System.currentTimeMillis();
			if (sourceAudioDevice != null) {
				sourceAudioDevice.initialize(audioProject);
			}
			fireSourceAudioDeviceChanged(oldDevice, sourceAudioDevice);
		}
	}

	private void fireSourceAudioDeviceChanged(SourceAudioDevice oldValue, SourceAudioDevice newValue) {
		if (audioTrackChangeListeners != null) {
			for (AudioTrackListener listener : new LinkedHashSet<AudioTrackListener>(audioTrackChangeListeners)) {
				listener.audioTrackSourceAudioDeviceChanged(this, oldValue, newValue);
			}
		}
	}

	public void addAudioTrackChangeListener(AudioTrackListener listener) {
		if (this.audioTrackChangeListeners == null) {
			audioTrackChangeListeners = new LinkedHashSet<AudioTrackListener>();
		}
		audioTrackChangeListeners.add(listener);
	}

	public void removeAudioTrackChangeListener(AudioTrackListener listener) {
		if (this.audioTrackChangeListeners == null) {
			return;
		}
		audioTrackChangeListeners.remove(listener);
	}

	void setAudioProject(AudioProject project) {
		this.audioProject = project;
		if (audioProject != null) {
			if (playbackAudioDeviceChain != null) {
				playbackAudioDeviceChain.setAudioProject(audioProject);
			}
			if (captureAudioDeviceChain != null) {
				captureAudioDeviceChain.setAudioProject(audioProject);
			}
			if (sourceAudioDevice != null) {
				sourceAudioDevice.initialize(audioProject);
			}
		}
	}

	@Override
	public String toString() {
		return getClass().getSimpleName() + " " + name;
	}

	public AudioDeviceChain getPlaybackAudioDeviceChain() {
		return playbackAudioDeviceChain;
	}

	public AudioDeviceChain getCaptureAudioDeviceChain() {
		return captureAudioDeviceChain;
	}

	public AudioProject getAudioProject() {
		return audioProject;
	}

	public void toggleSelectForRecording() throws IOException, AudioIOException {
		if (hasRecordableSourceAudioDevice()) {
			if (isSelectedForRecording()) {
				unselectForRecording();
			} else {
				selectForRecording();
			}
		}
	}

	private boolean hasRecordableSourceAudioDevice() {
		SourceAudioDevice sourceAudioDevice = getSourceAudioDevice();
		return sourceAudioDevice != null && sourceAudioDevice.isRecordable();
	}

	public int getNumber() {
		List<AudioTrack> list = new ArrayList<AudioTrack>(getAudioProject().getAudioTracks());
		return list.indexOf(this) + 1;
	}

	@Override
	public String getIdentifier() {
		return identifier;
	}

	public void setIdentifier(String identifier) {
		this.identifier = identifier;
	}

	public AudioDevice findAudioDeviceByIdentifier(String identifier) {
		AudioDevice audioDevice = findChainableAudioDeviceByIdentifier(identifier);
		if (audioDevice != null) {
			return audioDevice;
		}
		SourceAudioDevice sourceAudioDevice = getSourceAudioDevice();
		if (sourceAudioDevice != null && sourceAudioDevice.getIdentifier().equals(identifier)) {
			return sourceAudioDevice;
		}
		return null;
	}

	public ChainableAudioDevice findChainableAudioDeviceByIdentifier(String identifier) {
		ChainableAudioDevice audioDevice = getPlaybackAudioDeviceChain().findChainableAudioDeviceByIdentifier(identifier);
		if (audioDevice != null) {
			return audioDevice;
		}
		audioDevice = getCaptureAudioDeviceChain().findChainableAudioDeviceByIdentifier(identifier);
		if (audioDevice != null) {
			return audioDevice;
		}
		return null;
	}

	public AudioDeviceChain findAudioDeviceChainByIdentifier(String identifier) {
		if (playbackAudioDeviceChain.getIdentifier().equals(identifier)) {
			return playbackAudioDeviceChain;
		}
		if (captureAudioDeviceChain.getIdentifier().equals(identifier)) {
			return captureAudioDeviceChain;
		}
		return null;
	}

	@Override
	public void synchronizeBy(Synchronizable source) throws InstantiationException, IllegalAccessException {
		if (source instanceof AudioTrack && source.getIdentifier().equals(getIdentifier())) {
			AudioTrack sourceTrack = (AudioTrack) source;
			if (greaterOrBothZero(sourceTrack.nameChangeTime, nameChangeTime)) {
				setName(sourceTrack.getName());
				nameChangeTime = sourceTrack.nameChangeTime;
			}
			if (greaterOrBothZero(sourceTrack.descriptionChangeTime, descriptionChangeTime)) {
				setDescription(sourceTrack.getDescription());
				descriptionChangeTime = sourceTrack.descriptionChangeTime;
			}
			if (greaterOrBothZero(sourceTrack.enabledChangeTime, enabledChangeTime)) {
				setEnabled(sourceTrack.isEnabled());
				enabledChangeTime = sourceTrack.enabledChangeTime;
			}

			SourceAudioDevice sourceTrackSourceAudioDevice = sourceTrack.getSourceAudioDevice();
			if (sourceTrackSourceAudioDevice == null) {
				if (greaterOrBothZero(sourceTrack.sourceAudioDeviceReplaceTime, sourceAudioDeviceReplaceTime)) {
					setSourceAudioDevice(null);
					sourceAudioDeviceReplaceTime = sourceTrack.sourceAudioDeviceReplaceTime;
				}
			} else {
				SourceAudioDevice sourceAudioDevice = getSourceAudioDevice();
				if (greaterOrBothZero(sourceTrack.sourceAudioDeviceReplaceTime, sourceAudioDeviceReplaceTime)) {
					String sourceTrackSourceDeviceIdentifier = sourceTrackSourceAudioDevice.getIdentifier();
					if (sourceAudioDevice == null || !sourceAudioDevice.getIdentifier().equals(sourceTrackSourceDeviceIdentifier)) {
						if (sourceTrackSourceAudioDevice instanceof Copyable) {
							sourceAudioDevice = (SourceAudioDevice) ((Copyable<?>) sourceTrackSourceAudioDevice).copy();
						} else {
							sourceAudioDevice = sourceTrackSourceAudioDevice.getClass().newInstance();
						}
						sourceAudioDevice.setIdentifier(sourceTrackSourceDeviceIdentifier);
						// this status is only synced on new instances because
						// it is user specific and has no further relevance
						sourceAudioDevice.getUserInterfaceContext().setHidden(sourceTrackSourceAudioDevice.getUserInterfaceContext().isHidden());
						setSourceAudioDevice(sourceAudioDevice);
						sourceAudioDeviceReplaceTime = sourceTrack.sourceAudioDeviceReplaceTime;
					}
				}

				if (sourceAudioDevice != null) {
					sourceAudioDevice.synchronizeBy(sourceTrackSourceAudioDevice);
				}
			}

			playbackAudioDeviceChain.synchronizeBy(sourceTrack.playbackAudioDeviceChain);
			captureAudioDeviceChain.synchronizeBy(sourceTrack.captureAudioDeviceChain);
		}
	}

	@Override
	public Collection<BinaryDataset> getAllBinaryDatasets() {
		Collection<BinaryDataset> ret = new HashSet<BinaryDataset>();
		ret.addAll(playbackAudioDeviceChain.getAllBinaryDatasets());
		ret.addAll(captureAudioDeviceChain.getAllBinaryDatasets());
		if (sourceAudioDevice instanceof BinaryDatasetProvider) {
			ret.addAll(((BinaryDatasetProvider) sourceAudioDevice).getAllBinaryDatasets());
		}
		return ret;
	}

	Monitorable findMonitor(AudioDevice device) {
		if (device == sourceAudioDevice) {
			Monitorable sourceMonitor = playbackAudioDeviceChain.getSourceMonitor();
			if (sourceMonitor != null) {
				return sourceMonitor;
			}
			sourceMonitor = captureAudioDeviceChain.getSourceMonitor();
			if (sourceMonitor != null) {
				return sourceMonitor;
			}
			return null;
		} else {
			Monitorable sourceMonitor = playbackAudioDeviceChain.findMonitor(device);
			if (sourceMonitor != null) {
				return sourceMonitor;
			}
			sourceMonitor = captureAudioDeviceChain.findMonitor(device);
			return sourceMonitor;
		}
	}
}
