/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import static org.audiochain.model.ObjectStore.*;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.util.Calendar;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import org.audiochain.model.ObjectStore.ReadUpdateCallback;

/**
 * Used to transport files via the network. The file that is gonna be
 * transmitted, becomes zipped to use the network bandwidth efficiently. The
 * files modification info and the name of the file will be reconstructed at the
 * destination.
 */
public class BinaryFileDataset implements BinaryDataset {
	private static final String FILENAME_TIME_SEPARATOR = "#";
	private transient long length;
	private transient boolean lengthComputed;
	private transient final AudioProject associatedAudioProject;
	private transient final File fileToTransmit;

	// for instantiation by reflection
	public BinaryFileDataset() {
		this(null, null);
	}

	public BinaryFileDataset(AudioProject associatedAudioProject, File fileToTransmit) {
		this.associatedAudioProject = associatedAudioProject;
		this.fileToTransmit = fileToTransmit;
	}

	protected File getFile() {
		return fileToTransmit;
	}

	@Override
	public AudioProject getBinaryDatasetAssociatedAudioProject() {
		return associatedAudioProject;
	}

	@Override
	public String getBinaryDatasetIdentifier() {
		File file = getFile();
		// there is no reliable precision getting the last modification time on
		// several calls
		long lastModified = file.lastModified();
		Calendar cal = Calendar.getInstance();
		cal.setTimeInMillis(lastModified);
		cal.set(Calendar.MILLISECOND, 0);
		cal.set(Calendar.SECOND, 0);
		long roundedLastModified = cal.getTimeInMillis();
		return file.getName() + FILENAME_TIME_SEPARATOR + roundedLastModified;
	}

	@Override
	public InputStream getBinaryDatasetInputStream() throws IOException {
		return createInputStream();
	}

	@Override
	public long getBinaryDatasetLength() throws IOException {
		if (!lengthComputed) {
			createInputStream();
		}
		return length;
	}

	private InputStream createInputStream() throws IOException {
		File file = getFile();
		BufferedInputStream bis = new BufferedInputStream(new FileInputStream(file));
		File tmpZipFile = File.createTempFile(file.getName(), "-tmp.zip");
		FileOutputStream fos = new FileOutputStream(tmpZipFile);
		BufferedOutputStream bos = new BufferedOutputStream(fos);
		ZipOutputStream zos = new ZipOutputStream(bos);
		ZipEntry entry = new ZipEntry(file.getName());
		zos.putNextEntry(entry);
		byte[] b = new byte[1024];
		int len;
		while ((len = bis.read(b)) != -1) {
			zos.write(b, 0, len);
		}
		zos.closeEntry();
		zos.close();
		bos.close();
		fos.close();
		length = tmpZipFile.length();
		lengthComputed = true;
		tmpZipFile.deleteOnExit();
		return new BufferedInputStream(new FileInputStream(tmpZipFile));
	}

	@Override
	public void processBinaryDataset(AudioProject associatedAudioProject, String binaryDatasetIdentifier, InputStream is, long length, final BinaryDatasetProcessCallback callback) throws IOException,
			IllegalAccessException, ClassNotFoundException, InstantiationException, NoSuchMethodException, InvocationTargetException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		readExact(is, baos, length, new ReadUpdateCallback() {
			@Override
			public void readUpdate(long read, long total) {
				callback.binaryDatasetProcessUpdate(read, total);
			}
		});
		baos.close();
		byte[] bytes = baos.toByteArray();
		ByteArrayInputStream bais = new ByteArrayInputStream(bytes);

		String[] binaryDatasetFileAndLastModified = binaryDatasetIdentifier.split(FILENAME_TIME_SEPARATOR);
		String fileName = binaryDatasetFileAndLastModified[0];

		String projectPath = associatedAudioProject.getProjectPath();
		String tmpFilePath = projectPath + "." + fileName + ".tmp";
		File tmpFile = new File(tmpFilePath);
		BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(tmpFile));
		byte[] b = new byte[1024];
		int len;
		ZipInputStream zis = new ZipInputStream(bais);
		zis.getNextEntry();
		while ((len = zis.read(b)) != -1) {
			bos.write(b, 0, len);
		}
		bos.close();
		zis.closeEntry();
		zis.close();
		bais.close();

		String filePath = projectPath + fileName;
		File file = new File(filePath);
		boolean renamed = tmpFile.renameTo(file);
		assert renamed;
		// the timestamp must be set after the file is written / filled with
		// data
		long lastModified = Long.parseLong(binaryDatasetFileAndLastModified[1]);
		file.setLastModified(lastModified);
	}

}
