/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

public interface CommandLineUserInterface {
	
	/**
	 * Format that has to be used for formatting help Strings.
	 * The first String is the position for the propertyName, the second is for positioning the description, options and/or ranges.
	 * For example: "    AddFile <filename>        Add the <filename> to the selected track."
	 */
	public static final String HELP_FORMAT = "    %-25s %-15s%n";
	
	/**
	 * Format that has to be used for formatting help Strings.
	 * The first String is the position for the propertyName, the second is for positioning the default value of the property,
	 * the third is for positioning of description, options and/or ranges.
	 * For example: "    Band-Stop-CutOff          ◴1000.0         min: 0.0, max: 22050.0"
	 */
	public static final String HELP_FORMAT_WITH_DEFAULT_VALUE = "    %-25s %-15s %-15s%n";
	
	/**
	 * @return a String that is displayed when the user requests for the help command. The help String should have
	 * a format like the above HELP_FORMAT. This can be passed to the {@link java.util.Formatter} class.
	 */
	public String getHelpString();
	
	/**
	 * @return the String will be displayed within the track. Keep this String short.
	 */
	public String getDisplayString();
	
	/**
	 * @return the displayed String when the performance is running.
	 */
	public String getPerformanceDisplayString();
	
	/**
	 * This is the only way to enable the user to modify any audio device properties by it's name and a given String value.
	 * This value has to be transformed into the correct data type (float, int, boolean, etc) and passed to the audio device.
	 * Any properties that can be modified must be documented within the getHelpString() method so that the user can get
	 * the information of all possible properties.
	 * 
	 * @param propertyName a logical name that identifies the property. This name must be unique over all audio devices within one track.
	 * @param newValue a value that can be given by the user and must be parsed by this method. If the value can not be parsed or if it
	 * is not in any desired numeric range, an Exception can be thrown which is displayed to the user by it's message. To check
	 * ranges, a convenience Method of the Abstract command line implementation can be used.
	 * @return true, if the property was found, false, if the property is undefined. The command line interface, which calls this method
	 * stops after the first true return value.
	 * @throws Exception if the given newValue cannot be parsed, or other technical Exceptions should be displayed to the user. This Exception
	 * can be understood as a Veto.
	 */
	public boolean modifyAudioDevice(String propertyName, String newValue) throws Exception;
	
	public void addCommandLineUserInterfaceListener(CommandLineUserInterfaceListener listener);
	public void removeCommandLineUserInterfaceListener(CommandLineUserInterfaceListener listener);
}
