/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import static java.lang.Math.*;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class ObjectStore {
	private static final String HEX_VALUES = "0123456789ABCDEF";
	private static final String LINESEP = System.getProperty("line.separator");
	
	public static Object readObject(InputStream is) throws IOException, IllegalAccessException, ClassNotFoundException, InstantiationException, NoSuchMethodException, InvocationTargetException {
		return readObject(new BufferedReader(new InputStreamReader(is)));
	}

	public static Object readObject(BufferedReader br) throws IOException, IllegalAccessException, ClassNotFoundException, InstantiationException, NoSuchMethodException, InvocationTargetException {
		/* matches on "org.audiochain.model.AudioProject{659025413}(" */
		final Pattern classInCollectionPattern = Pattern.compile("^\\s*([\\w\\.]*)\\{(-?\\d*)\\}\\(\\s*$");
		
		/* matches on '"some string"' */
		final Pattern stringInCollectionPattern = Pattern.compile("^\\s*\"(.*)\"\\s*$");

		/* matches on "  selectedPlaybackMixer =   Some crazy Value   " */
		final Pattern fieldValuePattern = Pattern.compile("^\\s*([\\w\\_]*)\\s*=\\s*(.*\\S)\\s*$");

		/*
		 * matches on
		 * "          levelAudioDevice=          org.audiochain.devices.level.LevelAudioDevice{2019743861}("
		 */
		final Pattern fieldToObjectPattern = Pattern.compile("^\\s*([\\w\\_]*)\\s*=\\s*([\\w\\.]*)\\{(-?\\d*)\\}\\(\\s*$");

		/* matches on "  project = _REF15735326_  " */
		final Pattern fieldReferencePattern = Pattern.compile("^\\s*([\\w\\_]*)\\s*=\\s*_REF(-?\\d*)_\\s*$");

		/* matches on "  )  " */
		final Pattern objectEndPattern = Pattern.compile("^\\s*\\)\\s*$");

		/* matches on "  ]  " */
		final Pattern collectionEndPattern = Pattern.compile("^\\s*\\]\\s*$");

		/* matches on "  tracks  =  [  " */
		final Pattern collectionStartPattern = Pattern.compile("^\\s*([\\w\\_]*)\\s*=\\s*\\[\\s*$");

		/* matches on "  audioDeviceProperties = {  " */
		final Pattern mapStartPattern = Pattern.compile("^\\s*([\\w\\_]*)\\s*=\\s*\\{\\s*$");

		/* matches on "  }  " */
		final Pattern mapEndPattern = Pattern.compile("^\\s*\\}\\s*$");

		/*
		 * matches on
		 * "org.audiochain.devices.recording.RecordingAudioDevice=org.audiochain.devices.recording.RecordingAudioDeviceProjectProperties{4064057}("
		 */
		final Pattern mapEntryClassKeyToObjectValuePattern = Pattern.compile("^\\s*([\\w]+\\.[\\w\\.]+)\\s*=\\s*([\\w\\.]*)\\{(-?\\d*)\\}\\(\\s*$");

		/*
		 * matches on
		 * ""SomeString"=0123456789L"
		 */
		final Pattern mapEntryStringKeyToLongValuePattern = Pattern.compile("^\\s*\"(.*)\"\\s*=\\s*(\\d+)L\\s*$");
		/*
		 * matches on
		 * ""SomeString"=012.3456789F"
		 */
		final Pattern mapEntryStringKeyToFloatValuePattern = Pattern.compile("^\\s*\"(.*)\"\\s*=\\s*(\\d+\\.\\d+)F\\s*$");

		final Stack<Object> objectStack = new Stack<Object>();
		final Stack<Collection<Object>> collectionStack = new Stack<Collection<Object>>();
		final Map<Collection<Object>, Class<?>> collectionTypeMap = new HashMap<Collection<Object>, Class<?>>();
		final Stack<Map<Object, Object>> mapStack = new Stack<Map<Object, Object>>();
		final Map<Integer, Object> objectMap = new HashMap<Integer, Object>();

		Object ret = null;
		String line;
		int lineCount = 0;
		while ((line = br.readLine()) != null) {
			lineCount++;
			Matcher m;
			if ((m = mapEntryStringKeyToLongValuePattern.matcher(line)).matches()) {
				String stringKey = m.group(1);
				Long longValue = Long.parseLong(m.group(2));
				mapStack.peek().put(stringKey, longValue);
			} else if ((m = mapEntryStringKeyToFloatValuePattern.matcher(line)).matches()) {
				String stringKey = m.group(1);
				Float floatValue = Float.parseFloat(m.group(2));
				mapStack.peek().put(stringKey, floatValue);
			} else if ((m = mapEntryClassKeyToObjectValuePattern.matcher(line)).matches()) {
				String keyClassName = m.group(1);
				String objectClassName = m.group(2);
				Integer objectId = Integer.parseInt(m.group(3));
				Object newValueObject = createAndPushClass(objectStack, objectMap, objectClassName, objectId);
				Class<?> keyClass = Class.forName(keyClassName);
				if (!mapStack.isEmpty()) {
					mapStack.peek().put(keyClass, newValueObject);
				}
			} else if ((m = mapStartPattern.matcher(line)).matches()) {
				String fieldName = m.group(1);
				createAndPushMap(objectStack, mapStack, lineCount, fieldName);
			} else if ((m = classInCollectionPattern.matcher(line)).matches()) {
				String className = m.group(1);
				Integer objectId = Integer.parseInt(m.group(2));
				Object object = createAndPushClass(objectStack, objectMap, className, objectId);
				if (!collectionStack.isEmpty()) {
					Collection<Object> collection = collectionStack.peek();
					Class<?> collectionType = collectionTypeMap.get(collection);
					if (collectionType != null) {
						if (!collectionType.isAssignableFrom(object.getClass())) {
							throw new IllegalStateException("The object '" + objectId + "' of type '" + object.getClass() + "' cannot be assigned to a collection of type '" + collectionType.getName()
									+ "'. (Line " + lineCount + ")");
						}
					}
					collection.add(object);
				}
			} else if ((m = stringInCollectionPattern.matcher(line)).matches()) {
				String stringValue = m.group(1);
				if (!collectionStack.isEmpty()) {
					Collection<Object> collection = collectionStack.peek();
					Class<?> collectionType = collectionTypeMap.get(collection);
					if (collectionType != null) {
						if (!collectionType.isAssignableFrom(stringValue.getClass())) {
							throw new IllegalStateException("String cannot be assigned to a collection of type '" + collectionType.getName() + "'. (Line " + lineCount + ")");
						}
					}
					collection.add(stringValue);
				}
			} else if ((m = collectionStartPattern.matcher(line)).matches()) {
				String fieldName = m.group(1);
				createAndPushCollection(objectStack, collectionStack, lineCount, fieldName, collectionTypeMap);
			} else if ((m = fieldReferencePattern.matcher(line)).matches()) {
				String fieldName = m.group(1);
				Integer objectRefId = Integer.parseInt(m.group(2));
				resolveAndAssignReference(objectStack, objectMap, lineCount, fieldName, objectRefId);
			} else if ((m = fieldToObjectPattern.matcher(line)).matches()) {
				String fieldName = m.group(1);
				String className = m.group(2);
				Integer objectId = Integer.parseInt(m.group(3));
				createAndAssignNewObject(objectStack, objectMap, lineCount, fieldName, className, objectId);
			} else if ((m = fieldValuePattern.matcher(line)).matches()) {
				String fieldName = m.group(1);
				String fieldValue = m.group(2);
				assignFieldValue(objectStack, lineCount, fieldName, fieldValue);
			} else if ((m = objectEndPattern.matcher(line)).matches()) {
				ret = objectStack.pop();
				if(objectStack.isEmpty()) {
					/* end reached, with this break we can return, although the br does not return with null 
					 * this enables us to read from streams that will not be closed */
					break;
				}
			} else if ((m = collectionEndPattern.matcher(line)).matches()) {
				ret = collectionStack.pop();
			} else if ((m = mapEndPattern.matcher(line)).matches()) {
				ret = mapStack.pop();
			}
		}

		if (!objectStack.isEmpty()) {
			throw new IllegalStateException("Some objects could not be read completely: " + objectStack);
		}

		if (!collectionStack.isEmpty()) {
			throw new IllegalStateException("Some collections could not be read completely: " + collectionStack);
		}

		return ret;
	}

	private static void createAndAssignNewObject(final Stack<Object> objectStack, final Map<Integer, Object> objectMap, int lineCount, String fieldName, String className, Integer objectId)
			throws ClassNotFoundException, InstantiationException, IllegalAccessException, SecurityException, IllegalArgumentException, NoSuchMethodException, InvocationTargetException {
		// first get the current object
		Object object = objectStack.peek();
		// then add the new object to the stack
		Object newObject = createAndPushClass(objectStack, objectMap, className, objectId);
		boolean found = false;
		for (Field field : collectFields(object)) {
			field.setAccessible(true);
			if (Modifier.isStatic(field.getModifiers())) {
				continue;
			}
			if (Modifier.isTransient(field.getModifiers())) {
				continue;
			}
			if (fieldName.equals(field.getName())) {
				found = true;
				field.set(object, newObject);
			}
		}
		if (!found) {
			warn("Field '" + fieldName + "' not found in class '" + object.getClass().getName() + "'. (Line " + lineCount + ")");
		}
	}

	@SuppressWarnings({
			"unchecked", "rawtypes"
	})
	private static void assignFieldValue(final Stack<Object> objectStack, int lineCount, String fieldName, String fieldValue) throws IllegalAccessException {
		Object object = objectStack.peek();
		boolean found = false;
		for (Field field : collectFields(object)) {
			field.setAccessible(true);
			if (Modifier.isStatic(field.getModifiers())) {
				continue;
			}
			if (fieldName.equals(field.getName())) {
				if (Modifier.isTransient(field.getModifiers())) {
					warn("Field '" + fieldName + "' of class '" + object.getClass().getName() + "' became transient");
				}
				found = true;
				Class<?> fieldType = field.getType();
				if (fieldValue.equals("_NULL_")) {
					field.set(object, null);
				} else if (fieldType.isAssignableFrom(String.class)) {
					field.set(object, fieldValue);
				} else if (fieldType.isAssignableFrom(Integer.class) || fieldType.isAssignableFrom(Integer.TYPE)) {
					field.set(object, new Integer(fieldValue));
				} else if (fieldType.isAssignableFrom(Float.class) || fieldType.isAssignableFrom(Float.TYPE)) {
					field.set(object, new Float(fieldValue));
				} else if (fieldType.isAssignableFrom(Boolean.class) || fieldType.isAssignableFrom(Boolean.TYPE)) {
					field.set(object, new Boolean(fieldValue));
				} else if (fieldType.isAssignableFrom(Byte.class) || fieldType.isAssignableFrom(Byte.TYPE)) {
					field.set(object, new Byte(fieldValue));
				} else if (fieldType.isAssignableFrom(Short.class) || fieldType.isAssignableFrom(Short.TYPE)) {
					field.set(object, new Short(fieldValue));
				} else if (fieldType.isAssignableFrom(Character.class) || fieldType.isAssignableFrom(Character.TYPE)) {
					field.set(object, new Character(fieldValue.charAt(0)));
				} else if (fieldType.isAssignableFrom(Long.class) || fieldType.isAssignableFrom(Long.TYPE)) {
					field.set(object, new Long(fieldValue));
				} else if (fieldType.isAssignableFrom(Double.class) || fieldType.isAssignableFrom(Double.TYPE)) {
					field.set(object, new Double(fieldValue));
				} else if (fieldType.isAssignableFrom(Enum.class) || fieldType.isEnum()) {
					field.set(object, Enum.valueOf((Class<Enum>) fieldType, fieldValue));
				} else {
					throw new IllegalStateException("Field '" + fieldName + "' in class '" + object.getClass().getName() + "' contains an unknown value type '" + fieldType + "'. (Line " + lineCount
							+ ")");
				}
			}
		}
		if (!found) {
			warn("Field '" + fieldName + "' not found in class '" + object.getClass().getName() + "'. (Line " + lineCount + ")");
		}
	}

	private static void resolveAndAssignReference(final Stack<Object> objectStack, final Map<Integer, Object> objectMap, int lineCount, String fieldName, Integer objectRefId)
			throws IllegalAccessException {
		Object object = objectStack.peek();
		boolean found = false;
		for (Field field : collectFields(object)) {
			field.setAccessible(true);
			if (Modifier.isStatic(field.getModifiers())) {
				continue;
			}
			if (fieldName.equals(field.getName())) {
				if (Modifier.isTransient(field.getModifiers())) {
					warn("Reference '" + fieldName + "' of class '" + object.getClass().getName() + "' became transient. (Line " + lineCount + ")");
				}
				found = true;
				Object ref = objectMap.get(objectRefId);
				if (ref == null) {
					throw new IllegalStateException("Object (Reference ID " + objectRefId + ") not found for field '" + fieldName + "' of class " + object.getClass().getName()
							+ ". The following objects are present for now: " + objectMap);
				}
				field.set(object, ref);
			}
		}
		if (!found) {
			warn("Reference '" + fieldName + "' not found in class '" + object.getClass().getName() + "'. (Line " + lineCount + ")");
		}
	}

	private static void createAndPushCollection(final Stack<Object> objectStack, final Stack<Collection<Object>> collectionStack, int lineCount, String fieldName,
			Map<Collection<Object>, Class<?>> collectionTypeMap) throws IllegalAccessException {
		Object object = objectStack.peek();
		boolean found = false;
		for (Field field : collectFields(object)) {
			field.setAccessible(true);
			if (Modifier.isStatic(field.getModifiers())) {
				continue;
			}
			if (Modifier.isTransient(field.getModifiers())) {
				continue;
			}
			if (fieldName.equals(field.getName())) {
				found = true;
				Class<?> fieldType = field.getType();
				Collection<Object> collection;
				if (fieldType.isAssignableFrom(List.class)) {
					collection = new ArrayList<Object>();
				} else {
					collection = new LinkedHashSet<Object>();
				}
				field.set(object, collection);
				collectionStack.push(collection);

				// try to obtain the generic type of this collection to check
				// whether the later added objects are of a compatible type
				Type genericType = field.getGenericType();
				if (genericType instanceof ParameterizedType) {
					ParameterizedType pt = (ParameterizedType) genericType;
					for (Type arg : pt.getActualTypeArguments()) {
						if (arg instanceof Class) {
							Class<?> collectionType = (Class<?>) arg;
							collectionTypeMap.put(collection, collectionType);
						}
					}
				}
			}
		}
		if (!found) {
			throw new IllegalStateException("Collection '" + fieldName + "' not found in class '" + object.getClass().getName() + "'. (Line " + lineCount + ")");
		}
	}

	private static void createAndPushMap(final Stack<Object> objectStack, final Stack<Map<Object, Object>> mapStack, int lineCount, String fieldName) throws IllegalAccessException {
		Object object = objectStack.peek();
		boolean found = false;
		for (Field field : collectFields(object)) {
			field.setAccessible(true);
			if (Modifier.isStatic(field.getModifiers())) {
				continue;
			}
			if (Modifier.isTransient(field.getModifiers())) {
				continue;
			}
			if (fieldName.equals(field.getName())) {
				found = true;
				Map<Object, Object> map = new LinkedHashMap<Object, Object>();
				field.set(object, map);
				mapStack.push(map);
			}
		}
		if (!found) {
			throw new IllegalStateException("Map '" + fieldName + "' not found in class '" + object.getClass().getName() + "'. (Line " + lineCount + ")");
		}
	}

	private static Object createAndPushClass(final Stack<Object> objectStack, final Map<Integer, Object> objectMap, String className, Integer objectId) throws ClassNotFoundException,
			InstantiationException, IllegalAccessException, SecurityException, NoSuchMethodException, IllegalArgumentException, InvocationTargetException {
		
		// The following classes have been removed from the project (after version 1.1.2) and can be replaced by another implementation as an alternative
		Map<String,String> replaceClasses = new HashMap<String, String>();
		replaceClasses.put("org.audiochain.devices.level.LevelAudioDevice", "org.audiochain.devices.level.SamplePeakProgrammeMeterAudioDevice");
		replaceClasses.put("org.audiochain.devices.level.LevelUserInterfaceContext", "org.audiochain.devices.level.SamplePeakProgrammeMeterUserInterfaceContext");
		replaceClasses.put("org.audiochain.devices.compound.BasicLevelAudioDevice", "org.audiochain.devices.compound.BasicSppmAudioDevice");
		replaceClasses.put("org.audiochain.devices.compound.BasicLevelUserInterfaceContext", "org.audiochain.devices.compound.BasicSppmUserInterfaceContext");
		
		if (replaceClasses.containsKey(className)) {
			className = replaceClasses.get(className);
		}
		Class<?> klass = Class.forName(className);
		Constructor<?> constructor = getConstructor(klass);
		try {
			constructor.setAccessible(true);
		} catch (SecurityException e) {
			// nop
			e.printStackTrace();
		}
		Object object = constructor.newInstance();
		objectStack.push(object);
		objectMap.put(objectId, object);
		return object;
	}
	
	public static void writeObject(Object object, OutputStream os) throws IOException {
		writeObject(object, (Appendable) new PrintStream(os));
	}

	public static void writeObject(Object object, Appendable appendable) throws IOException {
		writeObject(object, appendable, new HashSet<Object>(), 0, 0);
		appendable.append(LINESEP);
	}

	private static void writeObject(Object object, Appendable appendable, Collection<Object> dumpedObjects, int indent, int objectIndent) throws IOException {
		if (!(object instanceof Serializable)) {
			throw new IllegalArgumentException("The Class " + object.getClass().getName() + " is not serializable.");
		}
		if (object instanceof Class<?>) {
			appendable.append(indent(objectIndent) + ((Class<?>) object).getName());
			return;
		} else if (object instanceof String) {
			appendable.append(indent(objectIndent) + "\"" + (String) object + "\"");
			return;
		} else {
			appendable.append(indent(objectIndent) + object.getClass().getName() + "{" + object.hashCode() + "}(" + LINESEP);
		}

		dumpedObjects.add(object);
		for (Field field : collectFields(object)) {
			field.setAccessible(true);
			if (Modifier.isStatic(field.getModifiers())) {
				continue;
			}
			if (Modifier.isTransient(field.getModifiers())) {
				continue;
			}
			try {
				Object value = field.get(object);
				if (value != null && dumpedObjects.contains(value)) {
					appendFieldName(appendable, indent, field);
					appendable.append("_REF").append(String.valueOf(value.hashCode())).append('_').append(LINESEP);
				} else if (value instanceof String) {
					appendFieldName(appendable, indent, field);
					appendable.append(String.valueOf(value)).append(LINESEP);
				} else if (value instanceof Enum) {
					appendFieldName(appendable, indent, field);
					Enum<?> enumObject = (Enum<?>) value;
					appendable.append(enumObject.name()).append(LINESEP);
				} else if (value instanceof Boolean) {
					appendFieldName(appendable, indent, field);
					appendable.append(String.valueOf(value)).append(LINESEP);
				} else if (value instanceof Number) {
					appendFieldName(appendable, indent, field);
					appendable.append(String.valueOf(value)).append(LINESEP);
				} else if (value instanceof Collection) {
					Collection<?> coll = (Collection<?>) value;
					if (!coll.isEmpty()) {
						appendFieldName(appendable, indent, field);
						appendable.append('[').append(LINESEP);
						LinkedHashSet<?> iter = new LinkedHashSet<Object>(coll);
						for (Object entry : iter) {
							int newIndent = indent + 2;
							writeObject(entry, appendable, dumpedObjects, newIndent, newIndent);
							appendable.append(LINESEP);
						}
						appendable.append(indent(indent + 1)).append(']').append(LINESEP);
					}
				} else if (value instanceof Map) {
					Map<?, ?> map = (Map<?, ?>) value;
					if (!map.isEmpty()) {
						appendFieldName(appendable, indent, field);
						appendable.append('{').append(LINESEP);
						for (Entry<?, ?> entry : map.entrySet()) {
							int newIndent = indent + 2;
							Object mapEntryKey = entry.getKey();
							Object mapEntryValue = entry.getValue();
							if (mapEntryKey instanceof String && mapEntryValue instanceof Long) {
								appendable.append(indent(newIndent));
								appendable.append('"');
								appendable.append((String) mapEntryKey);
								appendable.append('"');
								appendable.append('=');
								appendable.append(String.valueOf(((Long) mapEntryValue).longValue()));
								appendable.append('L').append(LINESEP);
							} else if (mapEntryKey instanceof String && mapEntryValue instanceof Float) {
								appendable.append(indent(newIndent));
								appendable.append('"');
								appendable.append((String) mapEntryKey);
								appendable.append('"');
								appendable.append('=');
								appendable.append(String.valueOf(((Float) mapEntryValue).floatValue()));
								appendable.append('F').append(LINESEP);
							} else {
								writeObject(mapEntryKey, appendable, dumpedObjects, newIndent, newIndent);
								appendable.append('=');
								writeObject(mapEntryValue, appendable, dumpedObjects, newIndent, 0);
								appendable.append(LINESEP);
							}
						}
						appendable.append(indent(indent + 1)).append('}').append(LINESEP);
					}
				} else if (value != null) {
					appendFieldName(appendable, indent, field);
					writeObject(value, appendable, dumpedObjects, indent + 1, 0);
					appendable.append(LINESEP);
				}
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			}
		}

		appendable.append(indent(indent)).append(')');
	}

	private static void appendFieldName(Appendable appendable, int indent, Field field) throws IOException {
		appendable.append(indent(indent + 1)).append(field.getName()).append('=');
	}

	private static String indent(int indent) {
		String indentation = "";
		for (int i = 0; i < indent; i++) {
			indentation += "  ";
		}
		return indentation;
	}

	static Collection<Field> collectFields(Object object) {
		Collection<Field> fields = new LinkedHashSet<Field>();
		Class<?> klass = object.getClass();
		collectFields(klass, fields);
		return fields;
	}

	private static void collectFields(Class<?> klasse, Collection<Field> fields) {
		Class<?> superClass = klasse.getSuperclass();
		if (superClass != null) {
			collectFields(superClass, fields);
		}
		fields.addAll(Arrays.asList(klasse.getDeclaredFields()));
	}

	private static Constructor<?> getConstructor(Class<?> klasse) throws SecurityException, NoSuchMethodException {
		Constructor<?> ret = klasse.getDeclaredConstructor(new Class[0]);
		if (ret == null) {
			Class<?> superClass = klasse.getSuperclass();
			if (superClass != null) {
				return getConstructor(superClass);
			} else {
				throw new NoSuchMethodException("Constructor not found for " + klasse.getName());
			}
		}
		return ret;
	}
	
	static Collection<Method> collectMethods(Object object) {
		Collection<Method> methods = new HashSet<Method>();
		Class<?> klass = object.getClass();
		collectMethods(klass, methods);
		return methods;
	}

	private static void collectMethods(Class<?> klasse, Collection<Method> methods) {
		Class<?> superClass = klasse.getSuperclass();
		if (superClass != null) {
			collectMethods(superClass, methods);
		}
		methods.addAll(Arrays.asList(klasse.getDeclaredMethods()));
	}
	
	private static void warn(String error) {
		System.err.println(error);
	}

	public static File createFileInSettingsDirectory(String fileName) throws IOException {
		String homePath = System.getProperty("user.home");
		File home = new File(homePath);
		File settingsDirectory;
		String configurationDirectory = System.getProperty("configurationDirectory");
		if (configurationDirectory != null) {
			configurationDirectory = configurationDirectory.replaceAll("~", homePath);
			settingsDirectory = new File(configurationDirectory);
		} else {
			settingsDirectory = new File(home, ".audiochain");
		}
		if (!settingsDirectory.exists()) {
			if (!settingsDirectory.mkdir()) {
				throw new IOException("Cannot create the settings directory '" + settingsDirectory.getAbsolutePath() + "'.");
			}
		}
		File file = new File(settingsDirectory, fileName);
		if (!file.exists()) {
			// the old way where properties are stored (before 1.1.4)
			File oldFile = new File(home, ".audiochain-" + fileName);
			if (oldFile.exists()) {
				if (!oldFile.renameTo(file)) {
					throw new IOException("The settings file '" + oldFile.getAbsolutePath() + "' cannot be moved to '" + file.getAbsoluteFile() + "'.");
				}
			}
		}
		return file;
	}

	public static void storeInSettingsDirectory(Object object, String fileName) throws IOException {
		PrintWriter pw = new PrintWriter(createFileInSettingsDirectory(fileName));
		ObjectStore.writeObject(object, pw);
		pw.close();
	}

	public static Object read(String fileName) throws IOException, IllegalAccessException, ClassNotFoundException, InstantiationException, NoSuchMethodException, InvocationTargetException {
		File file = createFileInSettingsDirectory(fileName);
		if (!file.exists()) {
			return null;
		}
		BufferedReader br = new BufferedReader(new FileReader(file));
		return ObjectStore.readObject(br);
	}
	
	public static boolean hasAnyWritableParentDirectory(File path) {
		if (path == null) {
			return false;
		}
		if (path.isDirectory() && path.canWrite()) {
			return true;
		}
		if (!path.exists()) {
			return hasAnyWritableParentDirectory(path.getParentFile());
		}
		return false;
	}
	
	public static void readExact(InputStream is, OutputStream os, long length, ReadUpdateCallback updateCallback) throws IOException {
		if (length <= 0) {
			return;
		}
		final long total = length;
		final long onePercentOfTotal = total / 100L;
		long lastUpdateReadBytes = 0;
		do {
			int chunkSize = (int) min(4096L, length);
			byte[] buf = new byte[chunkSize];
			int read = is.read(buf);
			assert read >= 0 : "Should read more than zero bytes. Total length is " + total + ", remaining bytes are " + length;
			length -= read;
			os.write(buf, 0, read);
			if (updateCallback != null) {
				long updateReadBytes = total - length;
				// to improve performance, only update if the process incremented at least one percent of the total value
				if (updateReadBytes >= (lastUpdateReadBytes + onePercentOfTotal)) {
					updateCallback.readUpdate(updateReadBytes, total);
					
					/*try {
						// this is for testing only - it forces the transmission to take 3 seconds
						long increment = updateReadBytes - lastUpdateReadBytes;
						long waitTime = 3000;
						long wait = waitTime * increment / total;
						wait = wait < 0 ? 1 : wait > 1000 ? 1000 : wait;
						Thread.sleep(wait);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}*/
					
					lastUpdateReadBytes = updateReadBytes;
				}
			}
		} while (length > 0);
	}
	
	public static interface ReadUpdateCallback {
		abstract public void readUpdate(long read, long total);
	}
	
	public static String toHex(Object o) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		writeObject(o, baos);
		baos.close();
		byte[] byteArray = baos.toByteArray();
		return toHex(byteArray);
	}

	public static String toHex(byte[] byteArray) {
		StringBuilder sb = new StringBuilder(2 * byteArray.length);
		for (byte b : byteArray) {
			sb.append(HEX_VALUES.charAt((b & 0xF0) >> 4));
			sb.append(HEX_VALUES.charAt((b & 0x0F)));
		}
		return sb.toString();
	}
	
	public static Object fromHex(String str) throws IOException, ClassNotFoundException, IllegalAccessException, InstantiationException, NoSuchMethodException, InvocationTargetException {
		byte[] byteArray = bytesFromHex(str);
		ByteArrayInputStream bais = new ByteArrayInputStream(byteArray);
		Object o = readObject(bais);
		bais.close();
		return o;
	}

	public static byte[] bytesFromHex(String str) {
		char[] charArray = str.toCharArray();
		if (charArray.length % 2 != 0) {
			throw new IllegalArgumentException("The given chars must have an even count.");
		}
		int len = charArray.length / 2;
		byte[] byteArray = new byte[len];
		for (int i = 0; i < len; i++) {
			char c1 = charArray[2 * i];
			char c2 = charArray[2 * i + 1];
			byte b1 = (byte) (HEX_VALUES.indexOf(c1) << 4);
			byte b2 = (byte) HEX_VALUES.indexOf(c2);
			byte b = (byte) (b1 + b2);
			byteArray[i] = b;
		}
		return byteArray;
	}
}
