/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import java.io.IOException;
import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.io.AudioIOException;

public class PerformanceManager implements PerformanceListener, ChainableAudioDeviceListener {
	private final AudioProject audioProject;
	private final int frameRate;
	private final int updateRate;
	private Performable performable;
	private long lastSecondSynchronizedFramePositionValue;
	private Collection<PerformanceListener> performanceListeners;
	private Collection<PerformanceMonitorListener> performanceMonitorListeners;
	private boolean ignoreFramePositionChange;
	private AudioDevice monitoredAudioDevice;
	private Monitorable currentMonitor;

	public PerformanceManager(AudioProject audioProject) {
		this.audioProject = audioProject;
		frameRate = audioProject.getFrameRate();
		// updates the slider (ie twice a second)
		updateRate = frameRate / 2;
		init();
	}

	private void init() {
		final AudioTrackListener audioTrackChangeListener = new AudioTrackAdapter() {
			@Override
			public void audioTrackSelectedForRecordingChanged(AudioTrack audioTrack, boolean oldValue, boolean newSelectedForRecordingValue) throws IOException, AudioIOException {
				if (newSelectedForRecordingValue) {
					trackSelectedForRecording(audioTrack);
				} else {
					closePerformable();
				}
			}
			
			@Override
			public void audioTrackEnabledChanged(AudioTrack audioTrack, boolean oldValue, boolean enabled) {
				if (!enabled && monitoredAudioDevice!=null) {
					Monitorable monitor = audioTrack.findMonitor(monitoredAudioDevice);
					if(monitor!=null&&monitor.equals(currentMonitor)) {
						setMonitoredAudioDevice(null);
					}
				}
			}
		};
		
		final AudioDeviceChainListener audioDeviceChainListener = new AudioDeviceChainAdapter(){
			@Override
			public void audioDeviceRemoved(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain) {
				if(audioDevice.equals(monitoredAudioDevice)) {
					setMonitoredAudioDevice(null);
				}
			}
		};

		Collection<AudioTrack> audioTracks = audioProject.getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			audioTrack.addAudioTrackChangeListener(audioTrackChangeListener);
			audioTrack.getPlaybackAudioDeviceChain().addAudioDeviceChainListener(audioDeviceChainListener);
			audioTrack.getCaptureAudioDeviceChain().addAudioDeviceChainListener(audioDeviceChainListener);
		}
		audioProject.getMasterAudioDeviceChain().addAudioDeviceChainListener(audioDeviceChainListener);
		audioProject.addAudioProjectChangeListener(new AudioProjectAdapter() {
			@Override
			public void audioTrackAdded(AudioTrack track) {
				track.addAudioTrackChangeListener(audioTrackChangeListener);
				track.getPlaybackAudioDeviceChain().addAudioDeviceChainListener(audioDeviceChainListener);
				track.getCaptureAudioDeviceChain().addAudioDeviceChainListener(audioDeviceChainListener);
			}

			@Override
			public void audioTrackRemoved(AudioTrack track) {
				track.removeAudioTrackChangeListener(audioTrackChangeListener);
				track.getPlaybackAudioDeviceChain().removeAudioDeviceChainListener(audioDeviceChainListener);
				track.getCaptureAudioDeviceChain().removeAudioDeviceChainListener(audioDeviceChainListener);
			}

			@Override
			public void audioProjectFramePositionChanged(long oldValue, long newFramePosition) throws IOException {
				if (ignoreFramePositionChange) {
					return;
				}
				if (isRunning() && !isRecording()) {
					performable.seek(newFramePosition);
				}
			}
		});
		audioProject.addAudioProjectLifecycleListener(new AudioProjectLifecycleAdapter() {
			@Override
			public void audioProjectClosed(AudioProject audioProject) {
				if (isRunning()) {
					stop();
				}
			}
		});
	}

	private void trackSelectedForRecording(AudioTrack track) throws IOException, AudioIOException {
		closePerformable();
		SourceAudioDevice sourceAudioDevice = track.getSourceAudioDevice();
		if (sourceAudioDevice == null) {
			throw new IllegalStateException("There is no " + SourceAudioDevice.class.getSimpleName() + " set for this " + AudioTrack.class.getSimpleName() + ".");
		}
		if (!sourceAudioDevice.isRecordable()) {
			throw new IllegalStateException("The " + AudioDevice.class.getSimpleName() + " '" + sourceAudioDevice.getName() + "' is not recordable.");
		}
		try {
			performable = sourceAudioDevice.getPerformable();
			if (performable != null) {
				addListenersToPerformable();
			}
		} catch (AudioIOException e) {
			track.unselectForRecording();
			throw e;
		}
	}

	public boolean isRecording() {
		return performable != null && performable.isRecording();
	}

	public boolean isRunning() {
		return performable != null && performable.isRunning();
	}

	public void start() throws IOException, AudioIOException {
		if (performable != null) {
			performable.joinStopped();
		}
		if (performable == null) {
			performable = new Playback(audioProject);
			addListenersToPerformable();
		}
		try {
			long framePosition = audioProject.getFramePosition();
			performable.start(framePosition);
		} catch (IOException e) {
			stop();
			performable = null;
			throw e;
		} catch (AudioIOException e) {
			stop();
			performable = null;
			throw e;
		}
	}

	private void addListenersToPerformable() {
		performable.addPerformanceListener(this);
		if (performanceListeners != null) {
			for (PerformanceListener listener : performanceListeners) {
				performable.addPerformanceListener(listener);
			}
		}
	}

	public void stop() {
		if (performable != null) {
			performable.stop();
		}
	}

	private synchronized void closePerformable() throws IOException {
		if (performable != null) {
			performable.stop();
			performable.close();
			// wait for call on performanceFinished which sets performable
			// to null
			while (performable != null) {
				Thread.yield();
			}
		}
	}

	public void toggleStart() throws IOException, AudioIOException {
		if (performable != null && performable.isRunning()) {
			performable.stop();
			performable.joinStopped();
		} else {
			start();
		}
	}

	public void startFromBeginning() throws IOException, AudioIOException {
		if (performable != null && performable.isRunning()) {
			performable.stop();
			performable.joinStopped();
		}
		audioProject.setFramePosition(0);
		start();
	}

	private void updateFramePosition() {
		if (performable != null) {
			long currentFramePosition = performable.getFrameObserver().getCurrentFramePosition();
			long newSecondSynchronizedFramePositionValue = currentFramePosition / updateRate;
			newSecondSynchronizedFramePositionValue *= updateRate;
			if (lastSecondSynchronizedFramePositionValue != newSecondSynchronizedFramePositionValue) {
				lastSecondSynchronizedFramePositionValue = newSecondSynchronizedFramePositionValue;
				try {
					ignoreFramePositionChange = true;
					audioProject.setFramePosition(currentFramePosition);
				} catch (IOException e) {
					e.printStackTrace();
				} finally {
					ignoreFramePositionChange = false;
				}
			}
		}
	}

	@Override
	public void performanceStarted() {
		setMonitoredAudioDevice(monitoredAudioDevice);
	}

	@Override
	public void performanceUpdated() {
		updateFramePosition();
	}

	@Override
	public void performanceStopped() {
		updateFramePosition();
	}

	@Override
	public void performanceFinished() {
		performable = null;
	}

	public Collection<Throwable> getLastPerformanceErrors() {
		if (performable != null) {
			return performable.getErrors();
		}
		return null;
	}

	public void addPerformanceListener(PerformanceListener listener) {
		if (performanceListeners == null) {
			performanceListeners = new LinkedHashSet<PerformanceListener>();
		}
		performanceListeners.add(listener);
	}

	public void removePerformanceListener(PerformanceListener listener) {
		if (performanceListeners == null) {
			return;
		}
		performanceListeners.remove(listener);
		if (performanceListeners.isEmpty()) {
			performanceListeners = null;
		}
	}

	public void addPerformanceMonitorListener(PerformanceMonitorListener listener) {
		if (performanceMonitorListeners == null) {
			performanceMonitorListeners = new LinkedHashSet<PerformanceMonitorListener>();
		}
		performanceMonitorListeners.add(listener);
	}

	public void removePerformanceMonitorListener(PerformanceMonitorListener listener) {
		if (performanceMonitorListeners == null) {
			return;
		}
		performanceMonitorListeners.remove(listener);
		if (performanceMonitorListeners.isEmpty()) {
			performanceMonitorListeners = null;
		}
	}

	public void setMonitoredAudioDevice(AudioDevice device) {
		AudioDevice oldDevice = this.monitoredAudioDevice;
		this.monitoredAudioDevice = device;
		Monitorable oldMonitor = currentMonitor;
		Monitorable monitor = null;
		if (device != null) {
			monitor = audioProject.findMonitor(device);
		}
		if (performable != null) {
			performable.setPerformanceMonitor(monitor);
		}
		currentMonitor = monitor;
		if (device instanceof ChainableAudioDevice) {
			((ChainableAudioDevice) device).addChainableAudioDeviceListener(this);
		}
		if (oldDevice instanceof ChainableAudioDevice) {
			((ChainableAudioDevice) oldDevice).removeChainableAudioDeviceListener(this);
		}
		if (oldDevice != device || oldMonitor != monitor) {
			firePerformanceMonitorChanged(oldMonitor, oldDevice, monitor, device);
		}
	}

	public AudioDevice getMonitoredAudioDevice() {
		return monitoredAudioDevice;
	}

	@Override
	public void audioDeviceEnabledChanged(ChainableAudioDevice chainableAudioDevice, boolean oldValue, boolean enabled) {
		if (!enabled && chainableAudioDevice.equals(monitoredAudioDevice)) {
			setMonitoredAudioDevice(null);
		}
	}

	protected void firePerformanceMonitorChanged(Monitorable oldMonitor, AudioDevice oldDevice, Monitorable newMonitor, AudioDevice newDevice) {
		if (performanceMonitorListeners == null) {
			return;
		}
		for (PerformanceMonitorListener listener : new LinkedHashSet<PerformanceMonitorListener>(performanceMonitorListeners)) {
			listener.performanceMonitorChanged(oldMonitor, oldDevice, newMonitor, newDevice);
		}
	}
}
