/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.SourceDataLine;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataUtil;
import org.audiochain.io.AudioIOException;
import org.audiochain.io.FrameObserver;
import org.audiochain.io.LineAudioDataWriter;
import org.audiochain.io.LineFrameObserver;
import org.audiochain.io.MixingAudioDataReader;


public class Playback implements Performable {
	public static enum PlaybackStatus {
		New, Initialized, Started, Playing, Stopped, Finished
	}

	protected final AudioProject project;
	protected long startFramePosition;
	private Collection<PerformanceListener> performanceListeners;
	private int[] playbackBuffer;
	private LineAudioDataWriter playbackLineWriter;
	private FrameObserver frameObserver;
	private AudioDataReader playbackReader;
	private Collection<Throwable> errors;
	private PlaybackStatus playbackStatus;
	private Map<AudioTrack, AudioDataReader> audioTrackReaderMap;
	private Monitorable monitor;

	protected Playback(AudioProject project) {
		this.project = project;
		project.addAudioProjectChangeListener(new AudioProjectAdapter(){
			@Override
			public void audioTrackRemoved(AudioTrack track) {
				if (audioTrackReaderMap != null) {
					AudioDataReader audioDataReader = audioTrackReaderMap.get(track);
					if (audioDataReader != null) {
						audioDataReader.stop();
					}
				}
			}
		});
		initPlayback();
	}

	protected void initPlayback() {
		playbackStatus = PlaybackStatus.New;
		playbackBuffer = new int[AudioMixer.getInstance().getPlaybackBufferSizeInSamples()];
		setNextPlaybackStatus(PlaybackStatus.Initialized);
	}

	@Override
	public void start(long framePosition) throws IOException, AudioIOException {
		if (playbackStatus != PlaybackStatus.Initialized) {
			throw new IllegalStateException(PlaybackStatus.class.getSimpleName() + " must be '" + PlaybackStatus.Initialized + "'.");
		}
		this.startFramePosition = framePosition;
		if (project.getPlaybackTracks().isEmpty()) {
			setNextPlaybackStatus(PlaybackStatus.Finished);
			firePlaybackFinished();
		} else {
			try {
				createPlaybackLine();
			} catch (RuntimeException e) {
				closePlaybackResources();
				throw e;
			} catch (IOException e) {
				closePlaybackResources();
				throw e;
			} catch (AudioIOException e) {
				closePlaybackResources();
				throw e;
			}
			new PlaybackThread().start();
			firePlaybackStarted();
		}
	}

	@Override
	public void close() throws IOException {
		// nop
	}

	@Override
	public void stop() {
		if (playbackReader != null) {
			playbackReader.stop();
		} else if (playbackStatus == PlaybackStatus.Initialized) {
			setNextPlaybackStatus(PlaybackStatus.Finished);
		}
	}

	@Override
	public void join() {
		while (playbackStatus != PlaybackStatus.Finished) {
			Thread.yield();
		}
	}

	@Override
	public void joinStopped() {
		while (playbackStatus != PlaybackStatus.Finished && playbackStatus != PlaybackStatus.Initialized) {
			Thread.yield();
		}
	}

	@Override
	public void seek(long framePosition) throws IOException {
		if (playbackReader != null) {
			playbackReader.seek(framePosition);
		}
	}

	private void createPlaybackLine() throws IOException, AudioIOException {
		AudioFormat projectAudioFormat = project.getAudioFormat();
		final AudioMixer mixer = AudioMixer.getInstance();
		playbackLineWriter = new LineAudioDataWriter(projectAudioFormat, mixer.getPlaybackMixer().getMixer());
		SourceDataLine playbackLine = playbackLineWriter.getLine();
		LineFrameObserver lineFrameObserver = new LineFrameObserver(playbackLine){
			@Override
			public void close() throws IOException {
				super.close();
				mixer.setPlaybackLineLatencyInFrames(getLatencyInFrames());
			}
		};
		// initialize with the last mixer stored setting
		lineFrameObserver.setLatencyInFrames(mixer.getPlaybackLineLatencyInFrames());

		audioTrackReaderMap = new LinkedHashMap<AudioTrack, AudioDataReader>();
		final float frameRate = project.getFrameRate();
		final int sampleSizeInBits = project.getSampleSizeInBits();
		
		for (AudioTrack audioTrack : project.getPlaybackTracks()) {
			SourceAudioDevice sourceAudioDevice = audioTrack.getSourceAudioDevice();
			AudioDataReader sourceAudioDataReader = sourceAudioDevice.createAudioDataReader(frameRate, sampleSizeInBits, lineFrameObserver);
			AudioDeviceChain playbackAudioDeviceChain = audioTrack.getPlaybackAudioDeviceChain();
			AudioDataReader audioDataReaderChain = playbackAudioDeviceChain.createReversedAudioDataReaderChain(sourceAudioDataReader, frameRate, sampleSizeInBits, lineFrameObserver);
			audioTrackReaderMap.put(audioTrack, audioDataReaderChain);
		}

		AudioDataReader mixedReader = new MixingAudioDataReader(audioTrackReaderMap.values());
		AudioDeviceChain masterAudioDeviceChain = project.getMasterAudioDeviceChain();
		AudioDataReader masterReader = masterAudioDeviceChain.createAudioDataReaderChain(mixedReader, frameRate, sampleSizeInBits, lineFrameObserver);
		lineFrameObserver.setSource(masterReader);
		frameObserver = lineFrameObserver;
		playbackReader = lineFrameObserver;
		playbackReader.seek(startFramePosition);
	}

	void playback() throws IOException {
		try {
			int[] silence = null;
			while (playbackStatus == PlaybackStatus.Started || playbackStatus == PlaybackStatus.Playing) {
				int len = playbackReader.read(playbackBuffer);
				if (len == -1) {
					setNextPlaybackStatus(PlaybackStatus.Stopped);
				} else {
					if (playbackStatus == PlaybackStatus.Started) {
						setNextPlaybackStatus(PlaybackStatus.Playing);
					}
					Monitorable monitor = this.monitor;
					if (monitor != null) {
						int[] monitorAudioData = monitor.getMonitorData();
						int length = monitor.getMonitorLength();
						if (length > 0 && monitorAudioData != null) {
							playbackLineWriter.write(monitorAudioData, 0, length);
						} else {
							if (silence == null) {
								silence = new int[playbackBuffer.length];
							}
							playbackLineWriter.write(silence, 0, len);
						}
					} else {
						playbackLineWriter.write(playbackBuffer, 0, len);
					}
					firePlaybackUpdated();
				}
			}
		} finally {
			// in case of exceptions we need to set the correct status here
			if (playbackStatus == PlaybackStatus.Started || playbackStatus == PlaybackStatus.Playing) {
				setNextPlaybackStatus(PlaybackStatus.Stopped);
			}
		}
	}

	class PlaybackThread extends Thread {
		@Override
		public void run() {
			setNextPlaybackStatus(PlaybackStatus.Started);
			setName(getClass().getSimpleName());
			try {
				playback();
			} catch (IOException e) {
				addError(e);
				throw new IllegalStateException(e);
			} finally {
				closePlaybackResources();
				firePlaybackFinished();
				setNextPlaybackStatus(PlaybackStatus.Finished);
			}
		}
	}

	void setNextPlaybackStatus(PlaybackStatus status) {
		if (status.ordinal() > playbackStatus.ordinal()) {
			playbackStatus = status;
		} else if (status == PlaybackStatus.Initialized && playbackStatus == PlaybackStatus.Stopped) {
			playbackStatus = status;
		} else {
			throw new IllegalStateException(PlaybackStatus.class.getSimpleName() + " '" + playbackStatus + "' cannot be followed by '" + status + "'.");
		}
	}

	private synchronized void closePlaybackResources() {
		try {
			if (playbackReader != null) {
				playbackReader.close();
				playbackReader = null;
			}
			if (playbackLineWriter != null) {
				playbackLineWriter.close();
				playbackLineWriter = null;
			}
		} catch (IOException e) {
			addError(e);
			throw new IllegalStateException(e);
		}
	}

	protected void firePlaybackStarted() {
		if (performanceListeners == null) {
			return;
		}
		while (playbackStatus == PlaybackStatus.Initialized) {
			Thread.yield();
		}
		firePerformanceStarted();
	}

	protected void firePerformanceStarted() {
		if (performanceListeners == null) {
			return;
		}
		for (PerformanceListener listener : new LinkedHashSet<PerformanceListener>(performanceListeners)) {
			listener.performanceStarted();
		}
	}

	protected void firePlaybackUpdated() {
		if (performanceListeners == null) {
			return;
		}
		for (PerformanceListener listener : new LinkedHashSet<PerformanceListener>(performanceListeners)) {
			listener.performanceUpdated();
		}
	}

	protected void firePerformanceUpdated() {
		if (performanceListeners == null) {
			return;
		}
		for (PerformanceListener listener : new LinkedHashSet<PerformanceListener>(performanceListeners)) {
			listener.performanceUpdated();
		}
	}

	protected void firePlaybackStopped() {
		if (performanceListeners == null) {
			return;
		}
		for (PerformanceListener listener : new LinkedHashSet<PerformanceListener>(performanceListeners)) {
			listener.performanceUpdated();
		}
	}

	protected void firePerformanceStopped() {
		if (performanceListeners == null) {
			return;
		}
		for (PerformanceListener listener : new LinkedHashSet<PerformanceListener>(performanceListeners)) {
			listener.performanceStopped();
		}
	}

	protected void firePlaybackFinished() {
		if (performanceListeners == null) {
			return;
		}
		for (PerformanceListener listener : new LinkedHashSet<PerformanceListener>(performanceListeners)) {
			listener.performanceFinished();
		}
		performanceListeners.clear();
		performanceListeners = null;
	}

	protected void firePerformanceFinished() {
		if (performanceListeners == null) {
			return;
		}
		for (PerformanceListener listener : new LinkedHashSet<PerformanceListener>(performanceListeners)) {
			listener.performanceFinished();
		}
		performanceListeners.clear();
		performanceListeners = null;
	}

	@Override
	public void addPerformanceListener(PerformanceListener listener) {
		if (performanceListeners == null) {
			performanceListeners = new LinkedHashSet<PerformanceListener>();
		}
		performanceListeners.add(listener);
	}

	@Override
	public void removePerformanceListener(PerformanceListener listener) {
		if (performanceListeners == null) {
			return;
		}
		performanceListeners.remove(listener);
		if (performanceListeners.isEmpty()) {
			performanceListeners = null;
		}
	}

	@Override
	public FrameObserver getFrameObserver() {
		return frameObserver;
	}

	public AudioProject getProject() {
		return project;
	}

	protected void addError(Throwable e) {
		if (errors == null) {
			errors = new ArrayList<Throwable>();
		}
		errors.add(e);
	}

	@Override
	public Collection<Throwable> getErrors() {
		Collection<Throwable> ret = new ArrayList<Throwable>();
		if (errors != null) {
			ret.addAll(errors);
		}
		return ret;
	}

	public PlaybackStatus getPlaybackStatus() {
		return playbackStatus;
	}

	@Override
	public boolean isRecording() {
		return false;
	}

	@Override
	public boolean isRunning() {
		return playbackStatus == PlaybackStatus.Playing || playbackStatus == PlaybackStatus.Started;
	}

	@Override
	public void setPerformanceMonitor(Monitorable monitor) {
		Monitorable oldMonitor = this.monitor;
		if (oldMonitor != null) {
			oldMonitor.setMonitorEnabled(false);
		}
		if (monitor != null) {
			monitor.setMonitorEnabled(true);
		}
		this.monitor = monitor;
		if ("true".equals(System.getProperty("dumpReaderChain"))) {
			try {
				dumpReaderChain();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	protected void dumpReaderChain() throws IOException {
		System.out.println();
		System.out.println("Playback reader chain:");
		AudioDataUtil.dumpAudioDataReaderChain(playbackReader, System.out);
	}
}
