/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import java.awt.Component;
import java.io.Serializable;

/**
 * <p>This interface is the seam between an {@link AudioDevice} and a user interface {@link Component}.
 * It decouples the audio operations from the gui and enable the ability to create non GUI interfaces to the
 * audio operations (for example the command line).
 * 
 * <p>You as the {@link AudioDevice} vendor may not trouble about the {@link UserInterfaceContext} and just extend the
 * GenericUserInterfaceContext for GUIs.
 * 
 * <p>The {@link UserInterfaceContext} implementations must be persistable. The created component must not be persisted and must be transient.
 * 
 * The {@link UserInterfaceContext} can also be used to store user interface settings which are not relevant for the {@link AudioDevice} itself.</p>
 */
public interface UserInterfaceContext<D extends AudioDevice> extends Serializable {
	
	/**
	 * <p>Provides a GUI component for this audio audioDevice. The returned component should visually integrate itself to the mixer seamless for
	 * a good look and feel. If the component provides additional dialogs, it should offer buttons to open these dialogs. Its on you to make
	 * these dialogs modal (all common dialogs in this program are modal) but its uniformly if all dialogs are modal.
	 * 
	 * <p>The created component will be embedded on the mixer. Once created, it must not resize itself to not confuse the user. The mixer automatically
	 * resizes itself, when the new component is displayed. If the component resizes itself, the mixer cannot detect this new size (its not smart to
	 * observ all component size changes at all time).
	 * 
	 * <p>Once created, this method should return always the same component instance for performance reasons.
	 * 
	 * @return a component for GUI presentation
	 */
	public Component getComponent(D audioDevice);
	
	/**
	 * The user may hide the audioDevice at the user interface. This state must be stored by setting this property.
	 * @param hidden
	 */
	public void setHidden(boolean hidden);
	public boolean isHidden();
	
	/**
	 * The mixer can register itself on this context to be notified, if the component becomes hidden or unhidden. The mixer resizes itself
	 * as necessary.
	 * @param listener this parameter should be attached once to the context listener chain, so you should use a set in the implementation.
	 */
	public void addUserInterfaceContextChangeListener(UserInterfaceContextListener listener);
	public void removeUserInterfaceContextChangeListener(UserInterfaceContextListener listener);
	
	/**
	 * <p>Provides the Command Line User Interface that is simply displayed by its getDisplayString() Method. If the String representation
	 * should be updated, this UserInterfaceContext implementation can fire the {@link UserInterfaceContextListener}.commandLineUserInterfaceUpdate() Method.
	 * @return any CommandLineUserInterface that can be displayed to the user by its getDisplayString() representation.
	 */
	public CommandLineUserInterface getCommandLineUserInterface(D audioDevice);
}
