/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;

import javax.swing.JFrame;
import javax.swing.JScrollPane;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.IntArrayAudioDataReader;


public class AudioDataPlotter extends Component implements ComponentListener, AudioDataReaderChainLink, MouseWheelListener, KeyListener {
	private static final long serialVersionUID = 1L;
	private AudioDataReader reader;
	private final float sampleMaximum; // full scale
	private int[] data;
	private int sampleWidthInPixel = 1;
	private int sampleHeightInPixel = 2;
	private float horizontalZoomFactor = 1F;
	private float verticalZoomFactor = 0.95F;
	private boolean solidLine = true;
	private boolean horizontalSampleLines = true;
	private boolean verticalSampleLines = false;
	private boolean frameLines = false;
	private Color leftColor = Color.BLUE;
	private Color leftSampleLineColor = leftColor;
	private Color rightColor = Color.RED;
	private Color rightSampleLineColor = rightColor;
	private Color zeroCrossingLineColor = Color.DARK_GRAY;
	private Color boundaryLineColor = Color.LIGHT_GRAY;
	private Color frameLineColor = Color.LIGHT_GRAY;
	private Color backgroundColor = Color.WHITE;
	private JScrollPane scroll;
	private Collection<Integer> pressedKeys = new HashSet<Integer>();

	public AudioDataPlotter(int sampleSize) {
		this.addComponentListener(this);
		this.addMouseWheelListener(this);
		this.sampleMaximum = (float) Math.pow(2D, sampleSize - 1);
	}

	public AudioDataPlotter(AudioDataReader reader, int sampleSize) {
		this(sampleSize);
		this.reader = reader;
	}

	@Override
	public int read(final int[] data) throws IOException {
		if (reader == null) {
			throw new IllegalStateException("no reader available");
		}
		int ret = reader.read(data);
		if (ret != -1) {
			addData(data, ret);
			new Thread() {
				@Override
				public void run() {
					resetWidth();
					repaint();
					refreshScroll();
				}
			}.start();
		}
		return ret;
	}

	@Override
	public void close() throws IOException {
		if (reader != null) {
			reader.close();
		}
	}

	@Override
	public void stop() {
		if (reader != null) {
			reader.stop();
		}
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
	}

	@Override
	public void paint(Graphics g) {
		if (data == null) {
			return;
		}

		final int visibleWidth = scroll == null ? getWidth() : scroll.getWidth();
		final int xstart = scroll == null ? 0 : scroll.getHorizontalScrollBar().getValue();
		final int xend = xstart + visibleWidth;
		final int height = getHeight();
		final int hhalf = height / 2;

		g.setColor(backgroundColor);
		g.fillRect(xstart, 0, visibleWidth, height);

		Color upperBoundaryColor = boundaryLineColor;
		Color lowerBoundaryColor = boundaryLineColor;

		final int from = getFrameOfPlotPosition(xstart);
		final int to = Math.min(getFrameOfPlotPosition(xend), data.length / 2);

		int x = from;
		Integer lastLeft = null;
		Integer lastRight = null;
		for (int i = from; i < to; i++) {
			final int ldata = data[2 * i];
			final int rdata = data[2 * i + 1];

			final int l = (int) (ldata * hhalf / -sampleMaximum * verticalZoomFactor + hhalf);
			final int r = (int) (rdata * hhalf / -sampleMaximum * verticalZoomFactor + hhalf);

			x += sampleWidthInPixel;

			final int x1 = (int) (x * horizontalZoomFactor);
			final int x2 = (int) ((x + sampleWidthInPixel) * horizontalZoomFactor);

			final int x21 = (int) ((x + sampleWidthInPixel - 1) * horizontalZoomFactor);
			final int x22 = (int) ((x + sampleWidthInPixel + 1) * horizontalZoomFactor);

			if (frameLines) {
				g.setColor(frameLineColor);
				g.drawLine(x2, 0, x2, height);
			}

			g.setColor(leftColor);
			if (ldata < Short.MIN_VALUE) {
				lowerBoundaryColor = leftColor;
				g.fillRect(x1, l, 5, 5);
			} else if (ldata > sampleMaximum) {
				upperBoundaryColor = leftColor;
				g.fillRect(x1, l, 5, 5);
			}

			if (solidLine && lastLeft != null) {
				g.drawLine(x2, l, x1, lastLeft);
			}

			if (!solidLine && !horizontalSampleLines && !verticalSampleLines) {
				g.drawLine(x1, l, x2, l);
			}

			if (horizontalSampleLines) {
				g.setColor(leftSampleLineColor);
				g.drawLine(x21, l, x22, l);
			}
			if (verticalSampleLines) {
				g.setColor(leftSampleLineColor);
				g.drawLine(x2, l - sampleHeightInPixel, x2, l + sampleHeightInPixel);
			}

			g.setColor(rightColor);
			if (rdata < Short.MIN_VALUE) {
				lowerBoundaryColor = rightColor;
				g.fillRect(x1, r, 5, 5);
			} else if (rdata > sampleMaximum) {
				upperBoundaryColor = rightColor;
				g.fillRect(x1, r, 5, 5);
			}

			if (solidLine && lastRight != null) {
				g.drawLine(x2, r, x1, lastRight);
			}
			if (!solidLine && !horizontalSampleLines && !verticalSampleLines) {
				g.drawLine(x1, r, x2, r);
			}
			if (horizontalSampleLines) {
				g.setColor(rightSampleLineColor);
				g.drawLine(x21, r, x22, r);
			}
			if (verticalSampleLines) {
				g.setColor(rightSampleLineColor);
				g.drawLine(x2, r - sampleHeightInPixel, x2, r + sampleHeightInPixel);
			}

			if (solidLine) {
				lastLeft = l;
				lastRight = r;
			}
		}

		g.setColor(zeroCrossingLineColor);
		g.drawLine(xstart, hhalf, xend, hhalf);

		g.setColor(upperBoundaryColor);
		int upperBoundary = (int) (-hhalf * verticalZoomFactor + hhalf);
		g.drawLine(xstart, upperBoundary - 1, xend, upperBoundary - 1);

		g.setColor(lowerBoundaryColor);
		int lowerBoundary = (int) (hhalf * verticalZoomFactor + hhalf);
		g.drawLine(xstart, lowerBoundary + 1, xend, lowerBoundary + 1);
	}

	public int addData(int[] newData, final int len) {
		if (data == null) {
			data = new int[len];
			System.arraycopy(newData, 0, data, 0, len);
			return len;
		}
		int oldLen = data.length;
		int newLen = oldLen + len;
		int[] ndata = new int[newLen];
		System.arraycopy(data, 0, ndata, 0, oldLen);
		System.arraycopy(newData, 0, ndata, oldLen, len);
		data = ndata;
		return newLen;
	}

	public void setData(int[] data) {
		this.data = data;
	}

	public int[] getData() {
		return data;
	}

	public Component getUIComponent() {
		if (scroll == null) {
			scroll = new JScrollPane();
			scroll.setViewportView(this);
		}
		return scroll;
	}

	public void setHeight(int height) {
		Dimension dim = new Dimension(getWidth(), height);
		setPreferredSize(dim);
		setSize(dim);
	}

	public void resetWidth() {
		Dimension dim = new Dimension((int) (data.length * sampleWidthInPixel / 2 * horizontalZoomFactor), getHeight());
		setPreferredSize(dim);
		setSize(dim);
	}

	public void setHorizontalScrollPosition(int position) {
		if (scroll != null) {
			scroll.getHorizontalScrollBar().setValue(position);
		}
	}

	public int getHorizontalScrollPosition() {
		if (scroll != null) {
			return scroll.getHorizontalScrollBar().getValue();
		}
		return 0;
	}

	public void refreshScroll() {
		if (scroll != null) {
			scroll.revalidate();
			scroll.repaint();
		}
	}

	@Override
	public void mouseWheelMoved(MouseWheelEvent e) {
		int rot = e.getWheelRotation();
		boolean up = rot < 0;

		int normalScroll = (int) (1000 * horizontalZoomFactor);
		int fastScroll = (int) (10000 * horizontalZoomFactor);

		for (int pressedKey : pressedKeys) {
			switch (pressedKey) {
			case 18: // alt
				int currentFrame = getFrameOfPlotPosition(e.getX());
				float before = horizontalZoomFactor;
				horizontalZoomFactor += up ? -0.1F * horizontalZoomFactor : 0.1F * horizontalZoomFactor;
				if (horizontalZoomFactor <= 0) {
					horizontalZoomFactor = 0.01F;
				} else if (horizontalZoomFactor >= 20) {
					horizontalZoomFactor = 20;
				}

				if (before != horizontalZoomFactor) {
					resetWidth();
					repaint();
					refreshScroll();
					scrollToFrame(currentFrame);
				}
				break;
			case 89: // y
				verticalZoomFactor += up ? -0.1F : 0.1F;
				repaint();
				break;
			case 16: // shift
				int h = getHeight();
				setHeight(h + (up ? (int) (-0.05 * h) : (int) (0.05 * h)));
				repaint();
				refreshScroll();
				if (scroll != null) {
					scroll.getVerticalScrollBar().setValue(getHeight() / 2 - scroll.getHeight() / 2);
				}
				refreshScroll();
				break;
			case 17: // ctrl
				setHorizontalScrollPosition(getHorizontalScrollPosition() + (up ? -fastScroll : fastScroll));
				refreshScroll();
				repaint();
				break;
			}
		}
		if (pressedKeys.isEmpty()) {
			setHorizontalScrollPosition(getHorizontalScrollPosition() + (up ? -normalScroll : normalScroll));
			refreshScroll();
			repaint();
		}
	}

	public int getFrameOfPlotPosition(int plotPosition) {
		int frame = (int) (plotPosition / horizontalZoomFactor);
		return frame;
	}

	public void scrollToFrame(int frame) {
		if (scroll != null) {
			int plotPosition = (int) (frame * horizontalZoomFactor);
			int val = (int) (plotPosition - scroll.getMousePosition().getX());
			scroll.getHorizontalScrollBar().setValue(val);
		}
	}

	@Override
	public void componentResized(ComponentEvent e) {
		repaint();
	}

	@Override
	public void componentMoved(ComponentEvent e) {

	}

	@Override
	public void componentShown(ComponentEvent e) {

	}

	@Override
	public void componentHidden(ComponentEvent e) {

	}

	public void setSampleWidthInPixel(int sampleWidthInPixel) {
		if (sampleWidthInPixel < 1) {
			sampleWidthInPixel = 1;
		}
		this.sampleWidthInPixel = sampleWidthInPixel;
	}

	public int getSampleWidthInPixel() {
		return sampleWidthInPixel;
	}

	public void setSolidLine(boolean solidLine) {
		this.solidLine = solidLine;
	}

	public boolean isSolidLine() {
		return solidLine;
	}

	public void setLeftColor(Color leftColor) {
		this.leftColor = leftColor;
	}

	public Color getLeftColor() {
		return leftColor;
	}

	public void setRightColor(Color rightColor) {
		this.rightColor = rightColor;
	}

	public Color getRightColor() {
		return rightColor;
	}

	public void setZeroCrossingLineColor(Color zeroCrossingLineColor) {
		this.zeroCrossingLineColor = zeroCrossingLineColor;
	}

	public Color getZeroCrossingLineColor() {
		return zeroCrossingLineColor;
	}

	public void setHorizontalZoomFactor(float horizontalZoomFactor) {
		this.horizontalZoomFactor = horizontalZoomFactor;
	}

	public float getHorizontalZoomFactor() {
		return horizontalZoomFactor;
	}

	public void setVerticalZoomFactor(float verticalZoomFactor) {
		this.verticalZoomFactor = verticalZoomFactor;
	}

	public float getVerticalZoomFactor() {
		return verticalZoomFactor;
	}

	public void setBoundaryLineColor(Color boundaryLineColor) {
		this.boundaryLineColor = boundaryLineColor;
	}

	public Color getBoundaryLineColor() {
		return boundaryLineColor;
	}

	@Override
	public void keyTyped(KeyEvent e) {
		switch (e.getKeyChar()) {
		case 's':
			setSolidLine(!isSolidLine());
			repaint();
			break;
		case 'h':
			setHorizontalSampleLines(!isHorizontalSampleLines());
			repaint();
			break;
		case 'v':
			setVerticalSampleLines(!isVerticalSampleLines());
			repaint();
			break;
		case 'f':
			setFrameLines(!isFrameLines());
			repaint();
			break;
		}
	}

	@Override
	public void keyPressed(KeyEvent e) {
		int pressedKey = e.getKeyCode();
		pressedKeys.add(pressedKey);
	}

	@Override
	public void keyReleased(KeyEvent e) {
		pressedKeys.remove(e.getKeyCode());
	}

	public boolean isHorizontalSampleLines() {
		return horizontalSampleLines;
	}

	public void setHorizontalSampleLines(boolean sampleLines) {
		this.horizontalSampleLines = sampleLines;
	}

	public Color getLeftSampleLineColor() {
		return leftSampleLineColor;
	}

	public void setLeftSampleLineColor(Color leftSampleLineColor) {
		this.leftSampleLineColor = leftSampleLineColor;
	}

	public Color getRightSampleLineColor() {
		return rightSampleLineColor;
	}

	public void setRightSampleLineColor(Color rightSampleLineColor) {
		this.rightSampleLineColor = rightSampleLineColor;
	}

	public Color getBackgroundColor() {
		return backgroundColor;
	}

	public void setBackgroundColor(Color backgroundColor) {
		this.backgroundColor = backgroundColor;
	}

	public void setFrameLines(boolean frameLines) {
		this.frameLines = frameLines;
	}

	public boolean isFrameLines() {
		return frameLines;
	}

	public void setFrameLineColor(Color frameLineColor) {
		this.frameLineColor = frameLineColor;
	}

	public Color getFrameLineColor() {
		return frameLineColor;
	}

	public void setVerticalSampleLines(boolean verticalSampleLines) {
		this.verticalSampleLines = verticalSampleLines;
	}

	public boolean isVerticalSampleLines() {
		return verticalSampleLines;
	}

	public void setSampleHeightInPixel(int sampleHeightInPixel) {
		this.sampleHeightInPixel = sampleHeightInPixel;
	}

	public int getSampleHeightInPixel() {
		return sampleHeightInPixel;
	}

	public JFrame showJFrame(boolean read) throws IOException {
		return showJFrame(read, 0);
	}

	public JFrame showJFrame(boolean read, int len) throws IOException {
		JFrame frame = new JFrame();
		frame.setBounds(50, 200, 1024, 200);
		frame.addKeyListener(this);
		frame.add(getUIComponent());
		frame.setVisible(true);
		setVerticalSampleLines(true);
		setHorizontalSampleLines(false);

		if (read && reader != null) {

			if (len > 0) {
				int[] data = new int[len];
				read(data);
			} else {
				int[] data = new int[4096];
				while (read(data) != -1) {
					// nop
				}
			}
			reader.close();
		}

		return frame;
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}
	
	public static void plot(int sampleSize, double scalingFactor, double[]... arrays) {
		for (int i = 0; i < arrays.length; i++) {
			double[] x = arrays[i];
			JFrame frame = plot(x, sampleSize, scalingFactor);
			frame.setBounds(10, 10 + i * frame.getHeight(), x.length + 15, frame.getHeight());
			frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		}
	}

	public static JFrame plot(double[] monodata, int sampleSize, double scalingFactor) {
		int n = monodata.length;
		// pull the samples up to a minimum sample factor to display values smaller 1 as an integral value
		int[] integralData = new int[2 * n];
		for (int i = 0; i < n; i++) {
			double sample = monodata[i] * scalingFactor;
			integralData[2 * i] = (int) (sample);
			integralData[2 * i + 1] = (int) (sample);
		}
		return plot(integralData, sampleSize);
	}

	public static JFrame plot(int[] stereoData, int sampleSize) {
		AudioDataPlotter plot = new AudioDataPlotter(new IntArrayAudioDataReader(stereoData), sampleSize);
		try {
			return plot.showJFrame(true);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}
}
