/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui;

import java.io.IOException;
import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.model.AudioChannels;
import org.audiochain.model.AudioDevice;
import org.audiochain.model.AudioDeviceChain;
import org.audiochain.model.AudioDeviceChainListener;
import org.audiochain.model.AudioDeviceListener;
import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectLifecycleListener;
import org.audiochain.model.AudioProjectListener;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.AudioTrackListener;
import org.audiochain.model.BinaryDataset;
import org.audiochain.model.BinarySynchronizable;
import org.audiochain.model.BinarySynchronizationListener;
import org.audiochain.model.ChainableAudioDevice;
import org.audiochain.model.ChainableAudioDeviceListener;
import org.audiochain.model.SourceAudioDevice;
import org.audiochain.model.UserInterfaceContextListener;

/**
 * This is the preliminary stage to an undo concept. For now, there is just a
 * change index, that is incremented for each change. To identify, if the
 * audioProject has any changes, the caller should remember the change index at a
 * "saved" status. Later the caller can ask this observer if any changes are
 * present by calling the has changes method by giving the last remembered
 * change index.
 * 
 * @author Florian Paul
 * 
 */
public class AudioProjectModificationObserver implements AudioProjectListener, AudioTrackListener, AudioDeviceChainListener, AudioDeviceListener,
UserInterfaceContextListener, ChainableAudioDeviceListener, AudioProjectLifecycleListener, BinarySynchronizationListener {

	private final AudioProject audioProject;
	private int changeIndex;
	private int changeIndexOfLastStore;
	private Collection<AudioProjectModificationObserverListener> listeners;

	public AudioProjectModificationObserver(AudioProject audioProject) {
		this.audioProject = audioProject;
		registerOnAudioProject(audioProject);
	}
	
	protected void registerOnAudioProject(AudioProject audioProject) {
		audioProject.addAudioProjectChangeListener(this);
		audioProject.getMasterAudioDeviceChain().addAudioDeviceChainListener(this);
		for (AudioDevice audioDevice : audioProject.getMasterAudioDeviceChain().getAudioDevices()) {
			registerOnAudioDevice(audioDevice);
		}
		Collection<AudioTrack> audioTracks = audioProject.getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			registerOnAudioTrack(audioTrack);
		}
		audioProject.addAudioProjectLifecycleListener(this);
	}

	protected void registerOnAudioTrack(AudioTrack audioTrack) {
		audioTrack.addAudioTrackChangeListener(this);

		AudioDeviceChain captureAudioDeviceChain = audioTrack.getCaptureAudioDeviceChain();
		captureAudioDeviceChain.addAudioDeviceChainListener(this);
		for (AudioDevice audioDevice : captureAudioDeviceChain.getAudioDevices()) {
			registerOnAudioDevice(audioDevice);
		}

		AudioDeviceChain playbackAudioDeviceChain = audioTrack.getPlaybackAudioDeviceChain();
		playbackAudioDeviceChain.addAudioDeviceChainListener(this);
		for (AudioDevice audioDevice : playbackAudioDeviceChain.getAudioDevices()) {
			registerOnAudioDevice(audioDevice);
		}

		SourceAudioDevice sourceAudioDevice = audioTrack.getSourceAudioDevice();
		if (sourceAudioDevice != null) {
			registerOnAudioDevice(sourceAudioDevice);
		}
	}
	
	protected void unregisterFromAudioProject(AudioProject audioProject) {
		audioProject.removeAudioProjectChangeListener(this);
		audioProject.getMasterAudioDeviceChain().removeAudioDeviceChainListener(this);
		for (AudioDevice audioDevice : audioProject.getMasterAudioDeviceChain().getAudioDevices()) {
			unregisterFromAudioDevice(audioDevice);
		}
		Collection<AudioTrack> audioTracks = audioProject.getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			unregisterFromAudioTrack(audioTrack);
		}
		audioProject.removeAudioProjectLifecycleListener(this);
	}

	protected void unregisterFromAudioTrack(AudioTrack audioTrack) {
		audioTrack.removeAudioTrackChangeListener(this);
		AudioDeviceChain captureAudioDeviceChain = audioTrack.getCaptureAudioDeviceChain();
		captureAudioDeviceChain.removeAudioDeviceChainListener(this);
		for (AudioDevice audioDevice : captureAudioDeviceChain.getAudioDevices()) {
			unregisterFromAudioDevice(audioDevice);
		}

		AudioDeviceChain playbackAudioDeviceChain = audioTrack.getPlaybackAudioDeviceChain();
		playbackAudioDeviceChain.removeAudioDeviceChainListener(this);
		for (AudioDevice audioDevice : playbackAudioDeviceChain.getAudioDevices()) {
			unregisterFromAudioDevice(audioDevice);
		}

		SourceAudioDevice sourceAudioDevice = audioTrack.getSourceAudioDevice();
		if (sourceAudioDevice != null) {
			unregisterFromAudioDevice(sourceAudioDevice);
		}
	}

	protected void registerOnAudioDevice(AudioDevice audioDevice) {
		audioDevice.addAudioDeviceListener(this);
		if (audioDevice instanceof ChainableAudioDevice) {
			((ChainableAudioDevice) audioDevice).addChainableAudioDeviceListener(this);
		}
		if (audioDevice instanceof BinarySynchronizable) {
			((BinarySynchronizable) audioDevice).addBinarySynchronizationListener(this);
		}
		audioDevice.getUserInterfaceContext().addUserInterfaceContextChangeListener(this);
	}

	protected void unregisterFromAudioDevice(AudioDevice audioDevice) {
		audioDevice.removeAudioDeviceListener(this);
		if (audioDevice instanceof ChainableAudioDevice) {
			((ChainableAudioDevice) audioDevice).removeChainableAudioDeviceListener(this);
		}
		if (audioDevice instanceof BinarySynchronizable) {
			((BinarySynchronizable) audioDevice).removeBinarySynchronizationListener(this);
		}
		audioDevice.getUserInterfaceContext().removeUserInterfaceContextChangeListener(this);
	}
	
	protected void fireProjectModified(AudioProject audioProject) {
		if (listeners != null) {
			for (AudioProjectModificationObserverListener listener : listeners) {
				listener.audioProjectModified(audioProject);
			}
		}
	}

	public void addAudioProjectModificationObserverListener(AudioProjectModificationObserverListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<AudioProjectModificationObserverListener>();
		}
		listeners.add(listener);
	}

	public void removeAudioProjectModificationObserverListener(AudioProjectModificationObserverListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	public boolean isModified() {
		return changeIndex != changeIndexOfLastStore;
	}

	public int getChangeIndex() {
		return changeIndex;
	}

	private void noticeChange() {
		changeIndex++;
		fireProjectModified(audioProject);
	}
	
	@Override
	public void performanceRelevantAudioDevicePropertyChanged(AudioDevice audioDevice) {
		
	}
	
	@Override
	public void persistableAudioDevicePropertyChanged(AudioDevice audioDevice) {
		noticeChange();
	}

	@Override
	public void audioDeviceAdded(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain) {
		registerOnAudioDevice(audioDevice);
		noticeChange();
	}

	@Override
	public void audioDeviceRemoved(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain) {
		unregisterFromAudioDevice(audioDevice);
		noticeChange();
	}

	@Override
	public void audioDeviceMoved(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain, boolean increment) {
		noticeChange();
	}

	@Override
	public void audioTrackNameChanged(AudioTrack audioTrack, String oldValue, String newValue) {
		noticeChange();
	}

	@Override
	public void audioTrackDescriptionChanged(AudioTrack audioTrack, String oldValue, String newValue) {
		noticeChange();
	}

	@Override
	public void audioTrackSourceAudioDeviceChanged(AudioTrack audioTrack, SourceAudioDevice oldValue, SourceAudioDevice newValue) {
		if (oldValue != null) {
			unregisterFromAudioDevice(oldValue);
		}
		if (newValue != null) {
			registerOnAudioDevice(newValue);
		}
		noticeChange();
	}

	@Override
	public void audioTrackSelectedForRecordingChanged(AudioTrack audioTrack, boolean oldValue, boolean newValue) {

	}

	@Override
	public void audioProjectNameChanged(String oldValue, String newValue) {
		noticeChange();
	}

	@Override
	public void audioProjectPathChanged(String oldValue, String newValue) {
		noticeChange();
	}

	@Override
	public void audioProjectTemplateFileNameChanged(String oldValue, String newValue) {
		noticeChange();
	}

	@Override
	public void audioProjectSampleRateChanged(int oldValue, int newValue) {
		noticeChange();
	}

	@Override
	public void audioProjectSampleSizeInBitsChanged(int oldValue, int newValue) {
		noticeChange();
	}

	@Override
	public void audioProjectBigEndianChanged(boolean oldValue, boolean newValue) {
		noticeChange();
	}

	@Override
	public void audioProjectChannelsChanged(AudioChannels oldValue, AudioChannels newValue) {
		noticeChange();
	}

	@Override
	public void audioTrackAdded(AudioTrack track) {
		registerOnAudioTrack(track);
		noticeChange();
	}

	@Override
	public void audioTrackRemoved(AudioTrack track) {
		unregisterFromAudioTrack(track);
		noticeChange();
	}

	@Override
	public void hiddenChanged(boolean oldValue, boolean newValue) {
		noticeChange();
	}

	@Override
	public void audioDeviceEnabledChanged(ChainableAudioDevice chainableAudioDevice, boolean oldValue, boolean newValue) {
		noticeChange();
	}

	@Override
	public void audioProjectFrameLengthChanged() {
		noticeChange();
	}

	@Override
	public void audioProjectOpened(AudioProject audioProject) {

	}

	@Override
	public void audioProjectClosed(AudioProject audioProject) {

	}

	@Override
	public void audioProjectActivated(AudioProject audioProject) {

	}

	@Override
	public void audioProjectStored(AudioProject audioProject) {
		changeIndexOfLastStore = changeIndex;
	}

	@Override
	public void audioProjectFramePositionChanged(long oldValue, long newValue) throws IOException {

	}

	@Override
	public void audioTrackHiddenChanged(AudioTrack audioTrack, boolean oldValue, boolean newValue) {
		noticeChange();
	}

	@Override
	public void audioTrackEnabledChanged(AudioTrack audioTrack, boolean oldValue, boolean newValue) {
		noticeChange();
	}

	@Override
	public void audioTrackMoved(AudioTrack track) {
		noticeChange();
	}

	@Override
	public void binaryChanged(BinaryDataset binaryDataset) {
		noticeChange();
	}

	public AudioProject getAudioProject() {
		return audioProject;
	}
}
