/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.MultipleGradientPaint.CycleMethod;
import java.awt.RadialGradientPaint;
import java.awt.geom.Point2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.image.BufferedImage;
import java.util.HashMap;
import java.util.Map;

import javax.swing.Icon;
import javax.swing.ImageIcon;

public class LightEmittingDiode {
	public static final int RED = 0xFF3300;
	public static final int YELLOW = 0xFFDD00;
	public static final int GREEN = 0x33FF00;
	private static float[] fractions = new float[] {
			0.1f, 0.3f, 0.7f, 0.8f
	};
	private float ledRadiusFactor = 1.0F;
	private float ledEdgeRadiusFactor = 0.1F;
	private int ledBaseColor;
	private boolean ledEnabled;

	private BufferedImage enabledImage;
	private BufferedImage disabledImage;

	private static Map<Integer, BufferedImage> enabledImageCache = new HashMap<Integer, BufferedImage>();
	private static Map<Integer, BufferedImage> disabledImageCache = new HashMap<Integer, BufferedImage>();
	private int currentImageHash;
	private final int width;
	private final int height;

	public LightEmittingDiode(int width, int height) {
		this(GREEN, width, height);
	}

	public LightEmittingDiode(int baseColor, int width, int height) {
		if (width <= 0 || height <= 0) {
			throw new IllegalArgumentException("The width and height must be greater zero.");
		}
		this.ledBaseColor = baseColor;
		this.width = width;
		this.height = height;
		createImages();
	}

	public float getLedRadiusFactor() {
		return ledRadiusFactor;
	}

	/**
	 * This factor affects the radius used for presenting the led. This
	 * radiusFactor should be set to a value greater 0.0F. If you set the
	 * radiusFactor to a value lower than 1.0F, ths LED gets a more round
	 * center. If it is greater than 1.0F, it becomes more homogeneous. The
	 * radius is calculated by the greater value of width vs height multiplied
	 * by the radiusFactor. In the case of small values (<0.4) you will get
	 * (funny or annoying) circles and fractal patterns.
	 * 
	 * @param radiusFactor
	 *            a value greater than 0.0F. The default value is 1.0F.
	 */
	public void setLedRadiusFactor(float radiusFactor) {
		if (this.ledRadiusFactor != radiusFactor) {
			this.ledRadiusFactor = radiusFactor;
			createImages();
		}
	}

	public int getLedBaseColor() {
		return ledBaseColor;
	}

	public void setLedBaseColor(int baseColor) {
		if (ledBaseColor != baseColor) {
			this.ledBaseColor = baseColor;
			enabledImage = null;
			disabledImage = null;
			createImages();
		}
	}

	private int getImageHash() {
		return width * 17 + height * 31 + getLedBaseColor() * 13 + (int) ledRadiusFactor * 7 + (int) ledEdgeRadiusFactor;
	}

	private void createImages() {
		int imageHash = getImageHash();
		if (enabledImage != null && disabledImage != null && currentImageHash == imageHash) {
			return;
		}

		// try to get the image from cache
		enabledImage = enabledImageCache.get(imageHash);
		disabledImage = disabledImageCache.get(imageHash);
		if (enabledImage != null && disabledImage != null) {
			currentImageHash = imageHash;
			return;
		}

		float wh = width / 2;
		float hh = height / 2;

		Point2D center = new Point2D.Float(wh, hh);
		float radius = Math.max(width * ledRadiusFactor, height * ledRadiusFactor);
		Color[] enabledColors = new Color[] {
				new Color(ledBaseColor), new Color(ledBaseColor & 0xCCCCCC), new Color(ledBaseColor & 0xAAAAAA), new Color(ledBaseColor & 0x666666)
		};
		Color[] disabledColors = new Color[] {
				new Color(ledBaseColor & 0x555555), new Color(ledBaseColor & 0x444444), new Color(ledBaseColor & 0x333333), new Color(ledBaseColor & 0x111111)
		};
		RadialGradientPaint enabledRadialGradientPaint = new RadialGradientPaint(center, radius, fractions, enabledColors, CycleMethod.REFLECT);
		RadialGradientPaint disabledRadialGradientPaint = new RadialGradientPaint(center, radius, fractions, disabledColors, CycleMethod.REFLECT);

		enabledImage = createBufferedImage(width, height, enabledRadialGradientPaint);
		disabledImage = createBufferedImage(width, height, disabledRadialGradientPaint);
		enabledImageCache.put(imageHash, enabledImage);
		disabledImageCache.put(imageHash, disabledImage);
		currentImageHash = imageHash;
	}

	private BufferedImage createBufferedImage(int width, int height, RadialGradientPaint radialGradientPaint) {
		BufferedImage ret = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		Graphics2D enabledGraphics = (Graphics2D) ret.getGraphics();
		enabledGraphics.setPaint(radialGradientPaint);
		float r = Math.min(width * ledEdgeRadiusFactor, height * ledEdgeRadiusFactor);
		enabledGraphics.fill(new RoundRectangle2D.Double(0, 0, width, height, r, r));
		return ret;
	}

	public void setLedBaseColor(Color baseColor) {
		setLedBaseColor(baseColor.getRGB());
	}

	public boolean isLedEnabled() {
		return ledEnabled;
	}

	public void setLedEnabled(boolean enabled) {
		this.ledEnabled = enabled;
	}

	public float getLedEdgeRadiusFactor() {
		return ledEdgeRadiusFactor;
	}

	public void setLedEdgeRadiusFactor(float ledEdgeRadiusFactor) {
		if (this.ledEdgeRadiusFactor != ledEdgeRadiusFactor) {
			this.ledEdgeRadiusFactor = ledEdgeRadiusFactor;
			createImages();
		}
	}

	public BufferedImage getBufferedImage() {
		return ledEnabled ? enabledImage : disabledImage;
	}

	public int getWidth() {
		return width;
	}

	public int getHeight() {
		return height;
	}

	public BufferedImage getEnabledImage() {
		return enabledImage;
	}

	public BufferedImage getDisabledImage() {
		return disabledImage;
	}

	public Icon getEnabledIcon() {
		return new ImageIcon(enabledImage);
	}

	public Icon getDisabledIcon() {
		return new ImageIcon(disabledImage);
	}
}
