/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui;

import java.util.ArrayList;
import java.util.List;


public class LightEmittingDiodeArray {
	private final List<LightEmittingDiode> ledArray;
	private final int[] arrayBaseColors;
	private final int ledCount;
	private final int ledWidth;
	private final int ledHeight;

	private float ledValue;
	private float ledSingleValue;
	private int enabledCount;
	private int enabledNumber;

	public LightEmittingDiodeArray(int[] arrayBaseColors, int ledWidth, int ledHeight) {
		ledArray = new ArrayList<LightEmittingDiode>();
		this.arrayBaseColors = arrayBaseColors;
		this.ledCount = arrayBaseColors.length;
		this.ledWidth = ledWidth;
		this.ledHeight = ledHeight;
		init();
	}

	private void init() {
		for (int baseColor : arrayBaseColors) {
			LightEmittingDiode led = new LightEmittingDiode(baseColor, ledWidth, ledHeight);
			ledArray.add(led);
		}
	}

	private void refreshLedValues() {
		int count = 0;
		for (LightEmittingDiode led : ledArray) {
			count++;
			boolean enable = count <= enabledCount;
			if (enable && !led.isLedEnabled()) {
				led.setLedEnabled(true);
			} else if (!enable && led.isLedEnabled()) {
				led.setLedEnabled(false);
			}
			if (enabledNumber == count) {
				led.setLedEnabled(true);
			}
		}
	}

	public float getLedValue() {
		return ledValue;
	}

	/**
	 * Enables the LEDs of the array. The given parameter is a value between 0
	 * and 1 and reflects the percentage of the enabled LED of the array,
	 * beginning from the bottom in case of a vertical presentation or from
	 * right in case of a horizontal presentation. For example if there are 10
	 * LEDs and this value is set to 0.4F, the first 4 LEDs will be enabled.
	 * 
	 * @param value
	 *            a "percentage" value between 0.0F and 1.0F.
	 */
	public void setLedValue(float value) {
		this.ledValue = value;
		int newEnabledCount = (int) (ledCount * value);
		if (enabledCount != newEnabledCount) {
			enabledCount = newEnabledCount;
			refreshLedValues();
		}
	}

	public float getLedSingleValue() {
		return ledSingleValue;
	}

	/**
	 * Enables a single LED that represents the given number of a value between
	 * 0 and 1. This feature can be used to realize a "peak hold". For example
	 * if there are 10 LEDs and this value is set to 0.7F, the seventh LED will
	 * be enabled, even if the setLedValue() was set to 0.4F.
	 * 
	 * @param value
	 *            a "percentage" value between 0.0F and 1.0F.
	 */
	public void setLedSingleValue(float ledSingleValue) {
		this.ledSingleValue = ledSingleValue;
		int newEnabledNumber = (int) (ledCount * ledSingleValue);
		if (newEnabledNumber != enabledNumber) {
			enabledNumber = newEnabledNumber;
			refreshLedValues();
		}
	}
	
	public LightEmittingDiode getLightEmittingDiode(int index) {
		return ledArray.get(index);
	}
}
