/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.cli;

import static org.audiochain.ui.UIFormatHelper.*;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.Manifest;
import java.util.regex.Matcher;

import javax.sound.sampled.AudioFormat;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioIOException;
import org.audiochain.io.FrameObserver;
import org.audiochain.model.AudioChannels;
import org.audiochain.model.AudioDevice;
import org.audiochain.model.AudioDeviceChain;
import org.audiochain.model.AudioDeviceRegistry;
import org.audiochain.model.AudioDeviceRegistryException;
import org.audiochain.model.AudioMixer;
import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.ChainableAudioDevice;
import org.audiochain.model.CommandLineUserInterface;
import org.audiochain.model.CommandLineUserInterfaceListener;
import org.audiochain.model.HardwareMixer;
import org.audiochain.model.PerformanceAdapter;
import org.audiochain.model.PerformanceManager;
import org.audiochain.model.SourceAudioDevice;
import org.audiochain.model.UserInterfaceContext;
import org.audiochain.ui.AudioDataPlotter;
import org.audiochain.ui.AudioProjectLifecycleObserver;
import org.audiochain.ui.ProjectMixdown;
import org.audiochain.ui.sync.CommandSyncSocket;
import org.audiochain.ui.sync.SyncManager;
import org.audiochain.ui.sync.SynchronizationListener;

public class CliAudioRecorder implements CommandLineUserInterfaceListener {
	private static final String LINESEP = System.getProperty("line.separator");
	private BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
	private AudioProject project;
	private AudioProject projectTemplate;
	private AudioTrack selectedTrack;
	private long customFramePosition;
	private PerformanceManager performanceManager;
	private Collection<CommandLineUserInterface> commandLineUserInterfaces = new LinkedHashSet<CommandLineUserInterface>();
	private final Collection<Command> commands = new LinkedHashSet<Command>();
	private final AudioProjectLifecycleObserver audioProjectLifecycleObserver = new AudioProjectLifecycleObserver();

	private void createCommands() {
		commands.add(new PatternCommand("i", "Show Info. State of tracks and project settings.") {
			@Override
			public void execute() {
				showInfo();
			}
		});
		commands.add(new PatternCommand("l", "List Tracks.") {
			@Override
			public void execute() {
				showTracks();
			}
		});
		commands.add(new PatternCommand("n", "Add a New track. Selects this new track automatically.") {
			@Override
			public void execute() {
				try {
					selectedTrack = project.addNewAudioTrack(projectTemplate);
					save(false);
				} catch (InstantiationException e) {
					println("Your template may be corrupt: " + e.getMessage());
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					println("Your template may be corrupt: " + e.getMessage());
					e.printStackTrace();
				}
				showTracks();
			}
		});
		commands.add(new PatternCommand("r <minutes>:<seconds>",
				"Start the Recording. There must be at least one selected track. Set frame to position of <minutes> and <seconds>. You can also use a dot instead of the colon." + LINESEP
						+ "If no value is specified, the position is the last set position. The recording will start immediately.") {
			@Override
			protected String getPattern() {
				return "^r\\s*(\\d*)[:\\.]?(\\d*)\\s*$";
			}

			@Override
			public void execute() throws IOException, AudioIOException {
				if (selectedTrack == null) {
					println("no track for recording selected.");
				} else if (selectedTrack.getSourceAudioDevice() == null) {
					println("no source device defined for this track.");
				} else {
					selectedTrack.selectForRecording();
					String minStr = m.group(1).trim();
					int min = minStr.isEmpty() ? 0 : Integer.parseInt(minStr);
					String secStr = m.group(2).trim();
					int sec = secStr.isEmpty() ? 0 : Integer.parseInt(secStr);
					if (!minStr.isEmpty() || !secStr.isEmpty()) {
						float rate = project.getAudioFormat().getFrameRate();
						customFramePosition = (long) (min * 60 * rate) + (long) (sec * rate);
						println("set frame position to " + customFramePosition + "");
					}
					startRecording(customFramePosition);
				}
			}
		});
		commands.add(new PatternCommand("rb", "Start the Recording from the Beginning. There must be at least one selected track.") {
			@Override
			public void execute() throws IOException, AudioIOException {
				if (selectedTrack == null) {
					println("no track for recording selected.");
				} else if (selectedTrack.getSourceAudioDevice() == null) {
					println("no source device defined for this track.");
				} else {
					selectedTrack.selectForRecording();
					startRecording(0);
				}
			}
		});
		commands.add(new PatternCommand("b", "Start the playBack from the Beginning.") {
			@Override
			public void execute() throws IOException, AudioIOException {
				startPlayback(0);
			}
		});
		commands.add(new PatternCommand("just enter", "Toggle between playback and stop. Playback starts from the current frame position.") {
			@Override
			public boolean matches(String command) {
				return command.isEmpty();
			}

			@Override
			public void execute() throws IOException, AudioIOException {
				boolean wasrunning = performanceManager.isRunning();
				if (wasrunning) {
					if (performanceManager.isRecording()) {
						println();
						println("stop recording ...");
						performanceManager.stop();
						unselectForRecording();
					} else {
						println();
						println("stop playback ...");
						performanceManager.stop();
					}
				} else {
					startPlayback(project.getFramePosition());
				}
			}
		});
		commands.add(new PatternCommand("<any number>", "Select a track.") {
			@Override
			protected String getPattern() {
				return "^\\s*(\\d+)\\s*(.*)$";
			}

			@Override
			public void execute() throws Exception {
				selectTrack(m);
			}
		});
		commands.add(new PatternCommand("show", "Show the PCM Signal of the selected track.") {
			@Override
			public void execute() {
				showPcmSignalOfSelectedTrack();
			}
		});
		commands.add(new PatternCommand("showall", "Show the Mixed PCM Signal of all tracks.") {
			@Override
			public void execute() {
				showMixedPcmSignal();
			}
		});
		commands.add(new PatternCommand("fr <value>", "Set frame to position <value>. If no value is specified, the position is the last set position. The playback will start immediately.") {
			@Override
			protected String getPattern() {
				return "^fr\\s*(\\d*)\\s*$";
			}

			@Override
			public void execute() throws IOException, AudioIOException {
				String position = m.group(1).trim();
				if (!position.isEmpty()) {
					customFramePosition = Long.parseLong(position);
					println("set frame position to " + customFramePosition + "");
				}

				startPlayback(customFramePosition);
			}
		});
		commands.add(new PatternCommand("w <minutes>:<seconds>", "Set frame to position of <minutes> and <seconds>. You can also use a dot instead of the colon." + LINESEP
				+ "If no value is specified, the position is the last set position. The playback will start immediately.") {
			@Override
			protected String getPattern() {
				return "^w\\s*(\\d*)[:\\.]?(\\d*)\\s*$";
			}

			@Override
			public void execute() throws IOException, AudioIOException {
				String minStr = m.group(1).trim();
				int min = minStr.isEmpty() ? 0 : Integer.parseInt(minStr);
				String secStr = m.group(2).trim();
				int sec = secStr.isEmpty() ? 0 : Integer.parseInt(secStr);
				if (!minStr.isEmpty() || !secStr.isEmpty()) {
					float rate = project.getAudioFormat().getFrameRate();
					customFramePosition = (long) (min * 60 * rate) + (long) (sec * rate);
					println("set frame position to " + customFramePosition + "");
				}

				startPlayback(customFramePosition);
			}
		});
		commands.add(new PatternCommand("sp <audio device>", "Set the Device <audio device> as source device for the selected track. For example: sp Drum") {
			@Override
			protected String getPattern() {
				return "^sp\\s+(\\w+)\\s*$";
			}

			@Override
			public void execute() throws InstantiationException, IllegalAccessException, AudioDeviceRegistryException, IOException, ClassNotFoundException {
				if (selectedTrack == null) {
					println("no track selected.");
					return;
				}
				String deviceName = m.group(1);

				Class<SourceAudioDevice> sourceAudioDeviceClass = AudioDeviceRegistry.getInstance().getSourceAudioDeviceClassByName(deviceName);
				if (sourceAudioDeviceClass != null) {
					SourceAudioDevice device = sourceAudioDeviceClass.newInstance();
					selectedTrack.setSourceAudioDevice(device);
					showTracks();
				} else {
					println("Device '" + deviceName + "' not found in the list of source audio device classes.");
				}
				save(false);
			}
		});
		commands.add(new PatternCommand("ep <audio device>", "Enable or add the Device <audio device> to the selected track for playback. For example: ep Basic") {
			@Override
			protected String getPattern() {
				return "^ep\\s+(.+)\\s*$";
			}

			@Override
			public void execute() throws InstantiationException, IllegalAccessException, AudioDeviceRegistryException, IOException, ClassNotFoundException {
				if (selectedTrack == null) {
					println("no track selected.");
					return;
				}
				enableDevice(m.group(1), true, selectedTrack.getPlaybackAudioDeviceChain());
				save(false);
			}
		});
		commands.add(new PatternCommand("dp <audio device>", "Disable Device <audio device> of the selected track for playback. For example: dp Delay") {
			@Override
			protected String getPattern() {
				return "^dp\\s+(.+)\\s*$";
			}

			@Override
			public void execute() throws InstantiationException, IllegalAccessException, AudioDeviceRegistryException, IOException, ClassNotFoundException {
				if (selectedTrack == null) {
					println("no track selected.");
					return;
				}
				enableDevice(m.group(1), false, selectedTrack.getPlaybackAudioDeviceChain());
				save(false);
			}
		});
		commands.add(new PatternCommand("rp <audio device>", "Remove Device <audio device> from the selected track for playback. For example: rp LevelMeter") {
			@Override
			protected String getPattern() {
				return "^rp\\s+(.+)\\s*$";
			}

			@Override
			public void execute() {
				if (selectedTrack == null) {
					println("no track selected.");
					return;
				}
				removeDevice(m.group(1), selectedTrack.getPlaybackAudioDeviceChain());
				save(false);
			}
		});
		commands.add(new PatternCommand("er <audio device>", "Enable or add the Device <audio device> to the selected track for recording.") {
			@Override
			protected String getPattern() {
				return "^er\\s+(.+)\\s*$";
			}

			@Override
			public void execute() throws InstantiationException, IllegalAccessException, AudioDeviceRegistryException, IOException, ClassNotFoundException {
				if (selectedTrack == null) {
					println("no track selected.");
					return;
				}
				enableDevice(m.group(1), true, selectedTrack.getCaptureAudioDeviceChain());
				save(false);
			}
		});
		commands.add(new PatternCommand("dr <audio device>", "Disable Device <audio device> of the selected track for recording.") {
			@Override
			protected String getPattern() {
				return "^dr\\s+(.+)\\s*$";
			}

			@Override
			public void execute() throws AudioDeviceRegistryException, InstantiationException, IllegalAccessException, IOException, ClassNotFoundException {
				if (selectedTrack == null) {
					println("no track selected.");
					return;
				}
				enableDevice(m.group(1), false, selectedTrack.getCaptureAudioDeviceChain());
				save(false);
			}
		});
		commands.add(new PatternCommand("rr <audio device>", "Remove Device <audio device> from the selected track for recording.") {
			@Override
			protected String getPattern() {
				return "^rr\\s+(.+)\\s*$";
			}

			@Override
			public void execute() {
				if (selectedTrack == null) {
					println("no track selected.");
					return;
				}
				removeDevice(m.group(1), selectedTrack.getCaptureAudioDeviceChain());
				save(false);
			}
		});
		commands.add(new PatternCommand("em <audio device>", "Enable or add the Device <audio device> to the Master device chain.") {
			@Override
			protected String getPattern() {
				return "^em\\s+(.+)\\s*$";
			}

			@Override
			public void execute() throws AudioDeviceRegistryException, InstantiationException, IllegalAccessException, IOException, ClassNotFoundException {
				enableDevice(m.group(1), true, project.getMasterAudioDeviceChain());
				save(false);
			}
		});
		commands.add(new PatternCommand("dm <audio device>", "Disable Device <audio device> of the Master device chain.") {
			@Override
			protected String getPattern() {
				return "^dm\\s+(.+)\\s*$";
			}

			@Override
			public void execute() throws AudioDeviceRegistryException, InstantiationException, IllegalAccessException, IOException, ClassNotFoundException {
				enableDevice(m.group(1), false, project.getMasterAudioDeviceChain());
				save(false);
			}
		});
		commands.add(new PatternCommand("rm <audio device>", "Remove Device <audio device> from the Master device chain.") {
			@Override
			protected String getPattern() {
				return "^rm\\s+(.+)\\s*$";
			}

			@Override
			public void execute() {
				removeDevice(m.group(1), project.getMasterAudioDeviceChain());
				save(false);
			}
		});
		commands.add(new PatternCommand("c <device-property> [value]", "Set the <value> for the <device-property>. For example: c level -3") {
			@Override
			protected String getPattern() {
				return "^c\\s*(\\S+)\\s*(.*)\\s*$";
			}

			@Override
			public void execute() {
				editAudioDevice(m.group(1), m.group(2));
				save(false);
			}
		});
		commands.add(new PatternCommand("name <value>", "Set the name of the selected track to <value>. For example: name Lead Guitar") {
			@Override
			protected String getPattern() {
				return "^name\\s+(.+)\\s*$";
			}

			@Override
			public void execute() {
				if (selectedTrack != null) {
					String name = m.group(1);
					selectedTrack.setName(name);
					showInfo();
					save(false);
				} else {
					println("no track selected.");
				}
			}
		});
		commands.add(new PatternCommand("descr <value>", "Set the Description of the selected track to <value>. For example: descr Wonderful Solo") {
			@Override
			protected String getPattern() {
				return "^descr\\s+(.+)\\s*$";
			}

			@Override
			public void execute() {
				if (selectedTrack != null) {
					String descr = m.group(1);
					selectedTrack.setDescription(descr);
					showInfo();
					save(false);
				} else {
					println("no track selected.");
				}
			}
		});
		commands.add(new PatternCommand("mixer", "Choose the hardware Mixer for playback and recording.") {
			@Override
			public void execute() throws IOException, AudioIOException {
				AudioMixer audioMixer = AudioMixer.getInstance();
				audioMixer.setAutomaticMixerDetection(false);
				println("choose mixer for playback");
				HardwareMixer selectedPlaybackMixer = audioMixer.getSelectedPlaybackMixer();
				HardwareMixer newSelectedPlaybackMixer = selectMixer(selectedPlaybackMixer, AudioMixer.getAvailablePlaybackMixers());
				println();
				println("choose mixer for recording");
				HardwareMixer selectedCaptureMixer = audioMixer.getSelectedCaptureMixer();
				HardwareMixer newSelectedCaptureMixer = selectMixer(selectedCaptureMixer, AudioMixer.getAvailableCaptureMixers());
				boolean testMixerCombination;
				testMixerCombination = AudioMixer.testMixerCombination(newSelectedPlaybackMixer, newSelectedCaptureMixer);
				if (testMixerCombination) {
					audioMixer.setSelectedPlaybackMixer(newSelectedPlaybackMixer);
					audioMixer.setSelectedCaptureMixer(newSelectedCaptureMixer);
					audioMixer.store();
				} else {
					println("mixer combination does not work, so using previously selected mixers");
				}
				save(false);
			}
		});
		commands.add(new PatternCommand("playbackbuffer <value>", "Set the Buffer size to <value> samples. Default is 2048.") {
			@Override
			protected String getPattern() {
				return "^playbackbuffer\\s*(\\d+)\\s*$";
			}

			@Override
			public void execute() throws AudioIOException {
				int bufferSize = Integer.parseInt(m.group(1));
				if (bufferSize <= 0 || bufferSize >= Integer.MAX_VALUE) {
					bufferSize = 2048;
				}
				AudioMixer.getInstance().setPlaybackBufferSizeInSamples(bufferSize);
				println("set buffer size to " + AudioMixer.getInstance().getPlaybackBufferSizeInSamples() + " samples");
				save(false);
			}
		});
		commands.add(new PatternCommand("capturebuffer <value>", "Set the Buffer size to <value> samples. Default is 2048.") {
			@Override
			protected String getPattern() {
				return "^capturebuffer\\s*(\\d+)\\s*$";
			}

			@Override
			public void execute() throws AudioIOException {
				int bufferSize = Integer.parseInt(m.group(1));
				if (bufferSize <= 0 || bufferSize >= Integer.MAX_VALUE) {
					bufferSize = 2048;
				}
				AudioMixer.getInstance().setCaptureBufferSizeInSamples(bufferSize);
				println("set buffer size to " + AudioMixer.getInstance().getCaptureBufferSizeInSamples() + " samples");
				save(false);
			}
		});
		commands.add(new PatternCommand("audio", "Select the project's default Audio quality. Do this before the first recording!") {
			@Override
			public void execute() {
				selectProjectAudioQuality();
				save(false);
			}
		});
		commands.add(new PatternCommand("mixdown", "Create a mixdown audio file of all tracks.") {
			@Override
			public void execute() {
				try {
					File file = new ProjectMixdown(project).createMixdown();
					println("mixdown file " + file.getAbsolutePath() + " created");
				} catch (Throwable e) {
					e.printStackTrace();
				}
			}
		});
		commands.add(new PatternCommand("syncserver <port>", "Start the sync server at the given port.") {
			@Override
			protected String getPattern() {
				return "^syncserver\\s*(\\d+)\\s*$";
			}

			@Override
			public void execute() {
				int port = Integer.parseInt(m.group(1));
				SyncManager sync = SyncManager.getInstance();
				if (sync.isSyncServerRunning()) {
					sync.stopSyncServer();
				}
				sync.setServerPort(port);
				sync.startSyncServer();
			}
		});
		commands.add(new PatternCommand("syncserver stop", "Stop the sync server if running.") {
			@Override
			protected String getPattern() {
				return "^syncserver\\s*stop\\s*$";
			}

			@Override
			public void execute() {
				SyncManager sync = SyncManager.getInstance();
				if (sync.isSyncServerRunning()) {
					sync.stopSyncServer();
				}
			}
		});
		commands.add(new PatternCommand("sync <host> <port>", "Open a sync connection to the given host and port.") {
			@Override
			protected String getPattern() {
				return "^sync\\s*(\\S+)\\s*(\\d+)\\s*$";
			}

			@Override
			public void execute() {
				String host = m.group(1);
				int port = Integer.parseInt(m.group(2));
				SyncManager sync = SyncManager.getInstance();
				try {
					sync.createRemoteConnection(host, port);
				} catch (UnknownHostException e) {
					println(e.getMessage());
				} catch (IOException e) {
					println(e.getMessage());
				}
			}
		});
		commands.add(new PatternCommand("sync list", "List all open sync connections.") {
			@Override
			protected String getPattern() {
				return "^sync\\s*list\\s*$";
			}

			@Override
			public void execute() {
				SyncManager sync = SyncManager.getInstance();
				Collection<CommandSyncSocket> allCommandSyncSockets = sync.getAllCommandSyncSockets();
				for (CommandSyncSocket commandSyncSocket : allCommandSyncSockets) {
					println("Session " + commandSyncSocket.getSessionIdentifier() + " is connected to " + commandSyncSocket.getSocket().getInetAddress() + " "
							+ commandSyncSocket.getSocket().getPort());
				}
			}
		});
		commands.add(new PatternCommand("sync close <session>", "Closes an open sync connection with the given session identifier.") {
			@Override
			protected String getPattern() {
				return "^sync\\s*close\\s*(\\S+)\\s*$";
			}

			@Override
			public void execute() {
				String sessionIdentifier = m.group(1);
				SyncManager sync = SyncManager.getInstance();
				Collection<CommandSyncSocket> allCommandSyncSockets = sync.getAllCommandSyncSockets();
				for (CommandSyncSocket commandSyncSocket : allCommandSyncSockets) {
					if (commandSyncSocket.getSessionIdentifier().equals(sessionIdentifier)) {
						commandSyncSocket.exit();
					}
				}
			}
		});
		commands.add(new PatternCommand("v", "Delete selected track. Deletes all recorded audio files too.") {
			@Override
			public void execute() {
				if (selectedTrack != null) {
					if (read("do you really want to delete the track " + getSelectedTrackNumber() + "? [y|N]", "n").equalsIgnoreCase("y")) {
						project.removeAudioTrack(selectedTrack);
						selectedTrack = null;
						showTracks();
						save(false);
					}
				} else {
					println("no track selected.");
				}
			}
		});
		commands.add(new PatternCommand("q", "Exit the application (Quit).") {
			@Override
			public void execute() throws IOException {
				performanceManager.stop();
				save(true);
				project.fireAudioProjectClosed();
				println("bye");
				System.exit(0);
			}
		});
		commands.add(new PatternCommand("help", "Show this Help.") {
			@Override
			public void execute() throws AudioDeviceRegistryException, IOException, ClassNotFoundException, InstantiationException, IllegalAccessException {
				showHelp();
			}
		});
		commands.add(new AliasCommand(commands) {
			boolean listAliases;

			@Override
			public boolean matches(String command) {
				listAliases = false;
				if (command.equals("alias")) {
					listAliases = true;
					return true;
				}
				return super.matches(command);
			}

			@Override
			public void execute() throws Exception {
				if (listAliases) {
					println(getAliasesAsString());
				} else {
					super.execute();
				}
			}
		});
	}

	private synchronized void startRecording(long framePosition) throws IOException, AudioIOException {
		performanceManager.stop();

		registerCommandLineUserInterfaceListener(project.getMasterAudioDeviceChain());
		registerCommandLineUserInterfaceListener(selectedTrack.getCaptureAudioDeviceChain());
		try {
			project.setFramePosition(framePosition);
			performanceManager.start();
		} catch (IOException e) {
			performanceManager.stop();
			throw e;
		} catch (AudioIOException e) {
			performanceManager.stop();
			throw e;
		}
	}

	private synchronized void startPlayback(long framePosition) throws IOException, AudioIOException {
		performanceManager.stop();

		registerCommandLineUserInterfaceListener(project.getMasterAudioDeviceChain());

		unselectForRecording();
		try {
			if (project.getPlaybackTracks().isEmpty()) {
				println("no tracks for playback available");
			} else {
				project.setFramePosition(framePosition);
				performanceManager.start();
			}
		} catch (IOException e) {
			performanceManager.stop();
			throw e;
		} catch (AudioIOException e) {
			performanceManager.stop();
			throw e;
		}
	}

	private void registerCommandLineUserInterfaceListener(AudioDeviceChain chain) {
		// activate all audioDevices and obtain all CLI-UIs
		for (ChainableAudioDevice device : chain.getAudioDevices()) {
			@SuppressWarnings("unchecked")
			UserInterfaceContext<AudioDevice> userInterfaceContext = (UserInterfaceContext<AudioDevice>) device.getUserInterfaceContext();
			CommandLineUserInterface commandLineUserInterface = userInterfaceContext.getCommandLineUserInterface(device);
			if (commandLineUserInterface != null) {
				commandLineUserInterface.addCommandLineUserInterfaceListener(this);
			}
		}
	}

	@Override
	public void commandLineUserInterfacePerformanceDisplayStringUpdate(CommandLineUserInterface commandLineUserInterface) {
		commandLineUserInterfaces.add(commandLineUserInterface);
		boolean rec = performanceManager.isRecording();
		long seconds = project.getFramePosition() / project.getFrameRate();
		print("\r" + (rec ? "Recording" : "Playback") + " " + formatTime(seconds));
		for (CommandLineUserInterface clui : new LinkedHashSet<CommandLineUserInterface>(commandLineUserInterfaces)) {
			print(" " + clui.getPerformanceDisplayString());
		}
	}

	private void selectTrack(Matcher m) throws Exception {
		int trackNumber = Integer.parseInt(m.group(1));
		AudioTrack trackAtNumber = getTrack(trackNumber);
		if (trackAtNumber != null) {
			String subcommand = m.group(2);
			if (subcommand != null && !subcommand.trim().isEmpty()) {
				final AudioTrack before = selectedTrack;
				selectedTrack = trackAtNumber;
				exec(subcommand);
				println("Command '" + subcommand + "' executed" + (getSelectedTrackNumber() != null ? (" on track " + getSelectedTrackNumber()) : "") + ".");
				selectedTrack = before;
				showTracks();
			} else {
				selectedTrack = trackAtNumber;
				showTracks();
			}
		}
	}

	private AudioTrack getTrack(int trackNumber) {
		Collection<AudioTrack> audioTracks = project.getAudioTracks();
		if (trackNumber < 1 || trackNumber > audioTracks.size()) {
			println("track " + trackNumber + " does not exit. There are " + audioTracks.size() + " tracks.");
			return null;
		}
		return audioTracks.toArray(new AudioTrack[0])[trackNumber - 1];
	}

	private void editAudioDevice(String propertyName, String newValue) {
		if (selectedTrack == null) {
			println("no track selected.");
			return;
		}

		Collection<AudioDevice> devices = new LinkedHashSet<AudioDevice>(selectedTrack.getPlaybackAudioDeviceChain().getAudioDevices());
		devices.addAll(selectedTrack.getCaptureAudioDeviceChain().getAudioDevices());
		SourceAudioDevice sourceAudioDevice = selectedTrack.getSourceAudioDevice();
		if (sourceAudioDevice != null) {
			devices.add(sourceAudioDevice);
		}
		for (AudioDevice audioDevice : devices) {
			try {
				@SuppressWarnings("unchecked")
				UserInterfaceContext<AudioDevice> userInterfaceContext = (UserInterfaceContext<AudioDevice>) audioDevice.getUserInterfaceContext();
				CommandLineUserInterface commandLineUserInterface = userInterfaceContext.getCommandLineUserInterface(audioDevice);
				if(commandLineUserInterface.modifyAudioDevice(propertyName, newValue)) {
					showTracks();
					return;
				}
			} catch (Throwable e) {
				e.printStackTrace();
				Throwable searchCause = e;
				Throwable rootCause;
				do {
					rootCause = searchCause;
					searchCause = rootCause.getCause();
				} while (searchCause != null);
				println(rootCause.getClass().getSimpleName() + ": " + rootCause.getMessage());
				return;
			}
		}
		println("Device property '" + propertyName + "' not found.");
	}

	private void enableDevice(String audioDeviceName, boolean enabled, AudioDeviceChain audioDeviceChain) throws AudioDeviceRegistryException, InstantiationException, IllegalAccessException,
			IOException, ClassNotFoundException {

		// try to find an existing device and set it enabled
		for (ChainableAudioDevice audioDevice : audioDeviceChain.getAudioDevices()) {
			String name = audioDevice.getName();
			if (name.equals(audioDeviceName)) {
				audioDevice.setEnabled(enabled);
				println(AudioDevice.class.getSimpleName() + " '" + audioDevice.getName() + "' " + (enabled ? "enabled." : "disabled."));
				showInfo();
				return;
			}
		}

		// add the AudioDevice, if enabled and not exists
		if (enabled) {
			Class<ChainableAudioDevice> availableClass = AudioDeviceRegistry.getInstance().getChainableAudioDeviceClassByName(audioDeviceName);
			if (availableClass != null) {
				ChainableAudioDevice audioDevice = availableClass.newInstance();
				audioDevice.setEnabled(enabled);
				audioDeviceChain.addAudioDevice(audioDevice);
				println(AudioDevice.class.getSimpleName() + " '" + audioDevice.getName() + "' " + (enabled ? "enabled." : "disabled."));
				showInfo();
				return;
			}
		}
		println("Device '" + audioDeviceName + "' not found.");
	}

	private void removeDevice(String audioDeviceName, AudioDeviceChain audioDeviceChain) {
		for (ChainableAudioDevice audioDevice : audioDeviceChain.getAudioDevices()) {
			if (audioDevice.getName().equals(audioDeviceName)) {
				audioDeviceChain.removeAudioDevice(audioDevice);
				println(AudioDevice.class.getSimpleName() + " '" + audioDevice.getName() + "' removed.");
				showInfo();
				return;
			}
		}
		println(AudioDevice.class.getSimpleName() + " '" + audioDeviceName + "' not found.");
	}

	private void showPcmSignalOfSelectedTrack() {
		if (selectedTrack == null) {
			println("no track selected.");
			return;
		}

		try {
			FrameObserver frameObserver = new FrameObserver() {
				@Override
				public long getCurrentFramePosition() {
					return 0;
				}

				@Override
				public void setFramePosition(long framePosition) {
				}

				@Override
				public long getLatencyInFrames() {
					return 0;
				}

				@Override
				public long getReaderFramePosition() {
					return 0;
				}
			};
			AudioFormat audioFormat = project.getAudioFormat();
			float frameRate = audioFormat.getFrameRate();
			int sampleSizeInBits = audioFormat.getSampleSizeInBits();
			AudioDataReader base = selectedTrack.getSourceAudioDevice().createAudioDataReader(frameRate, sampleSizeInBits, frameObserver);
			if (base != null) {
				AudioDataReader reader = selectedTrack.getPlaybackAudioDeviceChain().createAudioDataReaderChain(base, frameRate, sampleSizeInBits, frameObserver);
				AudioDataPlotter plotter = new AudioDataPlotter(reader, sampleSizeInBits);
				plotter.showJFrame(true);
			} else {
				println("no active audio file.");
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void showMixedPcmSignal() {
		try {
			unselectForRecording();
			AudioDataReader reader = new ProjectMixdown(project).createAudioDataReader();
			int sampleSizeInBits = project.getAudioFormat().getSampleSizeInBits();
			AudioDataPlotter plotter = new AudioDataPlotter(reader, sampleSizeInBits);
			plotter.showJFrame(true);
		} catch (IOException e) {
			e.printStackTrace();
		} catch (IllegalStateException e) {
			println(e.getMessage());
		} catch (AudioIOException e) {
			e.printStackTrace();
		}
	}

	private void unselectForRecording() throws IOException, AudioIOException {
		for (AudioTrack otherTrack : project.getAudioTracks()) {
			otherTrack.unselectForRecording();
		}
	}

	private void assertExists(File dir) {
		if (!dir.exists()) {
			String cmd = read("directory '" + dir.getAbsolutePath() + "' does not exist. Should I create it? [Y|n]", "y");
			if (cmd.equalsIgnoreCase("y")) {
				if (!dir.mkdirs()) {
					println("Cannot create directory '" + dir.getAbsolutePath() + "'. Exiting.");
					System.exit(-1);
				}
			} else {
				println("We need a project directory to work. Exiting.");
				System.exit(0);
			}
		}
	}

	private void showAudioQuality() {
		println("Audio Quality: " + project.getChannels() + " Channels, " + project.getSampleRate() + " Hz, " + project.getSampleSizeInBits() + " bit, " + project.getFrameRate() + " frames/sec, "
				+ project.getFrameSizeInBytes() + " bytes/frame, " + (project.isBigEndian() ? "BigEndian" : "LittleEndian"));
	}

	private void showInfo() {
		showAudioQuality();
		AudioMixer audioMixer = AudioMixer.getInstance();
		try {
			println("Mixer for Playback: " + audioMixer.getPlaybackMixer());
			println("Mixer for Recording: " + audioMixer.getCaptureMixer());
		} catch (AudioIOException e) {
			println(e.getMessage());
		}
		println("Playback Buffer Size: " + audioMixer.getPlaybackBufferSizeInSamples() + " samples");
		println("Capture Buffer Size: " + audioMixer.getCaptureBufferSizeInSamples() + " samples");
		print("Master Devices: ");
		printAudioDevices(project.getMasterAudioDeviceChain().getAudioDevices().toArray(new ChainableAudioDevice[0]));
		println();
		Collection<AudioTrack> audioTracks = project.getAudioTracks();
		if (audioTracks.isEmpty()) {
			print(getSelectedTrackNumber() == null ? "No track selected" : "Selected track " + getSelectedTrackNumber());
			print(", " + audioTracks.size() + " Tracks, ");
		}
		showTracks();
	}

	private void showTracks() {
		Integer selectedTrackNumber = getSelectedTrackNumber();
		Collection<AudioTrack> audioTracks = project.getAudioTracks();
		int maxtracks = audioTracks.size();
		int i = 1;
		for (AudioTrack track : audioTracks) {
			print(selectedTrackNumber != null && selectedTrackNumber == i ? (char) 10142 : ' ');
			String tracksymbol = maxtracks <= 10 ? "" + ((char) (10111 + i)) : "" + i;
			if (track.getName() != null && !track.getName().trim().isEmpty()) {
				print(tracksymbol + " " + track.getName() + " ");
			} else {
				print("Track " + tracksymbol + " ");
			}
			if (track.getDescription() != null && !track.getDescription().trim().isEmpty()) {
				print(track.getDescription());
				if (track.getDescription().trim().length() > 76) {
					println();
					print("    ");
				} else {
					print(" ");
				}
			}
			print(" PB: ");
			printAudioDevices(track.getPlaybackAudioDeviceChain().getAudioDevices().toArray(new ChainableAudioDevice[0]));

			print(", REC: ");
			printAudioDevices(track.getCaptureAudioDeviceChain().getAudioDevices().toArray(new ChainableAudioDevice[0]));

			print(", SRC: ");
			SourceAudioDevice sourceAudioDevice = track.getSourceAudioDevice();
			if (sourceAudioDevice != null) {
				printAudioDevices(sourceAudioDevice);
			} else {
				print("(no source device)");
			}

			println();
			i++;
		}
	}

	private static void printAudioDevices(AudioDevice... audioDevices) {
		int j = 0;
		for (AudioDevice audioDevice : audioDevices) {
			if (j++ > 0) {
				print(" ");
			}
			@SuppressWarnings("unchecked")
			UserInterfaceContext<AudioDevice> userInterfaceContext = (UserInterfaceContext<AudioDevice>) audioDevice.getUserInterfaceContext();
			CommandLineUserInterface commandLineUserInterface = userInterfaceContext.getCommandLineUserInterface(audioDevice);
			if (commandLineUserInterface != null) {
				String displayString = commandLineUserInterface.getDisplayString();
				boolean disabledAudioDevice = audioDevice instanceof ChainableAudioDevice && !((ChainableAudioDevice) audioDevice).isEnabled();
				if (disabledAudioDevice) {
					print("{");
				}
				print(displayString);
				if (disabledAudioDevice) {
					print("}");
				}
			}
		}
	}

	private Integer getSelectedTrackNumber() {
		if (selectedTrack != null) {
			int i = 1;
			for (AudioTrack track : project.getAudioTracks()) {
				if (selectedTrack.equals(track)) {
					return i;
				}
				i++;
			}
		}
		return null;
	}

	private HardwareMixer selectMixer(HardwareMixer currentMixer, Collection<HardwareMixer> availableMixers) {
		List<HardwareMixer> mixerList = new ArrayList<HardwareMixer>(availableMixers);
		int len = mixerList.size();
		for (int i = 0; i < len; i++) {
			println("Mixer " + (i + 1) + ": " + mixerList.get(i).getDescription());
		}
		if (currentMixer != null) {
			println("Currently selected mixer is " + currentMixer);
		} else {
			println("Currently selected mixer is the systems default mixer.");
		}
		int mixerNumber = readInt("Please select a mixer number (choose 0 for default mixer)", 0);

		if (mixerNumber < 0 || mixerNumber > len) {
			if (currentMixer != null) {
				println("Selecting mixer " + currentMixer);
			} else {
				println("Selecting default mixer.");
			}
			return currentMixer;
		} else if (mixerNumber == 0) {
			println("Selecting default mixer.");
			return null;
		} else {
			HardwareMixer newMixer = mixerList.get(mixerNumber - 1);
			println("Selecting mixer " + newMixer);
			return newMixer;
		}
	}

	private void selectProjectAudioQuality() {
		project.setChannels(AudioChannels.fromChannelCount(readInt("Please select the project's default channel count", project.getChannels().getChannelCount())));
		project.setSampleRate(readInt("Please select the project's default sample rate", project.getSampleRate()));
		project.setSampleSizeInBits(readInt("Please select the project's default sample size in bits", project.getSampleSizeInBits()));
		showAudioQuality();
	}

	private int readInt(String msg, int defaultValue) {
		String newValueStr = read(msg + " [" + defaultValue + "]", String.valueOf(defaultValue));
		int newValue;
		if (newValueStr == null) {
			newValue = defaultValue;
		} else {
			try {
				newValue = Integer.parseInt(newValueStr);
			} catch (NumberFormatException e) {
				println(newValueStr + " is no valid number. Using " + defaultValue);
				newValue = defaultValue;
			}
		}
		return newValue;
	}

	private void open(File file, File template) {
		try {
			if (template != null) {
				projectTemplate = AudioProject.load(template);
			} else {
				projectTemplate = null;
			}
		} catch (Exception e) {
			// no reason to exit
			e.printStackTrace();
		}

		if (file.isFile()) {
			try {
				project = AudioProject.load(file);
				project.fireAudioProjectOpened();
				project.fireAudioProjectActivated();
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(-1);
			}
		} else {
			String projectName;
			File dir;
			if (file.getName().toLowerCase().endsWith(AudioProject.FILEENDING)) {
				// a new not yet existing project file
				String name = file.getName();
				projectName = name.substring(0, name.length() - AudioProject.FILEENDING.length());
				dir = file.getParentFile();
				if (dir == null) {
					// current directory (workdir of the java process)
					dir = new File(new File("").getAbsolutePath());
				}
			} else if (file.getName().isEmpty() && (dir = new File(new File("").getAbsolutePath())).isDirectory()) {
				// current directory (workdir of the java process)
				projectName = dir.getName();
			} else {
				// existent or non existent relative or absolute path
				projectName = file.getName();
				dir = file;
			}

			assertExists(dir);

			String absoluteDirPath = dir.getAbsolutePath();
			loadOrCreateNewProject(projectName, absoluteDirPath);
		}
		performanceManager = new PerformanceManager(project);
		performanceManager.addPerformanceListener(new PerformanceAdapter() {
			@Override
			public void performanceStarted() {
				commandLineUserInterfaces.clear();
			}
		});
		audioProjectLifecycleObserver.registerAudioProject(project);
		showInfo();
	}

	private void loadOrCreateNewProject(String projectName, String absoluteDirPath) {
		File projectFile = new File(absoluteDirPath + System.getProperty("file.separator") + projectName + AudioProject.FILEENDING);
		if (projectFile.exists() && projectFile.isFile()) {
			// file already exists
			try {
				project = AudioProject.load(projectFile);
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(-1);
			}
		} else {
			try {
				project = AudioProject.createAudioProject(absoluteDirPath, projectTemplate);
			} catch (Exception e) {
				e.printStackTrace();
				project = new AudioProject(absoluteDirPath);
			}
			project.setName(projectName);
			project.setProjectFile(projectFile);
			println("New Project created.");
			println("Enter 'help' for a list of commands.");
		}
	}

	private void listen() {
		String line;
		try {
			createCommands();
			while ((line = br.readLine()) != null) {
				String[] cmds = line.split("#");
				for (String cmd : cmds) {
					try {
						exec(cmd);
					} catch (IllegalArgumentException e) {
						System.err.println(e.getMessage());
					} catch (Throwable e) {
						System.err.println("This error might be a bug:");
						e.printStackTrace();
					}
				}
			}
		} catch (IOException e) {
			println(e.getMessage());
		}
	}

	private void exec(String line) throws Exception {
		line = line.trim();
		boolean found = false;
		for (Command command : commands) {
			if (command.matches(line)) {
				found = true;
				command.execute();
				break;
			}
		}

		if (!found) {
			println("Unknown command: " + line);
		}
	}

	public static void main(String[] args) {
		File projectFile = null;
		File templateFile = null;

		String lastArg = null;
		for (String arg : args) {
			if ("--version".equals(arg) || "-v".equals(arg)) {
				printVersion();
				return;
			}
			if ("-t".equals(lastArg)) {
				templateFile = new File(arg);
			} else if (projectFile == null && !arg.startsWith("-")) {
				projectFile = new File(arg);
			} else if ("--help".equals(arg)) {
				usage();
				return;
			}

			lastArg = arg;
		}

		if (projectFile == null) {
			projectFile = new File("");
		}

		CliAudioRecorder cli = new CliAudioRecorder();
		cli.initSyncManager();
		cli.open(projectFile, templateFile);
		cli.listen();
	}

	private void initSyncManager() {
		SyncManager sync = SyncManager.getInstance();
		sync.setAudioProjectLifecycleObservable(audioProjectLifecycleObserver);
		sync.addSynchronizationListener(new SynchronizationListener() {
			@Override
			public void syncSocketDisconnected(CommandSyncSocket commandSyncSocket) {
				println("sync connection closed: " + commandSyncSocket.getSocket().getInetAddress());
			}

			@Override
			public void syncSocketConnected(CommandSyncSocket commandSyncSocket) {
				println("sync connection opened: " + commandSyncSocket.getSocket().getInetAddress());
			}

			@Override
			public void syncServerStopped() {
				println("sync service stopped");
			}

			@Override
			public void syncServerStarted(int port) {
				println("sync service started at port " + port);
			}

			@Override
			public void syncExceptionOccured(Exception e) {
				println("sync error: " + e.getMessage());
			}
		});
		sync.initialize();
	}

	private static void usage() {
		println("usage: java " + CliAudioRecorder.class.getName() + " [project-file" + AudioProject.FILEENDING + "]|[new project-path] [-t <project-template-file" + AudioProject.FILEENDING + ">]");
	}

	private static void printVersion() {
		InputStream is = CliAudioRecorder.class.getClassLoader().getResourceAsStream("META-INF/MANIFEST.MF");
		if (is == null) {
			println("No manifest file found.");
			return;
		}
		try {
			Manifest mf = new Manifest(is);
			Attributes mainAttributes = mf.getMainAttributes();
			// referenced / set in ant build
			println(mainAttributes.getValue("Git-Commit"));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void save(final boolean createBackupFile) {
		try {
			project.store(createBackupFile);
			project.fireAudioProjectStored();
		} catch (Exception e) {
			println("error while saving the project: " + e.getMessage());
			e.printStackTrace();
		}
	}

	private String read(String message, String defaultValue) {
		print(message + ": ");
		String ret = null;
		try {
			ret = br.readLine();
		} catch (IOException e) {
			println(e.getMessage());
		}
		if (ret == null || ret.trim().isEmpty()) {
			ret = defaultValue;
		}
		return ret;
	}

	private void showHelp() throws AudioDeviceRegistryException, IOException, ClassNotFoundException, InstantiationException, IllegalAccessException {
		final String commandFormatPattern = "%-25s %s%n";
		for (Command command : commands) {
			String description = command.getDescription();
			String[] descriptionLines = description.split(LINESEP);
			printf(commandFormatPattern, command.getSyntax(), descriptionLines[0]);
			for (int i = 1, n = descriptionLines.length; i < n; i++) {
				printf(commandFormatPattern, "", descriptionLines[i]);
			}
		}
		println();
		println("Additional Features:");
		printf(commandFormatPattern, "<any number><command>", "Select the track with number <any number> temporarily and execute <command> on it.");
		printf(commandFormatPattern, "", "For example: '5c mute true' mutes track 5 without a permanent selection of track 5.");
		printf(commandFormatPattern, "<command>#<command>#...", "Execute multiple commands. This is also combinable with <any number><command> and aliases.");

		println();
		println("Available " + AudioDevice.class.getSimpleName() + "s:");
		printf(commandFormatPattern, "<audio device> Class", AudioDevice.class.getSimpleName() + " implementation class name");
		printf("    %-25s %-15s %-15s%n", "<device-property>", "Default Value", "Options and Ranges");
		for (Class<?> clas : AudioDeviceRegistry.getInstance().getAudioDeviceClasses()) {
			try {
				AudioDevice audioDevice = (AudioDevice) clas.newInstance();
				audioDevice.initialize(project);
				@SuppressWarnings("unchecked")
				UserInterfaceContext<AudioDevice> userInterfaceContext = (UserInterfaceContext<AudioDevice>) audioDevice.getUserInterfaceContext();
				CommandLineUserInterface commandLineUserInterface = userInterfaceContext.getCommandLineUserInterface(audioDevice);
				if (commandLineUserInterface != null) {
					printf(commandFormatPattern, audioDevice.getName(), clas.getSimpleName() + (audioDevice instanceof SourceAudioDevice ? " (source device)" : ""));
					print(commandLineUserInterface.getHelpString());
				}
			} catch (InstantiationException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			}
		}
	}

	private static void println() {
		println("");
	}

	private static void println(String str) {
		System.out.println(str);
	}

	private static void print(String str) {
		System.out.print(str);
	}

	private static void print(char c) {
		System.out.print(c);
	}

	private static void printf(String format, Object... args) {
		System.out.printf(format, args);
	}
}
