/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui;

import java.awt.Color;
import java.awt.Font;
import java.awt.Rectangle;
import java.io.IOException;
import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map.Entry;

import javax.swing.UIManager;

import org.audiochain.model.ObjectStore;
import org.audiochain.ui.ProjectMixdown.AudioFileFormatType;


public class GuiSettings implements Serializable {
	private static final long serialVersionUID = 1L;
	private static final String PROPS_FILE = "gui.properties";

	private String lastOpenedFileName;
	private String lastNewProjectPathName;
	private int windowLocationX = 200;
	private int windowLocationY = 50;
	private int windowExtendedState;
	private int windowWidth = 1024;
	private int windowHeight = 600;
	private String projectTemplateFileName;
	private Collection<GuiSettingsDialogResult> dialogResults;
	private int backgroundRgbColorR = 45;
	private int backgroundRgbColorG = 45;
	private int backgroundRgbColorB = 45;
	private int foregroundRgbColorR = 255;
	private int foregroundRgbColorG = 255;
	private int foregroundRgbColorB = 255;
	private String fontName = "Sans";
	private int fontStyle = 0;
	private int fontSize = 10;
	private AudioFileFormatType lastMixdownAudioFileFormatType;
	// if there is no project opened at exit, there shall be no project opened when program becomes started again
	private boolean startEmpty;

	public void store() {
		try {
			ObjectStore.storeInSettingsDirectory(this, PROPS_FILE);
		} catch (IOException e) {
			throw new IllegalStateException(e);
		}
	}

	public static GuiSettings load() {
		GuiSettings ret;
		try {
			ret = (GuiSettings) ObjectStore.read(PROPS_FILE);
		} catch (Exception e) {
			throw new IllegalStateException(e);
		}
		if (ret == null) {
			ret = new GuiSettings();
		}
		return ret;
	}

	public void initializeUIManager() {
		Color backgroundColor = getBackgroundColor();
		Color foregroundColor = getForegroundColor();
		Font font = new Font(getFontName(), getFontStyle(), getFontSize());

		for (Entry<Object, Object> entry : new HashMap<Object, Object>(UIManager.getDefaults()).entrySet()) {
			Object key = entry.getKey();
			if (key instanceof String) {
				String k = (String) key;
				if (k.endsWith(".background")) {
					UIManager.put(key, backgroundColor);
				} else if (k.endsWith(".font")) {
					UIManager.put(key, font);
				} else if (k.endsWith(".foreground")) {
					UIManager.put(key, foregroundColor);
				} else if (k.endsWith(".caretForeground")) {
					UIManager.put(key, foregroundColor);
				} else if (k.endsWith(".selectionBackground")) {
					UIManager.put(key, foregroundColor);
				} else if (k.endsWith(".selectionForeground")) {
					UIManager.put(key, backgroundColor);
				}
			}
		}
	}

	public Rectangle getWindowBounds() {
		return new Rectangle(getWindowLocationX(), getWindowLocationY(), getWindowWidth(), getWindowHeight());
	}

	public String getLastOpenedFileName() {
		return lastOpenedFileName;
	}

	public void setLastOpenedFileName(String lastOpenedFileName) {
		this.lastOpenedFileName = lastOpenedFileName;
	}

	public int getWindowLocationX() {
		return windowLocationX;
	}

	public void setWindowLocationX(int windowLocationX) {
		this.windowLocationX = windowLocationX;
	}

	public int getWindowLocationY() {
		return windowLocationY;
	}

	public void setWindowLocationY(int windowLocationY) {
		this.windowLocationY = windowLocationY;
	}

	public int getWindowWidth() {
		return windowWidth;
	}

	public void setWindowWidth(int windowWidth) {
		this.windowWidth = windowWidth;
	}

	public int getWindowHeight() {
		return windowHeight;
	}

	public void setWindowHeight(int windowHeight) {
		this.windowHeight = windowHeight;
	}

	public String getProjectTemplateFileName() {
		return projectTemplateFileName;
	}

	public void setProjectTemplateFileName(String projectTemplateFileName) {
		this.projectTemplateFileName = projectTemplateFileName;
	}

	public Collection<GuiSettingsDialogResult> getDialogResults() {
		if (dialogResults == null) {
			dialogResults = new HashSet<GuiSettingsDialogResult>();
		}
		return dialogResults;
	}

	public void setDialogResults(Collection<GuiSettingsDialogResult> dialogResults) {
		this.dialogResults = dialogResults;
	}

	public Color getBackgroundColor() {
		return new Color(getBackgroundRgbColorR(), getBackgroundRgbColorG(), getBackgroundRgbColorB());
	}
	
	public Color getLightBackgroundColor() {
		return new Color(getBackgroundColor().getRGB() + 0x22222222);
	}

	public Color getForegroundColor() {
		return new Color(getForegroundRgbColorR(), getForegroundRgbColorG(), getForegroundRgbColorB());
	}

	public int getBackgroundRgbColorR() {
		return backgroundRgbColorR;
	}

	public void setBackgroundRgbColorR(int backgroundRgbColorR) {
		this.backgroundRgbColorR = backgroundRgbColorR;
	}

	public int getBackgroundRgbColorG() {
		return backgroundRgbColorG;
	}

	public void setBackgroundRgbColorG(int backgroundRgbColorG) {
		this.backgroundRgbColorG = backgroundRgbColorG;
	}

	public int getBackgroundRgbColorB() {
		return backgroundRgbColorB;
	}

	public void setBackgroundRgbColorB(int backgroundRgbColorB) {
		this.backgroundRgbColorB = backgroundRgbColorB;
	}

	public String getFontName() {
		return fontName;
	}

	public void setFontName(String fontName) {
		this.fontName = fontName;
	}

	public int getFontStyle() {
		return fontStyle;
	}

	public void setFontStyle(int fontStyle) {
		this.fontStyle = fontStyle;
	}

	public int getFontSize() {
		return fontSize;
	}

	public void setFontSize(int fontSize) {
		this.fontSize = fontSize;
	}

	public int getForegroundRgbColorR() {
		return foregroundRgbColorR;
	}

	public void setForegroundRgbColorR(int foregroundRgbColorR) {
		this.foregroundRgbColorR = foregroundRgbColorR;
	}

	public int getForegroundRgbColorG() {
		return foregroundRgbColorG;
	}

	public void setForegroundRgbColorG(int foregroundRgbColorG) {
		this.foregroundRgbColorG = foregroundRgbColorG;
	}

	public int getForegroundRgbColorB() {
		return foregroundRgbColorB;
	}

	public void setForegroundRgbColorB(int foregroundRgbColorB) {
		this.foregroundRgbColorB = foregroundRgbColorB;
	}

	public AudioFileFormatType getLastMixdownAudioFileFormatType() {
		return lastMixdownAudioFileFormatType;
	}

	public void setLastMixdownAudioFileFormatType(AudioFileFormatType lastMixdownAudioFileFormatType) {
		this.lastMixdownAudioFileFormatType = lastMixdownAudioFileFormatType;
	}

	public String getLastNewProjectPathName() {
		return lastNewProjectPathName;
	}

	public void setLastNewProjectPathName(String lastNewProjectPathName) {
		this.lastNewProjectPathName = lastNewProjectPathName;
	}

	public int getWindowExtendedState() {
		return windowExtendedState;
	}

	public void setWindowExtendedState(int windowExtendedState) {
		this.windowExtendedState = windowExtendedState;
	}

	public boolean isStartEmpty() {
		return startEmpty;
	}

	public void setStartEmpty(boolean startEmpty) {
		this.startEmpty = startEmpty;
	}
}
