/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui;

import static java.lang.Math.max;

import java.awt.Graphics;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.io.Serializable;

import javax.imageio.ImageIO;

public class ImageLayer implements Serializable {
	private static final long serialVersionUID = 1L;
	private String layerId;
	private String imageFileName;
	private boolean rotation;
	private boolean horizontalScaling;
	private boolean verticalScaling;
	// glue positions for stretching layers
	private int leftGlue;
	private int rightGlue;
	private int topGlue;
	private int bottomGlue;
	// boundary for moving layers
	private int leftBoundary;
	private int rightBoundary;
	private int topBoundary;
	private int bottomBoundary;
	private int relativePositionX;
	private int relativePositionY;
	private transient BufferedImage bufferedImage;
	private transient AffineTransform rotationAffineTransform;
	private transient AffineTransform scaleAffineTransform;
	private transient AffineTransform translationAffineTransform;
	private transient double componentWidth;
	private transient double componentHeight;

	BufferedImage getTransformedBufferedImage() {
		BufferedImage image = getBufferedImage();
		if (!hasAttributes()) {
			return image;
		}
		
		if (isStretching()) {
			image = stretchImage(image);
		}
		
		AffineTransform at = new AffineTransform();

		// scaling before rotation
		if (scaleAffineTransform != null) {
			at.concatenate(scaleAffineTransform);
		}

		if (rotationAffineTransform != null) {
			at.concatenate(rotationAffineTransform);
		}

		if (translationAffineTransform != null) {
			at.concatenate(translationAffineTransform);
		}

		int resultWidth = (int) Math.max(image.getWidth(), componentWidth);
		int resultHeight = (int) Math.max(image.getHeight(), componentHeight);
		image = transformImage(image, resultWidth, resultHeight, at);

		return image;
	}

	private BufferedImage stretchImage(BufferedImage bufferedImage) {
		int origWidth = bufferedImage.getWidth();
		int origHeight = bufferedImage.getHeight();
		
		if (componentWidth == 0 || componentHeight == 0) {
			componentWidth = bufferedImage.getWidth();
			componentHeight = bufferedImage.getHeight();
		}
		
		int compWidth = (int) componentWidth;
		int compHeight = (int) componentHeight;

		if (leftGlue == 0 && rightGlue == 0) {
			compWidth = origWidth;
		}
		if (topGlue == 0 && bottomGlue == 0) {
			compHeight = origHeight;
		}

		BufferedImage stretchImage = bufferedImage.getSubimage(leftGlue, topGlue, origWidth - rightGlue, origHeight - bottomGlue);

		int stretchImageWidth = origWidth - rightGlue - leftGlue;
		int stretchImageHeight = origHeight - bottomGlue - topGlue;
		int stretchRightEndPosition = compWidth - rightGlue;
		int stretchBottomEndPosition = compHeight - bottomGlue;

		int imageType = bufferedImage.getType();
		imageType = imageType == BufferedImage.TYPE_CUSTOM ? BufferedImage.TYPE_INT_ARGB : imageType;
		BufferedImage compositeImage = new BufferedImage(compWidth, compHeight, imageType);
		Graphics graphics = compositeImage.getGraphics();
		graphics.drawImage(stretchImage, leftGlue, topGlue, max(stretchRightEndPosition, leftGlue), max(stretchBottomEndPosition, topGlue), 0, 0, stretchImageWidth, stretchImageHeight, null);

		if (leftGlue > 0) {
			BufferedImage leftImage = bufferedImage.getSubimage(0, 0, leftGlue, origHeight);
			graphics.drawImage(leftImage, 0, topGlue, null);
		}

		if (rightGlue > 0) {
			BufferedImage rightImage = bufferedImage.getSubimage(origWidth - rightGlue, 0, rightGlue, origHeight);
			graphics.drawImage(rightImage, stretchRightEndPosition, topGlue, null);
		}

		if (topGlue > 0) {
			BufferedImage topImage = bufferedImage.getSubimage(0, 0, origWidth, topGlue);
			graphics.drawImage(topImage, leftGlue, 0, null);
		}

		if (bottomGlue > 0) {
			BufferedImage bottomImage = bufferedImage.getSubimage(0, origHeight - bottomGlue, origWidth, origHeight);
			graphics.drawImage(bottomImage, leftGlue, stretchBottomEndPosition, null);
		}

		bufferedImage = compositeImage;
		return bufferedImage;
	}

	private static BufferedImage transformImage(BufferedImage bufferedImage, int newWidth, int newHeight, AffineTransform at) {
		int imageType = bufferedImage.getType();
		imageType = imageType == BufferedImage.TYPE_CUSTOM ? BufferedImage.TYPE_INT_ARGB : imageType;
		BufferedImage targetImage = new BufferedImage(newWidth, newHeight, imageType);
		AffineTransformOp op = new AffineTransformOp(at, new RenderingHints(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY));
		op.filter(bufferedImage, targetImage);
		return targetImage;
	}

	void copyFrom(ImageLayer template) {
		layerId = template.layerId;
		imageFileName = template.imageFileName;
		rotation = template.rotation;
		horizontalScaling = template.horizontalScaling;
		verticalScaling = template.verticalScaling;
		leftGlue = template.leftGlue;
		rightGlue = template.rightGlue;
		topGlue = template.topGlue;
		bottomGlue = template.bottomGlue;
		leftBoundary = template.leftBoundary;
		rightBoundary = template.rightBoundary;
		topBoundary = template.topBoundary;
		relativePositionX = template.relativePositionX;
		relativePositionY = template.relativePositionY;
		bottomBoundary = template.bottomBoundary;
		bufferedImage = template.bufferedImage;
	}

	public void writeImage(File dir) throws IOException {
		if (bufferedImage == null) {
			throw new IllegalStateException("bufferedImage is null");
		}
		if (imageFileName == null) {
			throw new IllegalStateException("imageFileName is null");
		}
		ImageIO.write(bufferedImage, "PNG", new File(dir, imageFileName));
	}

	public boolean hasAttributes() {
		return isHorizontalScaling() || isVerticalScaling() || isRotation() || isStretching() || isMoving();
	}

	public boolean isStretching() {
		return getLeftGlue() > 0 || getRightGlue() > 0 || getTopGlue() > 0 || getBottomGlue() > 0;
	}

	public boolean isHorizontalMoving() {
		return leftBoundary>0 || rightBoundary>0;
	}

	public boolean isVerticalMoving() {
		return topBoundary>0||bottomBoundary>0;
	}
	
	public boolean isMoving() {
		return isHorizontalMoving() || isVerticalMoving();
	}

	AffineTransform getRotationAffineTransform() {
		return rotationAffineTransform;
	}

	void setRotationAffineTransform(AffineTransform rotationAffineTransform) {
		this.rotationAffineTransform = rotationAffineTransform;
	}

	AffineTransform getScaleAffineTransform() {
		return scaleAffineTransform;
	}

	void setScaleAffineTransform(AffineTransform scaleAffineTransform) {
		this.scaleAffineTransform = scaleAffineTransform;
	}

	AffineTransform getTranslationAffineTransform() {
		return translationAffineTransform;
	}

	void setTranslationAffineTransform(AffineTransform translationAffineTransform) {
		this.translationAffineTransform = translationAffineTransform;
	}

	public boolean isRotation() {
		return rotation;
	}

	public void setRotation(boolean rotation) {
		this.rotation = rotation;
	}

	public String getImageFileName() {
		return imageFileName;
	}

	public void setImageFileName(String imageFileName) {
		this.imageFileName = imageFileName;
	}

	public BufferedImage getBufferedImage() {
		return bufferedImage;
	}

	public void setBufferedImage(BufferedImage bufferedImage) {
		this.bufferedImage = bufferedImage;
	}

	public String getLayerId() {
		return layerId;
	}

	public void setLayerId(String layerId) {
		this.layerId = layerId;
	}

	public boolean isHorizontalScaling() {
		return horizontalScaling;
	}

	public void setHorizontalScaling(boolean horizontalScaling) {
		this.horizontalScaling = horizontalScaling;
	}

	public boolean isVerticalScaling() {
		return verticalScaling;
	}

	public void setVerticalScaling(boolean verticalScaling) {
		this.verticalScaling = verticalScaling;
	}

	public int getLeftGlue() {
		return leftGlue;
	}

	public void setLeftGlue(int leftGlued) {
		this.leftGlue = leftGlued;
	}

	public int getRightGlue() {
		return rightGlue;
	}

	public void setRightGlue(int rightGlued) {
		this.rightGlue = rightGlued;
	}

	public double getComponentHeight() {
		return componentHeight;
	}

	public void setComponentHeight(double componentHeight) {
		this.componentHeight = componentHeight;
	}

	public double getComponentWidth() {
		return componentWidth;
	}

	public void setComponentWidth(double componentWidth) {
		this.componentWidth = componentWidth;
	}

	public int getTopGlue() {
		return topGlue;
	}

	public void setTopGlue(int topGlue) {
		this.topGlue = topGlue;
	}

	public int getBottomGlue() {
		return bottomGlue;
	}

	public void setBottomGlue(int bottomGlue) {
		this.bottomGlue = bottomGlue;
	}

	public int getLeftBoundary() {
		return leftBoundary;
	}

	public void setLeftBoundary(int leftBoundary) {
		this.leftBoundary = leftBoundary;
	}

	public int getRightBoundary() {
		return rightBoundary;
	}

	public void setRightBoundary(int rightBoundary) {
		this.rightBoundary = rightBoundary;
	}

	public int getTopBoundary() {
		return topBoundary;
	}

	public void setTopBoundary(int topBoundary) {
		this.topBoundary = topBoundary;
	}

	public int getBottomBoundary() {
		return bottomBoundary;
	}

	public void setBottomBoundary(int bottomBoundary) {
		this.bottomBoundary = bottomBoundary;
	}

	public int getRelativePositionX() {
		return relativePositionX;
	}

	public void setRelativePositionX(int relativePositionX) {
		this.relativePositionX = relativePositionX;
	}

	public int getRelativePositionY() {
		return relativePositionY;
	}

	public void setRelativePositionY(int relativePositionY) {
		this.relativePositionY = relativePositionY;
	}
}
