/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui;

import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.imageio.ImageIO;

import org.audiochain.model.ObjectStore;


public class LayeredImage implements Serializable {
	private static final long serialVersionUID = 1L;
	private static Map<String, LayeredImage> layeredImageCache;
	private String imageId;
	private Collection<ImageLayer> layers = new ArrayList<ImageLayer>();

	private void copyFrom(LayeredImage template) {
		imageId = template.imageId;
		for (ImageLayer templateLayer : template.layers) {
			ImageLayer newLayer = new ImageLayer();
			layers.add(newLayer);
			newLayer.copyFrom(templateLayer);
		}
	}

	public Collection<ImageLayer> getLayers() {
		return layers;
	}

	public void setLayers(Collection<ImageLayer> layers) {
		this.layers = layers;
	}

	public String getImageId() {
		return imageId;
	}

	public void setImageId(String imageId) {
		this.imageId = imageId;
	}

	public BufferedImage getBufferedImage() {
		BufferedImage base = null;
		Graphics graphics = null;
		for (ImageLayer layer : layers) {
			BufferedImage bufferedImage = layer.getTransformedBufferedImage();
			int imageType = bufferedImage.getType();
			imageType = imageType == BufferedImage.TYPE_CUSTOM ? BufferedImage.TYPE_INT_ARGB : imageType;
			if (base == null) {
				base = new BufferedImage(bufferedImage.getWidth(), bufferedImage.getHeight(), imageType);
				graphics = base.getGraphics();
			} else if (base.getWidth() < bufferedImage.getWidth() || base.getHeight() < bufferedImage.getHeight()) {
				BufferedImage newBase = new BufferedImage(bufferedImage.getWidth(), bufferedImage.getHeight(), imageType);
				graphics = newBase.getGraphics();
				graphics.drawImage(base, 0, 0, null);
				base = newBase;
			}
			if (graphics != null) {
				graphics.drawImage(bufferedImage, 0, 0, null);
			}
		}
		return base;
	}

	public int getMinimumWidth() {
		int width = 0;
		for (ImageLayer layer : layers) {
			int w = layer.getLeftGlue() + layer.getRightGlue();
			width = Math.max(width, w);
		}
		return width;
	}

	public int getMinimumHeight() {
		int height = 0;
		for (ImageLayer layer : layers) {
			int h = layer.getTopGlue() + layer.getBottomGlue();
			height = Math.max(height, h);
		}
		return height;
	}

	public boolean isScalable() {
		for (ImageLayer layer : layers) {
			if (layer.isHorizontalScaling() || layer.isVerticalScaling() || layer.isStretching()) {
				return true;
			}
		}
		return false;
	}

	public boolean isMoveable() {
		for (ImageLayer layer : layers) {
			if (layer.isMoving()) {
				return true;
			}
		}
		return false;
	}

	public boolean isHorizontalMoveable() {
		for (ImageLayer layer : layers) {
			if (layer.isHorizontalMoving()) {
				return true;
			}
		}
		return false;
	}

	public boolean isVerticalMoveable() {
		for (ImageLayer layer : layers) {
			if (layer.isVerticalMoving()) {
				return true;
			}
		}
		return false;
	}

	public int getLeftMaximumBoundary() {
		int value = 0;
		for (ImageLayer layer : layers) {
			value = Math.max(value, layer.getLeftBoundary());
		}
		return value;
	}

	public int getTopMaximumBoundary() {
		int value = 0;
		for (ImageLayer layer : layers) {
			value = Math.max(value, layer.getTopBoundary());
		}
		return value;
	}

	public int getRightMaximumBoundary() {
		int value = 0;
		for (ImageLayer layer : layers) {
			value = Math.max(value, layer.getRightBoundary());
		}
		return value;
	}

	public int getBottomMaximumBoundary() {
		int value = 0;
		for (ImageLayer layer : layers) {
			value = Math.max(value, layer.getBottomBoundary());
		}
		return value;
	}

	/*
	 * Store for creating the files once. The developer should execute this once
	 * and copy the results to the source directory. The later reading at
	 * program runtime is done from classpath.
	 */
	public void store(File dir) throws IOException {
		if (dir != null) {
			if (!dir.exists()) {
				if (!dir.mkdirs()) {
					throw new IOException("unable to create dir " + dir);
				}
			}
		}
		Collection<ImageLayer> componentImageLayers = getLayers();
		for (ImageLayer componentImageLayer : componentImageLayers) {
			componentImageLayer.setImageFileName(getImageId() + "_" + componentImageLayer.getLayerId() + ".png");
			componentImageLayer.writeImage(dir);
		}
		File ciFile = new File(dir, getComponentImageFileName());
		BufferedWriter bw = new BufferedWriter(new FileWriter(ciFile));
		ObjectStore.writeObject(this, bw);
		bw.close();
	}

	public String getComponentImageFileName() {
		return getImageId() + ".ci";
	}

	public static LayeredImage read(String componentImageFilePathInClasspath) throws IOException, IllegalAccessException, ClassNotFoundException, InstantiationException, NoSuchMethodException, InvocationTargetException {
		if (layeredImageCache == null) {
			layeredImageCache = new HashMap<String, LayeredImage>();
		}
		LayeredImage cachedImage = layeredImageCache.get(componentImageFilePathInClasspath);
		if (cachedImage != null) {
			LayeredImage ret = new LayeredImage();
			ret.copyFrom(cachedImage);
			return ret;
		}
		InputStream is = LayeredImage.class.getResourceAsStream("/" + componentImageFilePathInClasspath);
		if (is == null) {
			throw new IllegalArgumentException("The resource '" + componentImageFilePathInClasspath + "' does not exist in classpath.");
		}
		BufferedReader br = new BufferedReader(new InputStreamReader(is, Charset.forName("UTF-8")));
		LayeredImage ret = (LayeredImage) ObjectStore.readObject(br);

		int pathEnd = componentImageFilePathInClasspath.lastIndexOf('/');
		String imageDirPathInClasspath = "";
		if (pathEnd > 1) {
			imageDirPathInClasspath = componentImageFilePathInClasspath.substring(0, pathEnd);
		}

		Collection<ImageLayer> layers = ret.getLayers();
		for (ImageLayer componentImageLayer : layers) {
			String imageFileName = componentImageLayer.getImageFileName();
			String imageFilePathInClasspath = imageDirPathInClasspath + "/" + imageFileName;
			InputStream imageIs = LayeredImage.class.getResourceAsStream("/" + imageFilePathInClasspath);
			if (imageIs == null) {
				throw new IllegalArgumentException("The resource '" + imageFilePathInClasspath + "' does not exist in classpath.");
			}
			componentImageLayer.setBufferedImage(ImageIO.read(imageIs));
		}

		layeredImageCache.put(componentImageFilePathInClasspath, ret);
		return ret;
	}
}
