/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Window;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.Popup;
import javax.swing.PopupFactory;
import javax.swing.SwingUtilities;

public class LayeredImageComponent extends JComponent {
	private static final long serialVersionUID = 1L;
	protected final LayeredImage layeredImage;
	private BufferedImage cachedBufferedImage;

	private JLabel valueLabel;
	private Popup valueLabelPopup;

	public LayeredImageComponent(LayeredImage layeredImage) {
		this.layeredImage = layeredImage;
		init();
	}

	public LayeredImageComponent(String componentImageFilePathInClasspath) {
		try {
			this.layeredImage = LayeredImage.read(componentImageFilePathInClasspath);
		} catch (IOException e) {
			throw new IllegalArgumentException(e);
		} catch (IllegalAccessException e) {
			throw new IllegalArgumentException(e);
		} catch (ClassNotFoundException e) {
			throw new IllegalArgumentException(e);
		} catch (InstantiationException e) {
			throw new IllegalArgumentException(e);
		} catch (NoSuchMethodException e) {
			throw new IllegalArgumentException(e);
		} catch (InvocationTargetException e) {
			throw new IllegalArgumentException(e);
		}
		init();
	}

	private void init() {
		BufferedImage bufferedImage = layeredImage.getBufferedImage();
		int w = bufferedImage.getWidth();
		int h = bufferedImage.getHeight();
		if (!layeredImage.isScalable()) {
			Dimension dim = new Dimension(w, h);
			setMinimumSize(dim);
			setMaximumSize(dim);
			setPreferredSize(dim);
		} else {
			int minimumWidth = layeredImage.getMinimumWidth();
			int minimumHeight = layeredImage.getMinimumHeight();
			setMinimumSize(new Dimension(minimumWidth, minimumHeight));
			setPreferredSize(new Dimension(w, h));
			addComponentListener(new ComponentAdapter() {
				@Override
				public void componentResized(ComponentEvent e) {
					scaleLayers();
				}
			});
		}
	}

	protected void refresh() {
		cachedBufferedImage = null;
		repaint();
	}

	@Override
	public void paint(Graphics g) {
		super.paint(g);
		if (cachedBufferedImage == null) {
			cachedBufferedImage = layeredImage.getBufferedImage();
		}
		g.drawImage(cachedBufferedImage, 0, 0, null);
	}

	public void rotateLayers(float angle) {
		Collection<ImageLayer> layers = layeredImage.getLayers();
		for (ImageLayer imageLayer : layers) {
			if (imageLayer.isRotation()) {
				AffineTransform at = new AffineTransform();
				imageLayer.setRotationAffineTransform(at);
				BufferedImage bufferedImage = imageLayer.getBufferedImage();
				int imageWidth = bufferedImage.getWidth();
				int imageHeight = bufferedImage.getHeight();
				double theta = Math.toRadians(angle);
				double w2 = imageWidth / 2D;
				double h2 = imageHeight / 2D;
				at.rotate(theta, w2, h2);
				this.cachedBufferedImage = null;
			}
		}
	}

	public void scaleLayers() {
		Collection<ImageLayer> layers = layeredImage.getLayers();
		double componentHeight = getHeight();
		double componentWidth = getWidth();
		for (ImageLayer imageLayer : layers) {
			imageLayer.setComponentWidth(componentWidth);
			imageLayer.setComponentHeight(componentHeight);
			boolean hscale = imageLayer.isHorizontalScaling();
			boolean vscale = imageLayer.isVerticalScaling();
			if (hscale || vscale) {
				AffineTransform at = new AffineTransform();
				imageLayer.setScaleAffineTransform(at);
				BufferedImage bufferedImage = imageLayer.getBufferedImage();
				double imageWidth = bufferedImage.getWidth();
				double imageHeight = bufferedImage.getHeight();
				if (hscale && vscale) {
					double sx = componentWidth / imageWidth;
					double sy = componentHeight / imageHeight;
					at.scale(sx, sy);
				} else if (hscale) {
					double sx = componentWidth / imageWidth;
					at.scale(sx, 1D);
				} else if (vscale) {
					double sy = componentHeight / imageHeight;
					at.scale(1D, sy);
				}
			}
		}
		this.cachedBufferedImage = null;
	}

	public BufferedImage getCachedBufferedImage() {
		if (cachedBufferedImage == null) {
			cachedBufferedImage = layeredImage.getBufferedImage();
		}
		return cachedBufferedImage;
	}
	
	protected void showValueLabel(String text, int relativeX, int relativeY) {
		showValueLabel(text, new Point(relativeX, relativeY));
	}
	
	protected void showValueLabel(String text, Point relativePoint) {
		hideValueLabel();
		if (valueLabel == null) {
			valueLabel = new JLabel();
		}
		updateValueLabel(text, relativePoint);
		Point popupLocation = getPopupLocation(relativePoint);
		valueLabelPopup = PopupFactory.getSharedInstance().getPopup(valueLabel, valueLabel, popupLocation.x, popupLocation.y);
		valueLabelPopup.show();
	}
	
	private Point getPopupLocation(Point relativePoint) {
		Point locationOnScreen = getLocationOnScreen();
		int x = locationOnScreen.x + relativePoint.x;
		int y = locationOnScreen.y + relativePoint.y;
		return new Point(x, y);
	}
	
	protected void updateValueLabel(String text, int relativeX, int relativeY) {
		updateValueLabel(text, new Point(relativeX, relativeY));
	}

	protected void updateValueLabel(String text, Point relativePoint) {
		if (valueLabel != null) {
			valueLabel.setText(text);
			Window windowAncestor = SwingUtilities.getWindowAncestor(valueLabel);
			if (windowAncestor != null) {
				windowAncestor.pack();
				Point popupLocation = getPopupLocation(relativePoint);
				windowAncestor.setLocation(popupLocation);
			}
		}
	}

	protected void hideValueLabel() {
		if (valueLabelPopup != null) {
			valueLabelPopup.hide();
			valueLabelPopup = null;
			valueLabel = null;
		}
	}
}
