/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.util.Collection;

public class MoveableImageComponent extends LayeredImageComponent {
	private static final long serialVersionUID = 1L;
	private int moveElementPositionX;
	private int moveElementPositionY;

	public MoveableImageComponent(String componentImageFilePathInClasspath) {
		super(componentImageFilePathInClasspath);
		init();
	}

	private void init() {
		if (layeredImage.isMoveable()) {
			MouseAdapter mouseAdapter = new MouseAdapter() {
				@Override
				public void mousePressed(MouseEvent e) {
					if (!isEnabled()) {
						return;
					}
					setPosition(e.getX(), e.getY());
					moveElementToPosition();
					refresh();
				}

				@Override
				public void mouseDragged(MouseEvent e) {
					if (!isEnabled()) {
						return;
					}
					setPosition(e.getX(), e.getY());
					moveElementToPosition();
					refresh();
				}
			};

			addMouseListener(mouseAdapter);
			addMouseMotionListener(mouseAdapter);
		}
	}

	private void setPosition(int x, int y) {
		int hmax = (int) getHorizontalMaximumPosition();
		int hmin = (int) getHorizontalMinimumPosition();
		int vmax = (int) getVerticalMaximumPosition();
		int vmin = (int) getVerticalMinimumPosition();
		if (x < hmin) {
			x = hmin;
		} else if (x > hmax) {
			x = hmax;
		}
		if (y < vmin) {
			y = vmin;
		} else if (y > vmax) {
			y = vmax;
		}
		moveElementPositionX = x;
		moveElementPositionY = y;
	}

	protected void setHorizontalValueInPercent(double percentValue) {
		double hmax = getHorizontalMaximumPosition();
		double hmin = getHorizontalMinimumPosition();
		double diff = hmax - hmin;
		double factor = percentValue / 100D;
		double value = diff * factor + hmin;
		int x = (int) value;
		setPosition(x, moveElementPositionY);
		moveElementToPosition();
	}

	protected void setVerticalValueInPercent(double percentValue) {
		double vmax = getVerticalMaximumPosition();
		double vmin = getVerticalMinimumPosition();
		double diff = vmax - vmin;
		double factor = percentValue / 100D;
		double value = diff * factor + vmin;
		int y = (int) value;
		setPosition(moveElementPositionX, y);
		moveElementToPosition();
	}

	protected double getHorizontalValueInPercent() {
		double hmax = getHorizontalMaximumPosition();
		double hmin = getHorizontalMinimumPosition();
		double diff = hmax - hmin;
		double factor = (moveElementPositionX - hmin) / diff;
		double percent = factor * 100D;
		return percent;
	}

	protected double getVerticalValueInPercent() {
		double vmax = getVerticalMaximumPosition();
		double vmin = getVerticalMinimumPosition();
		double diff = vmax - vmin;
		double factor = (moveElementPositionY - vmin) / diff;
		double percent = factor * 100D;
		return percent;
	}

	private synchronized void moveElementToPosition() {
		double x = moveElementPositionX;
		double y = moveElementPositionY;
		Collection<ImageLayer> layers = layeredImage.getLayers();
		for (ImageLayer layer : layers) {
			if (layer.isMoving()) {
				AffineTransform translationAffineTransform = new AffineTransform();
				layer.setTranslationAffineTransform(translationAffineTransform);
				if (layer.isHorizontalMoving() && layer.isVerticalMoving()) {
					translationAffineTransform.translate(x - layer.getRelativePositionX(), y - layer.getRelativePositionY());
				} else if (layer.isHorizontalMoving()) {
					translationAffineTransform.translate(x - layer.getRelativePositionX(), 0);
				} else if (layer.isVerticalMoving()) {
					translationAffineTransform.translate(0, y - layer.getRelativePositionY());
				}
			}
		}
	}

	protected double getVerticalMinimumPosition() {
		return layeredImage.getTopMaximumBoundary();
	}

	protected double getVerticalMaximumPosition() {
		int height = getHeight();
		if (height == 0) {
			height = getCachedBufferedImage().getHeight();
		}
		return height - layeredImage.getBottomMaximumBoundary();
	}

	protected double getHorizontalMinimumPosition() {
		return layeredImage.getLeftMaximumBoundary();
	}

	protected double getHorizontalMaximumPosition() {
		int width = getWidth();
		if (width == 0) {
			width = getCachedBufferedImage().getWidth();
		}
		return width - layeredImage.getRightMaximumBoundary();
	}
}
