/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui;

import static java.awt.event.InputEvent.*;
import static java.lang.Math.*;
import static org.audiochain.ui.gui.StringImageRenderer.*;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.KeyEventDispatcher;
import java.awt.KeyboardFocusManager;
import java.awt.Point;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import javax.swing.ToolTipManager;

import org.audiochain.model.Scale;


public class RotaryPotentiometerImageComponent extends LayeredImageComponent {
	private static final long serialVersionUID = 1L;

	private DecimalFormat decimalFormat = new DecimalFormat("+0.0#;-0.0#", DecimalFormatSymbols.getInstance(Locale.ENGLISH));
	private DecimalFormat decimalDecadeFormat = new DecimalFormat("+00;-00", DecimalFormatSymbols.getInstance(Locale.ENGLISH));
	private DecimalFormat decimalCentFormat = new DecimalFormat("+000;-000", DecimalFormatSymbols.getInstance(Locale.ENGLISH));
	private DecimalFormat decimalKiloFormat = new DecimalFormat("+0.#k;-0.#k", DecimalFormatSymbols.getInstance(Locale.ENGLISH));
	private DecimalFormat decimalPreciseFormat = new DecimalFormat("+0.0;-0.0", DecimalFormatSymbols.getInstance(Locale.ENGLISH));

	// initialize with NaN because any initial set value should trigger the
	// value change playbackListeners
	private float angle = Float.NaN;
	private float minAngle;
	private float maxAngle;
	private float minValue;
	private float maxValue;
	private Float midValue;
	private float value = Float.NaN;
	private int angleMagnet;
	private int lastMouseY;
	private boolean applyAngleByCoordinate;
	private StringBuilder keysTyped;
	private Scale scale = Scale.Linear;
	private Collection<RotaryPotentiometerValueChangeListener> listeners;
	private String unit;

	private boolean integralValue;
	private static boolean ctrlPressed;

	static {
		KeyboardFocusManager.getCurrentKeyboardFocusManager().addKeyEventDispatcher(new KeyEventDispatcher() {
			@Override
			public boolean dispatchKeyEvent(KeyEvent e) {
				int keyCode = e.getKeyCode();
				if (keyCode == 17) { // ctrl
					switch (e.getID()) {
					case KeyEvent.KEY_PRESSED:
						ctrlPressed = true;
						break;
					case KeyEvent.KEY_RELEASED:
						ctrlPressed = false;
						break;
					}
				}
				return false;
			}
		});
	}

	public RotaryPotentiometerImageComponent(String componentImageFilePathInClasspath, float minAngle, float maxAngle) {
		super(componentImageFilePathInClasspath);
		this.maxAngle = maxAngle;
		this.minAngle = minAngle;
		init();
		setMinValue(-1);
		setMaxValue(1);
	}

	public static RotaryPotentiometerImageComponent createLabel2RotaryPotentiometerImageComponent() {
		return new RotaryPotentiometerImageComponent("org/audiochain/ui/gui/layered-label2-knob.ci", -145, 145);
	}

	public static RotaryPotentiometerImageComponent createLabel3RotaryPotentiometerImageComponent() {
		RotaryPotentiometerImageComponent rotaryPotentiometerImageComponent = new RotaryPotentiometerImageComponent("org/audiochain/ui/gui/layered-label3-knob.ci", -145, 145);
		rotaryPotentiometerImageComponent.setMidValue(0F);
		return rotaryPotentiometerImageComponent;
	}
	
	private void init() {
		setFocusable(true);
		addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				grabFocus();
				lastMouseY = e.getY();
				applyAngleByCoordinate = isWithinDirectAdjustmentRadius(e.getX(), e.getY());
				if (applyAngleByCoordinate) {
					float newAngle = getAngleByCoordinate(e.getX(), e.getY());
					newAngle = getAngleForIntegralValue(newAngle);
					setAngle(newAngle);
				}
				showValueLabel(formatPrecise(getValue()), getWidth(), 0);
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				hideValueLabel();
			}
		});
		addMouseMotionListener(new MouseMotionAdapter() {
			@Override
			public void mouseDragged(MouseEvent e) {
				if (applyAngleByCoordinate) {
					float newAngle = getAngleByCoordinate(e.getX(), e.getY());
					if (midValue != null && !ctrlPressed) {
						// mangetic angle a tick around the magnet
						double angleRange = Math.abs(maxAngle - minAngle);
						double tick = angleRange * 0.03D;
						if (newAngle < angleMagnet + tick && newAngle > angleMagnet - tick) {
							newAngle = angleMagnet;
						}
					}

					newAngle = getAngleForIntegralValue(newAngle);

					setAngle(newAngle);
				} else {
					int currentMouseY = e.getY();
					float dragDiff = lastMouseY - currentMouseY;
					// more sensitive adjustment when ctrl pressed
					if (ctrlPressed) {
						dragDiff /= 20F;
					}
					if (Float.isNaN(angle)) {
						angle = 0;
					}
					float newAngle = angle + dragDiff;

					if (integralValue) {
						int newPotentialValue = Math.round(getValueByAngle(newAngle));
						int oldValue = (int) value;
						if (Math.abs(Math.abs(oldValue) - Math.abs(newPotentialValue)) == 0) {
							if (dragDiff < -15) {
								newPotentialValue--;
								lastMouseY = currentMouseY;
							} else if (dragDiff > 15) {
								newPotentialValue++;
								lastMouseY = currentMouseY;
							}
						} else {
							lastMouseY = currentMouseY;
						}
						newAngle = getAngleByValue(newPotentialValue);
					} else if (midValue != null && !ctrlPressed) {
						// mangetic angle a tick around the magnet
						double angleRange = Math.abs(maxAngle - minAngle);
						double tick = angleRange * 0.03D;
						if (newAngle < angleMagnet + tick && newAngle > angleMagnet - tick) {
							newAngle = angleMagnet;
						} else {
							lastMouseY = currentMouseY;
						}
					} else {
						lastMouseY = currentMouseY;
					}

					setAngle(newAngle);
				}
				updateValueLabel(formatPrecise(getValue()), getWidth(), 0);
			}
		});
		addMouseWheelListener(new MouseWheelListener() {
			@Override
			public void mouseWheelMoved(MouseWheelEvent e) {
				// the scroll amount depends on the whole angle range
				// the smaller the range, the smaller the amount of angle change
				float angleRange = Math.abs(maxAngle - minAngle);
				float amount = angleRange * 0.02F;
				boolean down = e.getWheelRotation() > 0;
				if (down) {
					amount = -amount;
				}
				if (ctrlPressed) {
					amount /= 20F;
				}
				if (Float.isNaN(angle)) {
					angle = 0;
				}
				float newAngle = angle + amount;

				if (midValue != null && !ctrlPressed) {
					// mangetic angle a tick around the magnet
					double tick = angleRange * 0.03D;
					// angle != angleMagnet because the next attemt to scroll
					// must
					// succeed to step over the magnet point
					if (newAngle < angleMagnet + tick && newAngle > angleMagnet - tick && angle != angleMagnet) {
						newAngle = angleMagnet;
					}
				}

				if (integralValue) {
					int newPotentialValue = (int) getValueByAngle(newAngle);
					int oldValue = (int) value;
					if (Math.abs(Math.abs(oldValue) - Math.abs(newPotentialValue)) == 0) {
						if (down) {
							newPotentialValue--;
						} else {
							newPotentialValue++;
						}
					}
					newAngle = getAngleByValue(newPotentialValue);
				}

				setAngle(newAngle);
				// force the tool tip to be refreshed immediately
				ToolTipManager.sharedInstance().mouseMoved(e);
			}
		});
		addKeyListener(new KeyAdapter() {
			@Override
			public void keyTyped(KeyEvent e) {
				// tries to accumulate the typed keys and parse them to a number
				// after a short while, it resets itself
				if ((e.getModifiersEx() & CTRL_DOWN_MASK) != 0) {
					char keyChar = e.getKeyChar();
					if (keysTyped == null) {
						keysTyped = new StringBuilder();
					}
					keysTyped.append(keyChar);

					try {
						float value = Float.parseFloat(keysTyped.toString().replace(',', '.'));
						setValue(value);
						new Thread() {
							@Override
							public void run() {
								try {
									sleep(2000);
									keysTyped = null;
								} catch (InterruptedException e) {
									// nop
								}
							}
						}.start();
					} catch (NumberFormatException ex) {
						if (keyChar != '-') {
							keysTyped = null;
						}
					}
				}
			}
		});
	}

	@Override
	public void paint(Graphics g) {
		super.paint(g);
		Font font = g.getFont();
		Font smallerFont = new Font(font.getName(), font.getStyle(), font.getSize() - 1);
		g.setFont(smallerFont);

		int w2 = getWidth() / 2;
		Color foregroundColor = GlobalGuiContext.getGuiSettings().getForegroundColor();

		Float midValue = this.midValue;
		if (midValue != null) {
			String midValueString = format(midValue);
			BufferedImage midLabel = drawNarrowString(midValueString, smallerFont, foregroundColor);
			g.drawImage(midLabel, w2 - midLabel.getWidth() / 2, 0, null);
		}

		String minValueString = format(minValue);
		BufferedImage minLabel = drawNarrowString(minValueString, smallerFont, foregroundColor);
		int minLabelX = w2 / 2 - minLabel.getWidth() / 2;
		if (minLabelX < 0) {
			minLabelX = 0;
		}
		g.drawImage(minLabel, minLabelX, getHeight() - minLabel.getHeight(), null);

		String maxValueString = format(maxValue);
		BufferedImage maxLabel = drawNarrowString(maxValueString, smallerFont, foregroundColor);
		int maxLabelX = w2 / 2 * 3 - maxLabel.getWidth() / 2;
		if (maxLabelX + maxLabel.getWidth() > getWidth()) {
			maxLabelX = getWidth() - maxLabel.getWidth();
		}
		g.drawImage(maxLabel, maxLabelX, getHeight() - maxLabel.getHeight(), null);

		g.setFont(font);
	}

	private float getAngleForIntegralValue(float newAngle) {
		if (integralValue) {
			newAngle = getAngleByValue(Math.round(getValueByAngle(newAngle)));
		}
		return newAngle;
	}

	private boolean isWithinDirectAdjustmentRadius(int x, int y) {
		final int centerX = getWidth() / 2;
		final int centerY = getHeight() / 2;
		final double w = getWidth() / 10D;
		final double w1 = 2D * w;
		final double w2 = 5D * w;

		double dx = x - centerX;
		double dy = y - centerY;
		double d = Math.sqrt(dx * dx + dy * dy);
		// if the mouse pointer is located within a defined sensitive
		// area, the poti can be adjusted by pointer instead of up-down
		// movements
		return d >= w1 && d <= w2;
	}

	private float getAngleByCoordinate(int x, int y) {
		int centerX = getWidth() / 2;
		int centerY = getHeight() / 2;
		double dx = x - centerX;
		double dy = y - centerY;
		double degrees = Math.toDegrees(Math.atan(dy / dx));
		return (float) (dx >= 0 ? degrees + 90 : degrees - 90);
	}

	public float getAngle() {
		return angle;
	}

	public void setAngle(float angle) {
		if (changeAngle(angle)) {
			setValueByAngle();
		}
	}

	private boolean changeAngle(float angle) {
		if (angle < minAngle) {
			angle = minAngle;
		}
		if (angle > maxAngle) {
			angle = maxAngle;
		}
		if (this.angle != angle) {
			this.angle = angle;
			rotateLayers(angle);
			repaint();
			return true;
		}
		return false;
	}

	private void setValueByAngle() {
		changeValue(getValueByAngle(angle));
	}

	private float getValueByAngle(float angle) {
		assertMaxGreaterMin();
		final float angleRange = abs(maxAngle - minAngle);
		final float absAngleInRange = angle - minAngle;
		// from 0.0 to 1.0
		final float angleFactor = absAngleInRange / angleRange;

		float value;
		switch (scale) {
		case Logarithmic:
			if (midValue != null) {
				double k;
				// angleFactor = 0.5 means middle position
				if (angleFactor >= 0.5) {
					k = (maxValue - midValue) / log(1.5);
				} else {
					k = (minValue - midValue) / log(0.5);
				}
				value = (float) (Math.log(angleFactor + 0.5) * k) + midValue;
			} else {
				// 10^(x/20) VS log10(x)*20
				// 6*log(x+0,4)/log(1,4)
				double k = 0.15; // find k
				value = (float) (maxValue * Math.log(angleFactor + k) / Math.log(1 + k));
				// TODO: implement Logarithmic find k
				throw new UnsupportedOperationException("Logarithmic scaling without mid value is not supported yet.");
			}
			break;
		case Linear:
		default:
			if (midValue != null) {
				// angleFactor = 0.5 means middle position
				if (angleFactor >= 0.5) {
					value = (angleFactor * 2 - 1) * (maxValue - midValue) + midValue;
				} else {
					value = (1 - angleFactor * 2) * (minValue - midValue) + midValue;
				}
			} else {
				final float valueRange = Math.abs(maxValue - minValue);
				final float absValueInRange = angleFactor * valueRange;
				value = absValueInRange + minValue;
			}
		}
		return value;
	}

	private void setAngleByValue() {
		changeAngle(getAngleByValue(value));
	}

	private float getAngleByValue(float value) {
		assertMaxGreaterMin();
		final float angleRange = Math.abs(maxAngle - minAngle);
		final float angle;
		switch (scale) {
		case Logarithmic:
			if (midValue != null) {
				double k;
				if (value >= midValue) {
					// exp((log(1,5)/6)*x)-0,5
					k = Math.log(1.5) / (maxValue - midValue);
				} else {
					k = Math.log(0.5) / (minValue - midValue);
				}

				float angleFactor = (float) (Math.exp((value - midValue) * k) - 0.5);
				angle = angleFactor * angleRange + minAngle;
			} else {
				// 6*log(x+0,4)/log(1,4)
				// double k=0.15; // find k
				// angle = (float) (maxValue * Math.exp(valueFactor + k) /
				// Math.exp(1 + k));
				// TODO: implement Logarithmic find k
				throw new UnsupportedOperationException("Logarithmic scaling without mid value is not supported yet.");
			}
			break;
		case Linear:
		default:
			if (midValue != null) {
				float angleFactor;
				if (value >= midValue) {
					angleFactor = (value - midValue) / (maxValue - midValue) / 2 + 0.5F;
				} else {
					angleFactor = 0.5F - (value - midValue) / (minValue - midValue) / 2;
				}
				angle = angleFactor * angleRange + minAngle;
			} else {
				final float valueRange = Math.abs(maxValue - minValue);
				final float absValueInRange = value - minValue;
				// from 0.0 to 1.0
				final float valueFactor = absValueInRange / valueRange;
				final float absAngleInRange = valueFactor * angleRange;
				angle = absAngleInRange + minAngle;
			}
			break;
		}
		return angle;
	}

	private void assertMaxGreaterMin() {
		if (maxAngle <= minAngle) {
			throw new IllegalStateException("The max angle must be greater than the min angle: max: " + maxAngle + ", min: " + minAngle);
		} else if (maxValue <= minValue) {
			throw new IllegalStateException("The max value must be greater than the min value: max: " + maxValue + ", min: " + minValue);
		} else if (midValue != null && midValue <= minValue) {
			throw new IllegalStateException("The mid value must be greater than the min value: mid: " + midValue + ", min: " + minValue);
		} else if (midValue != null && maxValue <= midValue) {
			throw new IllegalStateException("The max value must be greater than the mid value: max: " + maxValue + ", mid: " + midValue);
		}
	}

	public void setValue(float value) {
		if (changeValue(value)) {
			setAngleByValue();
		}
	}

	private boolean changeValue(float value) {
		if (integralValue) {
			// cut the floating point value to an integral value
			value = (int) value;
		}
		if (value < minValue) {
			value = minValue;
		}
		if (value > maxValue) {
			value = maxValue;
		}
		if (this.value != value) {
			float oldValue = this.value;
			this.value = value;
			fireValueChanged(value, oldValue);
			return true;
		}
		return false;
	}

	public void addValueChangeListener(RotaryPotentiometerValueChangeListener listener) {
		if (listeners == null) {
			listeners = new ArrayList<RotaryPotentiometerValueChangeListener>();
		}
		listeners.add(listener);
	}

	public void removeValueChangeListener(RotaryPotentiometerValueChangeListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	private void fireValueChanged(float newValue, float oldValue) {
		if (listeners == null) {
			return;
		}
		for (RotaryPotentiometerValueChangeListener listener : listeners) {
			listener.rotaryPotentiometerValueChanged(newValue, oldValue);
		}
	}

	public float getMinAngle() {
		return minAngle;
	}

	public void setMinAngle(float minAngle) {
		this.minAngle = minAngle;
	}

	public float getMaxAngle() {
		return maxAngle;
	}

	public void setMaxAngle(float maxAngle) {
		this.maxAngle = maxAngle;
	}

	public float getMinValue() {
		return minValue;
	}

	public void setMinValue(float minValue) {
		this.minValue = minValue;
		repaint();
	}

	public float getMaxValue() {
		return maxValue;
	}

	public void setMaxValue(float maxValue) {
		this.maxValue = maxValue;
		repaint();
	}

	public float getValue() {
		return value;
	}

	public int getAngleMagnet() {
		return angleMagnet;
	}

	public void setAngleMagnet(int angleMagnet) {
		this.angleMagnet = angleMagnet;
	}

	public Scale getScale() {
		return scale;
	}

	public void setScale(Scale scale) {
		if (scale == null) {
			scale = Scale.Linear;
		}
		this.scale = scale;
	}

	public Float getMidValue() {
		return midValue;
	}

	public void setMidValue(Float midValue) {
		this.midValue = midValue;
		repaint();
	}

	@Override
	public Point getToolTipLocation(MouseEvent event) {
		if (angle < 0) {
			return new Point(getWidth() / 2, getHeight() / 2);
		} else {
			return new Point(getWidth() / 2 - getWidth(), getHeight() / 2);
		}
	}

	public String format(float value) {
		if (value == 0) {
			return "0";
		} else if (value <= -1000F || value >= 1000F) {
			value /= 1000;
			return decimalKiloFormat.format(value);
		} else if (value <= -100F || value >= 100F) {
			return decimalCentFormat.format(value);
		} else if (value <= -10F || value >= 10F) {
			return decimalDecadeFormat.format(value);
		} else {
			return decimalFormat.format(value);
		}
	}

	public String formatPrecise(float value) {
		return decimalPreciseFormat.format(value) + (unit != null ? unit : "");
	}

	public boolean isIntegralValue() {
		return integralValue;
	}

	public void setIntegralValue(boolean integralValue) {
		this.integralValue = integralValue;
	}

	public String getUnit() {
		return unit;
	}

	public void setUnit(String unit) {
		this.unit = unit;
	}
}
