/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JFrame;
import javax.swing.JPanel;

public class StringImageRenderer {
	private static Map<String, BufferedImage> stringImageCache;

	/**
	 * This method draws a String and removes all empty pixel columns and rows
	 * to reduce the presentation size of the string.
	 * 
	 * @param string
	 * @param font
	 * @return
	 */
	public static BufferedImage drawNarrowString(String string, Font font, Color foregroundColor) {
		if (stringImageCache == null) {
			stringImageCache = new HashMap<String, BufferedImage>();
		}
		BufferedImage cachedImage = stringImageCache.get(string);
		if (cachedImage != null) {
			return cachedImage;
		}

		BufferedImage temp = new BufferedImage(1, 1, BufferedImage.TYPE_INT_ARGB);
		Graphics g = temp.getGraphics();
		g.setFont(font);
		FontMetrics fm = g.getFontMetrics();
		Rectangle2D bounds = fm.getStringBounds(string, g);

		int width = (int) bounds.getWidth();
		int height = (int) (bounds.getHeight());
		BufferedImage drawTemp = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		Graphics graphics = drawTemp.getGraphics();
		graphics.setFont(font);
		graphics.setColor(foregroundColor);
		graphics.drawString(string, 0, height);

		BufferedImage narrowTemp = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		final int[] emptyColumnPixels = new int[height * 4];
		final int[] emptyRowPixels = new int[width * 4];
		boolean lastColumnWasEmpty = true;
		int omittedCols = 0;
		WritableRaster targetRaster = narrowTemp.getRaster();
		WritableRaster sourceRaster = drawTemp.getRaster();
		// remove all empty columns
		for (int x = 0; x < width; x++) {
			int[] colulmnPixels = sourceRaster.getPixels(x, 0, 1, height, (int[]) null);
			if (Arrays.equals(emptyColumnPixels, colulmnPixels)) {
				if (lastColumnWasEmpty || x == width - 1) {
					// ignore this
					omittedCols++;
				} else {
					targetRaster.setPixels(x - omittedCols, 0, 1, height, colulmnPixels);
				}

				lastColumnWasEmpty = true;
			} else {
				targetRaster.setPixels(x - omittedCols, 0, 1, height, colulmnPixels);
				lastColumnWasEmpty = false;
			}
		}

		sourceRaster = targetRaster;
		int newWidth = width - omittedCols;
		if(newWidth<=0) {
			throw new IllegalArgumentException("The String '"+string+"' contains no visible chars.");
		}
		BufferedImage ret = new BufferedImage(newWidth, height, BufferedImage.TYPE_INT_ARGB);
		targetRaster = ret.getRaster();
		boolean lastRowWasEmpty = true;
		int omittedRows = 0;
		// remove all empty rows
		for (int y = 0; y < height; y++) {
			int[] rowPixels = sourceRaster.getPixels(0, y, width, 1, (int[]) null);
			if (Arrays.equals(emptyRowPixels, rowPixels)) {

				if (lastRowWasEmpty || y == height - 1) {
					// ignore this
					omittedRows++;
				} else {
					targetRaster.setPixels(0, y - omittedRows, newWidth, 1, rowPixels);
				}

				lastRowWasEmpty = true;
			} else {
				targetRaster.setPixels(0, y - omittedRows, newWidth, 1, rowPixels);
				lastRowWasEmpty = false;
			}
		}

		int newHeight = height - omittedRows;
		if (newHeight <= 0) {
			throw new IllegalArgumentException("The String '"+string+"' contains no visible chars.");
		}
		BufferedImage subimage = ret.getSubimage(0, 0, newWidth, newHeight);
		stringImageCache.put(string, subimage);
		return subimage;
	}

	public static void main(String[] args) {
		JFrame frame = new JFrame();

		JPanel panel = new JPanel() {
			private static final long serialVersionUID = 1L;

			@Override
			public void paint(Graphics g) {
				BufferedImage stringImage = StringImageRenderer.drawNarrowString("-1.0", g.getFont(), Color.WHITE);
				g.setColor(Color.BLUE);
				g.fillRect(0, 0, 300, 400);
				g.setColor(Color.RED);
				g.fillRect(0, 0, stringImage.getWidth(), stringImage.getHeight());
				g.drawImage(stringImage, 0, 0, null);
			}
		};
		frame.setContentPane(panel);
		frame.setSize(100, 100);
		frame.setLocationRelativeTo(null);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setVisible(true);
	}
}
