/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComponent;

import org.audiochain.model.AudioDevice;
import org.audiochain.model.AudioDeviceChain;
import org.audiochain.model.AudioDeviceChainListener;
import org.audiochain.model.AudioProject;
import org.audiochain.model.ChainableAudioDevice;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.gui.ImageButtonComponent;


public class AudioDeviceChainComponent extends Box {
	
	private static final long serialVersionUID = 1L;
	private final AudioDeviceChain audioDeviceChain;
	private final AudioProject audioProject;
	private final boolean reversed;
	private JComponent audioDevicesComponent;
	private Map<AudioDevice, ChainableAudioDeviceComponent> componentMap = new LinkedHashMap<AudioDevice, ChainableAudioDeviceComponent>();
	
	public AudioDeviceChainComponent(AudioProject audioProject, AudioDeviceChain audioDeviceChain, boolean reversed) {
		super(BoxLayout.PAGE_AXIS);
		this.audioProject = audioProject;
		this.audioDeviceChain = audioDeviceChain;
		this.reversed = reversed;
		init();
	}
	
	private void init() {
		setOpaque(false);
		Box chainDeviceToolbarBox = Box.createHorizontalBox();
		chainDeviceToolbarBox.add(Box.createHorizontalGlue());
		createAddDeviceButton(chainDeviceToolbarBox);
		add(chainDeviceToolbarBox);
		add(Box.createVerticalGlue());
		
		audioDevicesComponent = Box.createVerticalBox();
		add(audioDevicesComponent);
		
		audioDeviceChain.addAudioDeviceChainListener(new AudioDeviceChainListener() {
			@Override
			public void audioDeviceRemoved(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain) {
				updateAudioDeviceComponents();				
			}
			@Override
			public void audioDeviceMoved(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain, boolean increment) {
				updateAudioDeviceComponents();				
			}
			@Override
			public void audioDeviceAdded(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain) {
				updateAudioDeviceComponents();				
			}
		});
		updateAudioDeviceComponents();
	}
	
	private void createAddDeviceButton(Box toolbarBox) {
		ImageButtonComponent addControlButton = new ImageButtonComponent("org/audiochain/ui/gui/mixer/device-item.png");
		addControlButton.setToolTipText("Adds an audio device to this audio track.");
		addControlButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		addControlButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				addAudioDevice();
			}
		});
		toolbarBox.add(addControlButton);
	}
	
	private void addAudioDevice() {
		ChainableAudioDevice selectedAudioDevice = ChooseAudioDeviceDialog.selectChainableAudioDevice(this);
		if (selectedAudioDevice != null) {
			audioDeviceChain.addAudioDevice(selectedAudioDevice);
		}
	}
	
	private void removeAudioDevice(ChainableAudioDevice audioDevice) {
		if (Dialog.okCancel(this, "Removing Audio Device", "The "+AudioDevice.class.getSimpleName()+" '" + audioDevice.getName() + "' will be removed.")) {
			audioDeviceChain.removeAudioDevice(audioDevice);
		}
	}
	
	private void updateAudioDeviceComponents() {
		audioDevicesComponent.removeAll();
		audioDevicesComponent.revalidate();
		Collection<ChainableAudioDevice> audioDevices = reversed ? audioDeviceChain.getReversedAudioDevices() : audioDeviceChain.getAudioDevices();
		int lastElementIndex = audioDevices.size() - 1;
		int index = 0;
		for (final ChainableAudioDevice audioDevice : audioDevices) {
			ChainableAudioDeviceComponent component = componentMap.get(audioDevice);
			if (component == null) {
				component = new ChainableAudioDeviceComponent(audioProject, audioDevice);
				component.getRemoveButton().addMouseListener(new MouseAdapter() {
					@Override
					public void mouseClicked(MouseEvent e) {
						removeAudioDevice(audioDevice);
					}
				});
				component.getMoveUpButton().addMouseListener(new MouseAdapter() {
					@Override
					public void mouseClicked(MouseEvent e) {
						if(reversed) {
							audioDeviceChain.incrementAudioDevicePositionIndex(audioDevice);
						} else {
							audioDeviceChain.decrementAudioDevicePositionIndex(audioDevice);
						}
					}
				});
				component.getMoveDownButton().addMouseListener(new MouseAdapter() {
					@Override
					public void mouseClicked(MouseEvent e) {
						if(reversed) {
							audioDeviceChain.decrementAudioDevicePositionIndex(audioDevice);
						} else {
							audioDeviceChain.incrementAudioDevicePositionIndex(audioDevice);
						}
					}
				});
				componentMap.put(audioDevice, component);
			}
			component.getMoveUpButton().setVisible(index != 0);
			component.getMoveDownButton().setVisible(index != lastElementIndex);
			audioDevicesComponent.add(component);
			audioDevicesComponent.revalidate();
			index++;
		}

		Set<AudioDevice> obsoletCachedControlComponents = new LinkedHashSet<AudioDevice>(componentMap.keySet());
		obsoletCachedControlComponents.removeAll(audioDevices);
		for (AudioDevice audioDevice : obsoletCachedControlComponents) {
			componentMap.remove(audioDevice);
		}
	}
}
