/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import static org.audiochain.ui.gui.GlobalGuiContext.*;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;

import javax.swing.JFileChooser;

import org.audiochain.model.AudioProject;
import org.audiochain.ui.AudioProjectLifecycleObserver;
import org.audiochain.ui.AudioProjectModificationObserver;
import org.audiochain.ui.gui.AudioProjectFileFilter;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.gui.GlobalGuiContext;


class AudioProjectLifecycleManager extends AudioProjectLifecycleObserver {

	private final Map<AudioProject, AudioProjectModificationObserver> audioProjectModificationObserverMap;
	private AudioProject activeAudioProject;

	public AudioProjectLifecycleManager() {
		audioProjectModificationObserverMap = new LinkedHashMap<AudioProject, AudioProjectModificationObserver>();
	}

	public void newAudioProject() {
		CreateNewAudioProjectDialog createProjectDialog = new CreateNewAudioProjectDialog(this);
		AudioProject newProject = createProjectDialog.getAudioProject();
		if (newProject != null) {
			openAudioProject(newProject);
		}
	}
	
	public void editActiveAudioProject() {
		AudioProject project = activeAudioProject;
		if (project == null) {
			return;
		}
		@SuppressWarnings("unused")
		EditAudioProjectDialog editAudioProjectDialog = new EditAudioProjectDialog(this, project);
	}

	public void openAudioProject() {
		String fileName = GlobalGuiContext.getGuiSettings().getLastOpenedFileName();
		// convenience for the user
		if (fileName == null && activeAudioProject != null) {
			fileName = activeAudioProject.getProjectPath();
		}
		File path;
		if (fileName != null) {
			path = new File(fileName);
		} else {
			path = null;
		}
		
		File file = selectAudioFile(path, true);
		if (file != null) {
			loadAudioProject(file);
		}
	}

	private static File selectAudioFile(File file, boolean open) {
		JFileChooser jfc = file != null ? new JFileChooser(file) : new JFileChooser();
		jfc.setFileFilter(new AudioProjectFileFilter());
		int status = open ? jfc.showOpenDialog(null) : jfc.showSaveDialog(null);
		if (status == JFileChooser.APPROVE_OPTION) {
			return jfc.getSelectedFile();
		}
		return null;
	}

	public void loadLastOpenedAudioProject() {
		String lastOpenedFileName = getGuiSettings().getLastOpenedFileName();
		if (lastOpenedFileName != null) {
			loadAudioProject(new File(lastOpenedFileName));
		}
	}

	public void loadAudioProject(File file) {
		try {
			AudioProject project = AudioProject.load(file);
			if (isOpen(project)) {
				Dialog.message(this, "Open Project", "The project is already open.");
				return;
			}
			openAudioProject(project);
			getGuiSettings().setLastOpenedFileName(file.getAbsolutePath());
		} catch (Throwable e) {
			e.printStackTrace();
			String linesep = System.getProperty("line.separator");
			Dialog.asyncMessage(this, "Error while loading audio project", "Error while loading file " + file.getAbsolutePath() + linesep + linesep + "Cause: " + e);
		}
	}
	
	public void openAudioProject(AudioProject audioProject) {
		AudioProjectModificationObserver observer = new AudioProjectModificationObserver(audioProject);
		unregisterAudioProject(audioProject);
		audioProjectModificationObserverMap.put(audioProject, observer);
		fireAudioProjectOpened(audioProject);
		activateAudioProject(audioProject);
	}

	public void storeActiveAudioProject() throws IOException {
		AudioProject project = activeAudioProject;
		storeAudioProject(project);
	}

	public void storeActiveAudioProjectAs() throws IOException {
		AudioProject project = activeAudioProject;
		if (project == null) {
			return;
		}
		
		File projectFile = project.getProjectFile();
		if (projectFile == null) {
			projectFile = new File(project.getProjectPath());
		}

		File file = selectAudioFile(projectFile, false);
		if (file == null) {
			return;
		}

		project.setProjectFile(file);
		storeAudioProject(project);
	}

	private boolean storeAudioProject(AudioProject project) throws IOException {
		if (project == null) {
			return true;
		}

		if (project.getProjectFile() == null) {
			File file = selectAudioFile(new File(project.getProjectPath()), false);
			if (file == null) {
				return false;
			}
			project.setProjectFile(file);
		}
		project.store();
		fireAudioProjectStored(project);
		return true;
	}

	private boolean isOpen(AudioProject audioProject) {
		for (AudioProject project : audioProjectModificationObserverMap.keySet()) {
			if (project.getUniqueName().equals(audioProject.getUniqueName())) {
				return true;
			}
		}
		return false;
	}

	public boolean closeAllAudioProjects() throws IOException {
		for (AudioProject audioProject : new LinkedHashSet<AudioProject>(audioProjectModificationObserverMap.keySet())) {
			if(!closeAudioProject(audioProject)) {
				return false;
			}
		}
		return true;
	}

	public boolean closeActiveAudioProject() throws IOException {
		AudioProject project = activeAudioProject;
		return closeAudioProject(project);
	}

	private boolean closeAudioProject(AudioProject project) throws IOException {
		if (project != null) {
			AudioProjectModificationObserver observer = audioProjectModificationObserverMap.get(project);
			if (observer.isModified()) {
				Dialog.YesNoCancel yesNoCancel = Dialog.yesNoCancel(this, "Unsaved Modifications", "The project '" + project.getUniqueName() + "' has been modified. Do you want to save the changes?");
				switch (yesNoCancel) {
				case Yes:
					if (!storeAudioProject(project)) {
						return false;
					}
					break;
				case Cancel:
					return false;
				}
			}
			audioProjectModificationObserverMap.remove(project);
			fireAudioProjectClosed(project);

			if (!audioProjectModificationObserverMap.isEmpty()) {
				activateAudioProject(audioProjectModificationObserverMap.keySet().iterator().next());
			} else {
				activateAudioProject(null);
			}
		}
		return true;
	}

	public void activateAudioProject(AudioProject audioProject) {
		this.activeAudioProject = audioProject;
		fireAudioProjectActivated(audioProject);
	}

	public AudioProject getActiveAudioProject() {
		return activeAudioProject;
	}
	
	public AudioProjectModificationObserver getActiveAudioProjectModificationObserver() {
		if (activeAudioProject == null) {
			return null;
		}
		return audioProjectModificationObserverMap.get(activeAudioProject);
	}
	
	@Override
	public Collection<AudioProject> getAudioProjects() {
		LinkedHashSet<AudioProject> ret = new LinkedHashSet<AudioProject>(audioProjectModificationObserverMap.keySet());
		ret.addAll(super.getAudioProjects());
		return ret;
	}
}
