/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import static java.lang.Math.*;
import static javax.swing.SwingConstants.*;
import static org.audiochain.ui.gui.Dialog.*;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JTextField;

import org.audiochain.io.AudioIOException;
import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectAdapter;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.AudioTrackAdapter;
import org.audiochain.model.SourceAudioDevice;
import org.audiochain.ui.LightEmittingDiode;
import org.audiochain.ui.gui.CommonComponentFactory;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.gui.GlobalGuiContext;
import org.audiochain.ui.gui.ImageButtonComponent;
import org.audiochain.ui.gui.LightEmittingDiodeComponent;


public class AudioTrackComponent extends Box {
	private static final long serialVersionUID = 1L;
	private final AudioTrack audioTrack;
	private final AudioProject audioProject;
	private LightEmittingDiodeComponent recordingLedComponent;
	private Component cloneButton;
	private Component moveLeftButton;
	private Component moveRightButton;
	private Component moveLeftButtonOnFlatLine;
	private Component moveRightButtonOnFlatLine;
	private Component removeButton;
	private Component switchOnButton;
	private Component switchOffButton;
	private Component hideButton;
	private Component unhideButton;
	private Component verticalTrackNameLabel;
	private Container flatLineContainer;
	private Container mainContainer;
	private Component captureDeviceChainComponent;
	private Component playbackDeviceChainComponent;
	private JTextField trackNameField;

	public AudioTrackComponent(AudioProject audioProject, AudioTrack audioTrack) {
		super(BoxLayout.X_AXIS);
		this.audioProject = audioProject;
		this.audioTrack = audioTrack;
		init();
	}

	private void init() {
		setOpaque(false);

		flatLineContainer = Box.createVerticalBox();
		flatLineContainer.setVisible(false);
		switchOnButton = createSwitchOnButton();
		flatLineContainer.add(switchOnButton);
		unhideButton = createUnhideButton();
		flatLineContainer.add(unhideButton);
		moveLeftButtonOnFlatLine = createMoveLeftButton();
		flatLineContainer.add(moveLeftButtonOnFlatLine);
		moveRightButtonOnFlatLine = createMoveRightButton();
		flatLineContainer.add(moveRightButtonOnFlatLine);
		verticalTrackNameLabel = createVerticalTrackNameLabel();
		verticalTrackNameLabel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				audioTrack.setHidden(false);
			}
		});
		flatLineContainer.add(verticalTrackNameLabel);
		add(flatLineContainer);

		mainContainer = Box.createVerticalBox();
		add(mainContainer);
		add(CommonComponentFactory.createVerticalLine());

		mainContainer.add(createToolbar());
		mainContainer.add(new SourceAudioDeviceComponent(audioTrack));

		captureDeviceChainComponent = new AudioDeviceChainComponent(audioProject, audioTrack.getCaptureAudioDeviceChain(), true);
		captureDeviceChainComponent.setVisible(false);
		mainContainer.add(captureDeviceChainComponent);
		playbackDeviceChainComponent = new AudioDeviceChainComponent(audioProject, audioTrack.getPlaybackAudioDeviceChain(), true);
		mainContainer.add(playbackDeviceChainComponent);

		createTrackNameField();

		recordingLedComponent = createRecordingLed();
		mainContainer.add(recordingLedComponent);
		switchVisibleComponents();

		audioTrack.getAudioProject().addAudioProjectChangeListener(new AudioProjectAdapter() {
			@Override
			public void audioTrackAdded(AudioTrack track) {
				updateComponents();
			}

			@Override
			public void audioTrackRemoved(AudioTrack track) {
				updateComponents();
			}

			@Override
			public void audioTrackMoved(AudioTrack track) {
				updateComponents();
			}
		});
		audioTrack.addAudioTrackChangeListener(new AudioTrackAdapter() {
			@Override
			public void audioTrackHiddenChanged(AudioTrack audioTrack, boolean oldValue, boolean newValue) {
				updateStatus();
			}

			@Override
			public void audioTrackEnabledChanged(AudioTrack audioTrack, boolean oldValue, boolean newValue) {
				updateStatus();
			}

			@Override
			public void audioTrackNameChanged(AudioTrack audioTrack, String oldValue, String newValue) {
				verticalTrackNameLabel.repaint();
				trackNameField.setText(audioTrack.getName());
			}
			
			@Override
			public void audioTrackSourceAudioDeviceChanged(AudioTrack audioTrack, SourceAudioDevice oldValue, SourceAudioDevice newValue) {
				updateComponents();
			}
		});
		updateComponents();
		updateStatus();
	}

	private Component createToolbar() {
		Container toolbarBox = Box.createHorizontalBox();
		toolbarBox.add(Box.createHorizontalGlue());
		cloneButton = createCloneButton();
		toolbarBox.add(cloneButton);
		moveLeftButton = createMoveLeftButton();
		toolbarBox.add(moveLeftButton);
		moveRightButton = createMoveRightButton();
		toolbarBox.add(moveRightButton);
		hideButton = createHideButton();
		toolbarBox.add(hideButton);
		switchOffButton = createSwitchOffButton();
		toolbarBox.add(switchOffButton);
		removeButton = createRemoveButton();
		toolbarBox.add(removeButton);
		return toolbarBox;
	}

	private void updateComponents() {
		updateMoveLeftRightButtons();
		updateNumber();
	}

	private void updateStatus() {
		if (audioTrack.isEnabled()) {
			switchOnButton.setVisible(false);
			unhideButton.setVisible(true);
			if (audioTrack.isHidden()) {
				mainContainer.setVisible(false);
				flatLineContainer.setVisible(true);
			} else {
				mainContainer.setVisible(true);
				flatLineContainer.setVisible(false);
			}
		} else {
			switchOnButton.setVisible(true);
			mainContainer.setVisible(false);
			flatLineContainer.setVisible(true);
			unhideButton.setVisible(false);
		}
	}

	private Component createVerticalTrackNameLabel() {
		return new Component() {
			private static final long serialVersionUID = 1L;
			private static final double piHalf = PI / 2;
			private static final int distance = 2;

			@Override
			public void paint(Graphics g) {
				Graphics2D g2 = (Graphics2D) g;
				g2.rotate(piHalf);
				g2.setColor(getForeground());
				String trackName = audioTrack.getName();
				if (trackName != null) {
					g2.drawString(trackName, distance, -distance);
				}
			}

			@Override
			public Dimension getPreferredSize() {
				Graphics2D g2 = (Graphics2D) getGraphics();
				String trackName = audioTrack.getName();
				if(trackName==null) {
					return new Dimension(0, 0);
				}
				Rectangle2D stringBounds = g2.getFontMetrics().getStringBounds(trackName, g2);
				return new Dimension((int) stringBounds.getHeight() + distance, (int) stringBounds.getWidth() + distance);
			}

			@Override
			public Dimension getMaximumSize() {
				return new Dimension(getPreferredSize().width, Integer.MAX_VALUE);
			}
		};
	}

	private Component createSwitchOnButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/switchon-item.png");
		ret.setToolTipText("Switches this audio track on.");
		ret.setReleasedBorder(BorderFactory.createEmptyBorder());
		ret.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				audioTrack.setEnabled(true);
			}
		});
		return ret;
	}

	private ImageButtonComponent createSwitchOffButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/switchoff-item.png");
		ret.setToolTipText("Switches this audio track off.");
		ret.setReleasedBorder(BorderFactory.createEmptyBorder());
		ret.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				audioTrack.setEnabled(false);
			}
		});
		return ret;
	}

	private ImageButtonComponent createHideButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/pipe-item.png");
		ret.setToolTipText("Hides this audio track.");
		ret.setReleasedBorder(BorderFactory.createEmptyBorder());
		ret.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				audioTrack.setHidden(true);
			}
		});
		return ret;
	}

	private ImageButtonComponent createUnhideButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/duct-item.png");
		ret.setToolTipText("Unhides this audio track.");
		ret.setReleasedBorder(BorderFactory.createEmptyBorder());
		ret.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				audioTrack.setHidden(false);
			}
		});
		return ret;
	}

	private ImageButtonComponent createRemoveButton() {
		ImageButtonComponent removeButton = new ImageButtonComponent("org/audiochain/ui/gui/mixer/x-item.png");
		removeButton.setToolTipText("Closes and removes this audio track.");
		removeButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		removeButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				removeTrack();
			}
		});
		return removeButton;
	}

	private ImageButtonComponent createCloneButton() {
		ImageButtonComponent cloneButton = new ImageButtonComponent("org/audiochain/ui/gui/mixer/sheep-item.png");
		cloneButton.setToolTipText("Clones this audio track with all of it's audio devices.");
		cloneButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		cloneButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				try {
					audioTrack.getAudioProject().addNewAudioTrack(audioTrack);
				} catch (InstantiationException e) {
					Dialog.message(AudioTrackComponent.this, "Error While Cloning Audio Track", e);
				} catch (IllegalAccessException e) {
					Dialog.message(AudioTrackComponent.this, "Error While Cloning Audio Track", e);
				}
			}
		});
		return cloneButton;
	}

	private ImageButtonComponent createMoveLeftButton() {
		ImageButtonComponent moveLeftButton = new ImageButtonComponent("org/audiochain/ui/gui/mixer/triangleleft-item.png");
		moveLeftButton.setToolTipText("Moves this audio track to the left.");
		moveLeftButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		moveLeftButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				audioTrack.getAudioProject().decrementAudioTrackPositionIndex(audioTrack);
			}
		});
		return moveLeftButton;
	}

	private ImageButtonComponent createMoveRightButton() {
		ImageButtonComponent moveRightButton = new ImageButtonComponent("org/audiochain/ui/gui/mixer/triangleright-item.png");
		moveRightButton.setToolTipText("Moves this audio track to the right.");
		moveRightButton.setReleasedBorder(BorderFactory.createEmptyBorder());
		moveRightButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				audioTrack.getAudioProject().incrementAudioTrackPositionIndex(audioTrack);
			}
		});
		return moveRightButton;
	}

	private void updateMoveLeftRightButtons() {
		Collection<AudioTrack> audioTracks = audioTrack.getAudioProject().getAudioTracks();
		List<AudioTrack> list = new ArrayList<AudioTrack>(audioTracks);
		int lastTrackIndex = audioTracks.size() - 1;
		int index = list.indexOf(audioTrack);
		moveLeftButton.setVisible(index != 0);
		moveLeftButtonOnFlatLine.setVisible(index != 0);
		moveRightButton.setVisible(index != lastTrackIndex);
		moveRightButtonOnFlatLine.setVisible(index != lastTrackIndex);
	}

	private void createTrackNameField() {
		trackNameField = new JTextField(audioTrack.getName());
		trackNameField.setToolTipText("Name of this track.");
		trackNameField.setAlignmentX(CENTER_ALIGNMENT);
		trackNameField.setHorizontalAlignment(CENTER);
		Dimension dim = new Dimension(trackNameField.getPreferredSize());
		dim.width = Integer.MAX_VALUE;
		trackNameField.setMaximumSize(dim);
		trackNameField.setBorder(null);
		trackNameField.setOpaque(false);
		trackNameField.setFocusable(false);
		mainContainer.add(trackNameField);
		trackNameField.addKeyListener(new KeyAdapter() {
			@Override
			public void keyReleased(KeyEvent e) {
				if (e.getKeyCode() == 10) {
					trackNameField.setFocusable(false);
				}
				String text = trackNameField.getText();
				if (text.trim().isEmpty()) {
					text = null;
				}
				audioTrack.setName(text);
			}
		});
		trackNameField.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				trackNameField.setFocusable(true);
				trackNameField.grabFocus();
			}
		});
	}

	private LightEmittingDiodeComponent createRecordingLed() {
		final LightEmittingDiodeComponent recordingLedComponent = new LightEmittingDiodeComponent(68, 10);
		recordingLedComponent.setToolTipText("Selects this track for recording (type the number on your keyboard to select or unselect).");
		recordingLedComponent.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent evt) {
				try {
					audioTrack.toggleSelectForRecording();
				} catch (IOException e) {
					Dialog.message(this, "Select For Recording: IO Error", e);
				} catch (AudioIOException e) {
					Dialog.message(this, "Select For Recording: Audio Mixer Error", e);
				}
			}
		});
		AudioTrackAdapter listener = new AudioTrackAdapter() {
			@Override
			public void audioTrackSelectedForRecordingChanged(AudioTrack audioTrack, boolean oldValue, boolean newValue) {
				switchVisibleComponents();
			}

			@Override
			public void audioTrackSourceAudioDeviceChanged(AudioTrack audioTrack, SourceAudioDevice oldValue, SourceAudioDevice newValue) {
				SourceAudioDevice sourceAudioDevice = AudioTrackComponent.this.audioTrack.getSourceAudioDevice();
				if (sourceAudioDevice != null && sourceAudioDevice.isRecordable()) {
					recordingLedComponent.getLightEmittingDiode().setLedBaseColor(LightEmittingDiode.RED);
				} else {
					// make color to background color plus a bit brighter
					recordingLedComponent.getLightEmittingDiode().setLedBaseColor(GlobalGuiContext.getGuiSettings().getLightBackgroundColor());
					recordingLedComponent.setLedEnabled(false);
				}
			}
		};
		this.audioTrack.addAudioTrackChangeListener(listener);
		// initialize led color on startup
		listener.audioTrackSourceAudioDeviceChanged(null, null, null);
		return recordingLedComponent;
	}

	private void switchVisibleComponents() {
		boolean captureVisible = audioTrack.isSelectedForRecording();
		recordingLedComponent.setLedEnabled(captureVisible);
		playbackDeviceChainComponent.setVisible(!captureVisible);
		captureDeviceChainComponent.setVisible(captureVisible);
	}

	private void removeTrack() {
		if (okCancel(this, "Delete Track", "You are requesting to delete track '" + audioTrack.getName()
				+ "' from the project. This implies, that all audio data that is associated with the track, will also be deleted!")) {
			audioTrack.getAudioProject().removeAudioTrack(audioTrack);
		}
	}

	private void updateNumber() {
		recordingLedComponent.setText(String.valueOf(audioTrack.getNumber()));
		recordingLedComponent.repaint();
	}
}
