/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import static javax.swing.BorderFactory.*;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;

import org.audiochain.model.AudioDevice;
import org.audiochain.model.AudioProject;
import org.audiochain.model.ChainableAudioDevice;
import org.audiochain.model.ChainableAudioDeviceListener;
import org.audiochain.model.Monitorable;
import org.audiochain.model.PerformanceManager;
import org.audiochain.model.PerformanceMonitorListener;
import org.audiochain.model.UserInterfaceContext;
import org.audiochain.model.UserInterfaceContextListener;
import org.audiochain.ui.gui.CommonComponentFactory;
import org.audiochain.ui.gui.ImageButtonComponent;


public class ChainableAudioDeviceComponent extends Box {
	private static final long serialVersionUID = 1L;
	private final ChainableAudioDevice audioDevice;
	private final AudioProject audioProject;
	private Component removeButton;
	private Component hideButton;
	private Component switchOnButton;
	private Component switchOffButton;
	private Component moveUpButton;
	private Component moveDownButton;
	private Component monitorButton;
	private Component headLineBox;
	private Container mainContainer;
	
	public ChainableAudioDeviceComponent(AudioProject audioProject, ChainableAudioDevice audioDevice) {
		super(BoxLayout.Y_AXIS);
		this.audioProject = audioProject;
		this.audioDevice = audioDevice;
		init();
	}
	
	private void init() {
		headLineBox = createHeadLine();
		headLineBox.setVisible(false);
		add(headLineBox);
		mainContainer = Box.createHorizontalBox();
		mainContainer.add(Box.createHorizontalGlue());
		mainContainer.add(createDeviceComponent());
		mainContainer.add(Box.createHorizontalGlue());
		mainContainer.add(createToolbar());
		add(mainContainer);
		add(CommonComponentFactory.createHorizontalLine());
		audioDevice.getUserInterfaceContext().addUserInterfaceContextChangeListener(new UserInterfaceContextListener() {
			@Override
			public void hiddenChanged(boolean oldValue, boolean newValue) {
				updateStatus();
			}
		});
		audioDevice.addChainableAudioDeviceListener(new ChainableAudioDeviceListener() {
			@Override
			public void audioDeviceEnabledChanged(ChainableAudioDevice chainableAudioDevice, boolean oldValue, boolean newValue) {
				updateStatus();
			}
		});
		updateStatus();
	}

	private void updateStatus() {
		if(audioDevice.isEnabled()) {
			switchOnButton.setVisible(false);
			if(audioDevice.getUserInterfaceContext().isHidden()) {
				mainContainer.setVisible(false);
				headLineBox.setVisible(true);
			} else {
				mainContainer.setVisible(true);
				headLineBox.setVisible(false);
			}
		} else {
			switchOnButton.setVisible(true);
			mainContainer.setVisible(false);
			headLineBox.setVisible(true);
		}
	}

	private Component createHeadLine() {
		Container ret = Box.createHorizontalBox();
		ret.add(createLabel());
		ret.add(Box.createHorizontalGlue());
		switchOnButton = createSwitchOnButton();
		ret.add(switchOnButton);
		ret.addMouseListener(new MouseAdapter(){
			@Override
			public void mouseClicked(MouseEvent e) {
				if(mainContainer.isVisible()) {
					audioDevice.getUserInterfaceContext().setHidden(true);
				} else {
					audioDevice.getUserInterfaceContext().setHidden(false);
				}
			}
		});
		return ret;
	}

	private JLabel createLabel() {
		JLabel label = new JLabel(audioDevice.getName());
		label.setAlignmentX(CENTER_ALIGNMENT);
		return label;
	}

	private Component createDeviceComponent() {
		JComponent ret = Box.createVerticalBox();
		ret.setAlignmentY(TOP_ALIGNMENT);
		@SuppressWarnings("unchecked")
		
		UserInterfaceContext<ChainableAudioDevice> userInterfaceContext = (UserInterfaceContext<ChainableAudioDevice>) audioDevice.getUserInterfaceContext();
		Component audioDeviceComponent = userInterfaceContext.getComponent(audioDevice);
		ret.add(audioDeviceComponent);
		return ret;
	}

	private Component createToolbar() {
		JComponent ret = Box.createVerticalBox();
		ret.setAlignmentY(TOP_ALIGNMENT);
		
		hideButton = createHideButton();
		ret.add(hideButton);
		
		switchOffButton = createSwitchOffButton();
		ret.add(switchOffButton);
		
		removeButton = createRemoveButton();
		ret.add(removeButton);
		
		moveUpButton = createMoveUpButton();
		ret.add(moveUpButton);
		
		moveDownButton = createMoveDownButton();
		ret.add(moveDownButton);
		
		monitorButton = createMonitorButton();
		ret.add(monitorButton);
		
		return ret;
	}

	private static ImageButtonComponent createRemoveButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/x-item.png");
		ret.setToolTipText("Removes this audio device.");
		ret.setReleasedBorder(createEmptyBorder());
		return ret;
	}

	private Component createSwitchOnButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/switchon-item.png");
		ret.setToolTipText("Switches this audio device on.");
		ret.setReleasedBorder(createEmptyBorder());
		ret.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				audioDevice.setEnabled(true);
			}
		});
		return ret;
	}

	private ImageButtonComponent createSwitchOffButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/switchoff-item.png");
		ret.setToolTipText("Switches this audio device off.");
		ret.setReleasedBorder(createEmptyBorder());
		ret.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				audioDevice.setEnabled(false);
			}
		});
		return ret;
	}

	private ImageButtonComponent createHideButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/minus-item.png");
		ret.setToolTipText("Hides this audio device.");
		ret.setReleasedBorder(createEmptyBorder());
		ret.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				audioDevice.getUserInterfaceContext().setHidden(true);
			}
		});
		return ret;
	}

	private static ImageButtonComponent createMoveUpButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/triangleup-item.png");
		ret.setToolTipText("Locates this audio device before the previous audio audioDevice.");
		ret.setReleasedBorder(createEmptyBorder());
		return ret;
	}

	private static ImageButtonComponent createMoveDownButton() {
		ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/triangledown-item.png");
		ret.setToolTipText("Locates this audio device after the following audio audioDevice.");
		ret.setReleasedBorder(createEmptyBorder());
		return ret;
	}
	
	private ImageButtonComponent createMonitorButton() {
		final ImageButtonComponent ret = new ImageButtonComponent("org/audiochain/ui/gui/mixer/headphone-item.png");
		ret.setToolTipText("Listen to the output of this audio device.");
		ret.setReleasedBorder(createEmptyBorder());
		ret.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				final PerformanceManager performanceManager = audioProject.getPerformanceManager();
				if(audioDevice.equals(performanceManager.getMonitoredAudioDevice())) {
					performanceManager.setMonitoredAudioDevice(null);
				} else {
					performanceManager.addPerformanceMonitorListener(new PerformanceMonitorListener() {
						@Override
						public void performanceMonitorChanged(Monitorable oldMonitor, AudioDevice oldMonitoredAudioDevice, Monitorable newMonitor, AudioDevice newMonitoredAudioDevice) {
							if (audioDevice.equals(newMonitoredAudioDevice)) {
								ret.setBorder(createLineBorder(Color.ORANGE, 2));
							} else {
								ret.setBorder(createEmptyBorder());
								performanceManager.removePerformanceMonitorListener(this);
							}
						}
					});
					performanceManager.setMonitoredAudioDevice(audioDevice);
				}
			}
		});
		return ret;
	}

	public Component getRemoveButton() {
		return removeButton;
	}

	public Component getMoveUpButton() {
		return moveUpButton;
	}

	public Component getMoveDownButton() {
		return moveDownButton;
	}
}
