/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import static java.awt.event.KeyEvent.*;
import static org.audiochain.ui.gui.Dialog.*;
import static org.audiochain.ui.gui.GlobalGuiContext.*;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.URL;
import java.util.Collection;
import java.util.Enumeration;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Properties;

import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.KeyStroke;

import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectLifecycleListener;
import org.audiochain.model.PerformanceListener;
import org.audiochain.ui.AudioProjectModificationObserver;
import org.audiochain.ui.AudioProjectModificationObserverListener;
import org.audiochain.ui.gui.ComponentMoveMouseAdapter;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.gui.GuiSettings;
import org.audiochain.ui.sync.CommandSyncSocket;
import org.audiochain.ui.sync.SyncManager;
import org.audiochain.ui.sync.SynchronizationAdapter;

public class GuiAudioRecorderMenu extends JMenuBar implements AudioProjectLifecycleListener, AudioProjectModificationObserverListener, PerformanceListener {
	private static final long serialVersionUID = 1L;
	private final GuiAudioRecorder guiAudioRecorder;
	private JMenu projectMenu;
	private JMenuItem settingsMenuProjectSettingsItem;
	private JMenuItem settingsMenuAudioMixerSettingsItem;
	private JMenu settingsMenuSyncConnectionsMenu;
	private JMenuItem fileMenuSaveItem;
	private JMenuItem fileMenuSaveAsItem;
	private JMenuItem fileMenuMixdownItem;
	private JMenuItem fileMenuCloseProjectItem;
	private final Map<AudioProject, JMenuItem> audioProjectMenuItemMap;
	private final Map<AudioProject, AudioProjectModificationObserver> audioProjectModificationObserverMap;
	private final Map<JMenuItem, AudioProject> menuItemAudioProjectMap;
	private final Map<CommandSyncSocket, JMenuItem> syncConnectionMenuItemMap;

	public GuiAudioRecorderMenu(GuiAudioRecorder guiAudioRecorder) {
		this.guiAudioRecorder = guiAudioRecorder;
		audioProjectMenuItemMap = new LinkedHashMap<AudioProject, JMenuItem>();
		audioProjectModificationObserverMap = new LinkedHashMap<AudioProject, AudioProjectModificationObserver>();
		menuItemAudioProjectMap = new LinkedHashMap<JMenuItem, AudioProject>();
		syncConnectionMenuItemMap = new LinkedHashMap<CommandSyncSocket, JMenuItem>();
		init();
	}

	private void init() {
		guiAudioRecorder.getAudioProjectLifecycleManager().addAudioProjectLifecycleListener(this);
		MouseAdapter moveAdapter = new ComponentMoveMouseAdapter(guiAudioRecorder);
		addMouseListener(moveAdapter);
		addMouseMotionListener(moveAdapter);
		add(createFileMenu());
		projectMenu = new JMenu("Project");
		projectMenu.setMnemonic(VK_P);
		add(projectMenu);
		projectMenu.setEnabled(false);
		add(createSettingsMenu());
		add(createAboutMenu());
	}

	private JMenu createFileMenu() {
		JMenu fileMenu = new JMenu("File");
		fileMenu.setMnemonic(VK_F);

		JMenuItem fileMenuNewItem = new JMenuItem("New Project ...");
		fileMenuNewItem.setMnemonic(VK_N);
		fileMenuNewItem.setAccelerator(KeyStroke.getKeyStroke(VK_N, ActionEvent.CTRL_MASK));
		fileMenuNewItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				guiAudioRecorder.getAudioProjectLifecycleManager().newAudioProject();
			}
		});
		fileMenu.add(fileMenuNewItem);

		JMenuItem fileMenuOpenItem = new JMenuItem("Open Project ...");
		fileMenuOpenItem.setMnemonic(VK_O);
		fileMenuOpenItem.setAccelerator(KeyStroke.getKeyStroke(VK_O, ActionEvent.CTRL_MASK));
		fileMenuOpenItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				guiAudioRecorder.getAudioProjectLifecycleManager().openAudioProject();
			}
		});
		fileMenu.add(fileMenuOpenItem);

		fileMenuSaveItem = new JMenuItem("Save");
		fileMenuSaveItem.setMnemonic(VK_S);
		fileMenuSaveItem.setAccelerator(KeyStroke.getKeyStroke(VK_S, ActionEvent.CTRL_MASK));
		fileMenuSaveItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				storeProject();
			}
		});
		fileMenu.add(fileMenuSaveItem);
		fileMenuSaveItem.setEnabled(false);

		fileMenuSaveAsItem = new JMenuItem("Save As ...");
		fileMenuSaveAsItem.setMnemonic(VK_A);
		fileMenuSaveAsItem.setAccelerator(KeyStroke.getKeyStroke(VK_S, ActionEvent.ALT_MASK));
		fileMenuSaveAsItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					guiAudioRecorder.getAudioProjectLifecycleManager().storeActiveAudioProjectAs();
				} catch (IOException ex) {
					Dialog.message(GuiAudioRecorderMenu.this, "Error while saving project", ex);
				}
			}
		});
		fileMenu.add(fileMenuSaveAsItem);
		fileMenuSaveAsItem.setEnabled(false);

		fileMenuMixdownItem = new JMenuItem("Mixdown ...");
		fileMenuMixdownItem.setMnemonic(VK_M);
		fileMenuMixdownItem.setAccelerator(KeyStroke.getKeyStroke(VK_M, ActionEvent.CTRL_MASK));
		fileMenuMixdownItem.addActionListener(new ActionListener() {
			@Override
			@SuppressWarnings("unused")
			public void actionPerformed(ActionEvent e) {
				new MixdownDialog(this, guiAudioRecorder.getAudioProjectLifecycleManager().getActiveAudioProject());
			}
		});
		fileMenu.add(fileMenuMixdownItem);
		fileMenuMixdownItem.setEnabled(false);

		fileMenuCloseProjectItem = new JMenuItem("Close Project");
		fileMenuCloseProjectItem.setMnemonic(VK_C);
		fileMenuCloseProjectItem.setAccelerator(KeyStroke.getKeyStroke(VK_W, ActionEvent.CTRL_MASK));
		fileMenuCloseProjectItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					guiAudioRecorder.getAudioProjectLifecycleManager().closeActiveAudioProject();
				} catch (IOException ex) {
					Dialog.message(GuiAudioRecorderMenu.this, "Error while closing project", ex);
				}
			}
		});
		fileMenu.add(fileMenuCloseProjectItem);
		fileMenuCloseProjectItem.setEnabled(false);

		JMenuItem fileMenuQuitItem = new JMenuItem("Quit");
		fileMenuQuitItem.setMnemonic(VK_Q);
		fileMenuQuitItem.setAccelerator(KeyStroke.getKeyStroke(VK_Q, ActionEvent.CTRL_MASK));
		fileMenuQuitItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				guiAudioRecorder.exit();
			}
		});
		fileMenu.add(fileMenuQuitItem);

		return fileMenu;
	}

	private JMenu createSettingsMenu() {
		JMenu settingsMenu = new JMenu("Settings");
		settingsMenu.setMnemonic(VK_E);

		settingsMenuProjectSettingsItem = new JMenuItem("Project Settings");
		settingsMenuProjectSettingsItem.setMnemonic(VK_E);
		settingsMenuProjectSettingsItem.setAccelerator(KeyStroke.getKeyStroke(VK_E, ActionEvent.CTRL_MASK));
		settingsMenuProjectSettingsItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				guiAudioRecorder.getAudioProjectLifecycleManager().editActiveAudioProject();
			}
		});
		settingsMenu.add(settingsMenuProjectSettingsItem);
		settingsMenuProjectSettingsItem.setEnabled(false);

		settingsMenuAudioMixerSettingsItem = new JMenuItem("Audio Mixer Settings");
		settingsMenuAudioMixerSettingsItem.setMnemonic(VK_M);
		settingsMenuAudioMixerSettingsItem.addActionListener(new ActionListener() {
			@Override
			@SuppressWarnings("unused")
			public void actionPerformed(ActionEvent e) {
				new MixerSettingsDialog(GuiAudioRecorderMenu.this);
			}
		});
		settingsMenu.add(settingsMenuAudioMixerSettingsItem);

		JMenuItem settingsMenuResetQuestionDialogCheckboxesItem = new JMenuItem("Reset Question Dialog Checkboxes");
		settingsMenuResetQuestionDialogCheckboxesItem.setMnemonic(VK_Q);
		settingsMenuResetQuestionDialogCheckboxesItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				GuiSettings guiSettings = getGuiSettings();
				guiSettings.getDialogResults().clear();
				guiSettings.store();
			}
		});
		settingsMenu.add(settingsMenuResetQuestionDialogCheckboxesItem);

		JMenuItem settingsMenuRemoteSynchronizationSettingsItem = new JMenuItem("Remote Synchronization Settings");
		settingsMenuRemoteSynchronizationSettingsItem.setMnemonic(VK_Y);
		settingsMenuRemoteSynchronizationSettingsItem.addActionListener(new ActionListener() {
			@SuppressWarnings("unused")
			@Override
			public void actionPerformed(ActionEvent e) {
				new SynchronizationSettingsDialog(GuiAudioRecorderMenu.this);
			}
		});
		settingsMenu.add(settingsMenuRemoteSynchronizationSettingsItem);

		JMenuItem settingsMenuCreateSyncConnectionItem = new JMenuItem("Create Sync Connection ...");
		settingsMenuCreateSyncConnectionItem.setMnemonic(VK_R);
		settingsMenuCreateSyncConnectionItem.setAccelerator(KeyStroke.getKeyStroke(VK_R, ActionEvent.CTRL_MASK));
		settingsMenuCreateSyncConnectionItem.addActionListener(new ActionListener() {
			@SuppressWarnings("unused")
			@Override
			public void actionPerformed(ActionEvent e) {
				new CreateSynchronizationConnectionDialog(GuiAudioRecorderMenu.this);
			}
		});
		settingsMenu.add(settingsMenuCreateSyncConnectionItem);

		settingsMenuSyncConnectionsMenu = new JMenu("Sync Connections");
		settingsMenuSyncConnectionsMenu.setMnemonic(VK_C);
		settingsMenu.add(settingsMenuSyncConnectionsMenu);
		SyncManager.getInstance().addSynchronizationListener(new SynchronizationAdapter() {
			@Override
			public void syncSocketDisconnected(CommandSyncSocket commandSyncSocket) {
				updateSyncConnectionMenu();
			}

			@Override
			public void syncSocketConnected(CommandSyncSocket commandSyncSocket) {
				updateSyncConnectionMenu();
			}
		});
		settingsMenuSyncConnectionsMenu.setEnabled(false);

		return settingsMenu;
	}

	private void updateSyncConnectionMenu() {
		Collection<CommandSyncSocket> allCommandSyncSockets = SyncManager.getInstance().getAllCommandSyncSockets();
		for (final CommandSyncSocket commandSyncSocket : allCommandSyncSockets) {
			if (!syncConnectionMenuItemMap.containsKey(commandSyncSocket)) {
				Socket socket = commandSyncSocket.getSocket();
				InetAddress inetAddress = socket.getInetAddress();
				final String hostName = inetAddress.getHostName();
				final int port = socket.getPort();
				JMenu connectionMenu = new JMenu(hostName + ":" + port);
				syncConnectionMenuItemMap.put(commandSyncSocket, connectionMenu);
				settingsMenuSyncConnectionsMenu.add(connectionMenu);
				JMenuItem disconnectMenuItem = new JMenuItem("Close Connection");
				disconnectMenuItem.setMnemonic(VK_C);
				connectionMenu.add(disconnectMenuItem);
				disconnectMenuItem.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent e) {
						if (Dialog.okCancel(GuiAudioRecorderMenu.this, "Close Sync Connection", "The synchronization connection to '" + hostName + "' at port " + port + " will be closed now.")) {
							commandSyncSocket.exit();
						}
					}
				});
				JMenuItem listProjectsMenuItem = new JMenuItem("Open Remote Project ...");
				listProjectsMenuItem.setMnemonic(VK_O);
				connectionMenu.add(listProjectsMenuItem);
				listProjectsMenuItem.addActionListener(new ActionListener() {
					@SuppressWarnings("unused")
					@Override
					public void actionPerformed(ActionEvent e) {
						new OpenRemoteAudioProjectDialog(GuiAudioRecorderMenu.this, commandSyncSocket, guiAudioRecorder.getAudioProjectLifecycleManager());
					}
				});
			}
		}
		Collection<CommandSyncSocket> toRemove = new LinkedHashSet<CommandSyncSocket>();
		for (CommandSyncSocket commandSyncSocket : syncConnectionMenuItemMap.keySet()) {
			if (!allCommandSyncSockets.contains(commandSyncSocket)) {
				toRemove.add(commandSyncSocket);
			}
		}
		for (CommandSyncSocket commandSyncSocket : toRemove) {
			JMenuItem menuItem = syncConnectionMenuItemMap.remove(commandSyncSocket);
			settingsMenuSyncConnectionsMenu.remove(menuItem);
		}
		settingsMenuSyncConnectionsMenu.setEnabled(!syncConnectionMenuItemMap.isEmpty());
	}

	private JMenu createAboutMenu() {
		JMenu aboutMenu = new JMenu("About");
		JMenuItem licenseMenuItem = new JMenuItem("License information");
		licenseMenuItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				try {
					showTextFromFileDialog(GuiAudioRecorderMenu.this, "License information", "/license.txt", 15, 50, "Close");
				} catch (IOException ex) {
					throw new IllegalStateException(ex);
				}
			}
		});
		aboutMenu.add(licenseMenuItem);

		JMenuItem versionMenuItem = new JMenuItem("Version information");
		versionMenuItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				String linesep = System.getProperty("line.separator");
				StringBuilder sb = new StringBuilder();
				try {
					Enumeration<URL> manifests = getClass().getClassLoader().getResources("META-INF/MANIFEST.MF");
					while (manifests.hasMoreElements()) {
						URL manifest = manifests.nextElement();
						InputStream is = manifest.openStream();
						Properties props = new Properties();
						props.load(is);
						// referenced / set in ant build
						String softwareVersion = props.getProperty("Software-Version");
						if (softwareVersion != null) {
							sb.append(softwareVersion);
						}
						String gitCommit = props.getProperty("Git-Commit");
						if (gitCommit != null) {
							sb.append(linesep).append("Git-Commit ").append(gitCommit);
						}
					}
				} catch (IOException ex) {
					throw new IllegalStateException(ex);
				}
				showTextDialog(GuiAudioRecorderMenu.this, "Version information", sb.toString(), 2, 15, "Close");
			}
		});
		aboutMenu.add(versionMenuItem);

		return aboutMenu;
	}

	@Override
	public void audioProjectOpened(final AudioProject audioProject) {
		final JMenuItem projectMenuItem = new JMenuItem() {
			private static final long serialVersionUID = 1L;

			@Override
			public String getText() {
				AudioProjectModificationObserver observer = audioProjectModificationObserverMap.get(audioProject);
				if (observer != null) {
					StringBuilder sb = new StringBuilder();
					int mnemonic = getMnemonic();
					if (mnemonic != 0) {
						sb.append((char) mnemonic);
						sb.append(' ');
					}
					if (observer.isModified()) {
						sb.append('*');
					}
					sb.append(audioProject.getUniqueName());
					return sb.toString();
				} else {
					return audioProject.getUniqueName();
				}
			}
		};
		projectMenuItem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				guiAudioRecorder.getAudioProjectLifecycleManager().activateAudioProject(menuItemAudioProjectMap.get(projectMenuItem));
			}
		});

		audioProjectMenuItemMap.put(audioProject, projectMenuItem);
		menuItemAudioProjectMap.put(projectMenuItem, audioProject);
		projectMenu.add(projectMenuItem);
		projectMenu.setEnabled(true);
		AudioProjectModificationObserver observer = new AudioProjectModificationObserver(audioProject);
		audioProjectModificationObserverMap.put(audioProject, observer);
		observer.addAudioProjectModificationObserverListener(this);
		audioProject.getPerformanceManager().addPerformanceListener(this);
		updateProjectMenuItemMnemonics();
	}

	@Override
	public void audioProjectClosed(AudioProject audioProject) {
		JMenuItem menuItem = audioProjectMenuItemMap.remove(audioProject);
		menuItemAudioProjectMap.remove(menuItem);
		projectMenu.remove(menuItem);
		projectMenu.setEnabled(!audioProjectMenuItemMap.isEmpty());
		AudioProjectModificationObserver observer = audioProjectModificationObserverMap.get(audioProject);
		observer.removeAudioProjectModificationObserverListener(this);
		audioProjectModificationObserverMap.remove(audioProject);
		audioProject.getPerformanceManager().removePerformanceListener(this);
		updateProjectMenuItemMnemonics();
	}

	private void updateProjectMenuItemMnemonics() {
		int count = 0;
		for (JMenuItem item : menuItemAudioProjectMap.keySet()) {
			count++;
			int mnemonic = VK_0 + count;
			item.setMnemonic(mnemonic);
			if (count == 10) {
				item.setMnemonic(VK_0);
			} else if (count > 10) {
				item.setMnemonic(0);
			}
		}
	}

	@Override
	public void audioProjectActivated(AudioProject audioProject) {
		settingsMenuProjectSettingsItem.setEnabled(audioProject != null);
		AudioProjectModificationObserver observer = audioProjectModificationObserverMap.get(audioProject);
		fileMenuSaveItem.setEnabled(audioProject != null && observer.isModified());
		fileMenuSaveAsItem.setEnabled(audioProject != null);
		fileMenuMixdownItem.setEnabled(audioProject != null);
		fileMenuCloseProjectItem.setEnabled(audioProject != null);
	}

	@Override
	public void audioProjectStored(AudioProject audioProject) {
		AudioProjectModificationObserver observer = audioProjectModificationObserverMap.get(audioProject);
		fileMenuSaveItem.setEnabled(observer.isModified());
	}

	@Override
	public void audioProjectModified(AudioProject audioProject) {
		AudioProjectModificationObserver observer = audioProjectModificationObserverMap.get(audioProject);
		fileMenuSaveItem.setEnabled(observer.isModified());
	}

	private void storeProject() {
		try {
			guiAudioRecorder.getAudioProjectLifecycleManager().storeActiveAudioProject();
		} catch (IOException ex) {
			Dialog.message(this, "Error while saving project", ex);
		}
	}

	@Override
	public void performanceStarted() {
		fileMenuMixdownItem.setEnabled(false);
		settingsMenuAudioMixerSettingsItem.setEnabled(false);
	}

	@Override
	public void performanceUpdated() {

	}

	@Override
	public void performanceStopped() {

	}

	@Override
	public void performanceFinished() {
		fileMenuMixdownItem.setEnabled(true);
		settingsMenuAudioMixerSettingsItem.setEnabled(true);
	}
}
