/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import static javax.swing.SwingConstants.*;
import static org.audiochain.model.ObjectStore.*;
import static org.audiochain.ui.ProjectMixdown.*;
import static org.audiochain.ui.gui.GlobalGuiContext.*;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.GroupLayout;
import javax.swing.GroupLayout.Alignment;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JTextField;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.filechooser.FileFilter;

import org.audiochain.model.AudioProject;
import org.audiochain.ui.ProjectMixdown;
import org.audiochain.ui.ProjectMixdownListener;
import org.audiochain.ui.ProjectMixdown.AudioFileFormatType;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.gui.GlobalGuiContext;
import org.audiochain.ui.gui.GuiSettings;
import org.audiochain.ui.gui.ImageButtonComponent;
import org.audiochain.ui.gui.NotificationDialog;


public class MixdownDialog extends NotificationDialog {
	private final AudioProject audioProject;
	private final ProjectMixdown projectMixdown;
	private final FileFilter fileFilter;
	private JPanel mainPanel;
	private JLabel fileNameLabel;
	private JTextField fileNameField;
	private JComponent fileNameSearchButton;
	private JLabel startLabel;
	private JTextField startTimeField;
	private JLabel startTimeLabel;
	private JTextField  startFramesField;
	private JLabel startFramesLabel;
	private JLabel endLabel;
	private JTextField endTimeField;
	private JLabel endTimeLabel;
	private JTextField endFramesField;
	private JLabel endFramesLabel;
	private JButton closeButton;
	private JButton cancelButton;
	private JProgressBar progressBar;
	private Dialog dialog;
	private boolean pushStartTimeToFramesField;
	private boolean pushEndTimeToFramesField;
	
	public MixdownDialog(Object caller, AudioProject audioProject) {
		super(caller);
		this.audioProject = audioProject;
		this.projectMixdown = new ProjectMixdown(audioProject);
		this.fileFilter = createFileFilter();
		init();
	}

	protected void init() {
		initComponents();
		initValues();
		registerListeners();
		createLayout();
		dialog = new Dialog(caller, "Create Project Mixdown", applyButton, closeButton, applyButton, cancelButton, closeButton){
			@Override
			protected JComponent createContentComponent() {
				return mainPanel;
			}
		};
		dialog.setCloseManually(true);
		dialog.setHideShowMessageAgainCheckbox(true);
		dialog.open();
	}

	private void initComponents() {
		mainPanel = new JPanel();
		mainPanel.setOpaque(false);
		applyButton = new JButton("Create Mixdown");
		closeButton = new JButton("Close");
		cancelButton = new JButton("Cancel");
		cancelButton.setEnabled(false);
		fileNameLabel = new JLabel("Mixdown File:");
		fileNameField = new JTextField(30);
		fileNameSearchButton = new ImageButtonComponent("org/audiochain/ui/gui/open-directory.png");
		
		startLabel = new JLabel("Start Position:");
		startTimeField = new JTextField(5);
		startTimeField.setName("Start Position Time");
		startTimeLabel = new JLabel("mm:ss");
		startFramesField = new JTextField(8);
		startFramesLabel = new JLabel("frames");
		
		endLabel = new JLabel("End Position:");
		endTimeField = new JTextField(5);
		endTimeField.setName("End Position Time");
		endTimeLabel = new JLabel("mm:ss");
		endFramesField = new JTextField(8);
		endFramesLabel = new JLabel("frames");
		progressBar = new JProgressBar(HORIZONTAL);
		progressBar.setVisible(false);
		progressBar.setStringPainted(true);
	}

	private void initValues() {
		fileNameField.setText(projectMixdown.createDefaultMixdownFile(getGuiSettings().getLastMixdownAudioFileFormatType()).getAbsolutePath());
		startTimeField.setText(formatTime(0));
		startFramesField.setText(String.valueOf(0));
		endTimeField.setText(formatTime(audioProject.getSeconds()));
		endFramesField.setText(String.valueOf(audioProject.getMaxFrame()));
	}

	private void registerListeners() {
		closeButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				projectMixdown.cancel();
				dialog.close();
			}
		});
		cancelButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				projectMixdown.cancel();
			}
		});
		applyButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				createMixdown();
			}
		});
		fileNameSearchButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				String fileName = GlobalGuiContext.getGuiSettings().getLastOpenedFileName();
				JFileChooser jfc = fileName != null ? new JFileChooser(new File(fileName)) : new JFileChooser();
				jfc.setFileFilter(fileFilter);
				int status = jfc.showOpenDialog(null);
				if (status == JFileChooser.APPROVE_OPTION) {
					File path = jfc.getSelectedFile();
					fileNameField.setText(path.getAbsolutePath());
				}
			}
		});
		projectMixdown.addProjectMixdownListener(new ProjectMixdownListener() {
			
			private long startFrame;
			private long predictedEndFrame;
			
			@Override
			public void mixdownStarted(File file, AudioFileFormatType type, long startFrame, long predictedEndFrame) {
				this.startFrame = startFrame;
				this.predictedEndFrame = predictedEndFrame;
				progressBar.setMinimum((int)(startFrame));
				progressBar.setMaximum((int)predictedEndFrame);
				progressBar.setVisible(true);
				applyButton.setEnabled(false);
				cancelButton.setEnabled(true);
			}
			
			@Override
			public void mixdownUpdated(long currentFrame) {
				progressBar.setString("Creating Mixdown "+(100 * (currentFrame - startFrame) / (predictedEndFrame - startFrame))+"%");
				progressBar.setValue((int)currentFrame);
			}
			
			@Override
			public void mixdownFinished(long endFrame) {
				progressBar.setString("Mixdown Finished");
				applyButton.setEnabled(true);
				cancelButton.setEnabled(false);
			}
		});
		fileNameField.getDocument().addDocumentListener(new DocumentListener() {
			@Override
			public void removeUpdate(DocumentEvent e) {
				validateFileName();
			}
			
			@Override
			public void insertUpdate(DocumentEvent e) {
				validateFileName();
			}
			
			@Override
			public void changedUpdate(DocumentEvent e) {
				validateFileName();				
			}
		});
		startFramesField.getDocument().addDocumentListener(new DocumentListener() {
			@Override
			public void insertUpdate(DocumentEvent e) {
				evaluateStartFramesField();
			}

			@Override
			public void removeUpdate(DocumentEvent e) {
				evaluateStartFramesField();
			}

			@Override
			public void changedUpdate(DocumentEvent e) {
				evaluateStartFramesField();
			}
		});
		endFramesField.getDocument().addDocumentListener(new DocumentListener() {
			@Override
			public void insertUpdate(DocumentEvent e) {
				evaluateEndFramesField();
			}

			@Override
			public void removeUpdate(DocumentEvent e) {
				evaluateEndFramesField();
			}

			@Override
			public void changedUpdate(DocumentEvent e) {
				evaluateEndFramesField();
			}
		});
		DocumentListener timeFieldsValidator = new DocumentListener() {
			
			@Override
			public void removeUpdate(DocumentEvent e) {
				validatePositionTimeFields();
			}
			
			@Override
			public void insertUpdate(DocumentEvent e) {
				validatePositionTimeFields();
			}
			
			@Override
			public void changedUpdate(DocumentEvent e) {
				validatePositionTimeFields();				
			}
		};
		startTimeField.getDocument().addDocumentListener(timeFieldsValidator);
		endTimeField.getDocument().addDocumentListener(timeFieldsValidator);

		startTimeField.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				pushStartTimeToFramesField = true;
			}
		});
		endTimeField.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				pushEndTimeToFramesField = true;
			}
		});
		startFramesField.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				pushStartTimeToFramesField = false;
			}
		});
		endFramesField.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				pushEndTimeToFramesField = false;
			}
		});
	}
	
	private void createLayout() {
		GroupLayout layout = new GroupLayout(mainPanel);
		mainPanel.setLayout(layout);
		layout.setAutoCreateContainerGaps(true);
		layout.setAutoCreateGaps(true);
		layout.setHonorsVisibility(progressBar, false);
		
		layout.setHorizontalGroup(layout.createParallelGroup()
				.addComponent(notificationLabel)
				.addGroup(layout.createSequentialGroup()
					.addGroup(layout.createParallelGroup()
						.addComponent(fileNameLabel)
						.addComponent(startLabel)
						.addComponent(endLabel)
					)
					.addGroup(layout.createParallelGroup()
						.addGroup(layout.createSequentialGroup()
							.addComponent(fileNameField)
							.addComponent(fileNameSearchButton)
						)
						.addGroup(layout.createSequentialGroup()
							.addComponent(startTimeField)
							.addComponent(startTimeLabel)
							.addComponent(startFramesField)
							.addComponent(startFramesLabel)
						)
						.addGroup(layout.createSequentialGroup()
							.addComponent(endTimeField)
							.addComponent(endTimeLabel)
							.addComponent(endFramesField)
							.addComponent(endFramesLabel)
						)
					)
				)
				.addComponent(progressBar)
			);
			
			layout.setVerticalGroup(layout.createSequentialGroup()
				.addComponent(notificationLabel)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(fileNameLabel)
					.addComponent(fileNameField)
					.addComponent(fileNameSearchButton)
				)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(startLabel)
					.addComponent(startTimeField)
					.addComponent(startTimeLabel)
					.addComponent(startFramesField)
					.addComponent(startFramesLabel)
				)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(endLabel)
					.addComponent(endTimeField)
					.addComponent(endTimeLabel)
					.addComponent(endFramesField)
					.addComponent(endFramesLabel)
				)
				.addComponent(progressBar)
			);
	}
	
	private void createMixdown() {
		new Thread() {
			@Override
			public void run() {
				File file = new File(fileNameField.getText());
				try {
					if(fileFilter.accept(file)) {
						AudioFileFormatType type = ProjectMixdown.AudioFileFormatType.valueOf(file);
						long start = getStartFrame();
						long end = getEndFrame();
						projectMixdown.createMixdown(file, type, start, end);
						GuiSettings guiSettings = getGuiSettings();
						guiSettings.setLastMixdownAudioFileFormatType(type);
						guiSettings.store();
					}
				} catch (Throwable ex) {
					if(!file.delete()) {
						file.deleteOnExit();
					}
					ex.printStackTrace();
					Dialog.message(caller, "Error while creating project mixdown", ex);
				}
			}
		}.start();
	}
	
	private void validateFileName() {
		putNotification(fileNameField, null);
		File file = new File(fileNameField.getText());
		if (!file.exists() && !hasAnyWritableParentDirectory(file)) {
			putNotification(fileNameField, "The file can not be created because there is no writable parent directory.");
		} else if(file.isDirectory()) {
			putNotification(fileNameField, "The file name points to a directory.");
		} else if(!fileFilter.accept(file)) {
			putNotification(fileNameField, "The file name must be a type of "+fileFilter.getDescription());
		}
	}
	
	private Long getStartFrame() {
		try {
			return Long.parseLong(startFramesField.getText());
		} catch (NumberFormatException e) {
			putNotification(startFramesField, "The start frame must be a numeric (integral) value greater or equal to zero.");
			return null;
		}
	}
	
	private Long getEndFrame() {
		try {
			return Long.parseLong(endFramesField.getText());
		} catch (NumberFormatException e) {
			putNotification(endFramesField, "The end frame must be a numeric (integral) value greater than zero.");
			return null;
		}
	}

	private void evaluateStartFramesField() {
		if(pushStartTimeToFramesField) {
			putNotification(startFramesField, null);
		} else {
			validatePositionFramesFields();
		}
	}

	private void evaluateEndFramesField() {
		if(pushEndTimeToFramesField) {
			putNotification(endFramesField, null);
		} else {
			validatePositionFramesFields();
		}
	}
	
	private void validatePositionFramesFields() {
		putNotification(startFramesField, null);
		putNotification(endFramesField, null);
		
		Long startPositionFrames = getStartFrame();
		Long endPositionFrames = getEndFrame();
		final long maxFrame = audioProject.getMaxFrame();
		if (startPositionFrames!=null && maxFrame > 0 && startPositionFrames >= maxFrame) {
			putNotification(startFramesField, "The start frame must be lower than "+maxFrame+".");
		}
		if(startPositionFrames==null || startPositionFrames<0) {
			putNotification(startFramesField, "The start frame must be greater or equal to zero.");
		}
		
		if (endPositionFrames!=null && maxFrame > 0 && endPositionFrames > maxFrame) {
			putNotification(endFramesField, "The end frame must be lower or equal to "+maxFrame+".");
		}
		if(endPositionFrames==null || endPositionFrames<=0) {
			putNotification(endFramesField, "The end frame must be greater than zero.");
		}
		
		if(startPositionFrames!=null && endPositionFrames!=null && startPositionFrames>=endPositionFrames) {
			putNotification(startFramesField, "The start frame must be lower than the end frame.");
		}
		
		boolean hasErrors = hasNotification(startFramesField) || hasNotification(endFramesField);
		if(!hasErrors) {
			if(!pushStartTimeToFramesField && startPositionFrames!=null) {
				startTimeField.setText(formatTime(audioProject.toSeconds(startPositionFrames)));
			}
			if(!pushEndTimeToFramesField && endPositionFrames!=null) {
				endTimeField.setText(formatTime(audioProject.toSeconds(endPositionFrames)));
			}
		}
	}
	
	private void validatePositionTimeFields() {
		putNotification(startTimeField, null);
		Long startSeconds = parseTime(startTimeField.getText().trim());
		if (startSeconds == null) {
			putNotification(startTimeField, "The start time must be in format mm:ss (minutes and seconds)");
		}
		putNotification(endTimeField, null);
		Long endSeconds = parseTime(endTimeField.getText().trim());
		if (endSeconds == null) {
			putNotification(endTimeField, "The end time must be in format mm:ss (minutes and seconds)");
		}
		
		final long maxSeconds = audioProject.getSeconds();
		if(startSeconds!=null && startSeconds<0) {
			putNotification(startTimeField, "The start time must be greater or equal to "+formatTime(0)+".");
		}
		if(startSeconds!=null && startSeconds>=maxSeconds) {
			putNotification(startTimeField, "The start time must be lower than "+formatTime(maxSeconds)+".");
		}
		
		if(endSeconds!=null && endSeconds<=0) {
			putNotification(endTimeField, "The end time must be greater than "+formatTime(0)+".");
		}
		
		if(endSeconds!=null && endSeconds>maxSeconds) {
			putNotification(endTimeField, "The end time must be lower or equal to "+formatTime(maxSeconds)+".");
		}
		
		if(startSeconds!=null && endSeconds!=null && startSeconds>=endSeconds) {
			putNotification(startTimeField, "The start time must be lower than the end time.");
		}
		
		boolean hasErrors = hasNotification(startTimeField) || hasNotification(endTimeField);
		if(!hasErrors) {
			if(pushStartTimeToFramesField && startSeconds!=null) {
				startFramesField.setText(String.valueOf(audioProject.toFrames(startSeconds)));
			}
			if(pushEndTimeToFramesField && endSeconds!=null) {
				endFramesField.setText(String.valueOf(audioProject.toFrames(endSeconds)));
			}
		}
	}

	private static String formatTime(long seconds) {
		long min = seconds / 60L;
		long sec = seconds % 60L;
		return (min < 10 ? "0" : "") + min + ':' + (sec < 10 ? "0" : "") + sec;
	}

	private static Long parseTime(String time) {
		Pattern p = Pattern.compile("^(\\d\\d):(\\d\\d)$");
		Matcher m = p.matcher(time);
		if(m.matches()) {
			long min = Long.parseLong(m.group(1));
			long sec = Long.parseLong(m.group(2));
			if(sec>=60) {
				return null;
			}
			long ret = min * 60L + sec;
			return ret;
		} else {
			return null;
		}
	}

	@SuppressWarnings("unused")
	public static void main(String[] args) throws Exception {
		GlobalGuiContext.getGuiSettings().initializeUIManager();
		new MixdownDialog(new Object(), AudioProject.load(new File("testproject/testproject.audio")));
		System.exit(0);
	}
}
