/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import static org.audiochain.ui.gui.CommonComponentFactory.*;
import static org.audiochain.ui.gui.Dialog.*;
import static org.audiochain.ui.gui.GlobalGuiContext.*;

import java.io.File;
import java.net.SocketTimeoutException;
import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Vector;

import javax.swing.GroupLayout;
import javax.swing.GroupLayout.Alignment;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.audiochain.model.AudioProject;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.sync.CommandSyncSocket;


public class OpenRemoteAudioProjectDialog extends CreateNewAudioProjectDialog {
	private final CommandSyncSocket commandSyncSocket;
	private final AudioProjectLifecycleManager audioProjectLifecycleManager;
	private Vector<String> identifierList;
	private Vector<String> nameList;
	private JLabel projectListLabel;
	private JList projectList;
	private JScrollPane projectListScroll;

	public OpenRemoteAudioProjectDialog(Object caller, CommandSyncSocket commandSyncSocket, AudioProjectLifecycleManager audioProjectLifecycleManager) {
		super(caller);
		this.commandSyncSocket = commandSyncSocket;
		this.audioProjectLifecycleManager = audioProjectLifecycleManager;
		initGui();
	}
	
	@Override
	protected void init() {
		// nop, prevent super from initing anything
	}

	private void initGui() {
		createMainComponents();
		createApplyButton("Open Remote Audio Project");
		createProjectListComponents();
		createProjectPathComponents("Local Project Path:");
		createLayout();
		initValues();

		if (isButtonPressed(caller, "Open Remote Audio Project", mainPanel, applyButton)) {
			String projectPath = projectPathField.getText();
			if (projectPath == null || projectPath.trim().isEmpty()) {
				throw new IllegalStateException("The audio project path must not be null.");
			}
			File projectPathFile = new File(projectPath);
			if (!projectPathFile.exists()) {
				if (!projectPathFile.mkdirs()) {
					Dialog.message(caller, "Error while creating new Project Directory", "The project directory '" + projectPath + "' can not be created.");
					return;
				}
			}
			getGuiSettings().setLastNewProjectPathName(projectPath);
			
			int selectedIndex = projectList.getSelectedIndex();
			if (selectedIndex >= 0) {
				String identifier = identifierList.get(selectedIndex);
				try {
					AudioProject remoteAudioProject = commandSyncSocket.getRemoteAudioProject(identifier, projectPathFile.getAbsolutePath());
					audioProjectLifecycleManager.openAudioProject(remoteAudioProject);
				} catch (SocketTimeoutException ex) {
					Dialog.message(this, "Timeout While Getting Remote Audio Project", ex);
				}
			}
		}
	}
	
	private void initValues() {
		new Thread() {
			@Override
			public void run() {
				putNotification(projectList, "Loading remote projects ...");
				try {
					identifierList = new Vector<String>();
					Map<String, String> remoteAudioProjectIdentifierToNameMap = commandSyncSocket.listRemoteAudioProjects();
					Collection<AudioProject> localAudioProjects = audioProjectLifecycleManager.getAudioProjects();
					// remove all local project ids
					for (AudioProject audioProject : localAudioProjects) {
						remoteAudioProjectIdentifierToNameMap.remove(audioProject.getIdentifier());
					}
					for(Entry<String,String> entry : remoteAudioProjectIdentifierToNameMap.entrySet()) {
						String identifier = entry.getKey();
						String name = entry.getValue();
						String item = (name == null ? "untitled" : name) + " (" + identifier + ")";
						nameList.add(item);
						identifierList.add(identifier);
					}
					projectList.setListData(nameList);
					putNotification(projectList, null);
				} catch (SocketTimeoutException ex) {
					putNotification(projectList, ex.getMessage());
				}
			}
		}.start();
	}
	
	protected void createProjectListComponents() {
		projectListLabel = new JLabel("Available Projects:");
		nameList = new Vector<String>();
		projectList = new JList();
		projectList.setListData(nameList);
		projectList.setVisibleRowCount(5);
		projectListScroll = createScrollPane(projectList);
		projectList.addListSelectionListener(new ListSelectionListener() {
			@Override
			public void valueChanged(ListSelectionEvent e) {
				if (projectList.getSelectedIndex() >= 0) {
					putNotification(projectList, null);
				} else {
					if (nameList.isEmpty()) {
						putNotification(projectList, "No remote audio projects available.");
					} else {
						putNotification(projectList, "Please select an audio project!");
					}
				}
			}
		});
	}

	private void createLayout() {
		GroupLayout layout = new GroupLayout(mainPanel);
		mainPanel.setLayout(layout);
		layout.setAutoCreateContainerGaps(true);
		layout.setAutoCreateGaps(true);
		
		layout.setHorizontalGroup(layout.createParallelGroup()
			.addComponent(notificationLabel)
			.addGroup(layout.createSequentialGroup()
				.addGroup(layout.createParallelGroup()
					.addComponent(projectListLabel)
					.addComponent(projectPathLabel)
				)
				.addGroup(layout.createParallelGroup()
					.addComponent(projectListScroll)
					.addGroup(layout.createSequentialGroup()
						.addComponent(projectPathField)
						.addComponent(searchProjectPathButton)
					)
				)
			)
		);
		
		layout.setVerticalGroup(layout.createSequentialGroup()
			.addComponent(notificationLabel)
			.addGroup(layout.createSequentialGroup()
				.addGroup(layout.createParallelGroup(Alignment.CENTER)
					.addComponent(projectListLabel)
					.addComponent(projectListScroll)
				)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(projectPathLabel)
					.addComponent(projectPathField)
					.addComponent(searchProjectPathButton)
				)
			)
			// add additional gap to the dialog buttons, because the notification label causes an additional distance to the dialog title on the top
			.addGap(20)
		);
	}

	@Override
	protected boolean mandatoryFieldsValid() {
		return checkProjectPath() == null && projectList.getSelectedIndex() >= 0;
	}
}
