/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import static java.lang.Math.*;

import java.awt.Font;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.util.Collection;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;

import org.audiochain.io.AudioIOException;
import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectAdapter;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.AudioTrackAdapter;
import org.audiochain.model.PerformanceListener;
import org.audiochain.model.PerformanceManager;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.gui.ImageButtonComponent;
import org.audiochain.ui.gui.ImageSliderComponent;


public class PerformanceControlComponent extends Box implements PerformanceListener {
	private static final long serialVersionUID = 1L;
	private final AudioProject project;
	private final PerformanceManager performanceManager;
	private boolean userIsSeeking;
	private int lastSliderValue;
	private ImageSliderComponent slider;
	private JLabel timeLabel;
	private JLabel maxTimeLabel;
	private ImageButtonComponent startButton;
	private ImageButtonComponent recordingButton;
	private JComponent stopButton;
	private float frameRate;

	public PerformanceControlComponent(AudioProject project) {
		super(BoxLayout.X_AXIS);
		this.project = project;
		this.performanceManager = project.getPerformanceManager();
		frameRate = project.getAudioFormat().getFrameRate();
		init();
	}

	private void init() {
		setOpaque(false);

		startButton = new ImageButtonComponent("org/audiochain/ui/gui/mixer/playback-item.png");
		startButton.setToolTipText("Start playback from the current frame position (key p or b to start from beginning)");
		add(startButton);
		startButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				if (startButton.isSelected()) {
					return;
				}
				start();
			}
		});

		recordingButton = new ImageButtonComponent("org/audiochain/ui/gui/mixer/recording-item.png");
		recordingButton.setToolTipText("Start recording at the current frame position (key p or b to start from beginning)");
		recordingButton.setVisible(false);
		add(recordingButton);
		recordingButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				if (recordingButton.isSelected()) {
					return;
				}
				start();
			}
		});

		stopButton = new ImageButtonComponent("org/audiochain/ui/gui/mixer/stop-item.png");
		stopButton.setToolTipText("Stop the currently running playback or recording performance (key s or escape. To toggle start and stop use space or enter)");
		add(stopButton);
		stopButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				performanceManager.stop();
			}
		});

		timeLabel = new JLabel();
		timeLabel.setToolTipText("This is the current playback time of the current slider frame position (displayed in minutes and seconds).");
		Font font = timeLabel.getFont();
		Font greaterFont = new Font(font.getName(), font.getStyle(), font.getSize() + 4);
		timeLabel.setFont(greaterFont);
		add(timeLabel);
		maxTimeLabel = new JLabel();
		maxTimeLabel.setToolTipText("This is the maximum playback durance of the project (displayed in minutes and seconds).");
		maxTimeLabel.setFont(greaterFont);
		add(maxTimeLabel);

		slider = new ImageSliderComponent("org/audiochain/ui/gui/mixer/layered-slider.ci");
		slider.setToolTipText("This is a slider to display and select the current frame position for playback and recording (arrow keys move the slider).");
		add(slider);
		MouseAdapter sliderMouseAdapter = new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				if (!slider.isEnabled()) {
					return;
				}
				userIsSeeking = true;
				gatherSliderValues();
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				if (!slider.isEnabled()) {
					return;
				}
				gatherSliderValues();
				userIsSeeking = false;
			}

			@Override
			public void mouseDragged(MouseEvent e) {
				if (!slider.isEnabled()) {
					return;
				}
				gatherSliderValues();
			}
		};
		slider.addMouseListener(sliderMouseAdapter);
		slider.addMouseMotionListener(sliderMouseAdapter);

		setSliderAndTimeMaximum();
		setSliderAndTimeValue(min(project.getFramePosition(), project.getMaxFrame()));

		final AudioTrackAdapter trackChangeAdapter = new AudioTrackAdapter() {
			@Override
			public void audioTrackSelectedForRecordingChanged(AudioTrack audioTrack, boolean oldValue, boolean selected) {
				startButton.setVisible(!selected);
				recordingButton.setVisible(selected);
				if (selected) {
					/*
					 * asserts, that the start button is controlable via the
					 * keyboard. This is not the case when the focus hangs on a
					 * text field (i.e. the track's name field)
					 */
					recordingButton.grabFocus();
				} else {
					startButton.grabFocus();
				}
			}
		};
		
		for (AudioTrack track : project.getAudioTracks()) {
			track.addAudioTrackChangeListener(trackChangeAdapter);
		}

		project.addAudioProjectChangeListener(new AudioProjectAdapter() {
			@Override
			public void audioTrackAdded(AudioTrack track) {
				setSliderAndTimeMaximum();
				track.addAudioTrackChangeListener(trackChangeAdapter);
			}

			@Override
			public void audioTrackRemoved(AudioTrack track) {
				setSliderAndTimeMaximum();
				track.removeAudioTrackChangeListener(trackChangeAdapter);
			}

			@Override
			public void audioProjectFrameLengthChanged() {
				setSliderAndTimeMaximum();
			}
			
			@Override
			public void audioProjectFramePositionChanged(long oldFramePosition, long newFramePosition) throws IOException {
				if (!userIsSeeking) {
					setSliderAndTimeValue(newFramePosition);
				}
			}
		});
		
		performanceManager.addPerformanceListener(this);
	}

	@Override
	public void performanceStarted() {
		startButton.setSelected(true);
		recordingButton.setSelected(true);
		slider.setEnabled(!performanceManager.isRecording());
		stopButton.grabFocus();
	}

	@Override
	public void performanceUpdated() {
		
	}

	@Override
	public void performanceStopped() {
		setSliderAndTimeMaximum();
		startButton.setSelected(false);
		recordingButton.setSelected(false);
		slider.setEnabled(true);
	}

	@Override
	public void performanceFinished() {
		Collection<Throwable> errors = performanceManager.getLastPerformanceErrors();
		if (errors != null) {
			for (Throwable error : errors) {
				Dialog.message(this, "Error while performing", error);
			}
		}
		startButton.setSelected(false);
		recordingButton.setSelected(false);
		slider.setEnabled(true);
	}

	private void start() {
		try {
			project.setFramePosition(slider.getValue());
			performanceManager.start();
		} catch (IOException e) {
			e.printStackTrace();
			Dialog.message(this, "IO Error while starting Recording/Playback", e);
		} catch (RuntimeException e) {
			e.printStackTrace();
			Dialog.message(this, "Error while starting Recording/Playback", e);
		} catch (AudioIOException e) {
			e.printStackTrace();
			Dialog.message(this, "Audio Mixer Error while starting Recording/Playback", e);
		}
	}

	private void updateTimeLabel(long framePosition) {
		long currentSecond = framePosition / (long) frameRate;
		timeLabel.setText(formatTime(currentSecond));
	}

	private void setSliderAndTimeValue(long value) {
		slider.setValue((int) value);
		updateTimeLabel(value);
	}

	private void setSliderAndTimeMaximum() {
		long frameLength = project.getMaxFrame();
		slider.setMaximum((int) frameLength);
		float seconds = frameLength / frameRate;
		maxTimeLabel.setText("/" + formatTime((long) seconds));
		if (slider.getValue() > frameLength) {
			setSliderAndTimeValue(frameLength);
		}
	}

	private static String formatTime(long seconds) {
		long min = seconds / 60L;
		long sec = seconds % 60L;
		return (min < 10 ? "0" : "") + min + ':' + (sec < 10 ? "0" : "") + sec;
	}

	private void gatherSliderValues() {
		// only change the sliderValue, if it differs more than one second
		int newValue = slider.getValue();
		try {
			project.setFramePosition(newValue);
		} catch (IOException e) {
			e.printStackTrace();
			Dialog.message(this, "IO Error", e);
		}
		newValue /= frameRate;
		newValue *= frameRate;
		if (lastSliderValue != newValue) {
			lastSliderValue = newValue;
			updateTimeLabel(lastSliderValue);
		}
	}
}
