/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JLabel;

import org.audiochain.model.AudioDevice;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.AudioTrackAdapter;
import org.audiochain.model.SourceAudioDevice;
import org.audiochain.model.UserInterfaceContext;
import org.audiochain.ui.gui.CommonComponentFactory;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.gui.ImageButtonComponent;


public class SourceAudioDeviceComponent extends Box {
	private static final long serialVersionUID = 1L;
	private final AudioTrack audioTrack;
	private SourceAudioDevice audioDevice;
	private JLabel label;
	private Box headLineBox;
	private Box mainBox;
	private Box audioDeviceComponentBox;
	private ImageButtonComponent selectDeviceButton;
	
	public SourceAudioDeviceComponent(AudioTrack audioTrack) {
		super(BoxLayout.Y_AXIS);
		this.audioTrack = audioTrack;
		this.audioDevice = audioTrack.getSourceAudioDevice();
		init();
	}
	
	private void init() {
		label = new JLabel(createLabelText());
		label.setAlignmentX(CENTER_ALIGNMENT);
		headLineBox = Box.createHorizontalBox();
		headLineBox.add(label);
		headLineBox.add(Box.createHorizontalGlue());
		headLineBox.addMouseListener(new MouseAdapter(){
			@Override
			public void mouseClicked(MouseEvent e) {
				if(mainBox.isVisible()) {
					hideControl();
				} else {
					unhideControl();
				}
			}
		});
		headLineBox.setVisible(false);
		add(headLineBox);
		mainBox = Box.createHorizontalBox();
		mainBox.add(Box.createHorizontalGlue());
		audioDeviceComponentBox = Box.createVerticalBox();
		audioDeviceComponentBox.setAlignmentY(TOP_ALIGNMENT);
		updateDeviceComponentBox();
		mainBox.add(audioDeviceComponentBox);
		mainBox.add(Box.createHorizontalGlue());
		mainBox.add(createToolbarBox());
		add(mainBox);
		add(CommonComponentFactory.createHorizontalLine());
		audioTrack.addAudioTrackChangeListener(new AudioTrackAdapter(){
			@Override
			public void audioTrackSelectedForRecordingChanged(AudioTrack audioTrack, boolean oldValue, boolean selectedForRecording) {
				updateSelectDeviceButtonStatus();
			}

			@Override
			public void audioTrackSourceAudioDeviceChanged(AudioTrack audioTrack, SourceAudioDevice oldValue, SourceAudioDevice newDevice) {
				audioDevice = audioTrack.getSourceAudioDevice();
				updateDeviceComponentBox();
			}
		});
		updateHiddenStatus();
	}

	private void updateHiddenStatus() {
		if (audioDevice != null && audioDevice.getUserInterfaceContext().isHidden()) {
			mainBox.setVisible(false);
			headLineBox.setVisible(true);
		} else {
			mainBox.setVisible(true);
			headLineBox.setVisible(false);
		}
	}

	private String createLabelText() {
		if(audioDevice!=null) {
			return audioDevice.getName();
		}
		return null;
	}

	private void updateDeviceComponentBox() {
		audioDeviceComponentBox.removeAll();
		if (audioDevice != null) {
			@SuppressWarnings("unchecked")
			UserInterfaceContext<AudioDevice> userInterfaceContext = (UserInterfaceContext<AudioDevice>) audioDevice.getUserInterfaceContext();
			audioDeviceComponentBox.add(userInterfaceContext.getComponent(audioDevice));
			revalidate();
		}
		label.setText(createLabelText());
	}

	private Box createToolbarBox() {
		Box toolbarBox = Box.createVerticalBox();
		toolbarBox.setAlignmentY(TOP_ALIGNMENT);
		toolbarBox.add(createHideButton());
		selectDeviceButton = createSelectDeviceButton();
		updateSelectDeviceButtonStatus();
		toolbarBox.add(selectDeviceButton);
		return toolbarBox;
	}
	
	private void updateSelectDeviceButtonStatus() {
		selectDeviceButton.setVisible(!audioTrack.isSelectedForRecording());
	}

	private ImageButtonComponent createHideButton() {
		ImageButtonComponent button = new ImageButtonComponent("org/audiochain/ui/gui/mixer/minus-item.png");
		button.setToolTipText("Hides this audio device.");
		button.setReleasedBorder(BorderFactory.createEmptyBorder());
		button.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				hideControl();
			}
		});
		return button;
	}

	private ImageButtonComponent createSelectDeviceButton() {
		ImageButtonComponent button = new ImageButtonComponent("org/audiochain/ui/gui/mixer/source-device-item.png");
		button.setToolTipText("Selects the source device for this audio track.");
		button.setReleasedBorder(BorderFactory.createEmptyBorder());
		button.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				selectDevice();
			}
		});
		return button;
	}

	private void hideControl() {
		if(audioDevice!=null) {
			mainBox.setVisible(false);
			headLineBox.setVisible(true);
			// triggers listeners
			audioDevice.getUserInterfaceContext().setHidden(true);
		}
	}

	private void unhideControl() {
		if(audioDevice!=null) {
			mainBox.setVisible(true);
			headLineBox.setVisible(false);
			// triggers listeners
			audioDevice.getUserInterfaceContext().setHidden(false);
		}
	}
	
	private void selectDevice() {
		SourceAudioDevice selectedSourceAudioDevice = ChooseAudioDeviceDialog.selectSourceAudioDevice(this);
		if(selectedSourceAudioDevice!=null) {
			if(audioDevice!=null) {
				if(selectedSourceAudioDevice.getClass().equals(audioDevice.getClass())) {
					if(!Dialog.okCancel(this, "Overriding Source Device With Same Device Type", "You have selected the device '"+selectedSourceAudioDevice.getName()+"', but this type is already the source device for this track. I will override the currently set device with the same device type. All device settings and any associated files will be lost.")) {
						return;
					}
				} else {
					if(!Dialog.okCancel(this, "Overriding Source Device", "You have selected the device '"+selectedSourceAudioDevice.getName()+"'. I will override the currently set device '"+audioDevice.getClass().getSimpleName()+"'. All device settings and any associated files will be lost.")) {
						return;
					}
				}
			}
			audioTrack.setSourceAudioDevice(selectedSourceAudioDevice);
		}
	}
}
