/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.gui.mixer;

import static org.audiochain.ui.UIFormatHelper.*;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JProgressBar;
import javax.swing.SwingUtilities;

import org.audiochain.model.AudioProject;
import org.audiochain.model.BinaryDataset;
import org.audiochain.ui.gui.CommonComponentFactory;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.sync.BinarySyncSocket;
import org.audiochain.ui.sync.BinarySyncSocketListener;
import org.audiochain.ui.sync.CommandSyncSocket;
import org.audiochain.ui.sync.SyncManager;
import org.audiochain.ui.sync.SynchronizationListener;

public class SynchronizationMessageDialog implements SynchronizationListener, BinarySyncSocketListener {
	private Map<String, BinaryDatasetEntryComponent> binaryDatasetIdentifierToEntryComponentMap;
	private final Object caller;
	private Dialog dialog;
	private JComponent mainComponent;
	private JComponent entriesContainer;
	private Component glue;
	private boolean userHide;

	public SynchronizationMessageDialog(Object caller) {
		this.caller = caller;
	}

	private void init() {
		entriesContainer = Box.createVerticalBox();
		entriesContainer.setOpaque(false);
		glue = Box.createVerticalGlue();
		mainComponent = CommonComponentFactory.createScrollPane(entriesContainer);
		Dimension dim = new Dimension(450, 80);
		mainComponent.setMinimumSize(dim);
		mainComponent.setPreferredSize(dim);
		JButton hideButton = new JButton("Hide");
		hideButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				userHide = true;
				hideDialog();
			}
		});
		dialog = new Dialog(caller, "Binary Data Synchronization", hideButton){
			@Override
			protected JComponent createContentComponent() {
				return mainComponent;
			}
		};
		dialog.setCloseManually(true);
		dialog.setHideShowMessageAgainCheckbox(true);
		dialog.getDialog().setModal(false);
	}

	private void showDialog() {
		if (dialog == null) {
			init();
		}
		if (dialog.getDialog().isVisible()) {
			return;
		}
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				dialog.open();
			}
		});
	}

	private void hideDialog() {
		if (dialog != null) {
			dialog.close();
		}
	}

	@Override
	public void syncSocketConnected(CommandSyncSocket commandSyncSocket) {
		commandSyncSocket.getBinarySyncSocket().addBinarySyncSocketListener(this);
		Dialog.asyncMessage(caller, "Synchronization Connection Accepted", "A sync connection with address " + commandSyncSocket.getSocket().getInetAddress() + " has been connected to port "
				+ commandSyncSocket.getSocket().getLocalPort() + ".");
	}

	@Override
	public void syncSocketDisconnected(CommandSyncSocket commandSyncSocket) {
		BinarySyncSocket binarySyncSocket = commandSyncSocket.getBinarySyncSocket();
		// might not have been connected yet
		if (binarySyncSocket != null) {
			binarySyncSocket.removeBinarySyncSocketListener(this);
		}
		Dialog.asyncMessage(caller, "Synchronization Client Disconnected", "A client with address " + commandSyncSocket.getSocket().getInetAddress() + " has been disconnected.");
	}

	@Override
	public void syncExceptionOccured(Exception ex) {
		Dialog.asyncMessage(caller, "Error while running sync server", ex);
	}

	@Override
	public void syncServerStopped() {
		Dialog.asyncMessage(caller, "Sync Server stopped", "The sync server has been stopped.");
	}

	@Override
	public void syncServerStarted(final int port) {
		Dialog.asyncMessage(caller, "Sync Server started", "The sync server has started and is listening for new connections on port " + port + ".");
	}

	@Override
	public void binaryDatasetReceptionInitiated(BinarySyncSocket incomingSocket, AudioProject audioProject, String binaryDatasetIdentifier, long totalBytes) {
	}

	@Override
	public void binaryDatasetReceptionUpdate(BinarySyncSocket incomingSocket, AudioProject audioProject, String binaryDatasetIdentifier, long receivedBytes, long totalBytes, long remainTimeInMillis) {
		if (SyncManager.getInstance().displayIncoming()) {
			update(binaryDatasetIdentifier, receivedBytes, totalBytes, remainTimeInMillis);
		}
	}

	@Override
	public void binaryDatasetReceptionCompleted(BinarySyncSocket incomingSocket, AudioProject audioProject, String binaryDatasetIdentifier) {
		if (SyncManager.getInstance().displayIncoming()) {
			complete(binaryDatasetIdentifier);
		}
	}

	@Override
	public void binaryDatasetRemotelyAddedToQueue(BinarySyncSocket incomingSocket, AudioProject audioProject, String binaryDatasetIdentifier, long binaryDatasetLength) {
		if (SyncManager.getInstance().displayIncoming()) {
			update(binaryDatasetIdentifier, 0, binaryDatasetLength, -1);
		}
	}

	@Override
	public void binaryDatasetAddedToQueue(BinarySyncSocket binarySyncSocket, BinaryDataset binaryDataset) {
		if (SyncManager.getInstance().displayOutgoing()) {
			try {
				update(binaryDataset.getBinaryDatasetIdentifier(), 0, binaryDataset.getBinaryDatasetLength(), -1);
			} catch (IOException e) {
				Dialog.asyncMessage(caller, "Binary Data Sync Error", e);
			}
		}
	}

	@Override
	public void binaryDatasetTransmissionInitiated(BinarySyncSocket binarySyncSocket, BinaryDataset binaryDataset) {

	}

	@Override
	public void binaryDatasetTransmissionUpdate(BinarySyncSocket binarySyncSocket, BinaryDataset binaryDataset, long transmittedBytes, long totalBytes, long remainTimeInMillis) {
		if (SyncManager.getInstance().displayOutgoing()) {
			update(binaryDataset.getBinaryDatasetIdentifier(), transmittedBytes, totalBytes, remainTimeInMillis);
		}
	}

	@Override
	public void binaryDatasetTransmissionCompleted(BinarySyncSocket binarySyncSocket, BinaryDataset binaryDataset) {
		if (SyncManager.getInstance().displayOutgoing()) {
			complete(binaryDataset.getBinaryDatasetIdentifier());
		}
	}

	private void update(String binaryDatasetIdentifier, long currentLength, long totalLength, long remainTimeInMillis) {
		if (userHide) {
			return;
		}
		if (binaryDatasetIdentifierToEntryComponentMap == null) {
			binaryDatasetIdentifierToEntryComponentMap = new HashMap<String, BinaryDatasetEntryComponent>();
		}
		BinaryDatasetEntryComponent binaryDatasetEntryComponent = binaryDatasetIdentifierToEntryComponentMap.get(binaryDatasetIdentifier);
		if (binaryDatasetEntryComponent == null) {
			if (dialog == null) {
				init();
			}
			binaryDatasetEntryComponent = new BinaryDatasetEntryComponent(binaryDatasetIdentifier, totalLength);
			binaryDatasetIdentifierToEntryComponentMap.put(binaryDatasetIdentifier, binaryDatasetEntryComponent);
			entriesContainer.remove(glue);
			entriesContainer.add(binaryDatasetEntryComponent);
			entriesContainer.add(glue);
			mainComponent.validate();
		}
		binaryDatasetEntryComponent.update(currentLength, remainTimeInMillis);
		
		showDialog();
	}

	private void complete(String binaryDatasetIdentifier) {
		BinaryDatasetEntryComponent entry = binaryDatasetIdentifierToEntryComponentMap.remove(binaryDatasetIdentifier);
		if (entry != null) {
			entriesContainer.remove(entry);
			mainComponent.validate();
			if (binaryDatasetIdentifierToEntryComponentMap.isEmpty()) {
				userHide = false;
				hideDialog();
			}
		}
	}
}

class BinaryDatasetEntryComponent extends Box {
	private static final long serialVersionUID = 1L;
	private final JProgressBar progressBar;
	private final long binaryDatasetLength;
	private final JLabel remainTimeLabel;

	BinaryDatasetEntryComponent(String binaryDatasetIdentifier, long binaryDatasetLength) {
		super(BoxLayout.LINE_AXIS);
		this.binaryDatasetLength = binaryDatasetLength;
		JLabel identifierLabel = new JLabel(binaryDatasetIdentifier);
		JLabel sizeLabel = new JLabel(formatByteCount(binaryDatasetLength));
		progressBar = new JProgressBar();
		progressBar.setStringPainted(true);
		Dimension size = new Dimension(60, (int) progressBar.getPreferredSize().getHeight());
		progressBar.setPreferredSize(size);
		progressBar.setMaximumSize(size);
		progressBar.setBorderPainted(false);
		progressBar.setOpaque(false);
		setOpaque(false);
		add(identifierLabel);
		add(Box.createHorizontalGlue());
		add(sizeLabel);
		add(progressBar);
		remainTimeLabel = new JLabel();
		add(remainTimeLabel);
	}

	void update(long transmittedBytes, long remainTimeInMillis) {
		int percent = (int) (100 * transmittedBytes / binaryDatasetLength);
		progressBar.setValue(percent);
		remainTimeLabel.setText(formatTime(remainTimeInMillis < 0 ? remainTimeInMillis : (remainTimeInMillis / 1000)));
	}
}