/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.sync;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.model.AudioDevice;
import org.audiochain.model.AudioDeviceChain;
import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.BinaryAudioProject;
import org.audiochain.model.BinaryDataset;
import org.audiochain.model.ChainableAudioDevice;
import org.audiochain.model.ObjectStore;
import org.audiochain.model.SourceAudioDevice;

class AudioDeviceSyncObserver extends SyncAudioProjectModificationObserver {

	public AudioDeviceSyncObserver(AudioProject project, CommandSyncSocket commandSyncSocket) {
		super(project, commandSyncSocket);
	}
	
	@Override
	public void binaryChanged(BinaryDataset binaryDataset) {
		commandSyncSocket.getBinarySyncSocket().addBinaryDatasetToQueue(binaryDataset);
		if (!(binaryDataset instanceof BinaryAudioProject)) {
			commandSyncSocket.getBinarySyncSocket().addBinaryDatasetToQueue(new BinaryAudioProject(getAudioProject()));
		}
	}
	
	@Override
	public void audioDeviceEnabledChanged(ChainableAudioDevice chainableAudioDevice, boolean oldValue, boolean newValue) {
		commandSyncSocket.printCommand(audioDeviceEnabledChanged, getAudioProject().getIdentifier(), chainableAudioDevice.getIdentifier(), String.valueOf(newValue));
	}
	
	private final SyncCommand audioDeviceEnabledChanged = new SyncCommand("ADEC") {
		@Override
		protected void execute(String... parameters) {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			String deviceIdentifier = parameters[1];
			boolean newValue = Boolean.parseBoolean(parameters[2]);
			ChainableAudioDevice audioDevice = audioProject.findChainableAudioDeviceByIdentifier(deviceIdentifier);
			if (audioDevice != null) {
				audioDevice.setEnabled(newValue);
			}
		}
	};
	
	@Override
	public void performanceRelevantAudioDevicePropertyChanged(AudioDevice audioDevice) {
		try {
			String hexDevice = ObjectStore.toHex(audioDevice);
			commandSyncSocket.printCommand(audioDeviceChanged, getAudioProject().getIdentifier(), hexDevice);
		} catch (IOException e) {
			throw new IllegalStateException(e);
		}
	}

	private final SyncCommand audioDeviceChanged = new SyncCommand("ADC") {
		@Override
		protected void execute(String... parameters) throws SyncSocketException {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			String hexDevice = parameters[1];
			try {
				AudioDevice remoteDevice = (AudioDevice) ObjectStore.fromHex(hexDevice);
				String deviceIdentifier = remoteDevice.getIdentifier();
				AudioDevice localDevice = audioProject.findAudioDeviceByIdentifier(deviceIdentifier);
				if (localDevice != null) {
					localDevice.synchronizeBy(remoteDevice);
				}
			} catch (ClassNotFoundException e) {
				throw new SyncSocketException(e);
			} catch (InstantiationException e) {
				throw new SyncSocketException(e);
			} catch (IllegalAccessException e) {
				throw new SyncSocketException(e);
			} catch (IOException e) {
				throw new SyncSocketException(e);
			} catch (NoSuchMethodException e) {
				throw new SyncSocketException(e);
			} catch (InvocationTargetException e) {
				throw new SyncSocketException(e);
			}
		}
	};

	@Override
	public void audioDeviceAdded(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain) {
		registerOnAudioDevice(audioDevice);
		try {
			String hexDevice = ObjectStore.toHex(audioDevice);
			commandSyncSocket.printCommand(audioDeviceAdded, getAudioProject().getIdentifier(), audioDeviceChain.getIdentifier(), hexDevice);
		} catch (IOException e) {
			throw new IllegalStateException(e);
		}
	}
	
	private final SyncCommand audioDeviceAdded = new SyncCommand("ADA") {
		@Override
		protected void execute(String... parameters) throws SyncSocketException {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			String deviceChainIdentifier = parameters[1];
			String hexDevice = parameters[2];
			AudioDeviceChain audioDeviceChain = audioProject.findAudioDeviceChainByIdentifier(deviceChainIdentifier);
			if (audioDeviceChain != null) {
				try {
					ChainableAudioDevice newDevice = (ChainableAudioDevice) ObjectStore.fromHex(hexDevice);
					audioDeviceChain.addAudioDevice(newDevice);
				} catch (IOException e) {
					throw new SyncSocketException(e);
				} catch (ClassNotFoundException e) {
					throw new SyncSocketException(e);
				} catch (IllegalAccessException e) {
					throw new SyncSocketException(e);
				} catch (InstantiationException e) {
					throw new SyncSocketException(e);
				} catch (NoSuchMethodException e) {
					throw new SyncSocketException(e);
				} catch (InvocationTargetException e) {
					throw new SyncSocketException(e);
				}
			}
		}
	};

	@Override
	public void audioDeviceRemoved(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain) {
		unregisterFromAudioDevice(audioDevice);
		commandSyncSocket.printCommand(audioDeviceRemoved, getAudioProject().getIdentifier(), audioDevice.getIdentifier(), audioDeviceChain.getIdentifier());
	}
	
	private final SyncCommand audioDeviceRemoved = new SyncCommand("ADR") {
		@Override
		protected void execute(String... parameters) {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			String deviceIdentifier = parameters[1];
			String deviceChainIdentifier = parameters[2];
			ChainableAudioDevice audioDevice = audioProject.findChainableAudioDeviceByIdentifier(deviceIdentifier);
			AudioDeviceChain audioDeviceChain = audioProject.findAudioDeviceChainByIdentifier(deviceChainIdentifier);
			if (audioDevice != null && audioDeviceChain != null) {
				audioDeviceChain.removeAudioDevice(audioDevice);
			}
		}
	};

	@Override
	public void audioDeviceMoved(AudioDevice audioDevice, AudioDeviceChain audioDeviceChain, boolean increment) {
		commandSyncSocket.printCommand(audioDeviceMoved, getAudioProject().getIdentifier(), audioDevice.getIdentifier(), audioDeviceChain.getIdentifier(), String.valueOf(increment));
	}
	
	private final SyncCommand audioDeviceMoved = new SyncCommand("ADM") {
		@Override
		protected void execute(String... parameters) {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			String deviceIdentifier = parameters[1];
			String deviceChainIdentifier = parameters[2];
			boolean increment = Boolean.parseBoolean(parameters[3]);
			ChainableAudioDevice audioDevice = audioProject.findChainableAudioDeviceByIdentifier(deviceIdentifier);
			AudioDeviceChain audioDeviceChain = audioProject.findAudioDeviceChainByIdentifier(deviceChainIdentifier);
			if (audioDevice != null && audioDeviceChain != null) {
				if (increment) {
					audioDeviceChain.incrementAudioDevicePositionIndex(audioDevice);
				} else {
					audioDeviceChain.decrementAudioDevicePositionIndex(audioDevice);
				}
			}
		}
	};
	
	@Override
	public void audioTrackSourceAudioDeviceChanged(AudioTrack audioTrack, SourceAudioDevice oldDevice, SourceAudioDevice device) {
		super.audioTrackSourceAudioDeviceChanged(audioTrack, oldDevice, device);
		try {
			String hexDevice = ObjectStore.toHex(device);
			commandSyncSocket.printCommand(audioTrackSourceAudioDeviceChanged, getAudioProject().getIdentifier(), audioTrack.getIdentifier(), hexDevice);
		} catch (IOException e) {
			throw new IllegalStateException(e);
		}
	}
	
	private final SyncCommand audioTrackSourceAudioDeviceChanged = new SyncCommand("ATSADC") {
		@Override
		protected void execute(String... parameters) throws SyncSocketException {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			AudioTrack audioTrack = audioProject.findAudioTrackByIdentifier(parameters[1]);
			String hexDevice = parameters[2];
			try {
				SourceAudioDevice newDevice = (SourceAudioDevice) ObjectStore.fromHex(hexDevice);
				audioTrack.setSourceAudioDevice(newDevice);
			} catch (IOException e) {
				throw new SyncSocketException(e);
			} catch (ClassNotFoundException e) {
				throw new SyncSocketException(e);
			} catch (IllegalAccessException e) {
				throw new SyncSocketException(e);
			} catch (InstantiationException e) {
				throw new SyncSocketException(e);
			} catch (NoSuchMethodException e) {
				throw new SyncSocketException(e);
			} catch (InvocationTargetException e) {
				throw new SyncSocketException(e);
			}
		}
	};

	@Override
	public Collection<SyncCommand> getSyncCommands() {
		Collection<SyncCommand> syncCommands = new LinkedHashSet<SyncCommand>();
		syncCommands.add(audioDeviceChanged);
		syncCommands.add(audioDeviceEnabledChanged);
		syncCommands.add(audioDeviceAdded);
		syncCommands.add(audioDeviceRemoved);
		syncCommands.add(audioDeviceMoved);
		syncCommands.add(audioTrackSourceAudioDeviceChanged);
		return syncCommands;
	}

}
