/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.sync;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.ObjectStore;

class AudioTrackSyncObserver extends SyncAudioProjectModificationObserver {

	public AudioTrackSyncObserver(AudioProject project, CommandSyncSocket commandSyncSocket) {
		super(project, commandSyncSocket);
	}
	
	@Override
	public void audioTrackAdded(AudioTrack track) {
		super.audioTrackAdded(track);
		try {
			commandSyncSocket.printCommand(audioTrackAdded, getAudioProject().getIdentifier(), ObjectStore.toHex(track));
		} catch (IOException e) {
			commandSyncSocket.fireSyncSocketExceptionOccured(e);
		}
	}
	
	private final SyncCommand audioTrackAdded = new SyncCommand("ATA") {
		@Override
		protected void execute(String... parameters) throws SyncSocketException {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			try {
				AudioTrack track = (AudioTrack) ObjectStore.fromHex(parameters[1]);
				audioProject.addAudioTrack(track);
			} catch (IOException e) {
				throw new SyncSocketException(e);
			} catch (ClassNotFoundException e) {
				throw new SyncSocketException(e);
			} catch (IllegalAccessException e) {
				throw new SyncSocketException(e);
			} catch (InstantiationException e) {
				throw new SyncSocketException(e);
			} catch (NoSuchMethodException e) {
				throw new SyncSocketException(e);
			} catch (InvocationTargetException e) {
				throw new SyncSocketException(e);
			}
		}
	};
	
	@Override
	public void audioTrackRemoved(AudioTrack track) {
		super.audioTrackRemoved(track);
		commandSyncSocket.printCommand(audioTrackRemoved, getAudioProject().getIdentifier(), track.getIdentifier());
	}
	
	private final SyncCommand audioTrackRemoved = new SyncCommand("ATR") {
		@Override
		protected void execute(String... parameters) {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			String trackIdentifier = parameters[1];
			AudioTrack track = audioProject.findAudioTrackByIdentifier(trackIdentifier);
			audioProject.removeAudioTrack(track);
		}
	};
	
	@Override
	public void audioTrackNameChanged(AudioTrack audioTrack, String oldValue, String newValue) {
		commandSyncSocket.printCommand(audioTrackNameChanged, getAudioProject().getIdentifier(), audioTrack.getIdentifier(), newValue);
	}
	
	private final SyncCommand audioTrackNameChanged = new SyncCommand("ATNC") {
		@Override
		protected void execute(String... parameters) {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			String trackIdentifier = parameters[1];
			String name = parameters[2];
			AudioTrack audioTrack = audioProject.findAudioTrackByIdentifier(trackIdentifier);
			if (audioTrack != null) {
				audioTrack.setName(name);
			}
		}
	};

	@Override
	public void audioTrackEnabledChanged(AudioTrack audioTrack, boolean oldValue, boolean newValue) {
		commandSyncSocket.printCommand(audioTrackEnabledChanged, getAudioProject().getIdentifier(), audioTrack.getIdentifier(), String.valueOf(newValue));
	}
	
	private final SyncCommand audioTrackEnabledChanged = new SyncCommand("ATEC") {
		@Override
		protected void execute(String... parameters) {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			String trackIdentifier = parameters[1];
			boolean enabled = Boolean.parseBoolean(parameters[2]);
			AudioTrack audioTrack = audioProject.findAudioTrackByIdentifier(trackIdentifier);
			if (audioTrack != null) {
				audioTrack.setEnabled(enabled);
			}
		}
	};
	
	@Override
	public void audioTrackDescriptionChanged(AudioTrack audioTrack, String oldValue, String newValue) {
		commandSyncSocket.printCommand(audioTrackDescriptionChanged, getAudioProject().getIdentifier(), audioTrack.getIdentifier(), newValue);
	}
	
	private final SyncCommand audioTrackDescriptionChanged = new SyncCommand("ATDC") {
		@Override
		protected void execute(String... parameters) {
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(parameters[0]);
			if (audioProject == null) {
				return;
			}
			String trackIdentifier = parameters[1];
			String description = parameters[2];
			AudioTrack audioTrack = audioProject.findAudioTrackByIdentifier(trackIdentifier);
			if (audioTrack != null) {
				audioTrack.setDescription(description);
			}
		}
	};

	@Override
	public Collection<SyncCommand> getSyncCommands() {
		Collection<SyncCommand> syncCommands = new LinkedHashSet<SyncCommand>();
		syncCommands.add(audioTrackAdded);
		syncCommands.add(audioTrackRemoved);
		syncCommands.add(audioTrackNameChanged);
		syncCommands.add(audioTrackEnabledChanged);
		syncCommands.add(audioTrackDescriptionChanged);
		return syncCommands;
	}

}
