/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.sync;

import static org.audiochain.model.ObjectStore.*;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;
import java.net.Socket;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Queue;

import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectLifecycleObservable;
import org.audiochain.model.BinaryAudioProject;
import org.audiochain.model.BinaryDataset;
import org.audiochain.model.BinaryDataset.BinaryDatasetProcessCallback;
import org.audiochain.model.RemoteBinaryDatasetProvider;

public class BinarySyncSocket extends SyncSocket implements SyncCommandProvider {
	private final CommandSyncSocket commandSyncSocket;
	private final Queue<BinaryDataset> queue = new LinkedList<BinaryDataset>();
	private Collection<String> binaryDatasetIdentifiersToDistribute;
	private Collection<BinarySyncSocketListener> listeners;
	private final SyncCommand binaryDatasetQueued = new SyncCommand("bdq") {
		@Override
		protected void execute(String... parameters) throws SyncSocketException {
			String projectIdentifier = parameters[0];
			AudioProject audioProject = findAudioProjectByIdentifier(projectIdentifier);
			String binaryDatasetIdentifier = parameters[1];
			long binaryDatasetLength = Long.parseLong(parameters[2]);
			fireBinaryDatasetRemotelyAddedToQueue(audioProject, binaryDatasetIdentifier, binaryDatasetLength);
		}
	};

	BinarySyncSocket(Socket socket, OutputStream os, InputStream is, CommandSyncSocket commandSyncSocket, AudioProjectLifecycleObservable audioProjectLifecycleObserver) {
		super(socket, os, is, audioProjectLifecycleObserver, true);
		this.commandSyncSocket = commandSyncSocket;
		commandSyncSocket.setBinarySyncSocket(this);
		commandSyncSocket.fireBinarySyncSocketCreated(this);
	}

	BinarySyncSocket(Socket socket, CommandSyncSocket commandSyncSocket, AudioProjectLifecycleObservable audioProjectLifecycleObserver) throws IOException {
		super(socket, socket.getOutputStream(), socket.getInputStream(), audioProjectLifecycleObserver, false);
		this.commandSyncSocket = commandSyncSocket;
		commandSyncSocket.setBinarySyncSocket(this);
		commandSyncSocket.fireBinarySyncSocketCreated(this);
		os.write(getType());
		byte[] bytes = commandSyncSocket.getSessionIdentifier().getBytes();
		os.write(bytes.length);
		os.write(bytes);
		os.flush();
	}

	void addBinaryDatasetsToQueue(Collection<BinaryDataset> binaryDatasets) {
		if (binaryDatasets == null) {
			return;
		}
		for (BinaryDataset binaryDataset : binaryDatasets) {
			addBinaryDatasetToQueue(binaryDataset);
		}
	}

	void addBinaryDatasetToQueue(BinaryDataset binaryDataset) {
		if (binaryDataset == null) {
			return;
		}
		removeBinaryDatasetIdentifiersToDistribute(binaryDataset.getBinaryDatasetIdentifier());
		fireBinaryDatasetAddedToQueue(binaryDataset);
		queue.offer(binaryDataset);
	}

	@Override
	protected final int getType() {
		return 2;
	}

	@Override
	protected final void process() throws SyncSocketException {
		startBinaryDataQueue();
		try {
			String line;
			while ((line = readLine()) != null) {
				if (line.equals(EXIT)) {
					commitExit();
					return;
				}
				String projectIdentifier = line;
				String binaryDatasetClassName = readLine();
				final String binaryDatasetIdentifier = readLine();
				String lengthStr = readLine();
				long length = Long.parseLong(lengthStr);
				// create an instance to run the process method on it only (this
				// is an empty binary dataset that contains no data)
				BinaryDataset bd = (BinaryDataset) Class.forName(binaryDatasetClassName).newInstance();
				final AudioProject audioProject = findAudioProjectByIdentifier(projectIdentifier);
				fireBinaryDatasetReceptionInitiated(audioProject, binaryDatasetIdentifier, length);
				bd.processBinaryDataset(audioProject, binaryDatasetIdentifier, is, length, new BinaryDatasetProcessCallback() {
					long updateTime;
					long transmittedBytes;
					
					@Override
					public void binaryDatasetProcessUpdate(long processedBytes, long totalBytes) {
						// calculate remaining time
						long currentUpdateTime = System.nanoTime();
						long remainTimeInMillis = -1;
						if (updateTime > 0) {
							long lastUpdateTime = updateTime;
							long lastTransmittedBytes = transmittedBytes;
							long diffBytes = processedBytes - lastTransmittedBytes;
							long diffTime = currentUpdateTime - lastUpdateTime;
							double bytesPerMillis = diffBytes * 1000000 / diffTime;
							remainTimeInMillis = (long) ((totalBytes - processedBytes) / bytesPerMillis);
						}
						transmittedBytes = processedBytes;
						updateTime = currentUpdateTime;
						
						fireBinaryDatasetReceptionUpdate(audioProject, binaryDatasetIdentifier, processedBytes, totalBytes, remainTimeInMillis);
					}
				});
				if (bd instanceof RemoteBinaryDatasetProvider) {
					Collection<BinaryDataset> additionalBinaryDatasetsForRemote = ((RemoteBinaryDatasetProvider) bd).getAdditionalBinaryDatasetsForRemote();
					if (additionalBinaryDatasetsForRemote != null && !additionalBinaryDatasetsForRemote.isEmpty()) {
						addBinaryDatasetsToQueue(additionalBinaryDatasetsForRemote);
						// finally send the whole project for a new sync
						addBinaryDatasetToQueue(new BinaryAudioProject(audioProject));
					}
				}
				fireBinaryDatasetReceptionCompleted(audioProject, binaryDatasetIdentifier);
			}
		} catch (InstantiationException e) {
			throw new SyncSocketException(e);
		} catch (IllegalAccessException e) {
			throw new SyncSocketException(e);
		} catch (ClassNotFoundException e) {
			throw new SyncSocketException(e);
		} catch (IOException e) {
			throw new SyncSocketException(e);
		} catch (NoSuchMethodException e) {
			throw new SyncSocketException(e);
		} catch (InvocationTargetException e) {
			throw new SyncSocketException(e);
		} finally {
			closeSocket();
		}
	}

	private void writeBinaryDataset(final BinaryDataset binaryDataset) {
		try {
			fireBinaryDatasetTransmissionInitiated(binaryDataset);
			writeLine(binaryDataset.getBinaryDatasetAssociatedAudioProject().getIdentifier());
			writeLine(binaryDataset.getClass().getName());
			String binaryDatasetIdentifier = binaryDataset.getBinaryDatasetIdentifier();
			writeLine(binaryDatasetIdentifier);
			// getInputStream must be called before getLength to let the bd
			// create the length by the getInputStream call
			InputStream inputStream = binaryDataset.getBinaryDatasetInputStream();
			long length = binaryDataset.getBinaryDatasetLength();
			writeLine(String.valueOf(length));
			readExact(inputStream, os, length, new ReadUpdateCallback() {
				long updateTime;
				long transmittedBytes;
				
				@Override
				public void readUpdate(long read, long total) {
					// calculate remaining time
					long currentUpdateTime = System.nanoTime();
					long remainTimeInMillis = -1;
					if (updateTime > 0) {
						long lastUpdateTime = updateTime;
						long lastTransmittedBytes = transmittedBytes;
						long diffBytes = read - lastTransmittedBytes;
						long diffTime = currentUpdateTime - lastUpdateTime;
						double bytesPerMillis = diffBytes * 1000000 / diffTime;
						remainTimeInMillis = (long) ((total - read) / bytesPerMillis);
					}
					transmittedBytes = read;
					updateTime = currentUpdateTime;
					
					fireBinaryDatasetTransmissionUpdate(binaryDataset, read, total, remainTimeInMillis);
				}
			});
			int available = inputStream.available();
			assert available == 0 : "There are " + available + " non written bytes for binary sync of " + binaryDatasetIdentifier + ".";
			inputStream.close();
			os.flush();
			fireBinaryDatasetTransmissionCompleted(binaryDataset);
		} catch (IOException e) {
			fireSyncSocketExceptionOccured(e);
		}
	}

	private void startBinaryDataQueue() {
		new Thread((isServerSide() ? "Server" : "Client") + BinarySyncSocket.class.getSimpleName() + "DataQueue") {
			boolean runQueue = true;

			@Override
			public void run() {
				while (runQueue) {
					if (commandSyncSocket.getSocket().isClosed()) {
						try {
							closeSocket();
						} finally {
							runQueue = false;
						}
					} else if (BinarySyncSocket.this.getSocket().isClosed()) {
						try {
							commandSyncSocket.getSocket().close();
						} catch (IOException e) {
							fireSyncSocketExceptionOccured(e);
						} finally {
							runQueue = false;
						}
					} else if (!queue.isEmpty()) {
						writeBinaryDataset(queue.poll());
					} else {
						try {
							Thread.sleep(800);
						} catch (InterruptedException e) {
							fireSyncSocketExceptionOccured(e);
						}
					}
				}
				assert queue.isEmpty() : "Binary Queue was not empty.";
				queue.clear();
			}
		}.start();
	}

	private void closeSocket() {
		Socket socket = getSocket();
		if (!socket.isClosed()) {
			try {
				socket.close();
			} catch (IOException e) {
				fireSyncSocketExceptionOccured(e);
			}
		}
	}

	CommandSyncSocket getCommandSyncSocket() {
		return commandSyncSocket;
	}

	private void fireBinaryDatasetReceptionInitiated(AudioProject audioProject, String binaryDatasetIdentifier, long totalBytes) {
		if (listeners == null) {
			return;
		}
		for (BinarySyncSocketListener listener : new LinkedHashSet<BinarySyncSocketListener>(listeners)) {
			listener.binaryDatasetReceptionInitiated(this, audioProject, binaryDatasetIdentifier, totalBytes);
		}
	}

	private void fireBinaryDatasetReceptionUpdate(AudioProject audioProject, String binaryDatasetIdentifier, long receivedBytes, long totalBytes, long remainTimeInMillis) {
		if (listeners == null) {
			return;
		}
		for (BinarySyncSocketListener listener : new LinkedHashSet<BinarySyncSocketListener>(listeners)) {
			listener.binaryDatasetReceptionUpdate(this, audioProject, binaryDatasetIdentifier, receivedBytes, totalBytes, remainTimeInMillis);
		}
	}

	private void fireBinaryDatasetReceptionCompleted(AudioProject audioProject, String binaryDatasetIdentifier) {
		if (listeners == null) {
			return;
		}
		for (BinarySyncSocketListener listener : new LinkedHashSet<BinarySyncSocketListener>(listeners)) {
			listener.binaryDatasetReceptionCompleted(this, audioProject, binaryDatasetIdentifier);
		}
	}

	private void fireBinaryDatasetRemotelyAddedToQueue(AudioProject audioProject, String binaryDatasetIdentifier, long binaryDatasetLength) {
		if (listeners == null) {
			return;
		}
		for (BinarySyncSocketListener listener : new LinkedHashSet<BinarySyncSocketListener>(listeners)) {
			listener.binaryDatasetRemotelyAddedToQueue(this, audioProject, binaryDatasetIdentifier, binaryDatasetLength);
		}
	}

	private void fireBinaryDatasetAddedToQueue(BinaryDataset binaryDataset) {
		String projectIdentifier = binaryDataset.getBinaryDatasetAssociatedAudioProject().getIdentifier();
		String binaryDatasetIdentifier = binaryDataset.getBinaryDatasetIdentifier();
		long binaryDatasetLength;
		try {
			binaryDatasetLength = binaryDataset.getBinaryDatasetLength();
		} catch (IOException e) {
			e.printStackTrace();
			binaryDatasetLength = -1;
		}
		/*
		 * this is a control command, so it must be transferred in every case. There is no risk of command infinite ping pong at this point,
		 * because the further processing of this information is purely informative for the user.
		 */
		commandSyncSocket.forceWriteCommand(binaryDatasetQueued, projectIdentifier, binaryDatasetIdentifier, String.valueOf(binaryDatasetLength));
		if (listeners == null) {
			return;
		}
		for (BinarySyncSocketListener listener : new LinkedHashSet<BinarySyncSocketListener>(listeners)) {
			listener.binaryDatasetAddedToQueue(this, binaryDataset);
		}
	}

	private void fireBinaryDatasetTransmissionInitiated(BinaryDataset binaryDataset) {
		if (listeners == null) {
			return;
		}
		for (BinarySyncSocketListener listener : new LinkedHashSet<BinarySyncSocketListener>(listeners)) {
			listener.binaryDatasetTransmissionInitiated(this, binaryDataset);
		}
	}

	private void fireBinaryDatasetTransmissionUpdate(BinaryDataset binaryDataset, long transmittedBytes, long totalBytes, long remainTimeInMillis) {
		if (listeners == null) {
			return;
		}
		for (BinarySyncSocketListener listener : new LinkedHashSet<BinarySyncSocketListener>(listeners)) {
			listener.binaryDatasetTransmissionUpdate(this, binaryDataset, transmittedBytes, totalBytes, remainTimeInMillis);
		}
	}

	private void fireBinaryDatasetTransmissionCompleted(BinaryDataset binaryDataset) {
		if (listeners == null) {
			return;
		}
		for (BinarySyncSocketListener listener : new LinkedHashSet<BinarySyncSocketListener>(listeners)) {
			listener.binaryDatasetTransmissionCompleted(this, binaryDataset);
		}
	}

	public void addBinarySyncSocketListener(BinarySyncSocketListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<BinarySyncSocketListener>();
		}
		listeners.add(listener);
	}

	public void removeBinarySyncSocketListener(BinarySyncSocketListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	void addBinaryDatasetIdentifiersToDistribute(String binaryDatasetIdentifier) {
		if (binaryDatasetIdentifiersToDistribute == null) {
			binaryDatasetIdentifiersToDistribute = new LinkedHashSet<String>();
		}
		binaryDatasetIdentifiersToDistribute.add(binaryDatasetIdentifier);
	}

	private void removeBinaryDatasetIdentifiersToDistribute(String binaryDatasetIdentifier) {
		if (binaryDatasetIdentifiersToDistribute == null) {
			return;
		}
		binaryDatasetIdentifiersToDistribute.remove(binaryDatasetIdentifier);
	}

	Collection<String> getBinaryDatasetIdentifiersToDistribute() {
		Collection<String> ret = new LinkedHashSet<String>();
		if (binaryDatasetIdentifiersToDistribute != null) {
			ret.addAll(binaryDatasetIdentifiersToDistribute);
		}
		return ret;
	}

	@Override
	public Collection<SyncCommand> getSyncCommands() {
		Collection<SyncCommand> ret = new HashSet<SyncCommand>();
		ret.add(binaryDatasetQueued);
		return ret;
	}
}
