/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.sync;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.text.DateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedHashSet;
import java.util.Map;

import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectLifecycleAdapter;
import org.audiochain.model.AudioProjectLifecycleListener;
import org.audiochain.model.AudioProjectLifecycleObservable;
import org.audiochain.model.BinaryAudioProject;
import org.audiochain.model.IdentifierGenerator;

public class CommandSyncSocket extends SyncSocket {
	private static final String ALLOWED_SYNC_DIFF = "allowed.sync.difference";
	private static final String BINARY = "binary";
	private static final String READY = "ready";
	private static final String TIME = "TIME";
	private static final String TIMEASYNC = "TIMEASYNC";
	private final Collection<SyncCommand> syncCommands = new LinkedHashSet<SyncCommand>();
	private final RemoteAudioProjectProvider remoteAudioProjectProvider;
	private BinarySyncSocket binarySyncSocket;
	private String sessionIdentifier;
	private AudioProjectLifecycleListener audioProjectLifecycleListener = new AudioProjectLifecycleAdapter() {
		@Override
		public void audioProjectOpened(AudioProject audioProject) {
			createObserver(audioProject);
		}
	};
	private Collection<CommandSyncSocketListener> listeners;

	CommandSyncSocket(Socket socket, OutputStream os, InputStream is, final AudioProjectLifecycleObservable audioProjectLifecycleObserver) {
		super(socket, os, is, audioProjectLifecycleObserver, true);
		remoteAudioProjectProvider = new RemoteAudioProjectProvider(this);
		registerListeners();
		registerSyncCommands();
	}

	CommandSyncSocket(Socket socket, AudioProjectLifecycleObservable audioProjectLifecycleObserver) throws IOException {
		super(socket, socket.getOutputStream(), socket.getInputStream(), audioProjectLifecycleObserver, false);
		remoteAudioProjectProvider = new RemoteAudioProjectProvider(this);
		registerListeners();
		registerSyncCommands();
		String sessionIdentifier = IdentifierGenerator.generateIdentifier();
		setSessionIdentifier(sessionIdentifier);
		os.write(getType());
		os.write(sessionIdentifier.length());
		os.write(sessionIdentifier.getBytes());
		os.flush();
	}
	
	@Override
	protected final int getType() {
		return 1;
	}

	@Override
	protected final void process() throws IOException, SyncSocketException, SocketTimeoutException {
		try {
			if (isServerSide()) {
				initConnection();
			}
			String line;
			while ((line = readLine()) != null) {
				if (EXIT.equals(line)) {
					if (binarySyncSocket != null) {
						binarySyncSocket.exit();
					}
					commitExit();
					return;
				} else if (TIME.equals(line)) {
					// client side only
					writeLine(TIME + System.currentTimeMillis());
				} else if (line.startsWith(TIMEASYNC)) {
					exit();
					throw new SyncSocketException(line.substring(TIMEASYNC.length()));
				} else if (BINARY.equals(line)) {
					// client side only
					Socket socket = getSocket();
					InetAddress inetAddress = socket.getInetAddress();
					int port = socket.getPort();
					Socket sock = new Socket(inetAddress, port);
					binarySyncSocket = new BinarySyncSocket(sock, CommandSyncSocket.this, audioProjectLifecycleObserver);
					binarySyncSocket.setLogger(getLogger());
					binarySyncSocket.start();
					ready();
				} else if (READY.equals(line)) {
					waitUntilReady();
					Collection<AudioProject> audioProjects = getAudioProjects();
					for (AudioProject audioProject : audioProjects) {
						binarySyncSocket.addBinaryDatasetToQueue(new BinaryAudioProject(audioProject));
					}
				} else {
					boolean found = false;
					for (SyncCommand cmd : syncCommands) {
						if (cmd.execIfMatches(line)) {
							found = true;
							break;
						}
					}

					assert found || audioProjectLifecycleObserver.getAudioProjects().isEmpty() : "No " + SyncCommand.class.getSimpleName() + " found for line '" + line + "'.";
				}
			}
		} finally {
			audioProjectLifecycleObserver.removeAudioProjectLifecycleListener(audioProjectLifecycleListener);
		}
	}

	private void initConnection() throws IOException {
		long localTimeStart = System.currentTimeMillis();
		writeLine(TIME);
		String line = readLine(); // expecting something like 'TIME1366301333480'
		long localTimeResponse = System.currentTimeMillis();
		long remoteTime = Long.parseLong(line.substring(4));
		long networkLatency = localTimeResponse - localTimeStart;
		long diff = remoteTime - localTimeResponse;
		diff = diff < 0 ? diff * -1 : diff;
		diff -= networkLatency;
		diff = diff < 0 ? 0 : diff; // diff lower zero looks ugly to the user
		String allowedSyncDiff = System.getProperty(ALLOWED_SYNC_DIFF);
		if (allowedSyncDiff == null) {
			allowedSyncDiff = "1500"; // default
			System.setProperty(ALLOWED_SYNC_DIFF, allowedSyncDiff);
		}
		long allowedDiff = Long.parseLong(allowedSyncDiff);
		if (allowedDiff >= 0 && diff > allowedDiff) {
			DateFormat df = DateFormat.getDateTimeInstance();
			String msg = getTimeAsyncErrorMessage() + "Difference: " + diff + "ms, Server time: " + df.format(new Date(localTimeResponse)) + ", Client time: " + df.format(new Date(remoteTime)) + ", Network latency: " + networkLatency + "ms.";
			writeLine(TIMEASYNC + msg);
			fireSyncSocketExceptionOccured(new SyncSocketException(msg + " Alternatively set the server's JVM system property '" + ALLOWED_SYNC_DIFF + "' to a higher value or to a negative value to turn this check off."));
		} else {
			writeLine(BINARY);
		}
	}
	
	private static String getTimeAsyncErrorMessage() {
		return "The remote time differs from the local time. For an accurate synchronization, the system clocks must be nearly the same, because changes to the project are detected inter alia by the timestamp the changes are made. Please assert that the time difference is below or equal to "
				+ System.getProperty(ALLOWED_SYNC_DIFF) + "ms by adjusting either the remote or the local system clock. ";	
	}

	public final Map<String, String> listRemoteAudioProjects() throws SocketTimeoutException {
		return remoteAudioProjectProvider.listRemoteAudioProjects();
	}

	public final AudioProject getRemoteAudioProject(String identifier, String localProjectPath) throws SocketTimeoutException {
		return remoteAudioProjectProvider.getRemoteAudioProject(identifier, localProjectPath);
	}

	private void registerSyncCommands() {
		this.syncCommands.addAll(remoteAudioProjectProvider.getSyncCommands());
	}

	private void waitUntilReady() throws SocketTimeoutException {
		long start = System.currentTimeMillis();
		while (binarySyncSocket == null) {
			if (System.currentTimeMillis() - start > 20000) {
				throw new SocketTimeoutException("Binary Connection Timeout.");
			}
			try {
				Thread.sleep(50);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}

	void ready() {
		try {
			writeLine(READY);
		} catch (IOException e) {
			fireSyncSocketExceptionOccured(e);
		}
	}

	private void registerListeners() {
		audioProjectLifecycleObserver.addAudioProjectLifecycleListener(audioProjectLifecycleListener);
		Collection<AudioProject> audioProjects = audioProjectLifecycleObserver.getAudioProjects();
		for (AudioProject audioProject : audioProjects) {
			createObserver(audioProject);
		}
	}

	private void createObserver(AudioProject audioProject) {
		this.syncCommands.addAll(new AudioDeviceSyncObserver(audioProject, this).getSyncCommands());
		this.syncCommands.addAll(new AudioProjectSyncObserver(audioProject, this).getSyncCommands());
		this.syncCommands.addAll(new AudioTrackSyncObserver(audioProject, this).getSyncCommands());
	}

	public BinarySyncSocket getBinarySyncSocket() {
		return binarySyncSocket;
	}

	void setBinarySyncSocket(BinarySyncSocket binarySyncSocket) {
		this.binarySyncSocket = binarySyncSocket;
		syncCommands.addAll(binarySyncSocket.getSyncCommands());
	}

	public final String getSessionIdentifier() {
		return sessionIdentifier;
	}

	void setSessionIdentifier(String sessionIdentifier) {
		this.sessionIdentifier = sessionIdentifier;
	}

	void forceWriteCommand(SyncCommand syncCommand, String... parameters) {
		String commandString = syncCommand.getCommandString(parameters);
		try {
			writeLine(commandString);
		} catch (IOException e) {
			fireSyncSocketExceptionOccured(e);
		}
	}

	void printCommand(SyncCommand syncCommand, String... parameters) {
		if (mayWrite()) {
			forceWriteCommand(syncCommand, parameters);
		}
	}

	/**
	 * Prevent cyclic updates. Changes that are initiated within the socket
	 * client thread are not written back (reflected).
	 * 
	 * Any commands which caused changes within the socket server thread will be
	 * reflected to all other clients except the change initiating client.
	 */
	private boolean mayWrite() {
		Thread currentThread = Thread.currentThread();
		if (currentThread instanceof SyncSocketThread) {
			if (isServerSide()) {
				SyncSocketThread syncSocketThread = (SyncSocketThread) currentThread;
				if (syncSocketThread.syncSocket == this) {
					return false;
				}
				if (syncSocketThread.syncSocket == binarySyncSocket) {
					return false;
				}
			} else {
				return false;
			}
		}
		return true;
	}

	@Override
	public final void setLogger(Appendable logger) {
		super.setLogger(logger);
		if (binarySyncSocket != null) {
			binarySyncSocket.setLogger(logger);
		}
	}

	void addCommandSyncSocketListener(CommandSyncSocketListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<CommandSyncSocketListener>();
		}
		listeners.add(listener);
	}

	void removeCommandSyncSocketListener(CommandSyncSocketListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}
	
	void fireBinarySyncSocketCreated(BinarySyncSocket binarySyncSocket) {
		if (listeners == null) {
			return;
		}
		for(CommandSyncSocketListener listener : listeners) {
			listener.binarySyncSocketCreated(binarySyncSocket);
		}
	}
}
