/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.sync;

import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;

import org.audiochain.model.AudioProject;
import org.audiochain.model.BinaryAudioProject;

class RemoteAudioProjectProvider implements SyncCommandProvider {
	private final CommandSyncSocket commandSyncSocket;
	private static final String NULL_REPLACEMENT = "[NULL]";
	private static final String NAME_IDENTIFIER_SEPARATOR = "#";
	private boolean projectListRequestRunning;
	private AudioProject newLocalAudioProject;
	private Map<String, String> audioProjectIdentifierToNameMap;

	private final SyncCommand listProjects = new SyncCommand("lstpr") {
		@Override
		protected void execute(String... parameters) throws SyncSocketException {
			Collection<AudioProject> audioProjects = commandSyncSocket.audioProjectLifecycleObserver.getAudioProjects();
			Collection<String> projectNameAndIdentifiers = new ArrayList<String>();
			for (AudioProject audioProject : audioProjects) {
				String name = audioProject.getName();
				name = name == null ? NULL_REPLACEMENT : name;
				projectNameAndIdentifiers.add(name + NAME_IDENTIFIER_SEPARATOR + audioProject.getIdentifier());
			}
			/*
			 * write the response directly. Because the SyncSocketClientThread
			 * is not allowed to 'reflect' answers to the caller, we have to use
			 * the force write command to make the SyncSocketClientThread able
			 * to answer.
			 */
			commandSyncSocket.forceWriteCommand(listProjectResults, projectNameAndIdentifiers.toArray(new String[0]));
		}
	};

	private final SyncCommand listProjectResults = new SyncCommand("rlstpr") {
		@Override
		protected void execute(String... parameters) throws SyncSocketException {
			audioProjectIdentifierToNameMap = new LinkedHashMap<String, String>();
			for (String projectNameAndIdentifier : parameters) {
				String[] projectNameAndId = projectNameAndIdentifier.split(NAME_IDENTIFIER_SEPARATOR);
				String projectName = projectNameAndId[0];
				if (NULL_REPLACEMENT.equals(projectName)) {
					projectName = null;
				}
				String projectIdentifier = projectNameAndId[1];
				audioProjectIdentifierToNameMap.put(projectIdentifier, projectName);
			}

			projectListRequestRunning = false;
		}
	};

	private final SyncCommand getProject = new SyncCommand("getpr") {
		@Override
		protected void execute(String... parameters) throws SyncSocketException {
			String projectIdentifier = parameters[0];
			AudioProject audioProject = commandSyncSocket.findAudioProjectByIdentifier(projectIdentifier);
			commandSyncSocket.getBinarySyncSocket().addBinaryDatasetToQueue(new BinaryAudioProject(audioProject));
		}
	};

	public RemoteAudioProjectProvider(CommandSyncSocket commandSyncSocket) {
		this.commandSyncSocket = commandSyncSocket;
	}

	public Map<String, String> listRemoteAudioProjects() throws SocketTimeoutException {
		projectListRequestRunning = true;
		commandSyncSocket.printCommand(listProjects);
		long start = System.currentTimeMillis();
		while (projectListRequestRunning) {
			if (System.currentTimeMillis() - start > 20000) {
				throw new SocketTimeoutException("Connection timeout while gathering remote audio project list.");
			}
			try {
				Thread.sleep(50);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}

		return audioProjectIdentifierToNameMap;
	}

	public AudioProject getRemoteAudioProject(String identifier, String localProjectPath) throws SocketTimeoutException {
		// instantiate and register a new project with the same id. This will be
		// synced with the incoming remote project by the binary sync socket
		newLocalAudioProject = new AudioProject(localProjectPath);
		newLocalAudioProject.setIdentifier(identifier);
		commandSyncSocket.audioProjectLifecycleObserver.registerAudioProject(newLocalAudioProject);
		commandSyncSocket.printCommand(getProject, identifier);

		long start = System.currentTimeMillis();
		while (newLocalAudioProject.getLastSyncTime() == 0) {
			if (System.currentTimeMillis() - start > 20000) {
				commandSyncSocket.audioProjectLifecycleObserver.unregisterAudioProject(newLocalAudioProject);
				throw new SocketTimeoutException("Connection timeout while gathering remote audio project.");
			}
			try {
				Thread.sleep(50);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		// sync all files (init ping-pong sync)
		commandSyncSocket.getBinarySyncSocket().addBinaryDatasetToQueue(new BinaryAudioProject(newLocalAudioProject));
		return newLocalAudioProject;
	}

	@Override
	public Collection<SyncCommand> getSyncCommands() {
		Collection<SyncCommand> ret = new LinkedHashSet<SyncCommand>();
		ret.add(listProjects);
		ret.add(listProjectResults);
		ret.add(getProject);
		return ret;
	}
}
