/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.sync;

import java.io.IOException;
import java.io.Serializable;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.model.AudioProjectLifecycleObservable;
import org.audiochain.model.ObjectStore;

public final class SyncManager implements Serializable, SyncSocketListener, SyncServerListener, CommandSyncSocketListener {
	private static final long serialVersionUID = 1L;
	private static final String PROPS_FILE = "sync.properties";
	public static enum SyncProcessDisplayBehaviour {
		IncomingOnly, IncomingAndOutgoing, OutgoingOnly, Never
	}
	private boolean syncServerAtStartupEnabled;
	private boolean loggingEnabled;
	/*
	 * for "security" reasons, the port number defaults on any port between 1024
	 * and 65535 so it is not easy for an attacker to guess the type of service,
	 * when it is not listening on a given and known default port
	 */
	private int serverPort = (int) (Math.random() * (65535 - 1024) + 1024);
	private String lastRemoteHost;
	private int lastRemotePort;
	private SyncProcessDisplayBehaviour syncProcessDisplayBehaviour = SyncProcessDisplayBehaviour.IncomingOnly;
	private transient static SyncManager instance;
	private transient SyncServer syncServer;
	private transient AudioProjectLifecycleObservable audioProjectLifecycleObserver;
	private transient Appendable logger;
	private transient final Collection<CommandSyncSocket> commandSyncSockets;
	private transient final Collection<SynchronizationListener> listeners;

	SyncManager() {
		commandSyncSockets = new LinkedHashSet<CommandSyncSocket>();
		listeners = new LinkedHashSet<SynchronizationListener>();
	}

	public static synchronized SyncManager getInstance() {
		if (instance == null) {
			try {
				instance = (SyncManager) ObjectStore.read(PROPS_FILE);
			} catch (Exception e) {
				throw new IllegalStateException(e);
			}
			if (instance == null) {
				instance = new SyncManager();
				instance.store();
			}
			instance.init();
		}
		return instance;
	}

	private void init() {
		initLogger();
	}

	private void initLogger() {
		if (isLoggingEnabled()) {
			setLogger(System.out);
		} else {
			setLogger(null);
		}
	}

	public void setAudioProjectLifecycleObservable(AudioProjectLifecycleObservable audioProjectLifecycleObserver) {
		this.audioProjectLifecycleObserver = audioProjectLifecycleObserver;
	}

	public void initialize() {
		if (isSyncServerAtStartupEnabled()) {
			startSyncServer();
		}
	}

	private void store() {
		try {
			ObjectStore.storeInSettingsDirectory(this, PROPS_FILE);
		} catch (IOException e) {
			throw new IllegalStateException(e);
		}
	}

	public boolean isSyncServerRunning() {
		return syncServer != null;
	}

	public void startSyncServer() {
		if (isSyncServerRunning()) {
			throw new IllegalStateException(SyncServer.class.getSimpleName() + " is already running.");
		}
		if (audioProjectLifecycleObserver == null) {
			throw new IllegalStateException(SyncManager.class.getSimpleName() + " is not initialized.");
		}
		syncServer = new SyncServer(getServerPort(), audioProjectLifecycleObserver);
		syncServer.setLogger(getLogger());
		syncServer.addSyncServerListener(this);
		syncServer.start();
	}

	public void stopSyncServer() {
		if (!isSyncServerRunning()) {
			throw new IllegalStateException(SyncServer.class.getSimpleName() + " is not running.");
		}
		syncServer.finish();
	}

	public void setSyncServerEnabled(boolean enabled) {
		boolean running = isSyncServerRunning();

		if (enabled && !running) {
			startSyncServer();
		}

		if (!enabled && running) {
			stopSyncServer();
		}
	}

	public CommandSyncSocket createRemoteConnection(String host, int port) throws UnknownHostException, IOException {
		if (audioProjectLifecycleObserver == null) {
			throw new IllegalStateException(SyncManager.class.getSimpleName() + " is not initialized.");
		}
		Socket socket = new Socket(host, port);
		CommandSyncSocket syncSocket = new CommandSyncSocket(socket, audioProjectLifecycleObserver);
		syncSocket.setLogger(getLogger());
		syncSocket.addSyncSocketListener(this);
		syncSocket.addCommandSyncSocketListener(this);
		syncSocket.start();
		return syncSocket;
	}

	public Collection<CommandSyncSocket> getAllCommandSyncSockets() {
		return new LinkedHashSet<CommandSyncSocket>(commandSyncSockets);
	}

	@Override
	public void binarySyncSocketCreated(BinarySyncSocket binarySyncSocket) {
		binarySyncSocket.addSyncSocketListener(this);
	}

	@Override
	public void syncSocketOpened(SyncSocket syncSocket) {
		if (syncSocket instanceof CommandSyncSocket) {
			CommandSyncSocket commandSyncSocket = (CommandSyncSocket) syncSocket;
			commandSyncSockets.add(commandSyncSocket);
		} else if (syncSocket instanceof BinarySyncSocket) {
			BinarySyncSocket binarySyncSocket = (BinarySyncSocket) syncSocket;
			fireSyncSocketConnected(binarySyncSocket.getCommandSyncSocket());
		}
	}

	@Override
	public void syncSocketClosed(SyncSocket syncSocket) {
		if (syncSocket instanceof CommandSyncSocket) {
			CommandSyncSocket commandSyncSocket = (CommandSyncSocket) syncSocket;
			commandSyncSockets.remove(commandSyncSocket);
			fireSyncSocketDisconnected(commandSyncSocket);
		}
	}

	@Override
	public void syncSocketExceptionOccured(Exception ex) {
		fireSyncExceptionOccured(ex);
	}

	@Override
	public void syncServerStarted(int port) {
		fireSyncServerStarted(port);
	}

	@Override
	public void syncServerConnectionAccepted(SyncSocket syncSocket) {
		if (syncSocket instanceof CommandSyncSocket) {
			// to get info about a disconnect of this socket
			syncSocket.addSyncSocketListener(this);
		}
		syncSocketOpened(syncSocket);
	}

	@Override
	public void syncServerStopped() {
		syncServer = null;
		fireSyncServerStopped();
	}

	@Override
	public void syncServerExceptionOccured(Exception ex) {
		fireSyncExceptionOccured(ex);
	}

	public Appendable getLogger() {
		return logger;
	}

	public void setLogger(Appendable logger) {
		this.logger = logger;
		if (syncServer != null) {
			syncServer.setLogger(logger);
		}
		for (SyncSocket clientSocket : commandSyncSockets) {
			if (clientSocket != null) {
				clientSocket.setLogger(logger);
			}
		}
	}

	public boolean isLoggingEnabled() {
		return loggingEnabled;
	}

	public void setLoggingEnabled(boolean loggingEnabled) {
		if (this.loggingEnabled != loggingEnabled) {
			this.loggingEnabled = loggingEnabled;
			initLogger();
			store();
		}
	}

	public boolean isSyncServerAtStartupEnabled() {
		return syncServerAtStartupEnabled;
	}

	public void setSyncServerAtStartupEnabled(boolean syncServerAtStartupEnabled) {
		if (this.syncServerAtStartupEnabled != syncServerAtStartupEnabled) {
			this.syncServerAtStartupEnabled = syncServerAtStartupEnabled;
			store();
		}
	}

	public int getServerPort() {
		return serverPort;
	}

	public void setServerPort(int serverPort) {
		if (this.serverPort != serverPort) {
			this.serverPort = serverPort;
			restartServerIfRunning();
			store();
		}
	}

	/*
	 * When the server port changes, the server should be restarted using the
	 * new port.
	 */
	private void restartServerIfRunning() {
		if (isSyncServerRunning()) {
			syncServer.addSyncServerListener(new SyncServerAdapter() {
				@Override
				public void syncServerStopped() {
					startSyncServer();
				}
			});
			stopSyncServer();
		}
	}

	public void testServerPort(int testPort) throws IOException {
		if (serverPort != testPort) {
			ServerSocket server = null;
			try {
				server = new ServerSocket(testPort);
			} finally {
				if (server != null && !server.isClosed()) {
					server.close();
				}
			}
		}
	}

	public static void testRemoteConnection(String host, int port) throws UnknownHostException, IOException {
		Socket socket = null;
		try {
			socket = new Socket(host, port);
		} finally {
			if (socket != null && !socket.isClosed()) {
				socket.close();
			}
		}
	}

	public void addSynchronizationListener(SynchronizationListener listener) {
		listeners.add(listener);
	}

	public void removeSynchronizationListener(SynchronizationListener listener) {
		listeners.remove(listener);
	}

	private void fireSyncSocketConnected(CommandSyncSocket commandSyncSocket) {
		for (SynchronizationListener listener : listeners) {
			listener.syncSocketConnected(commandSyncSocket);
		}
	}

	private void fireSyncSocketDisconnected(CommandSyncSocket commandSyncSocket) {
		for (SynchronizationListener listener : listeners) {
			listener.syncSocketDisconnected(commandSyncSocket);
		}
	}

	private void fireSyncExceptionOccured(Exception ex) {
		for (SynchronizationListener listener : listeners) {
			listener.syncExceptionOccured(ex);
		}
	}

	private void fireSyncServerStarted(int port) {
		for (SynchronizationListener listener : listeners) {
			listener.syncServerStarted(port);
		}
	}

	private void fireSyncServerStopped() {
		for (SynchronizationListener listener : listeners) {
			listener.syncServerStopped();
		}
	}

	public int getLastRemotePort() {
		return lastRemotePort;
	}

	public void setLastRemotePort(int lastRemotePort) {
		this.lastRemotePort = lastRemotePort;
		store();
	}

	public String getLastRemoteHost() {
		return lastRemoteHost;
	}

	public void setLastRemoteHost(String lastRemoteHost) {
		this.lastRemoteHost = lastRemoteHost;
		store();
	}

	public void closeAllConnections() {
		for (CommandSyncSocket c : new LinkedHashSet<CommandSyncSocket>(commandSyncSockets)) {
			c.exit();
		}
	}

	public SyncProcessDisplayBehaviour getSyncProcessDisplayBehaviour() {
		return syncProcessDisplayBehaviour;
	}

	public void setSyncProcessDisplayBehaviour(SyncProcessDisplayBehaviour syncProcessDisplayBehaviour) {
		this.syncProcessDisplayBehaviour = syncProcessDisplayBehaviour;
		store();
	}
	
	public boolean displayIncoming() {
		return syncProcessDisplayBehaviour == SyncProcessDisplayBehaviour.IncomingAndOutgoing || syncProcessDisplayBehaviour == SyncProcessDisplayBehaviour.IncomingOnly;
	}
	
	public boolean displayOutgoing() {
		return syncProcessDisplayBehaviour == SyncProcessDisplayBehaviour.IncomingAndOutgoing || syncProcessDisplayBehaviour == SyncProcessDisplayBehaviour.OutgoingOnly;
	}
}
