/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.sync;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectLifecycleObservable;
import org.audiochain.model.BinaryAudioProject;
import org.audiochain.model.BinaryDataset;

final class SyncServer extends Thread implements SyncSocketListener {
	private boolean serverEnabled = true;
	private final int port;
	private final AudioProjectLifecycleObservable audioProjectLifecycleObserver;
	private final Collection<CommandSyncSocket> commandSyncSockets = new LinkedHashSet<CommandSyncSocket>();
	private Collection<SyncServerListener> listeners;
	private boolean startFired;
	private Appendable logger;

	public SyncServer(int port, AudioProjectLifecycleObservable audioProjectLifecycleObserver) {
		setName(SyncServer.class.getSimpleName());
		this.audioProjectLifecycleObserver = audioProjectLifecycleObserver;
		this.port = port;
	}

	@Override
	public void run() {
		try {
			while (serverEnabled) {
				ServerSocket server = new ServerSocket(port);
				server.setSoTimeout(5000);
				server.setReuseAddress(true);
				try {
					if (!startFired) {
						startFired = true;
						fireSyncServerStarted(port);
					}
					Socket socket = server.accept();
					SyncSocket syncSocket = createSyncSocket(socket);
					if (syncSocket != null) {
						syncSocket.addSyncSocketListener(this);
						syncSocket.start();
					} else {
						socket.close();
					}
				} catch (SocketTimeoutException e) {
					// this is what we aim to achieve every 5 seconds
				} finally {
					server.close();
				}
			}
		} catch (IOException e) {
			fireSyncServerExceptionOccured(e);
		} finally {
			if (startFired) {
				fireSyncServerStopped();
			}
			if (listeners != null) {
				listeners.clear();
			}
		}
	}

	private SyncSocket createSyncSocket(Socket socket) throws IOException {
		OutputStream os = socket.getOutputStream();
		InputStream is = socket.getInputStream();
		int type = is.read();
		if (type <= 0) {
			return null;
		}
		int sessionIdentifierLength = is.read();
		if (sessionIdentifierLength <= 0) {
			return null;
		}
		byte[] sessionIdentifierBytes = new byte[sessionIdentifierLength];
		is.read(sessionIdentifierBytes, 0, sessionIdentifierLength);
		String sessionIdentifier = new String(sessionIdentifierBytes);
		switch (type) {
		case 1:
			CommandSyncSocket commandSyncSocket = new CommandSyncSocket(socket, os, is, audioProjectLifecycleObserver);
			commandSyncSocket.setLogger(getLogger());
			commandSyncSocket.setSessionIdentifier(sessionIdentifier);
			commandSyncSockets.add(commandSyncSocket);
			return commandSyncSocket;
		case 2:
			CommandSyncSocket foundCommandSyncSocket = findCommandSyncSocketByIdentifier(sessionIdentifier);
			BinarySyncSocket binarySyncSocket = new BinarySyncSocket(socket, os, is, foundCommandSyncSocket, audioProjectLifecycleObserver);
			binarySyncSocket.addBinarySyncSocketListener(new BinaryDatasetDistributor());
			binarySyncSocket.setLogger(getLogger());
			return binarySyncSocket;
		default:
			throw new IllegalStateException("Socket Type " + type + " is undefined.");
		}
	}

	private CommandSyncSocket findCommandSyncSocketByIdentifier(String sessionIdentifier) {
		for (CommandSyncSocket commandSyncSocket : new LinkedHashSet<CommandSyncSocket>(commandSyncSockets)) {
			if (commandSyncSocket.getSessionIdentifier().equals(sessionIdentifier)) {
				return commandSyncSocket;
			}
		}
		throw new IllegalStateException("There is no running " + CommandSyncSocket.class.getSimpleName() + " for session " + sessionIdentifier);
	}

	public void finish() {
		if (!serverEnabled) {
			return;
		}
		serverEnabled = false;
		for(CommandSyncSocket cmd : commandSyncSockets) {
			cmd.exit();
		}
		commandSyncSockets.clear();
	}

	@Override
	public void syncSocketOpened(SyncSocket syncSocket) {
		if (syncSocket instanceof BinarySyncSocket) {
			BinarySyncSocket binarySyncSocket = (BinarySyncSocket) syncSocket;
			binarySyncSocket.getCommandSyncSocket().ready();
		}
		fireSyncServerConnectionAccepted(syncSocket);
	}

	@Override
	public void syncSocketClosed(SyncSocket syncSocket) {
		commandSyncSockets.remove(syncSocket);
	}

	@Override
	public void syncSocketExceptionOccured(Exception ex) {
		fireSyncServerExceptionOccured(ex);
	}

	private void fireSyncServerStarted(int port) {
		if (listeners == null) {
			return;
		}
		for (SyncServerListener listener : listeners) {
			listener.syncServerStarted(port);
		}
	}

	private void fireSyncServerConnectionAccepted(SyncSocket syncSocket) {
		if (listeners == null) {
			return;
		}
		for (SyncServerListener listener : listeners) {
			listener.syncServerConnectionAccepted(syncSocket);
		}
	}

	private void fireSyncServerStopped() {
		if (listeners == null) {
			return;
		}
		for (SyncServerListener listener : listeners) {
			listener.syncServerStopped();
		}
	}

	private void fireSyncServerExceptionOccured(Exception ex) {
		if (listeners == null) {
			return;
		}
		for (SyncServerListener listener : listeners) {
			listener.syncServerExceptionOccured(ex);
		}
	}

	public void addSyncServerListener(SyncServerListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<SyncServerListener>();
		}
		listeners.add(listener);
	}

	public void removeSyncServerListener(SyncServerListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	public Appendable getLogger() {
		return logger;
	}

	public void setLogger(Appendable logger) {
		this.logger = logger;
		for (SyncSocket syncSocket : commandSyncSockets) {
			syncSocket.setLogger(logger);
		}
	}

	public Collection<CommandSyncSocket> getCommandSyncSockets() {
		return new LinkedHashSet<CommandSyncSocket>(commandSyncSockets);
	}
	
	private class BinaryDatasetDistributor extends BinarySyncSocketAdapter {
		/**
		 * All incoming binary datasets must be distributed to all other clients.
		 */
		@Override
		public void binaryDatasetReceptionCompleted(BinarySyncSocket incomingSocket, AudioProject audioProject, String binaryDatasetIdentifier) {
			for (CommandSyncSocket cmd : commandSyncSockets) {
				BinarySyncSocket binarySyncSocket = cmd.getBinarySyncSocket();
				if (binarySyncSocket != incomingSocket) {
					// remember this identifier to be transferred, now or later if
					// not found at this time. The ids will be removed automatically by the binary sync socket when transferred
					binarySyncSocket.addBinaryDatasetIdentifiersToDistribute(binaryDatasetIdentifier);

					boolean anyDatasetAdded = false;
					// get all currently present binary datasets to find the ones
					// that can be transferred
					Collection<BinaryDataset> allBinaryDatasets = audioProject.getAllBinaryDatasets();
					Collection<String> binaryDatasetIdentifiersToDistribute = binarySyncSocket.getBinaryDatasetIdentifiersToDistribute();
					for (String potentialIdentifier : binaryDatasetIdentifiersToDistribute) {
						for (BinaryDataset binaryDataset : allBinaryDatasets) {
							String id = binaryDataset.getBinaryDatasetIdentifier();
							if (id.equals(potentialIdentifier)) {
								anyDatasetAdded = true;
								binarySyncSocket.addBinaryDatasetToQueue(binaryDataset);
							}
						}
					}
					if (anyDatasetAdded) {
						binarySyncSocket.addBinaryDatasetToQueue(new BinaryAudioProject(audioProject));
					}
				}
			}
		}
	}
}
