/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.ui.sync;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.nio.charset.Charset;
import java.util.Collection;
import java.util.LinkedHashSet;

import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectLifecycleObservable;

abstract class SyncSocket implements Runnable {
	protected static final String EXIT = "exit";

	static class SyncSocketThread extends Thread {
		final SyncSocket syncSocket;

		SyncSocketThread(SyncSocket syncSocket) {
			super(syncSocket, syncSocket.getDescription());
			this.syncSocket = syncSocket;
		}
	}

	private static final String LINESEP = System.getProperty("line.separator");
	private static final Charset UTF8 = Charset.forName("UTF-8");
	private static final int ETX = 3;
	private final Socket socket;
	protected final OutputStream os;
	protected final InputStream is;
	protected final AudioProjectLifecycleObservable audioProjectLifecycleObserver;
	private Collection<SyncSocketListener> listeners;
	private final boolean serverSide;
	private Appendable logger;
	private String logDescription;
	private boolean exitInitiator;

	public SyncSocket(Socket socket, OutputStream os, InputStream is, AudioProjectLifecycleObservable audioProjectLifecycleObserver, boolean serverSide) {
		this.socket = socket;
		this.os = os;
		this.is = is;
		this.audioProjectLifecycleObserver = audioProjectLifecycleObserver;
		this.serverSide = serverSide;
	}

	@Override
	public void run() {
		try {
			fireSyncSocketOpened();
			process();
		} catch (IOException e) {
			fireSyncSocketExceptionOccured(e);
		} finally {
			if (!socket.isClosed()) {
				try {
					socket.close();
				} catch (IOException e) {
					fireSyncSocketExceptionOccured(e);
				}
			}
			fireSyncSocketClosed();
			if (listeners != null) {
				listeners.clear();
			}
		}
	}

	void start() {
		new SyncSocketThread(this).start();
	}

	abstract protected void process() throws SyncSocketException, SocketTimeoutException, IOException;

	abstract protected int getType();

	protected void writeLine(String line) throws IOException {
		if (logger != null) {
			logger.append("<-" + getDescription() + "-> " + line).append(LINESEP);
		}
		if (socket.isClosed()) {
			throw new IllegalStateException("Socket is already closed.");
		}
		os.write((line).getBytes(UTF8));
		os.write(ETX);
		os.flush();
	}

	protected String readLine() throws IOException {
		if (socket.isClosed()) {
			throw new IllegalStateException("Socket is already closed.");
		}
		String line = read(is);
		if (logger != null) {
			logger.append("->" + getDescription() + "<- " + line).append(LINESEP);
		}
		return line;
	}

	private static String read(InputStream is) throws IOException {
		final int bufferIncrement = 64;
		int bufferSize = bufferIncrement;
		byte[] byteBuffer = new byte[bufferSize];
		int bufferIndex = 0;
		while (true) {
			final int oneByte = is.read();
			if (oneByte == -1) {
				return null;
			}
			if (oneByte == ETX) {
				return new String(byteBuffer, 0, bufferIndex, UTF8);
			}
			if (bufferIndex >= bufferSize) {
				// expand the buffer
				int newBufSize = bufferSize + bufferIncrement;
				byte[] newByteBuffer = new byte[newBufSize];
				System.arraycopy(byteBuffer, 0, newByteBuffer, 0, bufferSize);
				byteBuffer = newByteBuffer;
				bufferSize = newBufSize;
			}
			byteBuffer[bufferIndex++] = (byte) oneByte;
		}
	}

	public String getDescription() {
		if (logDescription == null) {
			logDescription = (getType() == 1 ? "cmd" : "bin") + " " + (isServerSide() ? "server" : "client") + " " + socket.getInetAddress().getHostAddress() + ":" + socket.getPort();
		}
		return logDescription;
	}

	public final void exit() {
		try {
			exitInitiator = true;
			writeLine(EXIT);
		} catch (IOException e) {
			fireSyncSocketExceptionOccured(e);
		}
	}

	protected void commitExit() throws IOException {
		if (!exitInitiator) {
			// send a return exit to the sender to commit the exit
			writeLine(EXIT);
		}
	}

	private void fireSyncSocketOpened() {
		if (listeners == null) {
			return;
		}
		for (SyncSocketListener listener : new LinkedHashSet<SyncSocketListener>(listeners)) {
			listener.syncSocketOpened(this);
		}
	}

	private void fireSyncSocketClosed() {
		if (listeners == null) {
			return;
		}
		for (SyncSocketListener listener : new LinkedHashSet<SyncSocketListener>(listeners)) {
			listener.syncSocketClosed(this);
		}
	}

	protected final void fireSyncSocketExceptionOccured(Exception ex) {
		ex.printStackTrace();
		if (listeners == null) {
			return;
		}
		for (SyncSocketListener listener : new LinkedHashSet<SyncSocketListener>(listeners)) {
			listener.syncSocketExceptionOccured(ex);
		}
	}

	final void addSyncSocketListener(SyncSocketListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<SyncSocketListener>();
		}
		listeners.add(listener);
	}

	final void removeSyncSocketListener(SyncSocketListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	public Socket getSocket() {
		return socket;
	}

	public boolean isServerSide() {
		return serverSide;
	}

	public OutputStream getOutputStream() {
		return os;
	}

	public InputStream getInputStream() {
		return is;
	}

	public AudioProject findAudioProjectByIdentifier(String identifier) {
		for (AudioProject audioProject : getAudioProjects()) {
			if (audioProject.getIdentifier().equals(identifier)) {
				return audioProject;
			}
		}
		return null;
	}

	protected Collection<AudioProject> getAudioProjects() {
		return audioProjectLifecycleObserver.getAudioProjects();
	}

	public Appendable getLogger() {
		return logger;
	}

	public void setLogger(Appendable logger) {
		this.logger = logger;
	}
}
